import string
from CmmValueLattice import CmmValueLattice

class CmmCmpInstruction(CmmInstruction) :

    def init(self) :
        self._cond_was_changed = 0
        self._make_following_jump_unconditional = 0

    def _make_following_jump_unconditional(self) :
        self._make_following_jump_unconditional = 1

    def _change_condition(self) :
        print ("_change_condition 0 " + str(self.body[0]) + "\n")
        tt = string.maketrans('<>', '><')
        c = self.body[0].translate(tt)
        if  c != self.body[0] :
            self.body[0] = c
            self._cond_was_changed = 1

            [reg, label, bit, bit_value] = self.get_jump_data()

            if bit == 30 or bit == 31 :
                self.set_jump_data([ reg, label, 61 - bit, bit_value ])

        print ("_change_condition 1 " + str(self.body[0]) + "\n")

    def condition_was_changed(self) :
        return self._cond_was_changed

    def get_op(self) :
        return self.body[0]

    def set_op(self, op) :
        self.body[0] = op

    def get_opA(self) :
        return self.body[1]

    def set_opA(self, opA) :
        self.body[1] = opA

    def get_opB(self) :
        return self.body[2]

    def set_opB(self, opB) :
        self.body[2] = opB

    def _get_auxiliary_var(self) :
        self.body.pop()

    def _set_auxiliary_var(self, var) :
        self.body += [var]

    def process_64_bit_spec_reg(self) :
        op = self.get_op()
        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        instr = []

        if opA.is_a_spec_reg() and opA_size == 64 :
            tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', opA_size))
            tmp_A_1 = tmp_A_0.new()

            instr += [CmmAssgnInstruction([ '=', tmp_A_0, opA ])]

            self.set_opA(tmp_A_1)

        if opB.is_a_spec_reg() and opB_size == 64 :
            tmp_B_0 = CmmSpecVar('','tmp', CmmType('u', opB_size))
            tmp_B_1 = tmp_B_0.new()

            instr += [CmmAssgnInstruction([ '=', tmp_B_0, opB ])]

            self.set_opB(tmp_B_1)

        return [ instr, self.cwl() ]

    def fix_use(self) :
        print ("RRR fix_use " + str(self) + " " + str(self.get_jump_data()) + "\n")

        op = self.get_op()
        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        instr = []

        if opA.is_a_var() :
            new_size = self.get_normalized_size(opA_size)

            if opA.is_nonalligned() or (opA.is_signed() and opA_size < new_size) :
                tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', new_size))
                tmp_A_1 = tmp_A_0.new()
                tmp_A_2 = CmmSpecVar('','tmp', CmmType('u', new_size))

                instr += [CmmAssgnDummyInstruction([ '=', tmp_A_0, CmmConst(0) ])]
                instr += [CmmBinAsInstruction([ "ext_" + str(new_size) + "_" + str(opA_size), tmp_A_2, tmp_A_1, opA ])]

                self.body[1] = tmp_A_2

        if opB.is_a_var() :
            new_size = self.get_normalized_size(opB_size)

            if opB.is_nonalligned() or (opB.is_signed() and opB_size < new_size) :
                tmp_B_0 = CmmSpecVar('','tmp', CmmType('u', new_size))
                tmp_B_1 = tmp_B_0.new()
                tmp_B_2 = CmmSpecVar('','tmp', CmmType('u', new_size))

                instr += [CmmAssgnDummyInstruction([ '=', tmp_B_0, CmmConst(0) ])]
                instr += [CmmBinAsInstruction([ "ext_" + str(new_size) + "_" + str(opB_size), tmp_B_2, tmp_B_1, opB ])]

                self.body[2] = tmp_B_2
        elif opB.is_a_constant() :
            opB_size = opB.get_size()
            [reg, label, bit, bit_value] = self.get_jump_data()

            if opB_size > 16 and (bit != 29 or (bit == 29 and opA_size >= opB_size)) :
                # bug : it could be 'sub' instead of 'cmp' and then you don't need to change_size of $opB which results in better code.
                tmp_B_0 = CmmSpecVar('', 'tmp', CmmType('u', opB_size))
                tmp_B_1 = tmp_B_0.new()

                instr += [CmmAssgnInstruction([ '=', tmp_B_0, opB ])]

                self.body[2] = tmp_B_1

        instr += [self.cwl()]
        ### 02.11.03 push @instr, CmmCmpInstruction->new( $self->{body} ) ;

        for lv in instr :
            print ("FIX_USE " + str(lv) + "\n")

        return [ instr ]

    def fix_memory_usage_0(self) :
        print ("RRR fix_memory_usage_0 " + str(self) + " " + str(self.get_jump_data()) + "\n")

        opA = self.body[1]
        opB = self.body[2]
        instr = []

        if opA.is_a_mem_ref() :
            tmp_A_0 = CmmSpecVar('','mem', CmmType('u', 64))
            tmp_A_1 = self.body[1] = tmp_A_0.new()

            instr += [CmmAssgnInstruction([ '=', tmp_A_0, opA ])]

        if opB.is_a_mem_ref() :
            tmp_B_0 = CmmSpecVar('','mem', CmmType('u', 64))
            tmp_B_1 = self.body[2] = tmp_B_0.new()

            instr += [CmmAssgnInstruction([ '=', tmp_B_0, opB ])]

        instr += [self.cwl()]
        ### 02.11.03 push @instr, CmmCmpInstruction->new( $self->{body} ) ;

        return [ instr ]

    def put_constant_at_opB(self) :
        print ("RRR put_constant_at_opB " + str(self) + " " + str(self.get_jump_data()) + "\n")

        opA = self.body[1]

        if opA.is_a_constant() :
            self._change_condition()
            self.body[1], self.body[2] = self.body[2], self.body[1]

        return [ self ]

    def fix_instr_with_opA_smaller_than_opB(self) :
        instr = []

        opA = self.body[1]
        opB = self.body[2]
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        if opA_size and opB_size and opA_size < opB_size :
            # bug : it could be 'sub' instead of 'cmp' and then you need to do nothing in this case which results in better code.
            if opB.is_a_constant() :
                tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', opB_size))
                tmp_A_1 = self.body[1] = tmp_A_0.new()

                instr += [CmmAssgnInstruction([ '=', tmp_A_0, opA ])]
            else :
                self._change_condition()
                self.body[1], self.body[2] = self.body[2], self.body[1]

        instr += [CmmCmpInstruction( self.body )]

        return [ instr ]

    def _process_ss(self) :
        instr = []

        op = self.body[0]
        opA = self.body[1]
        opB = self.body[2]
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        size = max(opA_size, opB_size)

        tmp_A_0 = None
        tmp_A_1 = None

        tmp_B_0 = None
        tmp_B_1 = None

        if opA_size < size :
            tmp_B_1 = opB

            tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', size))
            tmp_A_1 = tmp_A_0.new()

            instr  += [CmmAssgnInstruction([ "ext_" + str(size) + "_" + str(opA_size), tmp_A_0, opA])]
        elif opB_size < size :
            tmp_A_1 = opA

            tmp_B_0 = CmmSpecVar('','tmp', CmmType('u', size))
            tmp_B_1 = tmp_B_0.new()

            instr  += [CmmAssgnInstruction([ "ext_" + str(size) + "_" + str(opB_size), tmp_B_0, opB])]
        else :
            tmp_A_1 = opA
            tmp_B_1 = opB

        if op == '>' or op == '<=' :
            instr += [CmmBinAsInstruction([ '-', CmmSR('null'), tmp_B_1, tmp_A_1])]
        else : # '<' or '>=' or '==' or '!='
            instr += [CmmBinAsInstruction([ '-', CmmSR('null'), tmp_A_1, tmp_B_1])]

        if (op != '==' and op != '!=') :
            instr += [CmmGotoInstruction([ 'btj', CmmSR('pc') , None, None, None])]

        return [ instr ]

    def _process_su(self) :
        instr = []

        op = self.body[0]
        opA = self.body[1]
        opB = self.body[2]
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        size = max(opA_size, opB_size)

        zero = CmmSR('zero', CmmType('u', 8))
        null = CmmSR('null', CmmType('u', size))
        pc = CmmSR('pc')

        if opA_size > opB_size :
            if op == '>' or op == '<=' :
                instr += [CmmBinAsInstruction([ '-', null, opB, opA])]
            else : # '<' or '>=' or '==' or '!='
                instr += [CmmBinAsInstruction([ '-', null, opA, opB])]
        elif opA_size < opB_size :
            opB_size += 8

            tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', opB_size))
            tmp_A_1 = tmp_A_0.new()

            instr += [CmmAssgnInstruction([ "ext_" + str(opB_size) + "_" + str(opA_size), tmp_A_0, opA])]

            if op == '>' or op == '<=' :
                instr += [CmmBinAsInstruction([ '-', null, opB, tmp_A_1])]
            else : # '<' or '>=' or '==' or '!='
                instr += [CmmBinAsInstruction([ '-', null, tmp_A_1, opB])]
        else : # opA_size == opB_size
            tmp_E_0 = CmmSpecVar('','tmp', CmmType('u', 8))
            tmp_E_1 = tmp_E_0.new()

            tmp_A_0 = opA.new()
            tmp_A_1 = tmp_A_0.new()

            instr += [CmmAssgnInstruction([ 'ext_by_8', tmp_E_0, tmp_A_0])]

            if op == '>' or op == '<=' :
                instr += [CmmBinAsInstruction([ '-', null, opB, tmp_A_1])]
                instr += [CmmBinAsInstruction([ 'subb', null, zero, tmp_E_1])]
            else : # '<' or '>=' or '==' or '!='
                instr += [CmmBinAsInstruction([ '-', null, tmp_A_1, opB])]
                instr += [CmmBinAsInstruction([ 'subb', null, tmp_E_1, zero])]

        if op != '==' and op != '!=' :
            instr += [CmmGotoInstruction([ 'btj', pc, None, None, None])]

        return [ instr ]

    def _process_us(self) :
        instr = []

        op = self.body[0]
        opA = self.body[1]
        opB = self.body[2]
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        size = max(opA_size, opB_size)

        zero = CmmSR('zero', CmmType('u', 8))
        null = CmmSR('null', CmmType('u', size))
        pc = CmmSR('pc')

        if opA_size > opB_size :
            opA_size += 8

            tmp_B_0 = CmmSpecVar('','tmp', CmmType('u', opA_size))
            tmp_B_1 = tmp_B_0.new()

            instr += [CmmAssgnInstruction([ "ext_" + str(opA_size) + "_" + str(opB_size), tmp_B_0, opB])]

            if op == '>' or op == '<=' :
                instr += [CmmBinAsInstruction([ '-', null, tmp_B_1, opA])]
            else : # '<' or '>=' or '==' or '!='
                instr += [CmmBinAsInstruction([ '-', null, opA, tmp_B_1])]
        elif opA_size < opB_size :
            if op == '>' or op == '<=' :
                instr += [CmmBinAsInstruction([ '-', null, opB, opA])]
            else : # '<' or '>=' or '==' or '!='
                instr += [CmmBinAsInstruction([ '-', null, opA, opB])]
        else : # opA_size == opB_size
            tmp_E_0 = CmmSpecVar('','tmp', CmmType('u', 8))
            tmp_E_1 = tmp_E_0.new()

            tmp_B_0 = opB.new()
            tmp_B_1 = tmp_B_0.new()

            instr += [CmmAssgnInstruction([ 'ext_by_8', tmp_E_0, tmp_B_0])]

            if op == '>' or op == '<=' :
                instr += [CmmBinAsInstruction([ '-', null, tmp_B_1, opA])]
                instr += [CmmBinAsInstruction([ 'subb', null, tmp_E_1, zero])]
            else : # '<' or '>=' or '==' or '!='
                instr += [CmmBinAsInstruction([ '-', null, opA, tmp_B_1])]
                instr += [CmmBinAsInstruction([ 'subb', null, zero, tmp_E_1])]

        if op != '==' and op != '!=' :
            instr += [CmmGotoInstruction([ 'btj', pc, None, None, None])]

        return [ instr ]

    def _process_uu(self) :
        instr = []

        opA = self.body[1]
        opB = self.body[2]
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        opA_size = self.get_normalized_size(opA_size)
        opB_size = self.get_normalized_size(opB_size)

        print ("customize_size " + str(self.body) + " || " + str(opA_size) + " " + str(opB_size) + "\n")

    def customize_in_accordance_with_sizes(self) :
        print ("customize_in_accordance_with_sizes " + str(self) + "\n")

        op = self.get_op()
        op = 'eq_ne' if op == '==' or op == '!=' else 'g_l'

        opA = self.get_opA()
        opB = self.get_opB()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        opA_size = self.get_normalized_size(opA_size)
        opB_size = self.get_normalized_size(opB_size)

        var0 = CmmSpecVar('','tmp', CmmType('u', 32))
        var1 = CmmSpecVar('','tmp', CmmType('u', 32))

        self._set_auxiliary_var(var0.new())
        self._set_auxiliary_var(var1.new())

        return  [
                CmmAssgnDummyInstruction([ 'mviw', var0, opA.new() ]),
                CmmAssgnDummyInstruction([ 'mviw', var1, opB.new() ]),
                self
                ]

    def transform_to_pseudo_asm_form
        pass

    def fix_64__eq_ne_ss_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        return [
            self.cwl() ,
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_ss_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        var = var0
        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneB = CmmSR('one', CmmType('u', opB_size))

        r = var.get_reg()
        r.set_size(opB_size)

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', r, oneB, oneB ]),
            CmmBinAsInstruction([ 'shli', r, r, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', r, r, opA ]),
            CmmCmpInstruction([ 'cmp', r, opB ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_ss_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        opA_size = opA.get_size()
        opB_0_size = opB_0.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneB_0 = CmmSR('one', CmmType('u', opB_0_size))

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneB_0, oneB_0 ]),
            CmmAssgnInstruction([ 'move', var1, var0 ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmCmpInstruction([ 'cmp', var0, opB_0 ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmCmpInstruction([ 'cmp', var1, opB_1 ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_ss_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        var = var0
        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        oneA = CmmSR('one', CmmType('u', opA_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        r = var.get_reg()
        r.set_size(opA_size)

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', r, oneA, oneA ]),
            CmmBinAsInstruction([ 'shli', r, r, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', r, r, opB ]),
            CmmCmpInstruction([ 'cmp', opA, r ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_ss_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        opA_size = opA.get_size()
        opB_1_size = opB_1.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneB_1 = CmmSR('one', CmmType('u', opB_1_size))

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneB_1, oneB_1 ]),
            CmmCmpInstruction([ 'cmp', opA, opB_0 ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmCmpInstruction([ 'cmp', var0, opB_1 ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_ss_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opA_0_size = opA_0.get_size()
        opB_size = opB.get_size()
        oneA_0 = CmmSR('one', CmmType('u', opA_0_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
            CmmAssgnInstruction([ 'move', var1, var0 ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opB ]),
            CmmCmpInstruction([ 'cmp', opA_0, var0 ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmCmpInstruction([ 'cmp', opA_1, var1 ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_ss_64_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opA_1_size = opA_1.get_size()
        opB_size = opB.get_size()
        oneA_1 = CmmSR('one', CmmType('u', opA_1_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneA_1, oneA_1 ]),
            CmmCmpInstruction([ 'cmp', opA_0, opB ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmCmpInstruction([ 'cmp', opA_1, var0 ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_ss_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        # my ($label, $bit, $bit_value) = @{$self->get_jump_data()} ;

        return [
            CmmCmpInstruction([ 'cmp', opA_0, opB_0 ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmCmpInstruction([ 'cmp', opA_1, opB_1 ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_su_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()

        opA_size = opA.get_size()

        if opB.is_a_constant() :
            if opB & 0x80 :
                if bit_value :
                    return []
                else :
                    return [
                        CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                        ]
            else :
                return [
                    self.cwl(),
                    CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                    ]
        else :
            return [
                CmmGotoInstruction([ 'btjr', opA, CmmConst(3), CmmConst(opA_size) - 1, CmmConst(1) ] if bit_value\
                                else [ 'btj', opA, label, CmmConst(opA_size) - 1, CmmConst(1) ]),
                self.cwl() ,
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_su_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opB = self.get_opB()

        if opB.is_a_constant() :
            if bit_value :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            var = var0
            opA = self.get_opA()
            opB = self.get_opB()
            opA_size = opA.get_size()

            return [
                CmmGotoInstruction([ 'btjr', opA, CmmConst(4), CmmConst(opA_size) - 1, CmmConst(1) ] if bit_value\
                                else [ 'btj', opA, label, CmmConst(opA_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'or', var, CmmSR('zero', CmmType('u', 16)), opA ]),
                CmmCmpInstruction([ 'cmp', var, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_su_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opB = self.get_opB()

        if opB.is_a_constant() :
            # should be destroyed, following jump should be made unconditional
            if bit_value :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            var = var0
            opA = self.get_opA()
            opB = self.get_opB()
            opA_size = opA.get_size()

            return [
                CmmGotoInstruction([ 'btjr', opA, CmmConst(4), CmmConst(opA_size) - 1, CmmConst(1) ] if bit_value\
                                else [ 'btj', opA, label, CmmConst(opA_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'or', var, CmmSR('zero', CmmType('u', 32)), opA ]),
                CmmCmpInstruction([ 'cmp', var, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_su_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opB = self.get_opB()

        if opB.is_a_constant() :
            print("8_64 constant ! " + str(self.body) + "\n")
            # should be destroyed, following jump should be made unconditional
            if bit_value :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            print ("8_64 not a constant ! " + str(self.body) + "\n")
            var = var0
            opA = self.get_opA()
            opA_size = opA.get_size()
            opB = self.get_opB()
            [opB_1, opB_0] = opB.get_reg().splt_into_2()

            return [
                CmmGotoInstruction([ 'btjr', opA, CmmConst(6), CmmConst(opA_size) - 1, CmmConst(1) ] if bit_value\
                                else [ 'btj', opA, label, CmmConst(opA_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'or', var, CmmSR('zero', CmmType('u', 32)), opA ]),
                CmmCmpInstruction([ 'cmp', var, opB_0 ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                                else [ 'btj', reg, label, bit, bit_value ]),
                CmmCmpInstruction([ 'cmp', CmmSR('zero', CmmType('u', 32)), opB_1 ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_su_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        if opB.is_a_constant() :
            return [
                self.cwl(),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]
        else :
            return [
                CmmBinAsInstruction([ 'sub', zeroA, opA, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_su_16_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        if opB.is_a_constant() :
            if opB & 0x8000 :
                if bit_value :
                    return []
                else :
                    return [
                        CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                        ]
            else :
                return [
                    self.cwl(),
                    CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'sub', zeroA, opA, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_su_32_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opB = self.get_opB()

        if opB.is_a_constant() :
            opA = self.get_opA()
            opA_size = opA.get_size()
            var = var0
            r = var.get_reg()
            [r_1, r_0] = r.splt_into_2()
            opB_1 = opB >> 16
            opB_0 = opB & 0xffff

            if opB & 0x80000000 :
                if bit_value :
                    return []
                else :
                    return [
                        CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                        ]
            else :
                return [
                    CmmAssgnInstruction([ 'mviw', r_0, opB_0 ]),
                    CmmAssgnInstruction([ 'mviw', r_1, opB_1 ]),
                    CmmCmpInstruction([ 'cmp', opA, r ]),
                    CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                    ]
        else :
            opA = self.get_opA()
            opA_size = opA.get_size()
  
            return [
                CmmGotoInstruction([ 'btjr', opA, CmmConst(3), CmmConst(opA_size) - 1, CmmConst(1) ] if bit_value\
                                else [ 'btj', opA, label, CmmConst(opA_size) - 1, CmmConst(1) ]),
                self.cwl(),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_su_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opB = self.get_opB()

        if opB.is_a_constant() :
            # should be destroyed, following jump should be made unconditional
            if bit_value :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            opA = self.get_opA()
            opB = self.get_opB()
            [opB_1, opB_0] = opB.get_reg().splt_into_2()
            opA_size = opA.get_size()

            return [
                CmmGotoInstruction([ 'btjr', opA, CmmConst(5), CmmConst(opA_size) - 1, CmmConst(1) ] if bit_value\
                                else [ 'btj', opA, label, CmmConst(opA_size) - 1, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA, opB_0 ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                                else [ 'btj', reg, label, bit, bit_value ]),
                CmmCmpInstruction([ 'cmp', CmmSR('zero', CmmType('u', 32)), opB_1 ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_su_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opB = self.get_opB()
        opA_0_size = opA_0.get_size()
        zeroA_0 = CmmSR('zero', CmmType('u', opA_0_size))

        return [
            CmmBinAsInstruction([ 'sbi' if opB.is_a_constant() else 'sub', zeroA_0, opA_0, opB ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmBinAsInstruction([ 'sub', zeroA_0, opA_1, zeroA_0 ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_su_64_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opB = self.get_opB()
        opA_0_size = opA_0.get_size()
        opB_size = opB.get_size()
        zeroA_0 = CmmSR('zero', CmmType('u', opA_0_size))

        if opB.is_a_constant() :
            return [
                CmmBinAsInstruction([ 'cpiw', opA_0, opB ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                                else [ 'btj', reg, label, bit, bit_value ]),
                CmmBinAsInstruction([ 'cmp', opA_1, zeroA_0 ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]
        else :
            return [
                CmmBinAsInstruction([ 'sub', zeroA_0, opA_0, opB ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                                else [ 'btj', reg, label, bit, bit_value ]),
                CmmBinAsInstruction([ 'sub', zeroA_0, opA_1, zeroA_0 ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_su_64_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        instr = []

        if opB.is_a_constant() :
            r = var1.get_reg()
            opB_0 = opB & 0xffff
            opB_1 = opB >> 16

            [r_1, r_0] = r.splt_into_2()
            instr += [CmmAssgnInstruction([ 'mviw', r_0, opB_0 ]),
                      CmmAssgnInstruction([ 'mviw', r_1, opB_1 ])]

            opB = r

        return instr + [
            CmmCmpInstruction([ 'cmp', opA_0, opB ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', 32)) ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_su_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opB = self.get_opB()

        if opB.is_a_constant() :
            instr = []

            r_0 = var0.get_reg()
            r_1 = var1.get_reg()
            opB_0 = opB & 0xffffffff
            opB_1 = opB >> 32

            if opB_0 & 0xffff0000 :
                [r_0_1, r_0_0] = r_0.splt_into_2()
                instr += [CmmAssgnInstruction([ 'mviw', r_0_0, opB_0 & 0xffff ])]
                instr += [CmmAssgnInstruction([ 'mviw', r_0_1, opB_0 >> 16 ])]
            else :
                instr += [CmmAssgnInstruction([ 'mviw', r_0, opB_0 & 0xffff])]

            if opB_1 & 0xffff0000 :
                [r_1_1, r_1_0] = r_1.splt_into_2()
                instr += [CmmAssgnInstruction([ 'mviw', r_1_0, opB_1 & 0xffff ])]
                instr += [CmmAssgnInstruction([ 'mviw', r_1_1, opB_1 >> 16 ])]
            else :
                instr += [CmmAssgnInstruction([ 'mviw', r_1, opB_1 & 0xffff])]

            opA = self.get_opA()
            [opA_1, opA_0] = opA.get_reg().splt_into_2()
            # my [$label, $bit, $bit_value] = @{$self->get_jump_data()} ;
            opA_1_size = opA_1.get_size()

            return [
                CmmGotoInstruction([ 'btjr', opA_1, CmmConst(5) + len(instr), CmmConst(opA_1_size) - 1, CmmConst(1) ] if bit_value\
                                else [ 'btj', opA_1, label, CmmConst(opA_1_size) - 1, CmmConst(1) ])] +\
                instr +\
                [
                CmmCmpInstruction([ 'cmp', opA_0, r_0 ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                                else [ 'btj', reg, label, bit, bit_value ]),
                CmmCmpInstruction([ 'cmp', opA_1, r_1 ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]
        else :
            opA = self.get_opA()
            opB = self.get_opB()
            [opA_1, opA_0] = opA.get_reg().splt_into_2()
            [opB_1, opB_0] = opB.get_reg().splt_into_2()
            opA_1_size = opA_1.get_size()

            return [
                CmmGotoInstruction([ 'btjr', opA_1, CmmConst(5), CmmConst(opA_1_size) - 1, CmmConst(1) ] if bit_value\
                                else [ 'btj', opA_1, label, CmmConst(opA_1_size) - 1, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA_0, opB_0]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                                else [ 'btj', reg, label, bit, bit_value ]),
                CmmCmpInstruction([ 'cmp', opA_1, opB_1 ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_uu_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        return [
            self.cwl() ,
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_uu_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opB_size = opB.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if opB.is_a_constant() :
            if bit_value :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'sub', zeroB, opB, opA ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_uu_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opB_size = opB.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if opB.is_a_constant() :
            if bit_value :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'sub', zeroB, opB, opA ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_uu_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()

        if opB.is_a_constant() :
            if bit_value :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            [opB_1, opB_0] = opB.get_reg().splt_into_2()

            opB_0_size = opB_0.get_size()
            zeroB_0 = CmmSR('zero', CmmType('u', opB_0_size))

            return [
                CmmBinAsInstruction([ 'sub', zeroB_0, opB_0, opA ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                                else [ 'btj', reg, label, bit, bit_value ]),
                CmmBinAsInstruction([ 'sub', zeroB_0, opB_1, zeroB_0 ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_uu_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        if opB.is_a_constant() :
            return [
                self.cwl(),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]
        else :
            # CmmBinAsInstruction->new([ ($opB->is_a_constant() ? 'sbi' : 'sub'), $zeroA, $opA, $opB ]),
            return [
                CmmBinAsInstruction([ 'sub', zeroA, opA, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__eq_ne_uu_16_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()


        return [
            self.cwl(),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_uu_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opB = self.get_opB()
        opA_1_size = opA_1.get_size()
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))

        return [
            CmmCmpInstruction([ ('cpiw' if opB.is_a_constant() else 'cmp'), opA_0, opB ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmCmpInstruction([ 'cmp', opA_1, zeroA_1 ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_uu_64_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()

        return [
            CmmCmpInstruction([ 'cmp', opA_0, opB ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', 32)) ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__eq_ne_uu_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()

        return [
            CmmCmpInstruction([ 'cmp', opA_0, opB_0 ]),
            CmmGotoInstruction([ 'btjr', reg, CmmConst(3), bit, CmmConst(0) ] if bit_value\
                            else [ 'btj', reg, label, bit, bit_value ]),
            CmmCmpInstruction([ 'cmp', opA_1, opB_1 ]),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__g_l_ss_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        oneB = CmmSR('one', CmmType('u', opB_size))

        r0 = var0.get_reg()
        r1 = var1.get_reg()

        r0.set_size(opA_size)
        r1.set_size(opB_size)

        return [
            CmmBinAsInstruction([ 'sub', r1, opB, opA ] if bit == 31 else [ 'sub', r1, opA, opB ]),
            CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
            CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
            CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
            ]

    def fix_64__g_l_ss_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneB = CmmSR('one', CmmType('u', opB_size))

        r0 = var0.get_reg()
        r1 = var1.get_reg()

        r0.set_size(opB_size)
        r1.set_size(opB_size)

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', r1, oneB, oneB ]),
            CmmBinAsInstruction([ 'shli', r1, r1, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', r1, r1, opA ]),
            CmmBinAsInstruction([ 'sub', r1, opB, r1 ] if bit == 31 else [ 'sub', r1, r1, opB ]),
            CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
            CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
            CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
            ]

    def fix_64__g_l_ss_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        opA_size = opA.get_size()
        opB_0_size = opB_0.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneB_0 = CmmSR('one', CmmType('u', opB_0_size))

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var1, oneB_0, oneB_0 ]),
            CmmAssgnInstruction([ 'move', var0, var1 ]),
            CmmBinAsInstruction([ 'shli', var1, var1, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opA ]),
            CmmBinAsInstruction([ 'sub', var1, opB_0, var1 ]  if bit == 31 else [ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'subb', var1, opB_1, var0 ] if bit == 31 else [ 'subb', var1, var0, opB_1 ]),
            CmmBinAsInstruction([ 'addc', var0, oneB_0, oneB_0 ]),
            CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
            CmmGotoInstruction([ 'btj', var0, label, CmmConst(opB_0_size) - 1, bit_value ]),
            ]

    def fix_64__g_l_ss_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))
        oneA = CmmSR('one', CmmType('u', opA_size))

        r0 = var0.get_reg()
        r1 = var1.get_reg()

        r0.set_size(opA_size)
        r1.set_size(opA_size)

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', r1, oneA, oneA ]),
            CmmBinAsInstruction([ 'shli', r1, r1, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', r1, r1, opB ]),
            CmmBinAsInstruction([ 'sub', r1, r1, opA ] if bit == 31 else [ 'sub', r1, opA, r1 ]),
            CmmBinAsInstruction([ 'addc', r0, oneA, oneA ]),
            CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
            CmmGotoInstruction([ 'btj', r0, label, CmmConst(opA_size) - 1, bit_value ]),
            ]

    def fix_64__g_l_ss_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        opA_size = opA.get_size()
        opB_0_size = opB_0.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneB_0 = CmmSR('one', CmmType('u', opB_0_size))

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneB_0, oneB_0 ]),
            CmmBinAsInstruction([ 'sub', var1, opB_0, opA ]   if bit == 31 else [ 'sub', var1, opA, opB_0 ]),
            CmmBinAsInstruction([ 'subb', var1, opB_1, var0 ] if bit == 31 else [ 'subb', var1, var0, opB_1 ]),
            CmmBinAsInstruction([ 'addc', var0, oneB_0, oneB_0 ]),
            CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
            CmmGotoInstruction([ 'btj', var0, label, CmmConst(opB_0_size) - 1, bit_value ]),
            ]

    def fix_64__g_l_ss_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opB_size = opB.get_size()
        opA_0_size = opA_0.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))
        oneA_0 = CmmSR('one', CmmType('u', opA_0_size))

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var1, oneA_0, oneA_0 ]),
            CmmAssgnInstruction([ 'move', var0, var1 ]),
            CmmBinAsInstruction([ 'shli', var1, var1, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opB ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opA_0 ]  if bit == 31 else [ 'sub', var1, opA_0, var1 ]),
            CmmBinAsInstruction([ 'subb', var1, var0, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, var0 ]),
            CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
            CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
            CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
            ]

    def fix_64__g_l_ss_64_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opB_size = opB.get_size()
        opA_0_size = opA_0.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))
        oneA_0 = CmmSR('one', CmmType('u', opA_0_size))

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
            CmmBinAsInstruction([ 'sub', var1, opB, opA_0 ]   if bit == 31 else [ 'sub', var1, opA_0, opB ]),
            CmmBinAsInstruction([ 'subb', var1, var0, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, var0 ]),
            CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
            CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
            CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
            ]

    def fix_64__g_l_ss_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        opA_0_size = opA_0.get_size()
        oneA_0 = CmmSR('one', CmmType('u', opA_0_size))

        return [
            CmmBinAsInstruction([ 'sub', var1, opB_0, opA_0 ]  if bit == 31 else [ 'sub', var1, opA_0, opB_0 ]),
            CmmBinAsInstruction([ 'subb', var1, opB_1, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, opB_1 ]),
            CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
            CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
            CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
            ]

    def fix_64__g_l_su_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        oneB = CmmSR('one', CmmType('u', opB_size))

        r0 = var0.get_reg()
        r1 = var1.get_reg()

        r0.set_size(opA_size)
        r1.set_size(opB_size)

        if opB.is_a_constant() :
            if opB & (1 << opB_size) :
                if bit + bit_value == 31 :
                    return []
                else :
                    return [
                        CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                        ]
            else :
                if bit == 31 :
                    return [
                        CmmAssgnInstruction([ 'mvib', r1, opB ]),
                        CmmBinAsInstruction([ 'sub', r1, r1, opA ]),
                        CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
                        CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                        CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
                        ]
                else :
                    return [
                        CmmBinAsInstruction([ 'sbi', r1, opA, opB ]),
                        CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
                        CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                        CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
                        ]
        else :
            return [
                CmmGotoInstruction([ 'btjr', opB, CmmConst(4), CmmConst(opB_size) - 1, CmmConst(1) ] if bit + bit_value == 31\
                                else [ 'btj', opB, label, CmmConst(opB_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'sub', r1, opB, opA ] if bit == 31 else [ 'sub', r1, opA, opB ]),
                CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
                CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneB = CmmSR('one', CmmType('u', opB_size))

        r0 = var0.get_reg()
        r1 = var1.get_reg()

        r0.set_size(opB_size)
        r1.set_size(opB_size)

        if opB.is_a_constant() :
            if bit + bit_value == 31 :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            return [
                CmmGotoInstruction([ 'btjr', opB, CmmConst(8), CmmConst(opB_size) - 1, CmmConst(1) ] if bit + bit_value == 31\
                                else [ 'btj', opB, label, CmmConst(opB_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', r1, oneB, oneB ]),
                CmmBinAsInstruction([ 'shli', r1, r1, CmmConst(opA_size) ]),
                CmmBinAsInstruction([ 'or', r1, r1, opA ]),
                CmmBinAsInstruction([ 'sub', r1, opB, r1 ] if bit == 31 else [ 'sub', r1, r1, opB ]),
                CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
                CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        opA_size = opA.get_size()
        opB_0_size = opB_0.get_size()
        opB_1_size = opB_1.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneB_0 = CmmSR('one', CmmType('u', opB_0_size))

        if opB.is_a_constant() :
            if bit + bit_value == 31 :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            print ("reg = " + str(reg) + ", label = " + str(label) + ", bit = " + str(bit) + ", bit_value = " + str(bit_value) + "\n")
            return [
                CmmGotoInstruction([ 'btjr', opB_1, CmmConst(11), CmmConst(opB_1_size) - 1, CmmConst(1) ] if bit + bit_value == 31\
                                else [ 'btj', opB_1, label, CmmConst(opB_1_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var1, oneB_0, oneB_0 ]),
                CmmAssgnInstruction([ 'move', var0, var1 ]),
                CmmBinAsInstruction([ 'shli', var1, var1, CmmConst(opA_size) ]),
                CmmBinAsInstruction([ 'or', var1, var1, opA ]),
                CmmBinAsInstruction([ 'sub', var1, opB_0, var1 ]  if bit == 31 else [ 'sub', var1, var1, opB_0 ]),
                CmmBinAsInstruction([ 'subb', var1, opB_1, var0 ] if bit == 31 else [ 'subb', var1, var0, opB_1 ]),
                CmmBinAsInstruction([ 'addc', var0, oneB_0, oneB_0 ]),
                CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                CmmGotoInstruction([ 'btj', var0, label, CmmConst(opB_0_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))
        oneA = CmmSR('one', CmmType('u', opA_size))

        r0 = var0.get_reg()
        r1 = var1.get_reg()

        r0.set_size(opA_size)
        r1.set_size(opA_size)

        if opB.is_a_constant() :
            if bit == 31 :
                return [
                    CmmAssgnInstruction([ 'mviw', r1, opB ]),
                    CmmBinAsInstruction([ 'sub', r1, r1, opA ]),
                    CmmBinAsInstruction([ 'addc', r0, oneA, oneA ]),
                    CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                    CmmGotoInstruction([ 'btj', r0, label, CmmConst(opA_size) - 1, bit_value ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ 'sbi', r1, opA, opB ]),
                    CmmBinAsInstruction([ 'addc', r0, oneA, oneA ]),
                    CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                    CmmGotoInstruction([ 'btj', r0, label, CmmConst(opA_size) - 1, bit_value ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', r1, oneA, oneA ]),
                CmmBinAsInstruction([ 'shli', r1, r1, CmmConst(opB_size) ]),
                CmmBinAsInstruction([ 'or', r1, r1, opB ]),
                CmmBinAsInstruction([ 'sub', r1, r1, opA ] if bit == 31 else [ 'sub', r1, opA, r1 ]),
                CmmBinAsInstruction([ 'addc', r0, oneA, oneA ]),
                CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                CmmGotoInstruction([ 'btj', r0, label, CmmConst(opA_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_16_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        oneB = CmmSR('one', CmmType('u', opB_size))

        r0 = var0.get_reg()
        r1 = var1.get_reg()

        r0.set_size(opA_size)
        r1.set_size(opB_size)

        if opB.is_a_constant() :
            if opB & (1 << opB_size) :
                if bit + bit_value == 31 :
                    return []
                else :
                    return [
                        CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                        ]
            else :
                return [
                    CmmAssgnInstruction([ 'mviw', r1, opB ]),
                    CmmBinAsInstruction([ 'sub', r1, r1, opA ] if bit == 31 else [ 'sub', r1, opA, r1 ]),
                    CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
                    CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                    CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
                    ]
        else :
            return [
                CmmGotoInstruction([ 'btjr', opB, CmmConst(4), CmmConst(opB_size) - 1, CmmConst(1) ] if bit + bit_value == 31\
                                else [ 'btj', opB, label, CmmConst(opB_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'sub', r1, opB, opA ] if bit == 31 else [ 'sub', r1, opA, opB ]),
                CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
                CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_32_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))
        oneA = CmmSR('one', CmmType('u', opA_size))

        r0 = var0.get_reg()
        r1 = var1.get_reg()

        r0.set_size(opA_size)
        r1.set_size(opA_size)

        if opB.is_a_constant() :
            return [
                CmmAssgnInstruction([ 'mviw', r1, opB ]),
                CmmBinAsInstruction([ 'sub', r1, r1, opA ] if bit == 31 else [ 'sub', r1, opA, r1 ]),
                CmmBinAsInstruction([ 'addc', r0, oneA, oneA ]),
                CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                CmmGotoInstruction([ 'btj', r0, label, CmmConst(opA_size) - 1, bit_value ]),
                ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', r1, oneA, oneA ]),
                CmmBinAsInstruction([ 'shli', r1, r1, CmmConst(opB_size) ]),
                CmmBinAsInstruction([ 'or', r1, r1, opB ]),
                CmmBinAsInstruction([ 'sub', r1, r1, opA ] if bit == 31 else [ 'sub', r1, opA, r1 ]),
                CmmBinAsInstruction([ 'addc', r0, oneA, oneA ]),
                CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                CmmGotoInstruction([ 'btj', r0, label, CmmConst(opA_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_32_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        oneB = CmmSR('one', CmmType('u', opB_size))

        r0 = var0.get_reg()
        r1 = var1.get_reg()

        r0.set_size(opA_size)
        r1.set_size(opB_size)

        if opB.is_a_constant() :
            if opB & (1 << opB_size) :
                if bit + bit_value == 31 :
                    return []
                else :
                    return [
                        CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                        ]
            else :
                [r11, r10] = r1.splt_into_2()
                opB_0 = opB & 0xffff
                opB_1 = opB >> 16

                return [
                    CmmAssgnInstruction([ 'mviw', r10, opB_0 ]),
                    CmmAssgnInstruction([ 'mviw', r11, opB_1 ]),
                    CmmBinAsInstruction([ 'sub', r1, r1, opA ] if bit == 31 else [ 'sub', r1, opA, r1 ]),
                    CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
                    CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                    CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
                    ]
        else :
            return [
                CmmGotoInstruction([ 'btjr', opB, CmmConst(4), CmmConst(opB_size) - 1, CmmConst(1) ] if bit + bit_value == 31\
                                else [ 'btj', opB, label, CmmConst(opB_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'sub', r1, opB, opA ] if bit == 31 else [ 'sub', r1, opA, opB ]),
                CmmBinAsInstruction([ 'addc', r0, oneB, oneB ]),
                CmmBinAsInstruction([ 'xor', r0, r0, r1 ]),
                CmmGotoInstruction([ 'btj', r0, label, CmmConst(opB_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        opA_size = opA.get_size()
        opB_0_size = opB_0.get_size()
        opB_1_size = opB_1.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneB_0 = CmmSR('one', CmmType('u', opB_0_size))

        if opB.is_a_constant() :
            if bit + bit_value == 31 :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            return [
                CmmGotoInstruction([ 'btjr', opB_1, CmmConst(10), CmmConst(opB_1_size) - 1, CmmConst(1) ] if bit + bit_value == 31\
                                else [ 'btj', opB_1, label, CmmConst(opB_1_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var0, oneB_0, oneB_0 ]),
                CmmBinAsInstruction([ 'sub', var1, opB_0, opA ]   if bit == 31 else [ 'sub', var1, opA, opB_0 ]),
                CmmBinAsInstruction([ 'subb', var1, opB_1, var0 ] if bit == 31 else [ 'subb', var1, var0, opB_1 ]),
                CmmBinAsInstruction([ 'addc', var0, oneB_0, oneB_0]),
                CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                CmmGotoInstruction([ 'btj', var0, label, CmmConst(opB_0_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opB_size = opB.get_size()
        opA_0_size = opA_0.get_size()
        opA_1_size = opA_1.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))
        oneA_0 = CmmSR('one', CmmType('u', opA_0_size))

        if opB.is_a_constant() :
            if bit == 31 :
                return [
                    CmmAssgnInstruction([ 'mviw', var1, opB ]),
                    CmmBinAsInstruction([ 'sub', var1, var1, opA_0 ]),
                    CmmBinAsInstruction([ 'subb', var1, zeroA_1, opA_1 ]),
                    CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                    CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                    CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ 'sbi', var1, opA_0, opB ]),
                    CmmBinAsInstruction([ 'subb', var1, opA_1, zeroA_1 ]),
                    CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                    CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                    CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', var1, oneA_0, oneA_0 ]),
                CmmAssgnInstruction([ 'move', var0, var1 ]),
                CmmBinAsInstruction([ 'shli', var1, var1, CmmConst(opB_size) ]),
                CmmBinAsInstruction([ 'or', var1, var1, opB ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opA_0 ]  if bit == 31 else [ 'sub', var1, opA_0, var1 ]),
                CmmBinAsInstruction([ 'subb', var1, var0, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, var0 ]),
                CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_64_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opB_size = opB.get_size()
        opA_0_size = opA_0.get_size()
        opA_1_size = opA_1.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))
        oneA_0 = CmmSR('one', CmmType('u', opA_0_size))

        if opB.is_a_constant() :
            return [
                CmmAssgnInstruction([ 'mviw', var1, opB ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opA_0 ]     if bit == 31 else [ 'sub', var1, opA_0, var1 ]),
                CmmBinAsInstruction([ 'subb', var1, zeroA_1, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, zeroA_1 ]),
                CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
                ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', var1, oneA_0, oneA_0 ]),
                CmmAssgnInstruction([ 'move', var0, var1 ]),
                CmmBinAsInstruction([ 'shli', var1, var1, CmmConst(opB_size) ]),
                CmmBinAsInstruction([ 'or', var1, var1, opB ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opA_0 ]  if bit == 31 else [ 'sub', var1, opA_0, var1 ]),
                CmmBinAsInstruction([ 'subb', var1, var0, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, var0 ]),
                CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_64_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opB_size = opB.get_size()
        opA_0_size = opA_0.get_size()
        opA_1_size = opA_1.get_size()
        zeroB = CmmSR('zero', CmmType('u', opB_size))
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))
        oneA_0 = CmmSR('one', CmmType('u', opA_0_size))

        if opB.is_a_constant() :
            [r11, r10] = var1.get_reg().splt_into_2()
            opB_0 = opB & 0xffff
            opB_1 = opB >> 16

            return [
                CmmAssgnInstruction([ 'mviw', r10, opB_0 ]),
                CmmAssgnInstruction([ 'mviw', r11, opB_1 ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opA_0 ]     if bit == 31 else [ 'sub', var1, opA_0, var1 ]),
                CmmBinAsInstruction([ 'subb', var1, zeroA_1, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, zeroA_1 ]),
                CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
                ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                CmmBinAsInstruction([ 'sub', var1, opB, opA_0 ]   if bit == 31 else [ 'sub', var1, opA_0, opB ]),
                CmmBinAsInstruction([ 'subb', var1, var0, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, var0 ]),
                CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_su_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        opA_0_size = opA_0.get_size()
        opB_1_size = opB_1.get_size()
        oneA_0 = CmmSR('one', CmmType('u', opA_0_size))

        if opB.is_a_constant() :
            opB_0 = opB & 0xffffffff
            opB_1 = opB >> 32
            opB_1_size = 32

            if opB_1 & (1 << (opB_1_size - 1)) :
                if bit + bit_value == 31 :
                    return []
                else :
                    return [
                        CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                        ]
            else :
                [r01, r00] = var0.get_reg().splt_into_2()
                [r11, r10] = var1.get_reg().splt_into_2()
                opB_00 = opB_0 & 0xffff
                opB_01 = opB_0 >> 16
                opB_10 = opB_1 & 0xffff
                opB_11 = opB_1 >> 16

                return [
                    CmmAssgnInstruction([ 'mviw', r00, opB_00 ]),
                    CmmAssgnInstruction([ 'mviw', r01, opB_01 ]),
                    CmmAssgnInstruction([ 'mviw', r10, opB_10 ]),
                    CmmAssgnInstruction([ 'mviw', r11, opB_11 ]),
                    CmmBinAsInstruction([ 'sub', var0, var0, opA_0 ]  if bit == 31 else [ 'sub', var0, opA_0, var0 ]),
                    CmmBinAsInstruction([ 'subb', var1, var1, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, var1 ]),
                    CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                    CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                    CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
                    ]
        else :
            return [
                CmmGotoInstruction([ 'btjr', opB_1, CmmConst(4), CmmConst(opB_1_size) - 1, CmmConst(1) ] if bit + bit_value == 31\
                                else [ 'btj', opB_1, label, CmmConst(opB_1_size) - 1, CmmConst(1) ]),
                CmmBinAsInstruction([ 'sub', var1, opB_0, opA_0 ]  if bit == 31 else [ 'sub', var1, opA_0, opB_0 ]),
                CmmBinAsInstruction([ 'subb', var1, opB_1, opA_1 ] if bit == 31 else [ 'subb', var1, opA_1, opB_1 ]),
                CmmBinAsInstruction([ 'addc', var0, oneA_0, oneA_0 ]),
                CmmBinAsInstruction([ 'xor', var0, var0, var1 ]),
                CmmGotoInstruction([ 'btj', var0, label, CmmConst(opA_0_size) - 1, bit_value ]),
                ]

    def fix_64__g_l_uu_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        return [
            self.cwl(),
            CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
            ]

    def fix_64__g_l_uu_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opB_size = opB.get_size()

        if opB.is_a_constant() :
            if bit + bit_value == 31 :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'or', var1, CmmSR('zero', CmmType('u', opB_size)), opA ]),
                CmmCmpInstruction([ 'cmp', var1, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__g_l_uu_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()
        opB_1_size = opB_1.get_size()

        if opB.is_a_constant() :
            if bit + bit_value == 31 :
                return []
            else :
                return [
                    CmmGotoInstruction([ 'btj', CmmSR('zero'), label, CmmConst(0), CmmConst(0) ]),
                    ]
        else :
            G = CmmConst(31)
            L = CmmConst(30)

            if bit == L and bit_value == 0 :
                return [
                    CmmCmpInstruction([ 'cmp', opB_1, CmmSR('zero', CmmType('u', opB_1_size)) ]),
                    CmmGotoInstruction([ 'btjr', reg, CmmConst(3), G, CmmConst(1) ]),
                    CmmCmpInstruction([ 'cmp', opB_0, opA ]),
                    CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(0) ]),
                    ]
            elif bit == L and bit_value == 1 :
                return [
                    CmmCmpInstruction([ 'cmp', opB_1, CmmSR('zero', CmmType('u', opB_1_size)) ]),
                    CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                    CmmCmpInstruction([ 'cmp', opB_0, opA ]),
                    CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                    ]
            elif bit == G and bit_value == 0 :
                return [
                    CmmCmpInstruction([ 'cmp', opB_1, CmmSR('zero', CmmType('u', opB_1_size)) ]),
                    CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                    CmmCmpInstruction([ 'cmp', opB_0, opA ]),
                    CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(0) ]),
                    ]
            elif bit == G and bit_value == 1 :
                return [
                    CmmCmpInstruction([ 'cmp', opB_1, CmmSR('zero', CmmType('u', opB_1_size)) ]),
                    CmmGotoInstruction([ 'btjr', reg, CmmConst(3), G, CmmConst(1) ]),
                    CmmCmpInstruction([ 'cmp', opB_0, opA ]),
                    CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(1) ]),
                    ]

    def fix_64__g_l_uu_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        r1 = var1.get_reg()

        r1.set_size(opA_size)

        if opB.is_a_constant() :
            return [
                self.cwl(),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]
        else :
            return [CmmBinAsInstruction([ 'or', var1, zeroA, opB ])] if opB.get_reg().is_a_rf_reg() else []\
                + [
                CmmCmpInstruction([ 'cmp', opA, var1 ]),
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__g_l_uu_32_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()

        if opB.is_a_constant() :
            r1 = var1.get_reg()

            [r_11, r_10] = r1.splt_into_2()

            return [
                CmmAssgnInstruction([ 'mviw', r_10, opB & 0xffff]) ,
                CmmAssgnInstruction([ 'mviw', r_11, opB >> 16]) ,
                CmmCmpInstruction([ 'cmp', opA, r1 ]) ,
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]
        else :
            return [
                self.cwl() ,
                CmmGotoInstruction([ 'btj', reg, label, bit, bit_value ]),
                ]

    def fix_64__g_l_uu_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opA_1_size = opA_1.get_size()

        G = CmmConst(31)
        L = CmmConst(30)
        comp = 'cpiw' if opB.is_a_constant() else 'cmp'

        if bit == L and bit_value == 0 :
            return [
                CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', opA_1_size)) ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                CmmCmpInstruction([ comp, opA_0, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(0) ]),
                ]
        elif bit == L and bit_value == 1 :
            return [
                CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', opA_1_size)) ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), G, CmmConst(1) ]),
                CmmCmpInstruction([ comp, opA_0, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(1) ]),
                ]
        elif bit == G and bit_value == 0 :
            return [
                CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', opA_1_size)) ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), G, CmmConst(1) ]),
                CmmCmpInstruction([ comp, opA_0, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(0) ]),
                ]
        elif bit == G and bit_value == 1 :
            return [
                CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', opA_1_size)) ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                CmmCmpInstruction([ comp, opA_0, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                ]

    def fix_64__g_l_uu_64_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        opA_1_size = opA_1.get_size()

        G = CmmConst(31)
        L = CmmConst(30)

        instr = []

        if opB.is_a_constant() :
            r1 = var1.get_reg()
            [r11, r10] = r1.splt_into_2()
            opB_0 = opB & 0xffff
            opB_1 = opB >> 16

            instr = [
                CmmAssgnInstruction([ 'mviw', r10, opB_0 ]),
                CmmAssgnInstruction([ 'mviw', r11, opB_1 ]),
                ]

            opB = r1

        if bit == L and bit_value == 0 :
            return instr + [
                CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', opA_1_size)) ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA_0, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(0) ]),
                ]
        elif bit == L and bit_value == 1 :
            return instr + [
                CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', opA_1_size)) ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), G, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA_0, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(1) ]),
                ]
        elif bit == G and bit_value == 0 :
            return instr + [
                CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', opA_1_size)) ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), G, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA_0, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(0) ]),
                ]
        elif bit == G and bit_value == 1 :
            return instr + [
                CmmCmpInstruction([ 'cmp', opA_1, CmmSR('zero', CmmType('u', opA_1_size)) ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA_0, opB ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                ]

    def fix_64__g_l_uu_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        [reg, label, bit, bit_value] = self.get_jump_data()

        opA = self.get_opA()
        opB = self.get_opB()
        [opA_1, opA_0] = opA.get_reg().splt_into_2()
        [opB_1, opB_0] = opB.get_reg().splt_into_2()

        G = CmmConst(31)
        L = CmmConst(30)

        instr = []

        if opB.is_a_constant() :
            r0 = var0.get_reg()
            r1 = var1.get_reg()
            [r01, r00] = r0.splt_into_2()
            [r11, r10] = r1.splt_into_2()
            opB_0 = opB & 0xffffffff
            opB_1 = opB >> 32
            opB_00 = opB_0 & 0xffff
            opB_01 = opB_0 >> 16
            opB_10 = opB_1 & 0xffff
            opB_11 = opB_1 >> 16

            instr = [
                CmmAssgnInstruction([ 'mviw', r00, opB_00 ]),
                CmmAssgnInstruction([ 'mviw', r01, opB_01 ]),
                CmmAssgnInstruction([ 'mviw', r10, opB_10 ]),
                CmmAssgnInstruction([ 'mviw', r11, opB_11 ]),
                ]

            opB_0 = r0
            opB_1 = r1

        if bit == L and bit_value == 0 :
            return instr + [
                CmmCmpInstruction([ 'cmp', opA_1, opB_1 ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), L, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA_0, opB_0 ]),
                CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(0) ]),
                ]
        elif bit == L and bit_value == 1 :
            return instr + [
                CmmCmpInstruction([ 'cmp', opA_1, opB_1 ]),
                CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(1) ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), G, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA_0, opB_0 ]),
                CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(1) ]),
                ]
        elif bit == G and bit_value == 0 :
            return instr + [
                CmmCmpInstruction([ 'cmp', opA_1, opB_1 ]),
                CmmGotoInstruction([ 'btj', reg, label, L, CmmConst(1) ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), G, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA_0, opB_0 ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(0) ]),
                ]
        elif bit == G and bit_value == 1 :
            return instr + [
                CmmCmpInstruction([ 'cmp', opA_1, opB_1 ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                CmmGotoInstruction([ 'btjr', reg, CmmConst(3), L, CmmConst(1) ]),
                CmmCmpInstruction([ 'cmp', opA_0, opB_0 ]),
                CmmGotoInstruction([ 'btj', reg, label, G, CmmConst(1) ]),
                ]

    fix_64 = {
             'eq_ne' : {
                       'ss' : {
                              8  : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_ss_8_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_ss_8_16,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_ss_8_16,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_ss_8_64,
                              },
                              16 : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_ss_16_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_ss_8_8,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_ss_8_16,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_ss_8_64,
                              },
                              32 : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_ss_16_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_ss_16_8,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_ss_8_8,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_ss_32_64,
                              },
                              64 : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_ss_64_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_ss_64_8,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_ss_64_32,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_ss_64_64,
                              },
                       'su' : {
                              8  : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_su_8_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_su_8_16,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_su_8_32,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_su_8_64,
                              },
                              16 : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_su_16_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_su_16_16,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_su_8_32,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_su_8_64,
                              },
                              32 : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_su_16_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_su_16_8,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_su_32_32,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_su_32_64,
                              },
                              64 : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_su_64_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_su_64_16,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_su_64_32,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_su_64_64,
                              },
                       },
                       'uu' : {
                              8  : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_uu_8_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_uu_8_16,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_uu_8_32,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_uu_8_64,
                              },
                              16 : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_uu_16_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_uu_16_16,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_uu_8_32,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_uu_8_64,
                              },
                              32 : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_uu_16_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_uu_16_8,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_uu_16_16,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_uu_8_64,
                              },
                              64 : {
                                   8   : CmmCmpInstruction.fix_64__eq_ne_uu_64_8,
                                   16  : CmmCmpInstruction.fix_64__eq_ne_uu_64_8,
                                   32  : CmmCmpInstruction.fix_64__eq_ne_uu_64_32,
                                   64  : CmmCmpInstruction.fix_64__eq_ne_uu_64_64,
                              },
                       },
             'g_l' : {
                       'ss' : {
                              8  : {
                                   8   : CmmCmpInstruction.fix_64__g_l_ss_8_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_ss_8_16,
                                   32  : CmmCmpInstruction.fix_64__g_l_ss_8_16,
                                   64  : CmmCmpInstruction.fix_64__g_l_ss_8_64,
                              },
                              16 : {
                                   8   : CmmCmpInstruction.fix_64__g_l_ss_16_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_ss_8_8,
                                   32  : CmmCmpInstruction.fix_64__g_l_ss_8_16,
                                   64  : CmmCmpInstruction.fix_64__g_l_ss_8_64,
                              },
                              32 : {
                                   8   : CmmCmpInstruction.fix_64__g_l_ss_16_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_ss_16_8,
                                   32  : CmmCmpInstruction.fix_64__g_l_ss_8_8,
                                   64  : CmmCmpInstruction.fix_64__g_l_ss_32_64,
                              },
                              64 : {
                                   8   : CmmCmpInstruction.fix_64__g_l_ss_64_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_ss_64_8,
                                   32  : CmmCmpInstruction.fix_64__g_l_ss_64_32,
                                   64  : CmmCmpInstruction.fix_64__g_l_ss_64_64,
                              },
                       'su' : {
                              8  : {
                                   8   : CmmCmpInstruction.fix_64__g_l_su_8_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_su_8_16,
                                   32  : CmmCmpInstruction.fix_64__g_l_su_8_16,
                                   64  : CmmCmpInstruction.fix_64__g_l_su_8_64,
                              },
                              16 : {
                                   8   : CmmCmpInstruction.fix_64__g_l_su_16_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_su_16_16,
                                   32  : CmmCmpInstruction.fix_64__g_l_su_8_16,
                                   64  : CmmCmpInstruction.fix_64__g_l_su_8_64,
                              },
                              32 : {
                                   8   : CmmCmpInstruction.fix_64__g_l_su_16_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_su_32_16,
                                   32  : CmmCmpInstruction.fix_64__g_l_su_32_32,
                                   64  : CmmCmpInstruction.fix_64__g_l_su_32_64,
                              },
                              64 : {
                                   8   : CmmCmpInstruction.fix_64__g_l_su_64_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_su_64_16,
                                   32  : CmmCmpInstruction.fix_64__g_l_su_64_32,
                                   64  : CmmCmpInstruction.fix_64__g_l_su_64_64,
                              },
                       },
                       'uu' : {
                              8  : {
                                   8   : CmmCmpInstruction.fix_64__g_l_uu_8_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_uu_8_16,
                                   32  : CmmCmpInstruction.fix_64__g_l_uu_8_16,
                                   64  : CmmCmpInstruction.fix_64__g_l_uu_8_64,
                              },
                              16 : {
                                   8   : CmmCmpInstruction.fix_64__g_l_uu_16_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_uu_8_8,
                                   32  : CmmCmpInstruction.fix_64__g_l_uu_8_16,
                                   64  : CmmCmpInstruction.fix_64__g_l_uu_8_64,
                              },
                              32 : {
                                   8   : CmmCmpInstruction.fix_64__g_l_uu_16_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_uu_16_8,
                                   32  : CmmCmpInstruction.fix_64__g_l_uu_32_32,
                                   64  : CmmCmpInstruction.fix_64__g_l_uu_8_64,
                              },
                              64 : {
                                   8   : CmmCmpInstruction.fix_64__g_l_uu_64_8,
                                   16  : CmmCmpInstruction.fix_64__g_l_uu_64_8,
                                   32  : CmmCmpInstruction.fix_64__g_l_uu_64_32,
                                   64  : CmmCmpInstruction.fix_64__g_l_uu_64_64,
                              },
                       },
             }

    def fix_64_bit_arithmetic(self) :
        op = self.get_op()
        op = 'eq_ne' if op == '==' or op == '!=' else 'g_l'

        opA = self.get_opA()
        opB = self.get_opB()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        opA_size = self.get_normalized_size(opA_size)
        opB_size = self.get_normalized_size(opB_size)

        if (opA.is_signed() and opB.is_signed()) :
            instr = CmmCmpInstruction.fix_64[op]['ss'][opA_size][opB_size](self)
        elif (opA.is_signed() and opB.is_unsigned()) :
            instr = CmmCmpInstruction.fix_64[op]['su'][opA_size][opB_size](self)
        else :
            instr = CmmCmpInstruction.fix_64[op]['uu'][opA_size][opB_size](self)

        if (len(instr)) :
            # CmmInstruction->transfer_labels($self, $instr->[0]) ;

            next_goto = instr.pop()
            self.get_next().set_body(next_goto.get_body())
        else :
            self.transform('CmmPHInstruction', [])
            self.get_next().transform('CmmPHInstruction', [])

        self.set_final_asm_form(instr)

    def _cmp_fix_size(self) :
        opB = self.body[2]
        opB_size = opB.get_size()

        self.body += [self.get_asm_operand_size(opB_size, opB_size, '_')]

    def _cpiw_fix_size(self) :
        opA = self.body[1]
        opA_size = opA.get_size()

        if opA_size > 8 :
            print ("_cpiw_fix_size " + str(opA) + " , " + str(opA_size) + "\n")
            self.body += [self.get_asm_operand_size(opA_size, '_', '_')]
        else :
            self.set_body([ 'cpib', opA, CmmConst(0xff), self.get_opB() ])

    def _cpib_fix_size(self) :
        pass

    def fix_size_operand(self) :
        instr = []

        instr = self.get_final_asm_form()

        if instr :
            for lv in instr :
                return lv.fix_size_operand()
        else :
            if self.get_opB().is_a_constant() :
                self.set_op('cpiw')
                return self._cpiw_fix_size()
            else :
                self.set_op('cmp')
                return self._cmp_fix_size()

    def is_a_zero_comparison(self) :
        op = self.get_op()
        opB = self.get_opB()

        return (op == '==' or op == '!=') and opB.is_a_constant() and opB == 0

    def _get_flags_set(self) :
        return ['G', 'L', 'E']

    def evaluate(self) :
        print ("evaluate " + str(self) + " " + str(self.body) + "\n")

        [fst, sec] = [ x.get_value() if not x.is_a_constant() else x for x in self.get_operands() ]

        if fst.is_ordinary_constant() and sec.is_ordinary_constant() :
            print ("fst_size is " + str(fst.get_size()) + "\n")
            print ("sec_size is " + str(sec.get_size()) + "\n")
            print ("fst " + str(fst) + "\n")
            print ("sec " + str(sec) + "\n")
            if fst > sec :
                print "greater\n"
                self._set_G()
                self._reset_L()
                self._reset_E()
            elif fst < sec :
                print "less\n"
                self._reset_G()
                self._set_L()
                self._reset_E()
            else :
                print "equal\n"
                self._reset_G()
                self._reset_L()
                self._set_E()
        elif fst.is_max() or sec.is_max() :
            for lv in self.get_left_side_effect() :
                lv.set_value(CmmValueLattice('multiple_data'))

    def get_asm_code(self) :
        instr = self.get_final_asm_form()

        if instr :
            return [ y for x in instr for y in x.get_asm_code() ]
        else :
            code = ''

            body = [ x.get_reg() if hasattr(x, 'is_a_var') and x.is_a_var() else x for x in self.body ]

            code = "\t" + str(body.pop(0)) + "\t"

            if self.get_op() == 'cpib' :
                mask = self.get_opB()
                mask = "0x%02x" % mask

                code += body.pop(0) + ",\t" + str(mask) + ",\t" + str(pop(body))
            else :
                code += ",\t".join(body)

            return [code]
