import re
import copy
from CmmType import CmmType
from CmmSymbol import CmmSymbol

class CmmConst(CmmSymbol) :
    mask_63_56 = 0xff00000000000000
    mask_55_48 = 0x00ff000000000000
    mask_47_40 = 0x0000ff0000000000
    mask_39_32 = 0x000000ff00000000

    def __init__(self, value) :
        size = 0
        t = ''

        if type(value) == str :
            suffix_size = 0

            if re.match(value, "[Uu][Ll]$") or re.match(value, "[Ll][Uu]$"):
                suffix_size = 2
                t = 'u'
                size = 64
            elif re.match(value, "[Uu]$"):
                suffix_size = 1
                t = 'u'
            elif re.match(value, "[Ll]$"):
                suffix_size = 1
                size = 64

            print("V0 = " + value)
            if suffix_size > 0 :
                value = value[0:-suffix_size]
            print("V1 = " + value)

            if re.match(r'^0[Xx]', value):
                value = int(value, 16)
            elif re.match(r'^0', value):
                value = int(value, 8)
            else :
                value = int(value)

        if size == 0 :
           if value > 0xffffffff :
               size = 64
           elif value > 0xffff :
               size = 32
           elif value > 0xff :
               size = 16
           else :
               size = 8

        # bug: type may be signed too
        t = 'u'

        self.set_value(value)
        # printf("const = 0x%08x%08x, CmmConst size = %d\n", $value >> 32, $value & 0xffffffff, $size) ;
        self.set_type(CmmType(t, size))

    def new(self) :
        c = copy.deepcopy(self)
        return c

    def is_ordinary_constant(self) :
        return 1

    def is_a_constant(self) :
        return 1

    def is_max(self) :
        return 0

    def is_min(self) :
        return 0

    def splt_into_2(self) :
        value = self.get_value()
        size = self.get_size()
        mask = (CmmConst(1) << (size / 2)) - 1

        high = (value >> (size / 2)) & mask
        low = value & mask

        return [CmmConst(high) , CmmConst(low)]

    def __str__(self):
        return str(self.get_value())

    def __cmp__(self, other):
        if not isinstance(other, CmmConst):
            other = CmmConst(other)

        return self.get_value() - other.get_value()

    def __neg__(self):
        return CmmConst(-self.get_value())

    def __invert__(self):
        return CmmConst(~self.get_value())

    def __and__(self, other):
        return CmmConst(self.get_value() & (other.get_value() if isinstance(other, CmmConst) else other))

    def __or__(self, other):
        return CmmConst(self.get_value() | (other.get_value() if isinstance(other, CmmConst) else other))

    def __xor__(self, other):
        return CmmConst(self.get_value() ^ (other.get_value() if isinstance(other, CmmConst) else other))

    def __shl__(self, other):
        return CmmConst(self.get_value() << (other.get_value() if isinstance(other, CmmConst) else other))

    def __shr__(self, other):
        return CmmConst(self.get_value() >> (other.get_value() if isinstance(other, CmmConst) else other))

    def __mod__(self, other):
        return CmmConst(self.get_value() % (other.get_value() if isinstance(other, CmmConst) else other))

    def __add__(self, other):
        return CmmConst(self.get_value() + (other.get_value() if isinstance(other, CmmConst) else other))

    def __sub__(self, other):
        return CmmConst(self.get_value() - (other.get_value() if isinstance(other, CmmConst) else other))

    def __rand__(self, other):
        return CmmConst.__and__(CmmConst(other), self)

    def __ror__(self, other):
        return CmmConst.__or__(CmmConst(other), self)

    def __rxor__(self, other):
        return CmmConst.__xor__(CmmConst(other), self)

    def __rshl__(self, other):
        return CmmConst.__shl__(CmmConst(other), self)

    def __rshr__(self, other):
        return CmmConst.__shr__(CmmConst(other), self)

    def __rmod__(self, other):
        return CmmConst.__mod__(CmmConst(other), self)

    def __radd__(self, other):
        return CmmConst.__add__(CmmConst(other), self)

    def __rsub__(self, other):
        return CmmConst.__sub__(CmmConst(other), self)

    def __nonzero__(self):
        return self.get_value() != 0

    def __int__(self):
        return self.get_value()

    def __long__(self):
        return self.get_value()
