from CmmInstruction import CmmInstruction
from CmmValueLattice import CmmValueLattice

class CmmGotoInstruction(CmmInstruction) :
    def transform_to_pseudo_asm_form(self) :
        # my $self = shift ;

        # $self->{body} = [ 'mviw', CmmSR->new('pc'), @{$self->get_body()} ] ;
        pass

    def invert_comparison(self) :
        self.body[-1] = 1 - self.body[-1]

    def mirror_cond_in_jump(self) :
        if (self.body[3] == 30 or self.body[3] == 31) :
            self.body[3] = 61 - self.body[3]

    def set_bit(self, bit) :
        self.body[-2] = bit

    def get_bit(self) :
        return self.body[-2]

    def set_reg(self, reg) :
        self.body[1] = reg

    def get_reg(self) :
        return self.body[1]

    def set_instr_label(self, label) :
        self.body[2] = label

    def get_instr_label(self) :
        return self.body[2]

    def is_btj(self) :
        return self.body[0] == 'btj'

    def is_btjr(self) :
        return self.body[0] == 'btjr'

    def is_jsr(self) :
        return self.body[0] == 'jsr'

    def is_goto(self) :
        return 1

    def is_a_conditional_jump(self) :
        return not self.is_jsr()

    def compute_left(self) :
        if (self.body[0] == 'jsr') :
            return [self.body[1]]
        else :
            return []

    def _get_flags_used(self) :
        flag = ('C', 'E', 'L', 'G')
        index = None
        result = None

        if (self.is_btj() or self.is_btjr()) :
            index = self.body[3]

            if (self.body[1] == 'pc' and 28 <= index and index <= 31) :
                result = flag[index - 28]
        elif (self.is_jsr()) :
            index = self.body[3]

            if (4 <= index and index <= 7) :
                result = flag[index - 4]

        return ([result]) if result else ([])

    def evaluate(self) :
        print ("evaluate " + str(self) + " " + str(self.body) + "\n")

        op = self.body[0]
        (reg, addr, bit, v) = self.get_operands()
        flag = None
        reg_value = None

        if (op == 'jsr') :
            if (bit <= 3) :
                self.set_value(CmmValueLattice(1))
            else :
                (flag) = self.get_right_side_effect()

                if (flag.is_ordinary_constant()) :
                    self.set_value(CmmValueLattice((1) if (flag == v) else (0)))
                elif (flag.is_max()) :
                    self.set_value(CmmValueLattice('multiple_data'))
        else : # btj , btjr
            (flag) = self.get_right_side_effect()

            if (flag) :
                flag = flag.get_value()

                if (flag.is_ordinary_constant()) :
                    print ("flag = " + str(flag) + "\n")
                    print ("v = " + str(v) + "\n")
                    print (op + " gives ") + str((1) if (flag == v) else (0)) + "\n"
                    self.set_value(CmmValueLattice((1) if (flag == v) else (0)))
                elif (flag.is_max()) :
                    self.set_value(CmmValueLattice('multiple_data'))
            else :
                reg_value = reg.get_value()
                # bug : can be done better if we propagate some info about reg range

                if (reg_value.is_ordinary_constant()) :
                    self.set_value(CmmValueLattice((1) if ((reg_value & (1 << addr)) == v) else (0)))
                elif (reg_value.is_max()) :
                    self.set_value(CmmValueLattice('multiple_data'))

    def customize(self) :
        return [ self ]

    def get_asm_code(self) :
        print ("BTJ -1 " + str(self.body) + "\n")
        if (self.is_btj() or self.is_btjr()) :
            print ("BTJ 0 " + str((self.body[1].get_reg()) if hasattr(self.body[1], 'is_a_var') and self.body[1].is_a_var() else (self.body[1])) + " " + str(self.body) + "\n")
            if (not (type(self.body[1]) is int or type(self.body[1]) is str)) :
                print ("BTJ 1 " + str(self.body) + "\n")
                reg = (self.body[1].get_reg()) if self.body[1].is_a_var() else (self.body[1])

                bit = self.get_bit()
                if (reg.is_a_rf_reg() and bit > 7) :
                    print ("BTJ 2 " + str(self.body) + "\n")
                    new_bit = bit % 8
                    reg = reg.splt()

                    self.set_reg(reg[- ((bit >> 3) + 1)])
                    self.set_bit(new_bit)

        body = [ (x.get_reg()) if (hasattr(x, 'is_a_var') and x.is_a_var()) else (x)  for x in self.body]

        code = "\t" + str(body.pop(0)) + "\t"

        return code + ",\t".join([ str(x) for x in body])
