from CmmInstruction import CmmInstruction
from CmmValueLattice import CmmValueLattice

class CmmOrigAsmInstruction(CmmInstruction) :
    def is_orig_asm_instruction(self) :
        return 1

    def is_a_conditional_jump(self) :
        op = self.body[0]

        return op == 'btj' or op == 'btjr' or op == 'jsr'

    def _stld_fix_size(self) :
        mp = self.body[4]

        self.body[4] = mp.get_index()

    def _flo_fix_size(self) :
        opA = self.body[2]
        opA_size = opA.get_size()

        self.body += [self.get_asm_operand_size(opA_size, opA_size, '_')]

    def fix_size_operand(self) :
        asm_instr = self.body[0]

        if (asm_instr == 'stld') :
            return self._stld_fix_size()
        elif (asm_instr == 'flo') :
            return self._flo_fix_size()

    def compute_left(self) :
        op = self.body[0]

        instr = [
        'stld', 'flo',
        'jsr',
        'load', 'move', 'mviw' 'mvib',
        'add', 'addm', 'addc' 'addmc',
        'sub', 'subm', 'subb' 'submb',
        'or', 'orm',
        'and', 'andm',
        'xor', 'xorm',
        'shl', 'shlm',
        'shr', 'shrm',
        'adi', 'adim',
        'sbi', 'sbim',
        'ori', 'orim',
        'andi', 'andim',
        'xori', 'xorim',
        'shli', 'shlim',
        'shri', 'shrim',
        'bitc', 'bitcm',
        'bitci', 'bitcim']

        if op in instr:
            return [self.body[1]]
        else :
            return []

    def _get_flags_set(self) :
        op = self.body[0]
        result = []

        instr0 = [
        'flo',
        'or', 'orm',
        'and', 'andm',
        'xor', 'xorm',
        'shl', 'shlm',
        'shr', 'shrm',
        'ori', 'orim',
        'andi', 'andim',
        'xori', 'xorim',
        'shli', 'shlim',
        'shri', 'shrim',
        'bitc', 'bitcm',
        'bitci', 'bitcim']

        instr1 = [
        'add', 'addm', 'addc', 'addmc',
        'sub', 'subm', 'subb', 'submb',
        'adi', 'adim',
        'sbi', 'sbim']

        instr2 = ['cmp', 'cpiw', 'cpib']

        if op in instr0:
            result = ['E']
        elif op in instr1:
            result = ['E', 'C']
        elif op in instr2:
            result = ['G', 'L', 'E']

        return result

    def _get_flags_used(self) :
        flag = ('C', 'E', 'L', 'G')
        index = None
        result = None

        if (self.body[0] == 'btj' or self.body[0] == 'btjr') :
            index = self.body[3]

            if (self.body[1] == 'pc' and 28 <= index and index <= 31) :
                result = flag[index - 28]
        elif (self.body[0] == 'jsr') :
            index = self.body[3]

            if (4 <= index and index <= 7) :
                result = flag[index - 4]
        elif (self.body[0] == 'addc' or self.body[0] == 'addmc' or self.body[0] == 'subb' or self.body[0] == 'submb') :
            result = 'C'

        return [result] if result else []

    evaluate = {
        'add'    : lambda fst, sec: fst + sec,
        'addm'   : lambda fst, sec: fst + sec,
        'addc'   : lambda C, fst, sec: fst + sec + C,
        'addmc'  : lambda C, fst, sec: fst + sec + C,
        'adi'    : lambda fst, sec: fst + sec,
        'adim'   : lambda fst, sec: fst + sec,
        'sub'    : lambda fst, sec: fst - sec,
        'subm'   : lambda fst, sec: fst - sec,
        'subb'   : lambda C, fst, sec: fst - sec - C,
        'submb'  : lambda C, fst, sec: fst - sec - C,
        'sbi'    : lambda fst, sec: fst - sec,
        'sbim'   : lambda fst, sec: fst - sec,
        'and'    : lambda fst, sec: fst & sec,
        'andm'   : lambda fst, sec: fst & sec,
        'andi'   : lambda fst, sec: fst & sec,
        'andim'  : lambda fst, sec: fst & sec,
        'or'     : lambda fst, sec: fst | sec,
        'orm'    : lambda fst, sec: fst | sec,
        'ori'    : lambda fst, sec: fst | sec,
        'orim'   : lambda fst, sec: fst | sec,
        'xor'    : lambda fst, sec: fst ^ sec,
        'xorm'   : lambda fst, sec: fst ^ sec,
        'xori'   : lambda fst, sec: fst ^ sec,
        'xorim'  : lambda fst, sec: fst ^ sec,
        'shl'    : lambda fst, sec: fst << sec,
        'shlm'   : lambda fst, sec: fst << sec,
        'shli'   : lambda fst, sec: fst << sec,
        'shlim'  : lambda fst, sec: fst << sec,
        'shr'    : lambda fst, sec: fst >> sec,
        'shrm'   : lambda fst, sec: fst >> sec,
        'shri'   : lambda fst, sec: fst >> sec,
        'shrim'  : lambda fst, sec: fst >> sec,
        }

    mask = {
        0        : 0xff ,
        1        : 0xffff ,
        2        : 0xffffffff ,
        3        : 0xffffffffffffffffL ,
        4        : 0xffff ,
        5        : 0xff ,
        6        : 0xffffffff ,
        7        : 0xffffffff ,
        }

    def evaluate(self) :
        op = self.body[0]

        instr = [
        'add',
        'sub',
        'or',
        'and',
        'xor',
        'shl',
        'shr',
        'adi',
        'sbi',
        'ori',
        'andi',
        'xori',
        'shli',
        'shri']

        instrm = [ x + 'm' for x in instr0 ]

        if op in instr:
            [left] = self.get_left()
            [opA, opB] = [ x if x.is_a_constant() else x.get_value() for x in self.get_operands() ]
            lattice_elm = None

            if (opA.is_ordinary_constant() and opB.is_ordinary_constant()) :
                lattice_elm = CmmOrigAsmInstruction.evaluate[op](opA, opB)
            elif (opA.is_max() or opB.is_max()) :
                lattice_elm = CmmValueLattice('multiple_data')
            # else {
            # bug : can be done better, e.g. a & 0 = 0
            # $lattice_elm = CmmValueLattice->new('no_data') ;
            # }

            if (lattice_elm) :
                left.set_value(lattice_elm)

                opC = left.get_value() ; # mask with size of left

                if (opC.is_ordinary_constant()) :
                    self._reset_E() if opC else self._set_E()

                    if self._is_addition() :
                        self._reset_C() if lattice_elm < (1 << left.get_size()) else self._set_C()
                    elif (self._is_subtraction()) :
                        self._reset_C() if opA >= opB else self._set_C()
                elif (opC.is_max()) :
                    for lv in self.get_left_side_effect() :
                        lv.set_value(opC.new())

                # $self->set_value($opC->new()) ;
        elif op in instrm :
            [opC, opA, opB, size] = [ x if x.is_a_constant() else x.get_value() for x in self.get_operands() ]
            lattice_elm = None

            if (opA.is_ordinary_constant() and opB.is_ordinary_constant()) :
                lattice_elm = CmmOrigAsmInstruction.evaluate[op](opA, opB) & CmmOrigAsmInstruction.mask[size]
            elif (opA.is_max() or opB.is_max()) :
                lattice_elm = CmmValueLattice('multiple_data')
            # else {
            # bug : can be done better, e.g. a & 0 = 0
            # $lattice_elm = CmmValueLattice->new('no_data') ;
            # }

            if (lattice_elm) :
                opC = lattice_elm

                if (opC.is_ordinary_constant()) :
                    self._reset_E() if opC else self._set_E()

                    if (self._is_addition()) :
                        self._reset_C() if lattice_elm < (CmmOrigAsmInstruction.mask[size] + 1) else self._set_C()
                    elif (self._is_subtraction()) :
                        self._reset_C() if opA >= opB else self._set_C()
                elif (opC.is_max()) :
                    for lv in self.get_left_side_effect() :
                        lv.set_value(opC.new())

                # $self->set_value($opC->new()) ;
        elif (op == 'addc' or op == 'subb') :
            # elsif ($op eq 'addc' or $op eq 'addmc' or $op eq 'subb' or $op eq 'submb') {
            [left] = self.get_left()
            [C, opA, opB, size] = [ x if x.is_a_constant() else x.get_value() for x in self.get_operands_ext() ]
            lattice_elm = None

            if (C.is_ordinary_constant() and opA.is_ordinary_constant() and opB.is_ordinary_constant()) :
                lattice_elm = CmmOrigAsmInstruction.evaluate[op](C, opA, opB)
            elif (C.is_max() or opA.is_max() or opB.is_max()) :
                lattice_elm = CmmValueLattice('multiple_data')

            if (lattice_elm) :
                left.set_value(lattice_elm)

                opC = left.get_value() ; # mask with size of left

                if (opC.is_ordinary_constant()) :
                    self._reset_E() if opC else self._set_E()

                    if (self._is_addition()) :
                        self._reset_C() if lattice_elm < (1 << left.get_size()) else self._set_C()
                    elif (self._is_subtraction()) :
                        self._reset_C() if (opA >= opB + C) else self._set_C()
                elif (opC.is_max()) :
                    for lv in self.get_left_side_effect() :
                        lv.set_value(opC.new())

                # $self->set_value($opC->new()) ;
        elif (op == 'addmc' or op == 'submb') :
            [C, opC, opA, opB, size] = [ x if x.is_a_constant() else x.get_value() for x in self.get_operands_ext() ]
            lattice_elm = None

            if (C.is_ordinary_constant() and opA.is_ordinary_constant() and opB.is_ordinary_constant()) :
                lattice_elm = CmmOrigAsmInstruction.evaluate[op](C, opA, opB) & CmmOrigAsmInstruction.mask[size]
            elif (C.is_max() or opA.is_max() or opB.is_max()) :
                lattice_elm = CmmValueLattice('multiple_data')

            if (lattice_elm) :
                opC = lattice_elm

                if (opC.is_ordinary_constant()) :
                    self._reset_E() if opC else self._set_E()

                    if (self._is_addition()) :
                        self._reset_C() if lattice_elm < (CmmOrigAsmInstruction.mask[size] + 1) else self._set_C()
                    elif (self._is_subtraction()) :
                        self._reset_C() if opA >= opB + C else self._set_C()
                elif (opC.is_max()) :
                    for lv in self.get_left_side_effect() :
                        lv.set_value(opC.new())

                # $self->set_value($opC->new()) ;
        elif (op == 'bitc' or op == 'bitci') :
            [left] = self.get_left()
            operands = self.get_operands()
            [opA, opB, v] = [ x if x.is_a_constant() else x.get_value() for x in operands ]
            lattice_elm = None

            if (opA.is_ordinary_constant() and opB.is_ordinary_constant()) :
                opB &= (0x7 if operands[0].is_a_rf_reg() else 0x1f)

                if v :
                    lattice_elm = (opA | (1 << opB))
                else :
                    lattice_elm = (opA & (~ (1 << opB)))
            elif (opA.is_max() or opB.is_max()) :
                lattice_elm = CmmValueLattice('multiple_data')

            if (lattice_elm) :
                left.set_value(lattice_elm)

                opC = left.get_value() ; # mask with size of left

                if (opC.is_ordinary_constant()) :
                    self._reset_E() if opC else self._set_E()
                elif (opC.is_max()) :
                    for lv in self.get_left_side_effect() :
                        lv.set_value(opC.new())

                # $self->set_value($opC->new()) ;
        elif (op == 'bitcm' or op == 'bitcim') :
            operands = self.get_operands()
            [opC, opA, opB, v, dt, s02] = [ x if x.is_a_constant() else x.get_value() for x in operands ]
            lattice_elm = None

            if (opA.is_ordinary_constant() and opB.is_ordinary_constant()) :
                opB &= (0x7 if operands[0].is_a_rf_reg() else 0x1f)

                if v :
                    lattice_elm = (opA | (1 << opB))
                else :
                    lattice_elm = (opA & (~ (1 << opB)))

                lattice_elm &= 0xff ; # the destination of bitcm and bitcim is always 8 bit long.
            elif (opA.is_max() or opB.is_max()) :
                lattice_elm = CmmValueLattice('multiple_data')

            if (lattice_elm) :
                opC = lattice_elm

                if (opC.is_ordinary_constant()) :
                    self._reset_E() if opC else self._set_E()
                elif (opC.is_max()) :
                    for lv in self.get_left_side_effect() :
                        lv.set_value(opC.new())

                # $self->set_value($opC->new()) ;
        elif (op == 'flo') :
            pass
        elif (op == 'cmp' or op == 'cpiw' or op == 'cpib') :
            [opA, opB] = [ x.get_value() if not x.is_a_constant() else x for x in self.get_operands() ]

            if (opA.is_ordinary_constant() and opB.is_ordinary_constant()) :
                if (opA > opB) :
                    self._set_G()
                    self._reset_L()
                    self._reset_E()
                elif (opA < opB) :
                    self._reset_G()
                    self._set_L()
                    self._reset_E()
                else :
                    self._reset_G()
                    self._reset_L()
                    self._set_E()
            elif (opA.is_max() or opB.is_max()) :
                for lv in self.get_left_side_effect() :
                    lv.set_value(CmmValueLattice('multiple_data'))
        elif (op == 'jsr') :
            [reg, addr, bit, v] = self.get_operands()
            flag = None

            if (bit <= 3) :
                self.set_value(CmmValueLattice(1))
            else :
                (flag) = self.get_right_side_effect()

                if (flag.is_ordinary_constant()) :
                    self.set_value(CmmValueLattice(1 if flag == v else 0))
                elif (flag.is_max()) :
                    self.set_value(CmmValueLattice('multiple_data'))
        elif (op == 'btj' or op == 'btjr') :
            [reg, addr, bit, v] = self.get_operands()
            [flag] = self.get_right_side_effect()

            if (flag) :
                flag = flag.get_value()

                if (flag.is_ordinary_constant()) :
                    self.set_value(CmmValueLattice(1 if flag == v else 0))
                elif (flag.is_max()) :
                    self.set_value(CmmValueLattice('multiple_data'))
            else :
                reg_value = reg.get_value()
                # bug : can be done better if we propagate some info about reg range

                if (reg_value.is_ordinary_constant()) :
                    self.set_value(CmmValueLattice(1 if (reg_value & (1 << addr)) == v else 0))
                elif (reg_value.is_max()) :
                    self.set_value(CmmValueLattice('multiple_data'))

    def customize(self) :
        return [ self ]

    def get_asm_code(self) :
        code = ''

        body = [ x.get_reg() if (hasattr(x, 'is_a_var') and x.is_a_var()) else x for x in self.body ]

        code = "\t" + str(body.pop(0)) + "\t"

        if (len(body) < 5) :
            if (self.body[0] == 'mviw') :
                if (body[1].is_a_constant()) : const = "0x%04x" % (body[1].get_value(),)

                code += body.pop(0) + ",\t" + str(const) + ",\t" + str(pop(body))
            elif (self.body[0] == 'mvib') :
                if (body[1].is_a_constant()) : const = "0x%02x" % (body[1].get_value(),)

                code += body.pop(0) + ",\t" + str(const) + ",\t" + str(pop(body))
            else :
                code += ",\t".join([ str(x) for x in body ])
        else :
            z = body[0, 4]
            body[0, 4] = []
            code += ',\t'.join([ str(x) for x in z ])
            code += ',  ' + ',  '.join([ str(x) for x in body ])

        return code
            # code + "\t\t;; original assembly"
