import copy
import sys

from CmmEval import CmmEval
from CmmMisc import CmmMisc
from CmmError import CmmError
from CmmWarning import CmmWarning
from CmmInstruction import CmmInstruction
from CmmAssgnInstruction import CmmAssgnInstruction
from CmmGotoInstruction import CmmGotoInstruction
from CmmPHInstruction import CmmPHInstruction
from CmmNullInstruction import CmmNullInstruction
from CmmOrigAsmInstruction import CmmOrigAsmInstruction
from CmmFakeInstruction import CmmFakeInstruction

from CmmSR import CmmSR

from CmmLabel import CmmLabel
from CmmStartLabel import CmmStartLabel
from CmmFinishLabel import CmmFinishLabel
from CmmIfElseLabel import CmmIfElseLabel
from CmmWhileLabel import CmmWhileLabel
from CmmForLabel import CmmForLabel
from CmmDoLabel import CmmDoLabel
from CmmPcrLabel import CmmPcrLabel

from CmmBasicBlock import CmmBasicBlock
from CmmPiggyBank import CmmPiggyBank
from CmmValueLattice import CmmValueLattice

class CmmStatement(object) :
    eval_cnst = CmmEval('tokens_are_constants_only')

    def evaluate_cnst_expr(self, *args) :
        if not (CmmMisc.ref_to_array(args[0])) : args[0] = args[0].get_body()

        return CmmStatement.eval_cnst.evaluate(*args)

    def __init__(self, start, finish = None, gt = None, lbl = None) :
        if not (finish) : finish = start
        if not (gt) : gt = [ ]
        if not (lbl) : lbl = { }

        self.flowchart = { 'start' : start, 'finish' : finish}
        self.goto = gt
        self.break_ = [ ]
        self.continue_ = [ ]
        self.return_ = [ ]
        self.RETURN = [ ]
        self.case = { }
        self.label = lbl
        self._instruction = [ ]
        self._deleted = [ ]
        self._bb = [ ]
        self._bb_r = [ ]

    def new(self, copy_number) :
        # copy_number is used for hashing copys in CmmVar or CmmLabel objects
        # by not using it we lose space efficiency only (???)
        c = copy.deepcopy(self)

        return c

    def get_copy(self, copy_number, function_call_number) : # !!!!! DIFFERENT
        # this function different from the original in that
        # it doesn't change copy counter of destination of return expression
        c = deepcopy(self)

        return c

    def get_start(self) :
        return self.flowchart['start']

    def set_start(self, instr) :
        self.flowchart['start'] = instr

    def get_finish(self) :
        return self.flowchart['finish']

    def set_finish(self, instr) :
        self.flowchart['finish'] = instr

    @staticmethod
    def no_statement() :
        return CmmStatement( CmmNullInstruction([ 1 ]) )

    construct = [ 'NONE' ] * 2

    @classmethod
    def _update_construct(cls, construct_ = None) :
        if construct_ :
            cls.construct += [construct_]
        else :
            cls.construct.pop()

    @staticmethod
    def _in_switch_compound_statement() :
        return CmmStatement.construct[-1] == "COMPOUND" and CmmStatement.construct[-2] == "SWITCH"

    @staticmethod
    def _in_switch() :
        return filter(lambda x: x == "SWITCH", CmmStatement.construct)

    @staticmethod
    def _in_loop() :
        return filter(lambda x: x == "WHILE" or x == "DO" or x == "FOR", CmmStatement.construct)

    @staticmethod
    def _is_assignment_op(x) :
        return type(x) is str and x in ['=', '+=', '-=', '&=', '|=', '^=', '>>=', '<<=']

    @staticmethod
    def _check_lvalues(expression) :
        i = 0

        while (i < len(expression)) :
            if CmmStatement._is_assignment_op(expression[i]) :
                i += 1
                if not (expression[i].is_a_var() or expression[i].is_a_reg() or expression[i].is_a_mem_ref()) :
                    CmmError.error_message('ERR_INVALID_LVALUE_IN_ASSGNMNT')
            i += 1

    def process_if_statement(self, expression_top, selection_tail, function) :
        if CmmError.there_are_errors() :
            return CmmStatement.no_statement()
        else :
            cond_frst = expression_top.get_start()
            cond_last = expression_top.get_finish()

            cond_last.process_condition()

            frst_yes = self.get_start()
            last_yes = self.get_finish()

            placeholder = None

            if (selection_tail) :
                [tmp] = selection_tail

                self.goto += tmp.goto
                self.break_ += tmp.break_
                self.continue_ += tmp.continue_
                self.return_ += tmp.return_
                self.RETURN += tmp.RETURN

                for key, value in tmp[label].iteritems() :
                    self._register_label(key, value)

                frst_no = tmp.get_start()
                last_no = tmp.get_finish()

                else_begin_label = CmmIfElseLabel(function, 'begin')

                cond_last = cond_last.add_jump_after_comparison('failure', else_begin_label)

                frst_no.add_label(else_begin_label)
                self._register_label(else_begin_label, frst_no)

                if_end_label = CmmIfElseLabel(function, 'end')

                placeholder = CmmPHInstruction()
                CmmInstruction.concat_next(last_no, placeholder)

                placeholder.add_label(if_end_label)
                self._register_label(if_end_label, placeholder)

                if last_yes.is_placeholder() :
                    last_yes.transform(CmmAssgnInstruction, [ 'mviw', CmmSR('pc'), if_end_label ])
                    # $last_yes->transform('CmmGotoInstruction', [ 'mviw', CmmSR->new('pc'), $if_end_label ]) ;
                else :
                    [_, last_yes] = CmmInstruction.concat_next(last_yes,\
                                    CmmAssgnInstruction(['mviw', CmmSR('pc'), if_end_label ]))
                    # CmmGotoInstruction->new(['mviw', CmmSR->new('pc'), $if_end_label ])) ;

                self.goto += [last_yes]

                CmmInstruction.concat_next(last_yes, frst_no)

                CmmInstruction.concat_yes(cond_last, frst_yes)
                CmmInstruction.concat_no(cond_last, frst_no)
            else :
                if_end_label = CmmIfElseLabel(function, 'end')

                placeholder = CmmPHInstruction()
                CmmInstruction.concat_next(last_yes, placeholder)

                placeholder.add_label(if_end_label)
                self._register_label(if_end_label, placeholder)

                cond_last = cond_last.add_jump_after_comparison('failure', if_end_label)

                CmmInstruction.concat_yes(cond_last, frst_yes)
                CmmInstruction.concat_no(cond_last, placeholder)

            self.set_start(cond_frst)
            self.set_finish(placeholder)

            return self

    case = []

    @staticmethod
    def _start_case() :
        CmmStatement.case += [{}]

    @staticmethod
    def _finish_case() :
        CmmStatement.case.pop()

    @staticmethod
    def _add_case(case) :
        CmmStatement.case[-1][case] = 0

    @staticmethod
    def _exists_case(case) :
        return CmmStatement.case[-1].has_key(case)

    @staticmethod
    def _there_was_at_least_one_case() :
        return case[-1].keys()

    default_counter = []

    @staticmethod
    def _start_default() :
        CmmStatement.default_counter += [0]

    @staticmethod
    def _finish_default() :
        CmmStatement.default_counter.pop()

    @staticmethod
    def _add_default() :
        CmmStatement.default_counter[-1] += 1

    @staticmethod
    def _exists_default() :
        return len(CmmStatement.default_counter) and CmmStatement.default_counter[-1]

    def process_global_var_declaration(self) :
        return CmmStatement.no_statement()

    def process_local_var_declaration(self, var, filename, line) :
        if self._in_switch_compound_statement() :
            CmmError.error_message('ERR_UNEXPECTED_VAR_DECL')
            return CmmStatement.no_statement()
        else :
            for i in range(len(var)) :
                lv = var[i]
                identifier = lv[0]
                initializer = lv[1]

                if not initializer.flowchart['finish'].is_null() :
                    initializer.flowchart['finish']['body'][0:0] = ['=', identifier]

                var[i] = initializer

            j = 0

            while (j < var and var[j].flowchart['start'].is_null()) :
                j= j + 1

            if j < len(var) :
                start = var[j][flowchart]['start']

                for i in range(j + 1, len(var)) :
                    if not var[i].flowchart['start'].is_null() :
                        CmmInstruction.concat_next(var[j].flowchart['finish'], var[i].flowchart['start'])
                        j = i

                var[0].flowchart['start'] = start
                var[0].flowchart['finish'] = var[j].flowchart['finish']

            return var[0]

    def process_index(self, var) :
        self.flowchart['finish'].body[0:0] = ['=', var]

    def process_gotos(self) :
        label = self.label

        for instr in self.goto :
            print("INSTR " + str(instr) + ": " + str(instr.body) + " " + str(instr.body[1].get_name()))
            label_id = instr.get_instr_label().get_id()
            print("LABEL !!! " + str(label_id) + "\n")

            if label.has_key(label_id) :
                for lv in instr[child] :
                    del instr.child[lv].father[lv]
                    del instr.child[lv]

                CmmInstruction.concat_goto(instr, label[label_id])
            else :
                CmmError.error_message('ERR_UNDEFINED_LABEL', label_id.get_name())

    def process_returns(self, finish_label, var = None) :
        if var :
            for instr in self.return_ :
                instr.body[0:0] = ['=', var]
                print ("process_returns " + str(instr.body) + "\n")

        for instr in self.RETURN :
            print("RETURN INSTR " + str(instr) + ":" + str(instr.body))
            instr.set_instr_label(finish_label)
            self.goto += [instr]

        print("GOTO size is " + str(len(self.goto)))
        self.RETURN = []

    def check_returns(self) :
        return len(self.RETURN)

    def _process_cases(self, cond, end_point) :
        cond.child = self.case

        for x in cond.child :
            cond.child[x].father[x] = cond

        if not filter(lambda x: x[:len('default_')] == 'default_', cond.child.keys()) :
            CmmInstruction.concat_no(cond, end_point)

        self.case = {}

    def _process_breaks(self, end_point) :
        for instr in self.break_ :
            instr.set_instr_label(end_point.get_label())
            self.goto += [instr]

        self.break_ = []

    def _process_continues(self, cond) :
        for instr in self.continue_ :
            instr.set_instr_label(cond.get_label())
            self.goto += [instr]

        self.continue_ = []

    def process_switch_statement_header(self) :
        self._start_case()
        self._start_default()
        self._update_construct('SWITCH')

    def process_switch_statement_body(self, expression_top, filename, line) :
        self._update_construct()
        self._finish_default()

        if not self._there_was_at_least_one_case() :
            CmmError.error_message('ERR_NO_CASE_IN_SWITCH')

        self._finish_case()

        if CmmError.there_are_errors() :
            return CmmStatement.no_statement()
        else :
            cond_frst = expression_top.get_start()
            cond_last = expression_top.get_finish()

            cond_last.process_condition()

            placeholder = CmmPHInstruction()

            CmmInstruction.concat_next(self.get_finish(), placeholder)

            self.set_start(cond_frst)
            self.set_finish(placeholder)

            self._process_cases(cond_last, placeholder)
            self._process_breaks(placeholder)

            return self

    def process_case_statement(self, constant_expression, filename, line) :
        if self._in_switch_compound_statement() :
            case_value = CmmStatement.evaluate_cnst_expr(constant_expression, filename, line)

            if self._exists_case(case_value) :
                CmmError.error_message('ERR_DUPLICATE_CASE_VALUE', case_value)
            else :
                self._add_case(case_value)
                self.case[CmmInstruction.get_case_edge(case_value)] = self.get_start()
        else :
            CmmError.error_message('ERR_UNEXPECTED_CASE')

        return self

    def process_default_statement(self, filename, line) :
        if self._in_switch_compound_statement() :
            if self._exists_default() :
                CmmError.error_message('ERR_MULTIPLE_DEFAULTS')
            else :
                self._add_default()
                self.case[CmmInstruction.get_default_edge()] = self.get_start()
        else :
            CmmError.error_message('ERR_UNEXPECTED_DEFAULT')

        return self

    def process_while_statement_header(self) :
        self._update_construct('WHILE')

    def process_while_statement_body(self, condition, function) :
        self._update_construct()

        if CmmError.there_are_errors() :
            return CmmStatement.no_statement()
        else :
            cond_frst = condition.get_start()
            cond_last = condition.get_finish()

            cond_last.process_condition()

            frst_yes = self.get_start()
            last_yes = self.get_finish()

            while_begin_label = CmmWhileLabel(function, 'begin')

            cond_frst.add_label(while_begin_label)
            self._register_label(while_begin_label, cond_frst)

            if (last_yes.is_placeholder()) :
                last_yes.transform(CmmAssgnInstruction, [ 'mviw', CmmSR('pc'), while_begin_label ])
            else :
                [_, last_yes] = CmmInstruction.concat_next(last_yes,\
                                CmmAssgnInstruction(['mviw', CmmSR('pc'), while_begin_label ]))

            self.goto += [last_yes]

            while_end_label = CmmWhileLabel(function, 'end')
            placeholder = CmmPHInstruction()

            placeholder.add_label(while_end_label)
            self._register_label(while_end_label, placeholder)

            CmmInstruction.concat_next(last_yes, placeholder)

            cond_last = cond_last.add_jump_after_comparison('failure', while_end_label)

            CmmInstruction.concat_yes(cond_last, frst_yes)
            CmmInstruction.concat_no(cond_last, placeholder)

            self.set_start(cond_frst)
            self.set_finish(placeholder)

            self._process_continues(cond_frst)
            self._process_breaks(placeholder)

            return self

    def process_do_statement_header(self) :
        self._update_construct('DO')

    def process_do_statement_body(self, condition, function) :
        self._update_construct() ; # A bit dangerous. That would better be done before 'while' keyword.

        if CmmError.there_are_errors() :
            return CmmStatement.no_statement()
        else :
            cond_frst = condition.get_start()
            cond_last = condition.get_finish()

            cond_last.process_condition()

            frst_yes = self.get_start()
            last_yes = self.get_finish()

            do_begin_label = frst_yes.get_label()

            if not do_begin_label :
                do_begin_label = CmmDoLabel(function, 'begin')
                frst_yes.add_label(do_begin_label)
                self._register_label(do_begin_label, frst_yes)

            CmmInstruction.concat_next(last_yes, cond_frst)

            cond_last = cond_last.add_jump_after_comparison('success', do_begin_label)

            do_end_label = CmmDoLabel(function, 'end')
            placeholder = CmmPHInstruction()

            placeholder.add_label(do_end_label)
            self._register_label(do_end_label, placeholder)

            CmmInstruction.concat_no(cond_last, frst_yes)
            CmmInstruction.concat_yes(cond_last, placeholder)

            self.set_start(frst_yes)
            self.set_finish(placeholder)

            do_condition_label = CmmDoLabel(function, 'condition')

            cond_frst.add_label(do_condition_label)
            self._register_label(do_condition_label, cond_frst)

            self._process_continues(cond_frst)
            self._process_breaks(placeholder)

            return self

    def process_for_statement_header(self) :
        self._update_construct('FOR')

    def process_for_statement_body(self, for_expression_0, for_expression_1, for_expression_2, function) :
        self._update_construct()

        if CmmError.there_are_errors() :
            return CmmStatement.no_statement()
        else :
            init_frst = for_expression_0.get_start()
            init_last = for_expression_0.get_finish()
            cond_frst = for_expression_1.get_start()
            cond_last = for_expression_1.get_finish()
            incr_frst = for_expression_2.get_start()
            incr_last = for_expression_2.get_finish()

            cond_last.process_condition()

            frst_yes = self.get_start()
            last_yes = self.get_finish()

            CmmInstruction.concat_next(init_last, cond_frst)

            for_begin_label = CmmForLabel(function, 'begin')

            cond_frst.add_label(for_begin_label)
            self._register_label(for_begin_label, cond_frst)

            [_, incr_last] = CmmInstruction.concat_next(incr_last,\
                             CmmAssgnInstruction(['mviw', CmmSR('pc'), for_begin_label ]))

            self.goto += [incr_last]

            for_end_label = CmmForLabel(function, 'end')
            placeholder = CmmPHInstruction()

            placeholder.add_label(for_end_label)
            self._register_label(for_end_label, placeholder)

            cond_last = cond_last.add_jump_after_comparison('failure', for_end_label)

            CmmInstruction.concat_yes(cond_last, frst_yes)
            CmmInstruction.concat_no(cond_last, placeholder)

            CmmInstruction.concat_next(last_yes, incr_frst)
            CmmInstruction.concat_next(incr_last, placeholder)

            self.set_start(init_frst)
            self.set_finish(placeholder)

            self._process_continues(incr_frst)
            self._process_breaks(placeholder)

            return self

    @classmethod
    def process_compound_statement_open_bracket(cls) :
        print('process_compound_statement_open_bracket')
        cls._update_construct('COMPOUND')

    @classmethod
    def process_compound_statement_close_bracket(cls, var_declaration, statement) :
        print("********************************************************************************\n")
        statement_ = []
        print "start of process_compound_statement_close_bracket\n"

        cls._update_construct()

        if CmmError.there_are_errors() :
            return cls.no_statement()
        else :
            print("********************************************************************************\n")
            statement_ = var_declaration + statement

            if len(statement_):
                statement = statement_.pop(0)

                # concatenate flowcharts and eliminate placeholders

                print("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n")
                for lv in statement_ :
                    print("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n")
                    print("lv = " + str(lv))
                    start = lv.get_start()

                    CmmInstruction.concat_next(statement.get_finish(), start)

                    statement.goto += lv.goto
                    statement.break_ += lv.break_
                    statement.continue_ += lv.continue_
                    statement.return_ += lv.return_
                    statement.RETURN += lv.RETURN

                    for key in lv.case :
                        statement.case[key] = start

                    for key, value in lv.label.iteritems() :
                        statement._register_label(key, value)

                    statement.set_finish(lv.get_finish())

                print "finish of process_compound_statement_close_bracket 0\n"
                return statement
            else :
                print "finish of process_compound_statement_close_bracket 1\n"
                return cls.no_statement()

    def _register_label(self, label, instr) :
        self.label[label] = instr

        return label

    def process_labeled_statement(self, label, filename, line) :
        if self.label.has_key(label.get_id()) :
            CmmError.error_message('ERR_LABEL_REDEFINED', label.get_name())
        else :
            self.flowchart['start'].add_label(label)
            self._register_label(label, self.flowchart['start'])

        return self

    def add_start_label(self, function_name) :
        label = CmmStartLabel(function_name)

        self.flowchart['start'].add_label(label)
        self._register_label(label, self.flowchart['start'])

    def add_finish_label(self, function_name) :
        label = CmmFinishLabel(function_name)

        self.flowchart['finish'].add_label(label)
        self._register_label(label, self.flowchart['finish'])

    def process_goto_statement(self, label) :
        statement = CmmAssgnInstruction([ 'mviw', CmmSR('pc'), label ])
        statement = CmmStatement(statement, statement)

        statement.goto += [statement.get_finish()]

        return statement

    def process_continue_statement(self, filename, line) :
        if not CmmError.there_are_errors() :
            if not self._in_loop() :
                CmmError.error_message('ERR_UNEXPECTED_CONTINUE')

        statement = CmmAssgnInstruction([ 'mviw', CmmSR('pc') ])
        statement = CmmStatement(statement, statement)

        statement.continue_ += [statement.get_finish()]

        return statement

    def process_break_statement(self, filename, line) :
        if not CmmError.there_are_errors() :
            if not self._in_switch() or self._in_loop() :
                CmmError.error_message('ERR_UNEXPECTED_BREAK')

        statement = CmmAssgnInstruction([ 'mviw', CmmSR('pc') ])
        statement = CmmStatement(statement, statement)

        statement.break_ += [statement.get_finish()]

        return statement

    def process_return_statement(self) :
        print("????????????????????????????????????????????????????????????????????????????????\n")
        self.return_ += [self.get_finish()]

        [_, self.flowchart['finish']] = CmmInstruction.concat_next(self.flowchart['finish'],\
                                        CmmAssgnInstruction([ 'mviw', CmmSR('pc') ]))

        self.RETURN += [self.get_finish()]
        print("r len is " + str(len(self.RETURN)))

        print(str(self))
        return self

    @staticmethod
    def _concatenate(fst, sec) :
        return CmmInstruction.concat_next(fst, sec)

    @staticmethod
    def process_expression(expression) :
        start = None
        finish = None
        g = []
        l = {}
        L = {}

        postprocessing = []
        preprocessing = []

        expr = expression
        i = 0

        expr_ = []
        i_ = []

        while (i < len(expr)) :
            while (i < len(expr)) :
                # print ("process_expression 1 " + str(expr) + "\n")
                # print ("process_expression 2 " + str(expr) + "\n")

                if type(expr[i]) is list :
                    expr_ += [expr]
                    i_ += [i + 1]

                    expr = expr[i]
                    i = 0
                elif (not (type(expr[i]) is int or type(expr[i]) is str)) :
                    if (expr[i].is_a_mem_ref() or expr[i].is_a_var()) :
                        for lv in expr[i].get_statements_to_evaluate() :
                            if start :
                                CmmStatement._concatenate(finish, lv.flowchart['start'])
                            else :
                                start = lv.flowchart['start']

                            finish = lv.flowchart['finish']

                        i= i + 1
                    elif (expr[i].is_a_bracketed_expr()) :
                        for lv in expr[i].get_statement() : # at most one statement will be returned
                            if start :
                                CmmStatement._concatenate(finish, lv.flowchart['start'])
                            else :
                                start = lv.flowchart['start']

                            finish = lv.flowchart['finish']

                        expr[i : i + 1] = expr[i].get_expr()
                    elif (expr[i].is_a_pre_incr_expr()) :
                        op = expr[i].get_operand()
                        preprocessing += [[ '=' , op.new() , '+' , op.new() , CmmConst(1) ]]
                        expr[i] = op
                    elif (expr[i].is_a_pre_decr_expr()) :
                        op = expr[i].get_operand()
                        preprocessing += [[ '=' , op.new() , '-' , op.new() , CmmConst(1) ]]
                        expr[i] = op
                    elif (expr[i].is_a_post_incr_expr()) :
                        op = expr[i].get_operand()
                        postprocessing += [[ '=' , op.new() , '+' , op.new() , CmmConst(1) ]]
                        expr[i] = op
                    elif (expr[i].is_a_post_decr_expr()) :
                        op = expr[i].get_operand()
                        postprocessing += [[ '=' , op.new() , '-' , op.new() , CmmConst(1) ]]
                        expr[i] = op
                    elif (expr[i].is_a_select_expr()) :
                        st = expr[i].get_statement()

                        if (start) :
                            CmmStatement._concatenate(finish, st.flowchart['start'])
                        else :
                            start = st.flowchart['start']

                        finish = st.flowchart['finish']

                        g += st.goto

                        for key, value in st[label].iteritems() :
                            l[key] = value

                        expr[i] = expr[i].get_var()

                        i += 1
                    else :
                        i += 1
                else :
                    i += 1

            while (len(expr_) > 0 and i >= len(expr)) :
                expr = expr_.pop()
                i = i_.pop()

        # print ("process_expression 4 " + str(expr) + "\n")
        for lv in preprocessing + [expression] + postprocessing :
            expr = lv
            i = 0

            while (i < len(expr)) :
                while (i < len(expr)) :
                    if type(expr[i]) is list :
                        expr_ += [expr]
                        i_ += [i + 1]

                        expr = expr[i]
                        i = 0
                    elif (hasattr(expr[i], 'is_a_func_call') and expr[i].is_a_func_call()) :
                        if start :
                            CmmStatement._concatenate(finish, expr[i].get_start())
                        else :
                            start = expr[i].get_start()

                        finish = expr[i].get_finish()

                        expr[i] = expr[i].get_var() or CmmConst(0)

                        i += 1
                    else :
                        i += 1

                while (len(expr_) > 0 and i >= len(expr)) :
                    expr = expr_.pop()
                    i = i_.pop()

            if start :
                [_, finish] = CmmStatement._concatenate(finish, CmmInstruction(expr))
            else :
                start = finish = CmmInstruction(expr)

        # print ("process_expression 5 " + str(expr) + "\n")
        instr = start
        while instr :
            # print('>>> ' + str(instr) + ' <<<')
            instr.process_assignment_and_unary_operators()
            # print('>>> ' + str(instr) + ' <<<')
            instr = instr.get_next()
        # print ("process_expression 6 " + str(expr) + "\n")

        s = CmmStatement(start, finish, g, l)
        # return CmmStatement(start, finish, g, l)
        # print ("process_expression 7 " + str(s) + "\n")
        return s

    def process_pcr_statement(self, mem, pcr_statement_rvalue, pcr_var) :
        if CmmError.there_are_errors() :
            return CmmStatement.no_statement()
        else :
            req_type = pcr_statement_rvalue[-1] if len(pcr_statement_rvalue) > 1 else CmmConst(0)

            statement = [CmmStatement.process_expression([ '=', pcr_var, CmmPcrLabel(pcr_statement_rvalue[0]) ])]
            if req_type :
                statement += [CmmStatement.process_expression([ '=', pcr_var.new(), '|', pcr_var.new(), req_type & 0xff003fff ])]
            statement += [CmmStatement.process_expression([ '=', mem, pcr_var.new() ])]

            CmmStatement.process_compound_statement_open_bracket()
            return CmmStatement.process_compound_statement_close_bracket([], statement )

    @classmethod
    def process_expression_top(cls, expression) :
        print('process_expression_top 0')
        if CmmError.there_are_errors() :
            print('process_expression_top 1')
            return CmmStatement.no_statement()
        else :
            print('process_expression_top 2')
            for lv in expression :
                print(str(lv))
                CmmStatement._check_lvalues(lv)

            print('process_expression_top 3')
            if CmmError.there_are_errors() :
                print('process_expression_top 30')
                return CmmStatement.no_statement()
            else :
                print('process_expression_top 31')
                cls.process_compound_statement_open_bracket()
                print('process_expression_top 32')
                for x in expression:
                    print("! " + str(x) + " !")
                return CmmStatement.process_compound_statement_close_bracket([], [ CmmStatement.process_expression(x) for x in expression ])

    def process_assignment_expr_top(self, assignment_expression, filename, line) :
        return CmmStatement.no_statement() if CmmError.there_are_errors()\
                                           else CmmStatement.process_expression(assignment_expression)

    @staticmethod
    def process_cond_expression_top(cond_expression) :
        return CmmStatement.no_statement() if CmmError.there_are_errors()\
                                           else CmmStatement.process_expression(cond_expression)

    def process_assembly_statement(self, assembly_instruction) :
        return CmmStatement( CmmOrigAsmInstruction([ filter(lambda x: x != ",", assembly_instruction) ]) )

    def set_instruction(self, instruction) :
        self._instruction = instruction

    def get_instruction(self) :
        return self._instruction

    def get_first_instruction(self) :
        return self._instruction[0]

    def build_instruction_sequence(self) :
        index = 0
        instr = self.get_start()

        while instr :
            instr.set_coord(index)
            index += 1
            self._instruction += [instr]
            instr = instr.get_next()

    def compute_instruction_sequence(self) :
        index = 0

        for lv in self._instruction :
            lv.set_coord(index)
            index = index + 1

        self._deleted = []

    def _update_instruction_sequence(self, index) :
        if self._deleted :
            if index > self._deleted[0] :
                for i in range(self._deleted[0], len(self._instruction)) :
                    instr = self._instruction[i]

                    if instr.get_coord() == index :
                        index = i
                        self._deleted = [ index ]

                    instr.set_coord(i)
            else :
                self._deleted[0:0] = [ index ]
        else :
            self._deleted[0:0] = [ index ]

        return index

    def replace_in_instruction(self, index, to_replace, by) :
        return self._instruction[index].replace_in_instruction(to_replace, by)

    def substitute(self, instr, copy_) :
        bb = instr.get_bb()
        if bb :
            for lv in copy_ :
                lv.set_bb(bb)
                lv.set_instr_for_variables()

            bb.substitute_instr_by(instr, *copy_)

        if copy_ :
            [start, finish] = [copy_[0], copy_[-1]]
            CmmInstruction.substitute(instr, start, finish)
        else :
            start = finish = CmmInstruction.substitute(instr)
            if not start :
                instr.transform(CmmPHInstruction, [])
                return instr

        CmmInstruction.transfer_labels(instr, start)

        if self.get_start() == instr :
            self.set_start(start)

        if self.get_finish() == instr :
            self.set_finish(finish)

    def expand(self, var, address, args, ret_var = None) :
        start = self.get_start()
        new = CmmGotoInstruction([ 'jsr', var, address, CmmConst(0), CmmConst(0) ])

        self._instruction[0:0] = [ new ]
        CmmInstruction.concat_next(new, start)
        start = new

        for lv in reverse(args) :
            new = CmmFakeInstruction([ '=', lv, CmmConst(0) ])
            self._instruction[0:0] = [ new ]
            CmmInstruction.concat_next(new, start)
            start = new

        self.set_start(new)

        self.get_finish().transform(CmmAssgnInstruction, [ 'move', CmmSR('pc'), var ])

        if ret_var :
            print("insert fake instruction for return\n")
            new = CmmFakeInstruction([ '=', CmmSR('cmd'), ret_var ])
            self._instruction += [ new ]

            CmmInstruction.concat_next(self.get_finish(), new)

            self.set_finish(new)

    def remove_fake_instructions(self) :
        if self._instruction and self._instruction[-1].is_fake() :
            self._instruction.pop()
            self.set_finish(self._instruction[-1])

        while self._instruction and self._instruction[0].is_fake() :
            self._instruction.pop(0)
            self.set_start(self._instruction[0])

        return self.compute_instruction_sequence()

    def get_instruction_index(self, instr) :
        for i in range(len(self._instruction)) :
            if (self._instruction[i] == instr) :
                return i

        return -1

    def kill_instruction(self, instr) :
        tmp = instr.get_coord()
        print ("coord == " + str(tmp) + "\n")

        i = self._update_instruction_sequence(instr.get_coord())

        print "================================================================================\n"
        print ("instr being killed " + str(instr.body) + "\n")
        print "================================================================================\n"

        for child_name in instr.child :
            if instr.child[child_name] == instr :
                del instr.child[child_name]

        for father_name in instr.father :
            if instr.father[father_name] == instr :
                del instr.father[father_name]

        for child_name in instr.child :                        # the deleted instruction always has children
            child = instr.child[child_name]

            del child.father[child_name]

            for father_name in instr.father :
                father = instr.father[father_name]

                del father.child[father_name]

                if not filter(lambda x: x == child,  father.child.values()) :
                    CmmInstruction.concat_with_appropriate_edge(father, child, father_name)

            CmmInstruction.transfer_labels(instr, child)

        for lv in self._instruction :
            print("_my_customize 036 : |" + str(lv) + "| " + str(lv.body) + "\n")

        instr.destroy()

        for lv in self._instruction :
            print("_my_customize 037 : |" + str(lv) + "| " + str(lv.body) + "\n")

        print("i = " + str(i) + "\n")

        self._instruction[i : i + 1] = []

        for lv in self._instruction :
            print ("_my_customize 038 : |" + str(lv) + "| " + str(lv.body) + "\n")

        if i < len(self._instruction) :
            self._instruction[i - 1].set_next(self._instruction[i])

    def delete_leading_jsr(self) :
        self.kill_instruction(self._instruction[0])

####################################################################################################
################################ transform_to_SSA_form start #######################################
####################################################################################################

    def _start_basic_block_building(self, instr) :
        if not instr.get_bb() :
            bb = CmmBasicBlock()
            self._bb += [bb]
            bb.add_instruction(instr)
            instr.set_bb(bb)

    def _build_basic_blocks(self) :
        # 1. Delete unreachable instructions

        instr_ = self.get_instruction()

        for instr in instr_ :
            instr.set_color(0)

        q = [instr_[0]] ; # BFS

        for lv in q :
            lv.set_color(1)
            q += filter(lambda x: x.get_color() == 0, lv.get_children())

        for instr in instr_ :
            if (instr.get_color()) :
                instr.set_color(0)
            else :
                print ("_delete_unreachable_instructions " + str(instr.body) + "\n")
                self.kill_instruction(instr)

        # 2. Mark Basic Blocks Start Points

        instr_ = self.get_instruction()

        self._start_basic_block_building(instr_[0])

        for instr in instr_ :
            parents = instr.get_parents()
            if len(parents) > 1 :
                self._start_basic_block_building(instr)

            children = instr.get_children()
            if len(children) > 1 :
                for lv in children :
                    self._start_basic_block_building(lv)

        # 3. Build Basic Blocks (BFS)

        q = [instr_[0]]

        for lv in q :
            lv.set_color(1)

            bb0 = lv.get_bb()

            for instr in filter(lambda x: x.get_color() == 0, lv.get_children()) :
                if not instr.get_bb() :
                    bb0.add_instruction(instr)
                    instr.set_bb(bb0)

                q += [instr]

        # 4. Build Edges in Basic Blocks Graph

        for instr in instr_ :
            bb0 = instr.get_bb()

            names = instr.get_children_names()

            for tmp in instr.get_children() :
                bb1 = tmp.get_bb()

                if bb0 == bb1 and tmp != bb0.get_first_orig_instruction() :
                    names.pop(0)
                else :
                    CmmBasicBlock.concat(bb0, bb1, names.pop(0))

        instr_[-1].get_bb().set_exit()

    def _DFS_0(self, bb) :
        for lv in bb.get_orig_children() :
            if lv.init_semi() :
                lv.set_DFS_father(bb)
                self._bb += [lv]
                self._DFS_0(lv)

    def _DFS_1(self, bb) :
        for lv in bb.get_d_children() :
            self._bb += [lv]
            lv.set_level(bb.get_level() + 1)
            self._DFS_1(lv)

    def _build_dj_graph(self) :
        bucket = {}

        if not self._bb :
            return

        print ("_build_dj_graph " + str(self._bb) + "\n")

        for bb in self._bb :
            bb.init()

        CmmBasicBlock.reset_semi()

        # 1

        self._bb[1 : ] = []
        self._DFS_0( self._bb[0] )

        # 2 & 3

        tmp = bb_ = self._bb
        tmp.pop(0)

        for lv in bb_ :
            bucket[lv] = []

        rtmp = tmp
        rtmp.reverse()

        for bb in rtmp :
            for lv in bb.get_orig_parents() :
                min = lv.eval()
                semi = min.get_semi()

                if semi < bb.get_semi() :
                    bb.set_semi(semi)

            bucket[bb_[bb.get_semi()]] += [bb]

            bb_father = bb.get_DFS_father()
            bb.link(bb_father)

            v = bucket[bb_father].pop(0)

            while v :
                min = v.eval()

                if min.get_semi() < v.get_semi() :
                    v.set_min_semi(min)
                else :
                    v.set_d_edge(bb_father)

                v = bucket[bb_father].pop(0)

        # 4

        for bb in tmp :
            min_semi = bb.get_min_semi()
            if min_semi :
                bb.set_d_edge( min_semi.get_d_parents() )

        # 5

        self._bb[1 : ] = []
        print(self._bb)
        self._DFS_1( self._bb[0] )

    piggy_bank = CmmPiggyBank()
    root = None

    @staticmethod
    def _visit(v) :
        v.set_visited()

        for lv in v.get_j_children() :
            if lv.get_level() <= root.get_level() :
                if not lv.is_in_phi() :
                    lv.set_in_phi()

                    if not lv.is_init() :
                        piggy_bank.put(lv)

        for lv in v.get_d_children() :
            if not lv.is_visited() :
                CmmStatement._visit(lv)

    def _insert_phis(self, var, bb_list) :
        max_level = max([x.get_level() for x in bb_list])

        piggy_bank.set_level( max_level )

        for lv in bb_list :
            lv.set_init()
            piggy_bank.put(lv)

        CmmStatement.root = piggy_bank.get()
        while CmmStatement.root :
            CmmStatement._visit(CmmStatement.root)
            CmmStatement.root = piggy_bank.get()

        for lv in self._bb :
            if lv.is_in_phi() :
                lv.insert_phi(var.new())

            lv.reset_flags()

    count = {}
    stack = {}

    @staticmethod
    def _DFS_2(bb) :
        print "DFS DFS DFS DFS DFS DFS DFS DFS\n"

        t0 = bb.get_instructions()
        t1 = bb.get_instructions()

        for instr in bb.get_instructions() :
            print ("instr = " + str(instr.body) + "\n")
            if not instr.is_phi() :
                print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n")
                print("_DFS_2 " + str(instr) + " " + str(instr.body) + "\n")
                for lv in instr.get_operands_ext() :
                    print ("op 0 = " + str(lv) + "\n")
                    if lv.is_a_var() :
                        print ("op 1 = " + str(lv.get_orig_id()) + "\n")
                        def_ = CmmStatement.stack[lv.get_orig_id()]
                        def_ = def_ and def_[-1]
                        if def_ :
                            lv.set_SSA_index(def_.get_SSA_index())
                            lv.set_def(def_)
                            def_.set_use(lv)
                            print ("d " + str(def_) + ", u " + str(lv) + "\n")
                        elif lv.is_flag_var() :
                            CmmError.error_message('ERR_USE_OF_UNSET_FLAG')
                            return
                        else :
                            CmmError.error_message('ERR_USE_OF_UNDEFINED_VAR', lv.get_name())
                            return

            for lv in t0 :
                print("_rename 0 " + str(lv.body) + "\n")

            for lv in instr.get_left_ext() :
                if lv.is_a_var() :
                    print("left = " + str(lv.get_orig_id()) + "\n")
                    var_id = lv.get_orig_id()
                    CmmStatement.count[var_id] += 1
                    lv.set_SSA_index(CmmStatement.count[var_id])
                    CmmStatement.stack[var_id] += [lv]
                    print ("push " + str(CmmStatement.count[var_id]) + " to " + str(var_id) + "\n")

            for lv in t1 :
                print("_rename 1 " + str(lv.body) + "\n")

        names = bb.get_orig_children_names()

        for lv in bb.get_orig_children() :
            name = names.pop(0)

            for phi in lv.get_phi_functions() :
                [new_operand] = phi.get_left_ext()
                new_operand = new_operand.new()

                stack_ = CmmStatement.stack.has_key(new_operand.get_orig_id()) and\
                         CmmStatement.stack[new_operand.get_orig_id()]

                if stack_ :
                    top = stack_[-1]

                    new_operand.set_SSA_index(top.get_SSA_index())
                    new_operand.set_def(top)
                    top.set_use(new_operand)

                    phi.set_phi_operand(name, new_operand)
                else :
                    phi.set_phi_operand(name, CmmConst(0))

        for lv in bb.get_d_children() :
            CmmStatement._DFS_2(lv)

        for instr in bb.get_instructions() :
            for lv in instr.get_left_ext() :
                if lv.is_a_var() and CmmStatement.stack.has_key(lv.get_orig_id()) :
                    print("pop from " + str(lv.get_orig_id()) + "\n")
                    CmmStatement.stack[lv.get_orig_id()].pop()

    def _rename(self) :
        CmmStatement.count = {}
        CmmStatement.stack = {}

        print "RENAME RENAME RENAME RENAME RENAME RENAME RENAME RENAME RENAME S T A R T\n"

        if self._bb :
            for lv in self._bb[0].get_instructions() :
                print("_rename " + str(lv.body) + "\n")
            CmmStatement._DFS_2( self._bb[0] )

        if CmmError.there_are_errors() :
            CmmError.prnt()
            sys.exit()

        print "RENAME RENAME RENAME RENAME RENAME RENAME RENAME RENAME RENAME F I N I S H\n"

    def transform_to_SSA_form(self) :
        self._build_basic_blocks()
        self._build_dj_graph()

        instr_ = self.get_instruction()

        for instr in instr_ :
            for op in instr.get_left_ext() :
                if op.is_a_var() :
                    bb = instr.get_bb()
                    self._variable_left_instance[ op.get_id() ]._bb[bb] = bb
                    self._variable_left_instance[ op.get_id() ]._var = op

        for var_id in self._variable_left_instance :
            var = self._variable_left_instance[var_id]._var
            bb_list = self._variable_left_instance[var_id]._bb.values()
            self._insert_phis(var, bb_list)

            CmmStatement.count[var_id] = 0
            CmmStatement.stack[var_id] = []

        self._rename()

        for lv in self._bb :
            for x in lv.get_instructions() :
                x.set_instr_for_variables()

#==================================================================================================#
#=============================== transform_to_SSA_form finish =====================================#
#==================================================================================================#

####################################################################################################
################################ dead_code_elimination start #######################################
####################################################################################################

    def _DFS_0_r(self, bb) :
        for lv in bb.get_orig_parents() :
            if lv.init_semi() :
                lv.set_DFS_father(bb)
                self._bb_r += [lv]
                self._DFS_0_r(lv)

    def _DFS_1_r(self, bb) :
        for lv in bb.get_d_parents() :
            self._bb_r += [lv]
            lv.set_level(bb.get_level() + 1)
            self._DFS_1_r(lv)

    def _build_dj_graph_r(self) :
        bucket = {}

        # 1

        for bb in self._bb :
            bb.init()

        CmmBasicBlock.reset_semi()

        self._bb_r = [ self.get_instruction()[-1].get_bb() ]
        self._DFS_0_r( self._bb_r[0] )

        # 2 & 3

        tmp = bb_ = self._bb_r
        tmp.pop(0)

        for bb in bb_ :
            bucket[bb] = []

        for bb in reverse(tmp) :
            for lv in bb.get_orig_children() :
                min = lv.eval()

                semi = min.get_semi()
                if (semi < bb.get_semi()) :
                    bb.set_semi(semi)

            bucket[bb_[bb.get_semi()]] += [bb]

            bb_father = bb.get_DFS_father()
            bb.link(bb_father)

            v = bucket[bb_father].pop(0)

            while v :
                min = v.eval()

                if (min.get_semi() < v.get_semi()) :
                    v.set_min_semi(min)
                else :
                    v.set_d_edge_r(bb_father)

                v = bucket[bb_father].pop(0)

        # 4

        for bb in tmp :
            min_semi = bb.get_min_semi()
            if min_semi :
                bb.set_d_edge_r( min_semi.get_d_children() )

        # 5

        self._bb_r[1 : ] = []
        self._DFS_1_r( self._bb_r[0] )

    piggy_bank_ = CmmPiggyBank()
    root_ = None

    @staticmethod
    def _visit_r(v) :
        v.set_visited()

        for lv in v.get_j_parents() :
            if lv.get_level() <= CmmStatement.root_.get_level() :
                if not lv.is_in_phi() :
                    lv.set_in_phi()

                    if not lv.is_init() :
                        CmmStatement.piggy_bank_.put(lv)

        for lv in v.get_d_parents() :
            if not lv.is_visited() :
                CmmStatement._visit_r(lv)

    def _build_RIDF(self, bb) :
        RIDF = {}

        CmmStatement.piggy_bank_.set_level( bb.get_level() )
        bb.set_init()
        CmmStatement.piggy_bank_.put(bb)

        CmmStatement.root_ = CmmStatement.piggy_bank_.get()

        while CmmStatement.root_ :
            CmmStatement._visit_r(CmmStatement.root_)
            CmmStatement.root_ = CmmStatement.piggy_bank_.get()

        for lv in self._bb_r :
            if lv.is_in_phi() :
                RIDF[lv] = lv

            lv.reset_flags()

        bb.set_RIDF(RIDF.values())

    def _build_RIDFs(self) :
        self._build_dj_graph_r()

        for lv in self._bb_r :
            self._build_RIDF(lv)

    def _find_dead_instructions(self, worklist) :
        list = []

        for instr in worklist :
            print ("INSTRustion " + str(instr.body) + "\n")
            for op in instr.get_operands_ext() :
                print ("_find_dead_instructions op " + str(op) + "\n")
                if op.is_a_var() :
                    for def_ in op.get_def() :
                        print("_find_dead_instructions def_ " + str(def_) + "\n")
                        ins = def_.get_instr()
                        print("instr " + str(ins.body) + "\n")

                        if ins.is_dead() :
                            ins.set_alive()
                            worklist += [ins]

            # take care of non conditional jumps

            for ins in instr.get_parents() :
                print ("parent " + str(ins.body) + " next >>> " + str(ins.get_next()) + " <<< YYOO\n")
                if ins.is_goto() and not ins.is_a_conditional_jump() and ins.get_next() != instr :
                    list += [ins]

            for bb in instr.get_bb().get_RIDF() :
                ins = bb.get_last_orig_instruction()

                if ins.is_dead() :
                    ins.set_alive()
                    worklist += [ins]

        instr = list.pop(0)
        while instr :
            if filter(lambda x: not x.is_dead(), instr.get_bb().get_instructions()) :
                instr.set_alive()

                for ins in instr.get_parents() :
                    if ins.is_goto() and not ins.is_a_conditional_jump() and ins.get_next() != instr :
                        list += [ins]
            else :
                for bb in instr.get_bb().get_RIDF() :
                    ins = bb.get_last_orig_instruction()

                    if ins.is_a_conditional_jump() :
                        instr_ = self.get_instruction()

                        for i in reversed(range(len(instr_))) :
                            if (instr[i] == ins) :
                                break

                        if instr[i + 1] == instr.get_bb().get_first_orig_instruction() :
                            ins.invert_comparison()
                            ins.set_instr_label(instr.get_instr_label())
                            break

            instr = list.pop(0)

    def _delete_unused_labels(self) :
        label = {}
        instr = {}

        for instr_ in self.get_instruction() :
            for lv in instr_.get_labels() :
                if lv.is_ifelse_label() or lv.is_while_label() or lv.is_for_label() or lv.is_do_label() :
                    label[lv] = lv
                    instr[lv] = instr_

        for instr_ in self.get_instruction() :
            if (instr_.is_goto()) :
                label_ = instr_.get_instr_label()

                del label[label_]
                del instr[label_]

        for lv in label :
            print("delete unused label " + str(label[lv]) + "\n")
            instr[lv].delete_label(label[lv])

    def _delete_redundant_placeholders(self) :
        for bb in self._bb :
            for instr in bb.get_instructions() :
                if instr.is_placeholder() :
                    if not instr.get_label() :
                        instr.get_bb().delete_instruction(instr)
                        self.kill_instruction(instr)

    def __tmp_print(self) :
        instr = self.get_instruction()

        for lv in instr :
            print(str(lv.get_body()))

    def coalesce(self, left, right) :
        sign = left.is_signed()

        new_left_def = {}
        new_left_use = {}

        left_size = left.get_size()
        right_size = right.get_size()

        left_handler = left.get_handler()
        d_handler = None

        usage = right.get_usage() | left.get_usage()
        print ("coalesce " + str(left) + " " + str(right) + " usage is " + str(usage) + "\n")
        right_def = right.get_def()
        left_use = left.get_use()

        (msb, lsb) = right.get_bit_range()

        if left_size < right_size :
            right.set_bit_range(lsb - left_size + 1, lsb)

        for d in right_def :
            d.set_dynamic_field(lsb - left_size + 1, lsb)

        for u in left.get_web('use') :
            print ("define new_left_use{" + str(u.get_handler()) + "}\n")
            new_left_use[u.get_handler()] = new = right.new()

            instr = u.get_instr()

            new.set_instr(instr)

            if sign :
                print ("set " + str(new) + " signed\n")
                new.set_signed()
            else :
                print ("set " + str(new) + " unsigned\n")
                new.set_unsigned()

            print ("replace in instruction u0 " + str(instr.body) + "\n")
            instr.replace_in_instruction(u, new)
            print ("replace in instruction u1 " + str(instr.body) + "\n")

        for d in left.get_web('def') :
            d_handler = d.get_handler()

            if d_handler != left_handler :
                new = right.new()
                new_left_def[d_handler] = [ d, new ]

                instr = d.get_instr()

                new.set_instr(instr)

                print("replace in instruction d0 " + str(instr.body) + "\n")
                instr.replace_in_instruction(d, new)
                print("replace in instruction d1 " + str(instr.body) + "\n")

        for lv in new_left_def :
            [old_d, d] = new_left_def[lv]

            for x in old_d.get_use() :
                u = new_left_use[x.get_handler()]
                d.set_use(u)
                u.set_def(d)

            d.set_dynamic_field(lsb - left_size + 1, lsb)

        left.delete_ud_chain()
        right.delete_du_chain()

        for d in right_def :
            for lv in left_use :
                u = new_left_use[lv.get_handler()]
                d.set_use(u)
                u.set_def(d)

        for lv in right_def[0].get_web() :
            print ("set usage of " + str(lv) + " [ " + str(lv.get_handler()) + " ] eq to " + str(usage) + "\n")
            lv.set_usage(usage)

    def delete_dead_instructions(self) :
        number_of_killed_instructions = 0

        # 1. Delete dead instructions

        for bb in self._bb :
            for instr in bb.get_instructions() :
                if (instr.is_dead()) :
                    print ("instruction_is_dead " + str(instr.body) + "\n")
                    if (instr.is_phi()) :
                        instr.get_bb().delete_phi(instr)
                    else :
                        instr.get_bb().delete_instruction(instr)

                        self.kill_instruction(instr)
                        number_of_killed_instructions += 1
                else :
                    print ("instruction_is_not_dead " + str(instr.body) + "\n")

        # 2. Delete redundant 'goto' instructions, unused labels and redundant placeholders

        i = len(self.get_instruction()) - 1
        while i > 0 :
            instr = self._instruction[i - 1]

            if instr.is_goto() and not instr.is_jsr() :
                label = instr.get_instr_label()

                next_instr = self._instruction[i]

                if next_instr.is_labeled_with(label) :
                    if not filter(lambda x: x != instr, next_instr.get_parents()) :
                        if label.is_ifelse_label() or label.is_while_label() or label.is_for_label() or label.is_do_label() :
                            next_instr.delete_label(label)

                    instr.get_bb().delete_instruction(instr)

                    self.kill_instruction(instr)
                    number_of_killed_instructions += 1
                    i += 1
                elif next_instr.is_goto() and not next_instr.is_a_conditional_jump() :
                    next_label = next_instr.get_instr_label()

                    if label.is_else_label() and next_label.is_if_label() :
                        self._instruction[i + 1].delete_label(label)
                        self.kill_instruction(next_instr)

                        instr.replace_in_instruction(label, next_label)
                        instr.invert_comparison()

                        number_of_killed_instructions += 1
            i -= 1

        self._delete_unused_labels()
        self._delete_redundant_placeholders()

        # 3. Delete empty Basic Blocks and all edges in BB_CFG

        i = 0
        while i < len(self._bb) :
            bb = self._bb[i]

            for child_name in bb.get_children_names() :
                child = bb.get_child(child_name)

                bb.delete_child(child_name)
                child.delete_parent(child_name)

            if not bb.get_instructions() :
                for child_name in bb.get_orig_children_names() :
                    child = bb.get_child(child_name)

                    for lv in child.get_phi_functions() :
                        lv.delete_phi_operand(child_name)

                self._bb[i : i + 1] = []
                i -= 1

            i += 1

        # 4. Concatenate Basic Blocks

        print "Concatenate Basic Blocks\n"

        i = 0
        while i < len(self._bb) :
            bb = self._bb[i]

            instr = bb.get_first_orig_instruction()
            if instr :
                print ("first instruction " + str(instr.body) + "\n")
                parents = instr.get_parents()
                if (len(parents) == 1 and len(parents[0].get_children()) == 1) :
                    for lv in bb.get_phi_functions() :
                        [left] = lv.get_left()
                        [right] = lv.get_right()

                        if left != right : # they are equal before register allocation
                            self.coalesce(left, right)

                        bb.delete_phi(lv)

                    print(str(parents[0].body) + "\n")

                    parent_bb = parents[0].get_bb()

                    for lv in bb.get_orig_instructions() :
                        lv.set_bb(parent_bb)
                        parent_bb.add_instruction(lv)

                    self._bb[i : i + 1] = []
                    i -= 1

            i += 1

        # 5. Build Edges in Basic Blocks Graph

        for instr in self.get_instruction() :
            bb0 = instr.get_bb()

            names = instr.get_children_names()

            for tmp in instr.get_children() :
                bb1 = tmp.get_bb()

                if bb0 == bb1 and tmp != bb0.get_first_orig_instruction() :
                    names.pop(0)
                else :
                    CmmBasicBlock.concat(bb0, bb1, names.pop(0))

        # 6. Build new DJ graph

        self._build_dj_graph()

        return number_of_killed_instructions

    def dead_code_elimination(self) :
        init = []

        for lv in self._bb :
            for x in lv.get_instructions() :
                print ("DCE " + str(x) + " " + str(x.body) + " | " + str(x._dead) + "\n")
                if x.is_orig_asm_instruction() or x.is_function_call_instruction() or x.is_store() or x.is_placeholder() or x.is_cmd() :
                    init += [x]
                else :
                    dead = 1

                    for y in x.get_dest_ext() :
                        print ("Left " + str(y) + "\n")
                        if y.is_a_reg() or y.is_an_arg() :
                            dead = 0
                            break

                    if dead :
                        x.set_dead()
                        print ("ABC " + str(x) + " " + str(x.body) + " | " + str(x._dead) + "\n")
                    else :
                        print ("ALIVE " + str(x) + " " + str(x.body) + " | " + str(x._dead) + "\n")
                        init += [x]

        print "dead_code_elimination start\n"
        for lv in self.get_instruction() :
            print('d ' if lv.is_dead() else 'a ',)
            print(str(lv.body) + "\n")
        print "dead_code_elimination finish\n"

        self._build_RIDFs()
        print "DEAD_CODE_ELIMINATION start\n"
        for lv in self.get_instruction() :
            print('d ' if lv.is_dead() else 'a ',)
            print(str(lv.body) + "\n")
        print "DEAD_CODE_ELIMINATION finish\n"
        self._find_dead_instructions(init)
        print "Dead_code_elimination start\n"
        for lv in self.get_instruction() :
            print('d ' if lv.is_dead() else 'a ',)
            print(str(lv.body) + "\n")
        print "Dead_code_elimination finish\n"
        self.delete_dead_instructions()

#==================================================================================================#
#=============================== dead_code_elimination finish =====================================#
#==================================================================================================#

####################################################################################################
####################### sparse_conditional_constant_propagation start ##############################
####################################################################################################

    exec_edge = {} ;
    exec_node = {} ;
    node_worklist = () ;
    var_def_worklist = () ;

    @staticmethod
    def _scc_visit(instr) :
        print(str(instr.body) + "\n")

        left_ext = instr.get_left_ext()
        old_value = [ x.get_value() for x in left_ext ]

        instr.evaluate()

        i = 0

        print("left_ext " + str(left_ext) + "\n")

        for lv in left_ext :
            if CmmValueLattice.compare(old_value[i], lv.get_value()) < 0 :
                CmmStatement.var_def_worklist += [lv]
            i = i + 1

        cn = instr.get_children_names()
        children = instr.get_children()

        print "children cn\n"
        if len(cn) == 1 :
            if not CmmStatement.exec_edge.has_key(cn[0]) :
                CmmStatement.exec_edge[cn[0]] = children[0]

                if not CmmStatement.exec_node.has_key(children[0]) :
                    bb = children[0].get_bb()

                    if (bb.get_first_orig_instruction() == children[0]) :
                        for lv in bb.get_phi_functions() :
                            CmmStatement.exec_node[lv] = lv
                            CmmStatement.node_worklist += [lv]

                    CmmStatement.exec_node[children[0]] = children[0]
                    CmmStatement.node_worklist += [children[0]]
        elif len(cn) == 2 and instr.is_a_conditional_jump() :
            i = 0

            diff = instr.evaluated_to_different_values()
            jump = instr.evaluated_to_jump()
            label = instr.get_instr_label()

            for lv in cn :
                if not CmmStatement.exec_edge.has_key(lv) :
                    child = instr.get_child(lv)

                    if (diff or (jump and child.is_labeled_with(label)) or (not jump and not child.is_labeled_with(label))) :
                        CmmStatement.exec_edge[lv] = children[i]

                        if not CmmStatement.exec_node.has_key(children[i]) :
                            bb = children[i].get_bb()

                            if bb.get_first_orig_instruction() == children[i] :
                                for x in bb.get_phi_functions() :
                                    CmmStatement.exec_node[x] = x
                                    CmmStatement.node_worklist += [x]

                            CmmStatement.exec_node[children[i]] = children[i]
                            CmmStatement.node_worklist += [children[i]]
                i += 1
        else : # switch/case
            # FIXME: bug this should be completed
            pass

    def transform_phi_function_to_const_assignment(self, bb, instr, right) :
        [left] = instr.get_left_ext()
        left_use = {}

        for lv in left.get_use() :
            for x in lv.get_instr() :
                left_use[x] = x

        left_use_instructions = left_use.values()

        next_instr = None
        i = 0

        for ins in bb.get_orig_instructions() :
            if filter(lambda x: x == ins, left_use_instructions) :
                next_instr = ins
                break

            i = i + 1

        ins = CmmAssgnInstruction([ 'move', left.new(), right ])
        bb.insert_instr_at_position(ins, i)

        if not next_instr :
            next_instr = bb.get_last_orig_instruction().get_next()

        j = 0

        while self._instruction[j] != next_instr :
            j += 1

        self._instruction[j : j] = ins

        CmmInstruction.transfer_parents(next_instr, ins)
        CmmInstruction.concat_next(ins, next_instr)

        left.set_instr(ins)
        right.set_instr(ins)
        ins.set_bb(bb)

    def sparse_conditional_constant_propagation(self) :
        instr = self.get_first_instruction()

        CmmStatement.exec_edge = {}
        CmmStatement.exec_node = { instr : instr }
        CmmStatement.node_worklist = [ instr ]
        CmmStatement.var_def_worklist = []

        visit = {}

        print "SCC !\n"

        while CmmStatement.node_worklist or CmmStatement.var_def_worklist :
            instr = CmmStatement.node_worklist.pop(0)
            if instr :
                CmmStatement._scc_visit(instr)

            var = CmmStatement.var_def_worklist.pop(0)
            if var :
                visit = {}

                for lv in var.get_use() :
                    instr = lv.get_instr()
                    visit[instr] = instr

                for instr in visit.itervalues() :
                    if CmmStatement.exec_node.has_key(instr) :
                        CmmStatement._scc_visit(instr)

        unreachable = []

        for bb in self._bb :
            for instr in bb.get_instructions() :
                if CmmStatement.exec_node.has_key(instr) :
                    if instr.is_phi() :
                        op = [ y for x in instr.get_operands_ext()
                                     for y in x
                                         if CmmStatement.exec_node.has_key(y.get_instr()) ]

                        if instr.evaluated_to_ordinary_constant() and len(op) == 1 :
                            instr.clean_up_right_side_defs()
                            self.transform_phi_function_to_const_assignment(bb, instr, op)
                            print ("INSTR 0 " + str(instr.body) + " SET DEAD\n")
                            instr.set_dead()
                    else :
                        for lv in instr.get_children_names() :
                            if not CmmStatement.exec_edge.has_key(lv) :
                                child = instr.get_child(lv)

                                child.delete_parent(lv)
                                instr.delete_child(lv)

                                if not child.get_parents() :
                                    unreachable += [child]

                        if instr.is_a_conditional_jump() and not instr.is_orig_asm_instruction() :
                            child_ = instr.get_children()
                            if (instr.is_btj() or instr.is_btjr()) and len(child_) == 1 :
                                if instr.get_next() == child_[0] :
                                    instr.set_dead()
                                else :
                                    instr.clean_up_right_side_defs()
                                    self.substitute(instr, [ CmmAssgnInstruction([ 'mviw', 'pc', instr.get_instr_label() ]) ] )
                            elif instr.is_jsr() and instr.evaluated_to_jump() :
                                instr.clean_up_right_side_defs()
                                instr.set_bit(CmmConst(0))
                else :
                    print ("INSTR 1 " + str(instr.body) + " SET DEAD\n")
                    instr.set_dead()

        self.delete_dead_instructions()

#==================================================================================================#
#====================== sparse_conditional_constant_propagation finish ============================#
#==================================================================================================#

####################################################################################################
############################### transform_from_SSA_form start ######################################
####################################################################################################

    def transform_from_SSA_form(self) :
        bb = None
        phi = None
        left = None
        left_SSA_index = None
        left_def = None
        left_use = None
        op = None
        op_def
        op_use
        op_def = None
        op_use = None
        def_ = None
        use = None
        left_use
        phi_operands

        for bb in self._bb :
            for phi in bb.get_phi_functions() :
                [left] = phi.get_left()
                left_SSA_index = left.get_SSA_index()
                left_use = left.get_use()
                phi_operands = phi.get_operands()

                for op in phi_operands :
                    for lv in op.get_web() :
                        lv.set_SSA_index(left_SSA_index)

                for op in phi_operands :
                    for op_def in op.get_def() :
                        for left_use in left_use :
                            left_use.set_def(op_def)
                            op_def.set_use(left_use)

                    op.delete_du_chain()

                left.delete_ud_chain()

                if 0 :
                    left_def = left
                    [left_use] = left.get_use()

                    for op in phi.get_operands() :
                        op_def_ = op.get_def()
                        op_use_ = op_def_[0].get_use()

                        for lv in op_def_ :
                            lv.set_SSA_index(left_SSA_index)

                        for lv in op_use_ :
                            lv.set_SSA_index(left_SSA_index)

                        for op_def in op_def_ :
                            for use in left_def.get_use() :
                                use.set_def(op_def)
                                op_def.set_use(use)

                        for op_use in op_use_ :
                            for def_ in left_use.get_def() :
                                def_.set_use(op_use)
                                op_use.set_def(def_)

                        left_def = op_def_[0]

                        op.delete_du_chain()

                    left.delete_ud_chain()

#==================================================================================================#
#============================== transform_from_SSA_form finish ====================================#
#==================================================================================================#

    def _fix_64_bit_arithmetic(self) :
        for instr in self.get_instruction() :
            instr.fix_64_bit_arithmetic()

    def _fix_pcr(self) :
        i = len(self._instruction) - 1

        while i > 0 :
            instr = self._instruction[i]

            label = instr.is_pcr()
            if label :
                self._instruction[i - 1].fix_pcr(label)
                self.kill_instruction(instr)
                i -= 1

            i -= 1

    def _fix_size_operand(self) :
        for lv in self.get_instruction() :
            lv.fix_size_operand()

    def postprocess_asm_form(self) :
        for lv in self.get_instruction() :
            print ("postprocess_asm_form 0 " + str(lv) + " " + str(lv.get_labels()) + " " + str(lv.body) + "\n")
        self._fix_64_bit_arithmetic()
        for lv in self.get_instruction() :
            print ("postprocess_asm_form 1 " + str(lv) + " " + str(lv.get_labels()) + " " + str(lv.body) + "\n")
        self._fix_pcr()
        self._fix_size_operand()

    def generate_asm_code(self) :
        code = ''
        number_of_instructions = 0

        for instr in self._instruction :
            for lv in reversed(instr.get_labels()) :
                code += lv.get_id() + "\n"

            print("generate_asm_code " + str(instr) + " " + str(instr.body) + "\n")
            tmp = instr.get_asm_code()
            code += ''.join([ str(x) + "\n" for x in tmp])
            number_of_instructions += len(tmp)

        return [code, number_of_instructions]
