import copy
from CmmSymbol import CmmSymbol
from CmmType import CmmType
from CmmConst import CmmConst
from CmmValueLattice import CmmValueLattice

class CmmVar(CmmSymbol) :
    _handler = 0

    def _get_next_handler(self) :
        h = CmmVar._handler
        CmmVar._handler = CmmVar._handler + 1
        return h

    def __init__(self, function, name, type = None, scope = None, value = None, msb_bit = None, lsb_bit = None) :
        if not (type) : type = CmmType('u', 0)
        if not (scope) : scope = 0
        if not (value) : value = CmmValueLattice('no_data')
        # $value = CmmConst->new(0L) unless (defined $value) ;

        self._handler = self._get_next_handler()
        self._function = function
        self._name = name
        self._type = type
        self._types = []
        self._scope = scope
        self._value = value
        self._usage = 0
        self._statements_to_evaluate = []
        self._SSA_index = 0
        self._def = []
        self._use = []
        self._dead = 0
        self._instr = None
        self._reg = None
        self._cc = 0L
        self._copy = {}

        if (msb_bit) : self.set_bit_range(msb_bit, lsb_bit)

        self._init()
        self._set_id()

    def new(self, copy_number = None) :
        ##### was changed 05.21.02
        if (copy_number) :
            if (self._copy.has_key(copy_number)) :
                return self._copy[copy_number]
            else :
                c = copy.deepcopy(self)

                c._handler = self._get_next_handler()
                c._type = c._type.new()
                c._types = []
                c._statements_to_evaluate = []
                c._def = []
                c._use = []

                if (c._reg) :
                    c._reg = c._reg.new()

                c._set_cc(copy_number)
                c._set_id()
                self._copy[copy_number] = c
        else :
            c = copy.deepcopy(self)

            c._handler = self._get_next_handler()
            c._type = c._type.new()
            c._types = []
            c._statements_to_evaluate = []
            c._def = []
            c._use = []

            if (c._reg) :
                c._reg = c._reg.new()

        return c

    def set_name(self, name) :
        self._name = name

    def get_name(self) :
        return self._name

    def set_scope(self, scope) :
        self._scope = scope

    def get_scope(self) :
        return self._scope

    def change_cc(self, value) :
        self._cc = value
        self._set_id()

    def _set_cc(self, value) :
        self._cc = value

    def _get_cc(self) :
        return self._cc

    def set_SSA_index(self, SSA_index) :
        self._SSA_index = SSA_index
        self._set_id()

    def get_SSA_index(self) :
        return self._SSA_index

    def _set_id(self) :
        # print ("function is " . $self->{_function} . "\n") ;
        # print ("name is " . $self->{_name} . "\n") ;
        ##### print ("cc is " . $self->{_cc} . "\n") ;
        # print ("scope is " . $self->{_scope} . "\n") ;
        # print ("SSA index is " . $self->{_SSA_index} . "\n") ;

        ##### $self->{_id} = $self->{_function} . '#' . $self->{_cc} . '#' . $self->{_scope} . '#' . $self->{_name} ;
        # $self->{_id} = $self->{_function} . '#' . $self->{_scope} . '#' . $self->{_name} . '#' . $self->{_SSA_index} ;
        self._id = self._function + '#' + str(self._scope) + '#' + str(self._cc) + '#' + str(self._name) + '#' + str(self._SSA_index)

    def get_id(self, function = None, name = None, scope = None) :
        if (function == None) :
            return self._id
        else :
            ##### $function . '#' . 0 . '#' . $scope . '#' . $name ;
            # $function . '#' . $scope . '#' . $name . '#' . 0 ;
            return function + '#' + str(scope) + '#' + str(0) + '#' + name + '#' + str(0)

    def get_orig_id(self) :
        # $self->{_function} . '#' . $self->{_scope} . '#' . $self->{_name} . '#' . 0 ;
        return self._function + '#' + str(self._scope) + '#' + str(self._cc) + '#' + str(self._name) + '#' + str(0)

    def set_usage(self, usage) :
        self._usage = usage

    def get_usage(self) :
        return self._usage

    def set_bit_range(self, *args) :
        self._type.set_bit_range(*args)
        # push @{$self->{_statements_to_evaluate}}, @statements ;

    def get_bit_range(self, *args) :
        return self._type.get_bit_range(*args)

    def get_statements_to_evaluate(self) :
        z = self._statements_to_evaluate[ : ]
        self._statements_to_evaluate[ : ] = []
        return z

    def get_type(self) :
        return self._type

    def set_type(self, type) :
        self._type = type

    def get_sign(self) :
        return self._type.get_sign()

    def set_def(self, df) :
        self._def += [df]

    def get_def(self) :
        return self._def

    def set_use(self, use) :
        self._use += [use]

    def get_use(self) :
        return self._use

    def get_web(self, subset = None) :
        d = None
        u = None
        defi = {}
        use = {}

        df = self.get_def()
        use = self.get_use()
        prev = []

        if (len(use)) :
            while (len(use) != len(prev)) :
                prev = use

                for lv in use :
                    for d in lv.get_def() :
                        defi[d.get_handler()] = d

                df = defi.values()

                for lv in df :
                    for u in lv.get_use() :
                        use[u.get_handler()] = u

                use = use.values()
        else :
            while (len(df) != len(prev)) :
                prev = df

                for lv in df :
                    for u in lv.get_use() :
                        use[u.get_handler()] = u

                use = use.values()

                for lv in use :
                    for d in lv.get_def() :
                        defi[d.get_handler()] = d

                df = defi.values()

        if (subset) :
            return df if (subset == 'def') else use
        else :
            return [df, use]

    def set_dead(self) :
        self._dead = 1

    def is_dead(self) :
        return self._dead

    def delete_ud_chain(self) :
        use = None

        self.set_dead()

        for use in self.get_use() :
            use.delete_dead_def()

        self._use = []

    def delete_dead_def(self) :
        i = 0
        numb_of_defs = len(self._def)

        while i < numb_of_defs :
            if (self._def[i].is_dead()) :
                self._def[i : i + 1] = []
                i = i - 1
                numb_of_defs = numb_of_defs - 1
                # return ;
            i = i + 1

    def delete_du_chain(self) :
        df = None

        self.set_dead()

        for df in self.get_def() :
            df.delete_dead_use()

        self._def = []

    def delete_dead_use(self) :
        i = 0
        numb_of_uses = len(self._use)

        while i < numb_of_uses :
            if (self._use[i].is_dead()) :
                self._use[i : i + 1] = []
                i = i - 1
                numb_of_uses = numb_of_uses - 1
                # return ;
            i = i + 1

    def set_signed(self) :
        return self._type.set_signed()

    def is_signed(self) :
        return self._type.is_signed()

    def set_unsigned(self) :
        return self._type.set_unsigned()

    def is_unsigned(self) :
        return self._type.is_unsigned()

    def is_a_var(self) :
        return 1

    def is_flag_var(self) :
        return 0

    def is_an_arg(self) :
        return self._scope == 1

    def is_a_return_var(self) :
        return self._scope == 3

    def get_function_name(self) :
        return self._function

    def get_SSA_id(self, function = None, name = None, scope = None, SSA_index = None) :
        if (function == None) :
            return self._SSA_id
        else :
            ##### $function . '#' . 0 . '#' . $scope . '#' . $name ;
            return function + '#' + str(scope) + '#' + name + '#' + str(SSA_index)

    def set_instr(self, instr) :
        self._instr = instr

    def get_instr(self) :
        return self._instr

    def set_reg(self, reg) :
        size = self.get_size()
        reg_size = reg.get_size()

        if (size < reg_size) :
            reg_1 = None

            (reg_1, reg) = reg.splt_into_2() # 32

            if (size < reg.get_size()) :
                (reg_1, reg) = reg.splt_into_2() # 16

                if (size < reg.get_size()) :
                    (reg_1, reg) = reg.splt_into_2() # 8

        self._reg = reg

    def get_reg(self) :
        return self._reg

    def set_size(self, size) :
        reg = self.get_reg()

        (reg.set_size(size)) if reg else super(CmmVar, self).set_size(size)

    def get_size(self) :
        reg = self.get_reg()

        return (reg.get_size()) if reg else super(CmmVar, self).get_size()

    def can_be_put_in_RF(self) :
        return self.get_usage() != 1

    def put_in_RF(self) :
        self.set_usage(2)

    def destroy_instr(self) :
        self._instr = None

    def set_static_field(self, *args) :
        return self._type.set_static_field(*args)

    def set_dynamic_field(self, *args) :
        return self._type.set_dynamic_field(*args)

    def get_static_fields(self) :
        return self._type.get_static_fields()

    def get_dynamic_fields(self) :
        return self._type.get_dynamic_fields()

    def is_nonalligned(self) :
        return self._type.is_nonalligned()

    def get_mask(self) :
        return (CmmConst(1) << self._type.get_size()) - 1

    def get_shift_correction_value(self) :
        return (8 - (self._type.get_lsb() & 0x7)) & 0x7

    def _get_number_of_bytes_the_var_intersects_with(self) :
        return (self._type.get_lsb() >> 3) - (self._type.get_msb() >> 3) + 1

    def get_zero(self) :
        num_of_bytes = self._get_number_of_bytes_the_var_intersects_with()
        mask = None

        if (num_of_bytes <= 1) :
            mask = CmmConst(0xffL)
        elif (num_of_bytes <= 2) :
            mask = CmmConst(0xffffL)
        elif (num_of_bytes <= 4) :
            mask = CmmConst(0xffffffffL)
        elif (num_of_bytes <= 8) :
            mask = CmmConst(0xffffffffffffffffL)
        else :
            mask = CmmConst(0xffffffffffffffffffL)

        return ((self.get_mask() << self.get_shift_correction_value()) ^ mask) & mask

    def set_envelope(self) :
        (msb, lsb) = self.get_bit_range()
        num_of_bytes = self._get_number_of_bytes_the_var_intersects_with()
        envelope_size = None

        if (num_of_bytes <= 1) :
            envelope_size = 8
        elif (num_of_bytes <= 2) :
            envelope_size = 16
        elif (num_of_bytes <= 4) :
            envelope_size = 32
        elif (num_of_bytes <= 8) :
            envelope_size = 64
        else :
            envelope_size = 72

        msb = (msb >> 3) << 3
        lsb = msb + envelope_size - 1

        self.set_bit_range(msb, lsb)

    def get_envelope_size(self) :
        return self.get_size()

    def _init(self) :
        pass

    def get_handler(self) :
        return self._handler

    def __eq__(self, other) :
        # print "_f_equal form CmmVar\n" ;
        return self.get_handler() == other

    def __ne__(self, other) :
        # print "_f_nequal form CmmVar\n" ;
        return self.get_handler() != other
