/*
 * Copyright (c) 2007 George Verbitsky
 * All rights reserved.
 */

%token
################################################################################
"{"					'{'
# "<%"				'{'
"}"					'}'
# "%>"				'}'
"["					'['
# "<:"				'['
"]"					']'
# ":>"				']'
"("					'('
")"					')'
";"					';'
":"					':'
# "..."				ELLIPSIS
"?"					'?'
"."					'.'
"+"					'+'
"-"					'-'
# "*"					'*'
# "/"					'/'
# "%"					'%'
"^"					'^'
"&"					'&'
"|"					'|'
"~"					'~'
"!"					'!'
"="					'='
"<"					'<'
">"					'>'
"+="				ADDEQ
"-="				SUBEQ
# "*="				MULEQ
# "/="				DIVEQ
# "%="				MODEQ
"^="				XOREQ
"&="				ANDEQ
"|="				OREQ
"<<"				SHLEFT
">>"				SHRIGHT
"<<="				SHLEQ
">>="				SHREQ
"=="				EQ
"!="				NOTEQ
"<="				LTEQ
">="				GTEQ
"&&"				ANDAND
"||"				OROR
"++"				PLUSPLUS
"--"				MINUSMINUS
","					','
# "->"				ARROW

%keyword
################################################################################
mem					MEM

%sync_keyword
################################################################################
# auto				AUTO
break				BREAK
# case				CASE
char				CHAR
# const				CONST
continue			CONTINUE
# default				DEFAULT
do					DO
# double				DOUBLE
else				ELSE
# enum				ENUM
# extern				EXTERN
# float				FLOAT
for					FOR
goto				GOTO
if					IF
int					INT
long				LONG
# register			REGISTER
return				RETURN
short				SHORT
signed				SIGNED
# sizeof				SIZEOF
# static				STATIC
struct				STRUCT
# switch				SWITCH
# typedef				TYPEDEF
union				UNION
unsigned			UNSIGNED
void				VOID
# volatile			VOLATILE
while				WHILE
inline				INLINE
applet				APPLET
interrupt			INTERRUPT

%token2
################################################################################
[a-zA-Z_][a-zA-Z_0-9]*															IDENTIFIER
([0-9]*|0[xX][0-9a-fA-F]+)(([uU][lL]?)|([lL][uU]?))?							INTEGER
# ((([0-9]*\.[0-9]+)|([0-9]+\.))([eE][-+]?[0-9]+)?|[0-9]+[eE][-+]?[0-9]+)[fFlL]?	FLOATING
# '([^'\n]|\\[abfnrtv\\'"?])'														CHARACTER
\"([^"\n]|\\\"|\\\n)*\"															STRING

# %other_token
# TYPEDEF_NAME

%start
file

%%

/* B.2.1 Expressions. */

primary_expression:
	memref
	| var_name_or_reg_name
	# | identifier			'idetifier' was substituted with var_name_or_reg_name
	# | integer				this one was moved to postfix_expression
	# | CHARACTER
	# | FLOATING
	# | STRING
	# | '(' expression ')'	this one was moved to postfix_expression
	;

memref:
	MEM '[' expression ']'
	;

identifier:
	IDENTIFIER
	;

integer:
	INTEGER
	;

# 'postfix_expression' was redefined.
# There is no pointers (and pointers to functions in oartuicular).
# So we do not want id()->() be a syntaically correct construct.
# There is no arrays. So we don't want id[]->() be a syntaically correct construct.
# Nothing wrong with id().id construct,
# but we don't want id.id().id be a syntaically correct construct.
# because there is no function pointers in the language.
# So id().id although potentially valid won't be permitted for now.
# 
# postfix_expression:
	# primary_expression
	# | postfix_expression '[' expression ']'
	# | postfix_expression '(' argument_expression_list ')'
	# | postfix_expression '(' ')'
	# | postfix_expression '.' identifier
	# | postfix_expression ARROW identifier
	# | postfix_expression PLUSPLUS
	# | postfix_expression MINUSMINUS
	# ;

# assignment_expression_list was substituted with experession
# because they are syntactically identical
postfix_expression:
	primary_expression
	| identifier '(' expression ')'
	| identifier '(' ')'
	| '(' expression ')'
	| primary_expression PLUSPLUS
	| primary_expression MINUSMINUS
	| integer
	;

# argument_expression_list:
	# assignment_expression
	# | argument_expression_list ',' assignment_expression
	# ;

# there is no casting allowed in the language, and
# no multiplication, division, and modulo operations
# that is why multiplicative expression is essentially unary_expression
# and that is why the new rule for multiplicative_expression
# was introduced instead of three rules for
# unary_expression, cast_expression and multiplicative_expression

multiplicative_expression:
	postfix_expression
	| PLUSPLUS multiplicative_expression
	| MINUSMINUS multiplicative_expression
	| unary_operator multiplicative_expression
	;

# unary_expression:
	# postfix_expression
	# | PLUSPLUS unary_expression
	# | MINUSMINUS unary_expression
	# | unary_operator cast_expression
	# | SIZEOF unary_expression
	# | SIZEOF '(' type_name ')'
	;

unary_operator:
	# '&'
	# | '*'
	  '+'
	| '-'
	| '~'
	| '!'
	;

# cast_expression:
	# unary_expression
	# | '(' type_name ')' cast_expression
	# ;

# multiplicative_expression:
	# cast_expression
	# | multiplicative_expression '*' cast_expression
	# | multiplicative_expression '/' cast_expression
	# | multiplicative_expression '%' cast_expression
	# ;

additive_expression:
	multiplicative_expression
	| additive_expression '+' multiplicative_expression
	| additive_expression '-' multiplicative_expression
	;

shift_expression:
	additive_expression
	| shift_expression SHLEFT additive_expression
	| shift_expression SHRIGHT additive_expression
	;

relational_expression:
	shift_expression
	| relational_expression '<' shift_expression
	| relational_expression '>' shift_expression
	| relational_expression LTEQ shift_expression
	| relational_expression GTEQ shift_expression
	;

equality_expression:
	relational_expression
	| equality_expression EQ relational_expression
	| equality_expression NOTEQ relational_expression
	;

and_expression:
	equality_expression
	| and_expression '&' equality_expression
	;

exclusive_or_expression:
	and_expression
	| exclusive_or_expression '^' and_expression
	;

inclusive_or_expression:
	exclusive_or_expression
	| inclusive_or_expression '|' exclusive_or_expression
	;

logical_and_expression:
	inclusive_or_expression
	| logical_and_expression ANDAND inclusive_or_expression
	;

logical_or_expression:
	logical_and_expression
	| logical_or_expression OROR logical_and_expression
	;

conditional_expression:
	logical_or_expression
	| logical_or_expression '?' expression ':' conditional_expression
	;

conditional_expression_top:
	conditional_expression
	;

# multiplicative_expression is an unary_expression.
# See the definition and comment for multiplicative_expression.
# the rule for unary_expression was eliminated
# This is the reason for substituting unary_expression
# with multiplicative_expression below

assignment_expression:
	conditional_expression
	# | unary_expression assignment_operator assignment_expression
	| multiplicative_expression assignment_operator assignment_expression
	;

assignment_operator:
	'='
	# | MULEQ
	# | DIVEQ
	# | MODEQ
	| ADDEQ
	| SUBEQ
	| SHLEQ
	| SHREQ
	| ANDEQ
	| XOREQ
	| OREQ
	;

expression_top:
	expression
	;

expression:
	assignment_expression
	| expression ',' assignment_expression
	| error ',' assignment_expression
	;

# constant_expression:
	# conditional_expression
	# ;

declaration:
	data_type init_declarator_list ';'
	# FIXME: not clear why this rule exists | data_type ';'
	| data_type error error_sync
	;

data_type:
	declaration_specifiers
	;

declaration_specifiers:
	# storage_class_specifier declaration_specifiers
	# | storage_class_specifier
	  type_specifier declaration_specifiers
	| type_specifier
	# | type_qualifier declaration_specifiers
	# | type_qualifier
	;

init_declarator_list:
	init_declarator
	| init_declarator_list ',' init_declarator
	;

init_declarator:
	declarator
	| declarator '=' initializer
	;

# storage_class_specifier:
	# typedef
	# | extern
	# | static
	# | auto
	# | register
	;

type_specifier:
	void
	| char
	| short
	| int
	| long
	# | float
	# | double
	| signed
	| unsigned
	# | struct_or_union_specifier
	# | enum_specifier
	# | TYPEDEF_NAME
	;

# struct_or_union_specifier:
	# struct_or_union identifier '{' struct_declaration_list '}'
	# | struct_or_union '{' struct_declaration_list '}'
	# | struct_or_union identifier
	# ;

# struct_or_union:
	# struct
	# | union
	# ;

# struct_declaration_list:
	# struct_declaration
	# | struct_declaration_list struct_declaration
	# ;

# struct_declaration:
	# specifier_qualifier_list struct_declarator_list ';'
	# ;

specifier_qualifier_list:
	type_specifier specifier_qualifier_list
	| type_specifier
	# | type_qualifier specifier_qualifier_list
	# | type_qualifier
	;

# struct_declarator_list:
	# struct_declarator
	# | struct_declarator_list ',' struct_declarator
	# ;

# struct_declarator:
	# declarator
	# |  ':' constant_expression
	# | declarator ':' constant_expression
	# ;

# enum_specifier:
	# enum identifier '{' enumerator_list '}'
	# | enum '{' enumerator_list '}'
	# | enum identifier
	# ;

# enumerator_list:
	# enumerator
	# | enumerator_list ',' enumerator
	# ;

# enumerator:
	# identifier
	# | identifier '=' constant_expression
	# ;

# type_qualifier:
	# const
	# | volatile
	# ;

# declarator and direct_declarator are coalescing
# because there is no pointers in the language.

# declarator:
	# pointer direct_declarator
	# | direct_declarator
	# ;

# direct_declarator:
	# identifier
	# | '(' declarator ')'
	# | direct_declarator '[' constant_expression ']'
	# | direct_declarator '[' ']'
	# | direct_declarator '(' parameter_type_list ')'
	# | direct_declarator '(' identifier_list ')'
	# | direct_declarator '(' ')'
	# ;

# we split declarator into declarator and func_declarator
# the last one used only with function_definition
declarator:
	identifier
	# no pointers and so no pointers to functions | '(' declarator ')'
	# | declarator '[' constant_expression ']'
	# | declarator '[' ']'
	# | declarator '(' parameter_type_list ')'
	# | declarator '(' identifier_list ')'
	# | declarator '(' ')'
	;

func_declarator:
	  identifier '(' parameter_type_list ')'
	| identifier '(' ')'
	;

# pointer:
	# '*' type_qualifier_list
	# | '*'
	# | '*' type_qualifier_list pointer
	# | '*' pointer
	# ;

# type_qualifier_list:
	# type_qualifier
	# | type_qualifier_list type_qualifier
	# ;

parameter_type_list:
	parameter_list
	# | parameter_list ',' ELLIPSIS
	;

parameter_list:
	parameter_declaration
	| parameter_list ',' parameter_declaration
	;

parameter_declaration:
	data_type declarator
	# | declaration_specifiers abstract_declarator
	# | declaration_specifiers
	;

identifier_list:
	identifier
	| identifier_list ',' identifier
	;

type_name:
	specifier_qualifier_list
	# | specifier_qualifier_list abstract_declarator
	;

# abstract_declarator and direct_abstract_declarator are coalescing
# because there is no pointers in the language.

# abstract_declarator:
	# pointer
	# | direct_abstract_declarator
	# | pointer direct_abstract_declarator
	# ;

# direct_abstract_declarator:
	# '(' abstract_declarator ')'
	# | '[' ']'
	# | '[' constant_expression ']'
	# | direct_abstract_declarator '[' ']'
	# | direct_abstract_declarator '[' constant_expression ']'
	# | '(' ')'
	# | '(' parameter_type_list ')'
	# | direct_abstract_declarator '(' ')'
	# | direct_abstract_declarator '(' parameter_type_list ')'
	# ;

# abstract_declarator:
	# '(' abstract_declarator ')'
	# | '[' ']'
	# | '[' constant_expression ']'
	# | abstract_declarator '[' ']'
	# | abstract_declarator '[' constant_expression ']'
	# | '(' ')'
	# | '(' parameter_type_list ')'
	# | abstract_declarator '(' ')'
	# | abstract_declarator '(' parameter_type_list ')'
	# ;

initializer:
	# FIXME: it should be assignmenmt_expression(or assignment_expression_top) instead of conditional_expression_top
	# FIXME: initializer_list should be permitted
	conditional_expression_top
	# assignment_expression
	# | '{' initializer_list '}'
	# | '{' initializer_list ',' '}'
	;

# initializer_list:
	# initializer
	# | initializer_list ',' initializer
	;

/* B.2.3 Statements. */

statement:
	labeled_statement
	| compound_statement
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	| arch_dependent_statement
	| error error_sync
	;

labeled_statement:
	identifier ':' statement
	# | case constant_expression ':' statement
	# | default ':' statement
	;

compound_statement:
	'{' '}'
	| '{' statement_list '}'
	| '{' declaration_list '}'
	| '{' declaration_list statement_list '}'
	;

declaration_list:
	declaration
	| declaration_list declaration
	;

statement_list:
	statement
	| statement_list statement
	;

expression_statement:
	';'
	| expression_top ';'
	;

selection_statement:
	if '(' expression_top ')' statement
	| if '(' expression_top ')' statement else statement
	# FIXME: | switch '(' expression ')' statement
	;

iteration_statement:
	while '(' expression_top ')' statement
	| do statement while '(' expression_top ')' ';'
	| for '(' ';' ';' ')' statement
	| for '(' expression_top ';' ';' ')' statement
	| for '(' ';' expression_top ';' ')' statement
	| for '(' expression_top ';' expression_top ';' ')' statement
	| for '(' ';' ';' expression_top ')' statement
	| for '(' expression_top ';' ';' expression_top ')' statement
	| for '(' ';' expression_top ';' expression_top ')' statement
	| for '(' expression_top ';' expression_top ';' expression_top ')' statement
	# if the following rule is opened there will be 
	# problems with push_construct action
	# | do statement while '(' error ';'
	| for '(' error ';' ';' ')' statement
	| for '(' ';' error ';' ')' statement
	| for '(' error ';' expression_top ';' ')' statement
	| for '(' expression_top ';' error ';' ')' statement
	| for '(' error ';' error ';' ')' statement
	| for '(' ';' ';' error ')' statement
	| for '(' error ';' ';' expression_top ')' statement
	| for '(' expression_top ';' ';' error ')' statement
	| for '(' error ';' ';' error ')' statement
	| for '(' ';' error ';' expression_top ')' statement
	| for '(' ';' expression_top ';' error ')' statement
	| for '(' ';' error ';' error ')' statement
	| for '(' expression_top ';' expression_top ';' error ')' statement
	| for '(' expression_top ';' error ';' expression_top ')' statement
	| for '(' expression_top ';' error ';' error ')' statement
	| for '(' error ';' expression_top ';' expression_top ')' statement
	| for '(' error ';' expression_top ';' error ')' statement
	| for '(' error ';' error ';' expression_top ')' statement
	| for '(' error ';' error ';' error ')' statement
	;

jump_statement:
	goto identifier ';'
	| continue ';'
	| break ';'
	| return ';'
	| return expression_top ';'
	;

/* B.2.4 External definitions. */

file:
	translation_unit
	;

translation_unit:
	external_declaration
	| translation_unit external_declaration
	;

external_declaration:
	function_definition
	| declaration
	| error ext_error_sync
	;

function_definition:
# FIXME: supress temporarily old-style function definition
	# declaration_specifiers declarator declaration_list compound_statement
	  # substitute declarator with func_declarator
	  # because that will make syntax of function definition stricter
	  # declaration_specifiers declarator compound_statement
# func_type was added
	### func_type declaration_specifiers func_declarator compound_statement
	### | declaration_specifiers func_declarator compound_statement
	function_definition_prefix function_definition_args compound_statement
	# | declaration_specifiers func_declarator compound_statement
	# | declarator declaration_list compound_statement
	# | declarator compound_statement
	;

function_definition_prefix:
	func_type data_type identifier
	| data_type identifier
	;

function_definition_args:
	  '(' parameter_type_list ')'
	| '(' ')'
	;

func_type:
	inline
	| applet
	| interrupt
	;

inline:
	INLINE
	;

applet:
	APPLET
	;

interrupt:
	INTERRUPT
	;

# auto:
	# AUTO AUTO
	# | AUTO
	# ;

break:
	BREAK BREAK
	| BREAK
	;

# case:
	# CASE CASE
	# | CASE
	# ;

char:
	CHAR CHAR
	| CHAR
	;

# const:
	# CONST CONST
	# | CONST
	# ;

continue:
	CONTINUE CONTINUE
	| CONTINUE
	;

# default:
	# DEFAULT DEFAULT
	# | DEFAULT
	# ;

do	:
	DO DO
	| DO
	;

# double:
	# DOUBLE DOUBLE
	# | DOUBLE
	# ;

else:
	ELSE ELSE
	| ELSE
	;

# enum:
	# ENUM ENUM
	# | ENUM
	# ;

# extern:
	# EXTERN EXTERN
	# | EXTERN
	# ;

# float:
	# FLOAT FLOAT
	# | FLOAT
	# ;

for:
	FOR FOR
	| FOR
	;

goto:
	GOTO GOTO
	| GOTO
	;

if	:
	IF IF
	| IF
	;

int:
	INT INT
	| INT
	;

long:
	LONG LONG
	| LONG
	;

# register:
	# REGISTER REGISTER
	# | REGISTER
	# ;

return:
	RETURN RETURN
	| RETURN
	;

short:
	SHORT SHORT
	| SHORT
	;

signed:
	SIGNED SIGNED
	| SIGNED
	;

# static:
	# STATIC STATIC
	# | STATIC
	# ;

struct:
	STRUCT STRUCT
	| STRUCT
	;

# switch:
	# SWITCH SWITCH
	# | SWITCH
	# ;

# typedef:
	# TYPEDEF TYPEDEF
	# | TYPEDEF
	# ;

union:
	UNION UNION
	| UNION
	;

unsigned:
	UNSIGNED UNSIGNED
	| UNSIGNED
	;

void:
	VOID VOID
	| VOID
	;

# volatile:
	# VOLATILE VOLATILE
	# | VOLATILE
	# ;

while:
	WHILE WHILE
	| WHILE
	;

error_sync:
	';'
	# | AUTO
	| BREAK
	# | CASE
	| CHAR
	# | CONST
	| CONTINUE
	# | DEFAULT
	| DO
	# | DOUBLE
	| ELSE
	# | ENUM
	# | EXTERN
	# | FLOAT
	| FOR
	| GOTO
	| IF
	| INT
	| LONG
	# | REGISTER
	| RETURN
	| SHORT
	| SIGNED
	# | STATIC
	| STRUCT
	# | SWITCH
	# | TYPEDEF
	| UNION
	| UNSIGNED
	| VOID
	# | VOLATILE
	| WHILE
	;

ext_error_sync:
	';'
	| CHAR
	# | CONST
	# | DOUBLE
	# | ENUM
	# | EXTERN
	# | FLOAT
	| INT
	| LONG
	| SHORT
	| SIGNED
	# | STATIC
	| STRUCT
	# | TYPEDEF
	| UNION
	| UNSIGNED
	| VOID
	;
%%
