from CmmConst import CmmConst
from CmmInstruction import CmmInstruction

class CmmAssgnInstruction(CmmInstruction) :
    def get_opA(self) :
        return self.body[2]

    def set_opA(self, opA) :
        self.body[2] = opA

    def get_opC(self) :
        return self.body[1]

    def set_opC(self, opC) :
        self.body[1] = opC

    def is_copy(self) :
        return self.body[0] == 'move'

    def set_instr_label(self, label) :
        while len(self.body) < 3:
            self.body.append(None)
        self.body[2] = label

    def get_instr_label(self) :
        return self.body[2] if len(self.body) > 2 else None

    def is_a_too_big_to_small_assignment(self) :
        if self.body[0] == 'move' :
            opC = self.body[1]
            opA = self.body[2]

            if opC.is_a_var() and opA.is_a_var() :
                opC_size = opC.get_size()
                opA_size = opA.get_size()

                return (opC_size < opA_size and opC_size < 32)

    def fix_bitfield_def(self) :
        opC = self.body[1]
        opA = self.body[2]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if opC.is_a_var() and opC.is_nonalligned() :
            if opA.is_a_var() and opC.get_id() == opA.get_id() :
                self.set_dead()

                return [ self ]
            else :
                instr = []

                mask = opC.get_mask()
                zero = opC.get_zero()
                shft = opC.get_shift_correction_value()

                opC.set_envelope()

                tmp_C_0 = opC.new()

                instr += [CmmBinAsInstruction([ '&', tmp_C_0, opC.new(), zero ])]

                if (opA.is_a_constant()) :
                    tmp_C_1 = tmp_C_0.new()

                    if opC_size < opA_size :
                        CmmWarning.warning_message('WARN_OVERFLOW_IN_IMPLICIT_CONSTANT_CONVERSION', '', 0)

                    instr += [CmmBinAsInstruction([ '|', opC, tmp_C_1, (opA & mask) << shft ])]
                else :
                    tmp_C_1 = CmmSpecVar('','tmp', CmmType('u', opC.get_envelope_size()))
                    tmp_C_2 = tmp_C_1.new()
                    tmp_C_3 = tmp_C_1.new()
                    tmp_C_4 = tmp_C_1.new()
                    tmp_C_5 = tmp_C_0.new()

                    if opC_size < opA_size :
                        instr += [CmmBinAsInstruction([ '&', tmp_C_1, opA, mask ])]
                        if shft : instr += [CmmBinAsInstruction([ '<<', tmp_C_3, tmp_C_2, shft ])]
                    else :
                        if shft : instr += [CmmBinAsInstruction([ '<<', tmp_C_3, opA, shft ])]

                    instr += [CmmBinAsInstruction([ '|', opC, tmp_C_5, tmp_C_4 ])]

                return instr
        else :
            return [ self ]

    def process_64_bit_spec_reg(self) :
        print('H' * 80)
        opC = self.body[1]
        opA = self.body[2]
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        pre = []
        post = []

        print('F' * 80)
        if opC.is_a_spec_reg() and opC_size == 64 :
            print('F' * 80)
            if opC == 'cmd' and opA.is_a_constant() and ((opA & 0xffffff) == opA) :
                return [ CmmCmdInstruction([ 'cmdi', opA ]) ]
            else :
                tmp_C_0 = CmmSpecVar('','tmp', CmmType('u', opC_size))
                tmp_C_1 = tmp_C_0.new()

                self.set_opC(tmp_C_0)

                post += [CmmAssgnInstruction([ '=', opC, tmp_C_1 ])]

        print('G' * 80)
        if opA.is_a_spec_reg() and opA_size == 64 :
            print('G' * 80)
            if opC_size < 64 :
                tmp_A_0 = CmmSpecVar('', 'tmp', CmmType('u', opA_size))
                tmp_A_1 = tmp_A_0.new()

                self.set_opA(tmp_A_1)

                pre += [CmmAssgnInstruction([ '=', tmp_A_0, opA ])]

        print('H' * 80)
        return pre + [ self.cwl() ] + post

    def fix_use(self) :
        opC = self.body[1]
        opA = self.body[2]
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        pre = []
        post = []

        if opA.is_a_var() :
            new_size = self.get_normalized_size(max(opC_size, opA_size))

            if opA.is_nonalligned() or (opA.is_signed() and opA_size < new_size) :
                pass
                # my $tmp_A_0 = CmmSpecVar->new('','tmp', CmmType->new('u', $new_size)) ;
                # my $tmp_A_1 = $tmp_A_0->new() ;

                # push @instr, CmmAssgnDummyInstruction->new([ '=', $tmp_A_0, CmmConst->new(0) ]) ;
                # push @instr, CmmBinAsInstruction->new([ "ext_${new_size}_${opA_size}", $opC, $tmp_A_1, $opA ]) ;

                # [ @instr ]
        elif (opA.is_a_constant()) :
            if opC_size < opA_size :
                CmmWarning.warning_message('WARN_OVERFLOW_IN_IMPLICIT_CONSTANT_CONVERSION', '', 0)
                # CmmWarning->warning_message('WARN_OVERFLOW_IN_IMPLICIT_CONSTANT_CONVERSION', $opC->get_filename(), $opC->get_line()) ;
                # $self->{body}[2] = $opA & ((1 << $opC_size) - 1) ;
                # $opA = $self->{body}[2] ;
                self.set_opA(opA & ((1 << opC_size) - 1))
        # else {
        # [ $self->cwl() ]
        # ### 02.11.03 [ CmmAssgnInstruction->new( $self->{body} ) ]
        # }

        return pre + [ self.cwl() ] + post

    def fix_memory_usage_0(self) :
        instr = []

        opC = self.body[1]
        opA = self.body[2]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if (not opC.is_a_var() or opC_size < opA_size) and opA.is_a_mem_ref() :
            print ("fix_memory_usage_0 " + str(self.body) + "\n")
            tmp_A_0 = CmmSpecVar('','mem', CmmType('u', 64))
            tmp_A_1 = tmp_A_0.new()

            # $self->set_opA($tmp_A_1) ;
            self.body[2] = tmp_A_1

            instr += [CmmAssgnInstruction([ '=', tmp_A_0, opA ])]

            # $tmp_A_0->set_use($tmp_A_1) ;
            # $tmp_A_1->set_def($tmp_A_0) ;

        instr += [self.cwl()]
        ### 02.11.03 push @instr, CmmAssgnInstruction->new( $self->{body} ) ;

        return instr

    def fix_memory_usage_1(self) :
        print('J' * 80)
        print(self)
        opC = self.body[1]
        opA = self.body[2]
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        instr = []

        print('J' * 80)
        if opC.is_a_mem_ref() and (not opA.is_a_var() or opA_size < 64) :
            tmp_C_0 = CmmSpecVar('', 'tmp', CmmType('u', opC_size))
            tmp_C_1 = tmp_C_0.new()

            # $self->set_opC($tmp_C_0) ;
            self.body[1] = tmp_C_0

            instr += [CmmAssgnInstruction([ '=', opC, tmp_C_1 ])]

            # $tmp_C_0->set_use($tmp_C_1) ;
            # $tmp_C_1->set_def($tmp_C_0) ;

        instr[0:0] = [ self.cwl() ]
        ### 02.11.03 unshift @instr, CmmAssgnInstruction->new( $self->{body} ) ;

        return instr

    def fix_instr_with_opC_smaller_than_opA(self) :
        instr = []

        opC = self.body[1]
        opA = self.body[2]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if opC_size and opA_size and opC_size < opA_size and opA.is_a_mem_ref() :
            tmp_A_0 = self.body[1] = CmmSpecVar('','tmp', CmmType('u', opA_size))
            tmp_A_1 = tmp_A_0.new()

            instr += [CmmAssgnInstruction([ '=', opC, tmp_A_1 ])]

        instr[0:0] = CmmAssgnInstruction( self.body )

        return instr

    def fix_assignment_with_small_right_side_0(self) :
        opC = self.body[1]
        opA = self.body[2]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        desirable_size = 32 if opC_size == 64 else opC_size

        if opC != 'cmd' and opA_size < desirable_size :
            if opA.is_a_constant() :
                if opC_size == 64 :
                    if opA == 0 :
                        self.body[2] = CmmSR('zero', CmmType('u', 64))

                return [ self ]
            elif opA.is_a_label() :
                return [ self ]
            else :
                return [ CmmBinAsInstruction([ '|', opC, CmmSR('zero', CmmType('u', desirable_size)), opA ]) ]
        else :
            return [ self ]

    def customize_in_accordance_with_sizes(self) :
        opC = self.get_opC()
        opA = self.get_opA()
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if opC.is_a_spec_reg() :
            if opC == 'cmd' :
                if opA.is_a_constant() :
                    if (opA | 0xffffff) != 0xffffff :
                        if opA_size > opC_size :
                            opA = (opA & ((1 << opC_size) - 1))

                        tmp_A = CmmSpecVar('','tmp', CmmType('u', opA_size))

                        return [
                            CmmAssgnInstruction([ '=', tmp_A, opA ]),
                            CmmAssgnInstruction([ '=', opC, tmp_A ]),
                            ]
                    else :
                        return [
                            CmmCmdInstruction([ 'cmdi', opA ])
                            ]
                else :
                    return [ self.cwl() ]
            else :        # any other special register
                return [ self.cwl() ]
        else :
            return [ self.cwl() ]

    def transform_to_pseudo_asm_form(self) :
        if self.body[1].is_a_var() or self.body[1].is_a_reg() :
            if self.body[2].is_a_var() or self.body[2].is_a_reg() :
                self.body = [ 'move', self.body[1], self.body[2] ]
            elif self.body[2].is_a_constant() or self.body[2].is_a_label() : # label is added for 'mviw' pc, label
                # created by loop processing procedures
                self.body = [ 'mviw', self.body[1], self.body[2] ]
            else :
                self.body = [ 'load', CmmConst(0), self.body[1], self.body[2].get_mem_address_var() ]
        else :
            self.body = [ 'store', CmmConst(0), self.body[2], self.body[1].get_mem_address_var() ]

    # my %_one_one = {} ;
    # my %_one_zero = {} ;
    # my %_shl_one = {} ;
    # my %_shr_one = {} ;
    # my %_two_one = {} ;
    # my %_two_zero = {} ;
    # my %_shl_one_one = {} ;
    # my %_shr_one_one = {} ;
    # my %_shl_one_zero = {} ;
    # my %_shr_one_zero = {} ;
    # my %_shr_shl = {} ;

    _one_one = {
        0x80000000                                        : 31,
        0x40000000                                        : 30,
        0x20000000                                        : 29,
        0x10000000                                        : 28,
        0x08000000                                        : 27,
        0x04000000                                        : 26,
        0x02000000                                        : 25,
        0x01000000                                        : 24,
        0x00800000                                        : 23,
        0x00400000                                        : 22,
        0x00200000                                        : 21,
        0x00100000                                        : 20,
        0x00080000                                        : 19,
        0x00040000                                        : 18,
        0x00020000                                        : 17,
        0x00010000                                        : 16,
        0x00008000                                        : 15,
        0x00004000                                        : 14,
        0x00002000                                        : 13,
        0x00001000                                        : 12,
        0x00000800                                        : 11,
        0x00000400                                        : 10,
        0x00000200                                        : 9,
        0x00000100                                        : 8,
        0x00000080                                        : 7,
        0x00000040                                        : 6,
        0x00000020                                        : 5,
        0x00000010                                        : 4,
        0x00000008                                        : 3,
        0x00000004                                        : 2,
        0x00000002                                        : 1,
        0x00000001                                        : 0,
        }

    _two_one = {
        0xc0000000                                        : [30, 31],
        0xa0000000                                        : [29, 31],
        0x60000000                                        : [29, 30],
        0x90000000                                        : [28, 31],
        0x50000000                                        : [28, 30],
        0x30000000                                        : [28, 29],
        0x88000000                                        : [27, 31],
        0x48000000                                        : [27, 30],
        0x28000000                                        : [27, 29],
        0x18000000                                        : [27, 28],
        0x84000000                                        : [26, 31],
        0x44000000                                        : [26, 30],
        0x24000000                                        : [26, 29],
        0x14000000                                        : [26, 28],
        0x0c000000                                        : [26, 27],
        0x82000000                                        : [25, 31],
        0x42000000                                        : [25, 30],
        0x22000000                                        : [25, 29],
        0x12000000                                        : [25, 28],
        0x0a000000                                        : [25, 27],
        0x06000000                                        : [25, 26],
        0x81000000                                        : [24, 31],
        0x41000000                                        : [24, 30],
        0x21000000                                        : [24, 29],
        0x11000000                                        : [24, 28],
        0x09000000                                        : [24, 27],
        0x05000000                                        : [24, 26],
        0x03000000                                        : [24, 25],
        0x80800000                                        : [23, 31],
        0x40800000                                        : [23, 30],
        0x20800000                                        : [23, 29],
        0x10800000                                        : [23, 28],
        0x08800000                                        : [23, 27],
        0x04800000                                        : [23, 26],
        0x02800000                                        : [23, 25],
        0x01800000                                        : [23, 24],
        0x80400000                                        : [22, 31],
        0x40400000                                        : [22, 30],
        0x20400000                                        : [22, 29],
        0x10400000                                        : [22, 28],
        0x08400000                                        : [22, 27],
        0x04400000                                        : [22, 26],
        0x02400000                                        : [22, 25],
        0x01400000                                        : [22, 24],
        0x00c00000                                        : [22, 23],
        0x80200000                                        : [21, 31],
        0x40200000                                        : [21, 30],
        0x20200000                                        : [21, 29],
        0x10200000                                        : [21, 28],
        0x08200000                                        : [21, 27],
        0x04200000                                        : [21, 26],
        0x02200000                                        : [21, 25],
        0x01200000                                        : [21, 24],
        0x00a00000                                        : [21, 23],
        0x00600000                                        : [21, 22],
        0x80100000                                        : [20, 31],
        0x40100000                                        : [20, 30],
        0x20100000                                        : [20, 29],
        0x10100000                                        : [20, 28],
        0x08100000                                        : [20, 27],
        0x04100000                                        : [20, 26],
        0x02100000                                        : [20, 25],
        0x01100000                                        : [20, 24],
        0x00900000                                        : [20, 23],
        0x00500000                                        : [20, 22],
        0x00300000                                        : [20, 21],
        0x80080000                                        : [19, 31],
        0x40080000                                        : [19, 30],
        0x20080000                                        : [19, 29],
        0x10080000                                        : [19, 28],
        0x08080000                                        : [19, 27],
        0x04080000                                        : [19, 26],
        0x02080000                                        : [19, 25],
        0x01080000                                        : [19, 24],
        0x00880000                                        : [19, 23],
        0x00480000                                        : [19, 22],
        0x00280000                                        : [19, 21],
        0x00180000                                        : [19, 20],
        0x80040000                                        : [18, 31],
        0x40040000                                        : [18, 30],
        0x20040000                                        : [18, 29],
        0x10040000                                        : [18, 28],
        0x08040000                                        : [18, 27],
        0x04040000                                        : [18, 26],
        0x02040000                                        : [18, 25],
        0x01040000                                        : [18, 24],
        0x00840000                                        : [18, 23],
        0x00440000                                        : [18, 22],
        0x00240000                                        : [18, 21],
        0x00140000                                        : [18, 20],
        0x000c0000                                        : [18, 19],
        0x80020000                                        : [17, 31],
        0x40020000                                        : [17, 30],
        0x20020000                                        : [17, 29],
        0x10020000                                        : [17, 28],
        0x08020000                                        : [17, 27],
        0x04020000                                        : [17, 26],
        0x02020000                                        : [17, 25],
        0x01020000                                        : [17, 24],
        0x00820000                                        : [17, 23],
        0x00420000                                        : [17, 22],
        0x00220000                                        : [17, 21],
        0x00120000                                        : [17, 20],
        0x000a0000                                        : [17, 19],
        0x00060000                                        : [17, 18],
        0x80010000                                        : [16, 31],
        0x40010000                                        : [16, 30],
        0x20010000                                        : [16, 29],
        0x10010000                                        : [16, 28],
        0x08010000                                        : [16, 27],
        0x04010000                                        : [16, 26],
        0x02010000                                        : [16, 25],
        0x01010000                                        : [16, 24],
        0x00810000                                        : [16, 23],
        0x00410000                                        : [16, 22],
        0x00210000                                        : [16, 21],
        0x00110000                                        : [16, 20],
        0x00090000                                        : [16, 19],
        0x00050000                                        : [16, 18],
        0x00030000                                        : [16, 17],
        0x80008000                                        : [15, 31],
        0x40008000                                        : [15, 30],
        0x20008000                                        : [15, 29],
        0x10008000                                        : [15, 28],
        0x08008000                                        : [15, 27],
        0x04008000                                        : [15, 26],
        0x02008000                                        : [15, 25],
        0x01008000                                        : [15, 24],
        0x00808000                                        : [15, 23],
        0x00408000                                        : [15, 22],
        0x00208000                                        : [15, 21],
        0x00108000                                        : [15, 20],
        0x00088000                                        : [15, 19],
        0x00048000                                        : [15, 18],
        0x00028000                                        : [15, 17],
        0x00018000                                        : [15, 16],
        0x80004000                                        : [14, 31],
        0x40004000                                        : [14, 30],
        0x20004000                                        : [14, 29],
        0x10004000                                        : [14, 28],
        0x08004000                                        : [14, 27],
        0x04004000                                        : [14, 26],
        0x02004000                                        : [14, 25],
        0x01004000                                        : [14, 24],
        0x00804000                                        : [14, 23],
        0x00404000                                        : [14, 22],
        0x00204000                                        : [14, 21],
        0x00104000                                        : [14, 20],
        0x00084000                                        : [14, 19],
        0x00044000                                        : [14, 18],
        0x00024000                                        : [14, 17],
        0x00014000                                        : [14, 16],
        0x0000c000                                        : [14, 15],
        0x80002000                                        : [13, 31],
        0x40002000                                        : [13, 30],
        0x20002000                                        : [13, 29],
        0x10002000                                        : [13, 28],
        0x08002000                                        : [13, 27],
        0x04002000                                        : [13, 26],
        0x02002000                                        : [13, 25],
        0x01002000                                        : [13, 24],
        0x00802000                                        : [13, 23],
        0x00402000                                        : [13, 22],
        0x00202000                                        : [13, 21],
        0x00102000                                        : [13, 20],
        0x00082000                                        : [13, 19],
        0x00042000                                        : [13, 18],
        0x00022000                                        : [13, 17],
        0x00012000                                        : [13, 16],
        0x0000a000                                        : [13, 15],
        0x00006000                                        : [13, 14],
        0x80001000                                        : [12, 31],
        0x40001000                                        : [12, 30],
        0x20001000                                        : [12, 29],
        0x10001000                                        : [12, 28],
        0x08001000                                        : [12, 27],
        0x04001000                                        : [12, 26],
        0x02001000                                        : [12, 25],
        0x01001000                                        : [12, 24],
        0x00801000                                        : [12, 23],
        0x00401000                                        : [12, 22],
        0x00201000                                        : [12, 21],
        0x00101000                                        : [12, 20],
        0x00081000                                        : [12, 19],
        0x00041000                                        : [12, 18],
        0x00021000                                        : [12, 17],
        0x00011000                                        : [12, 16],
        0x00009000                                        : [12, 15],
        0x00005000                                        : [12, 14],
        0x00003000                                        : [12, 13],
        0x80000800                                        : [11, 31],
        0x40000800                                        : [11, 30],
        0x20000800                                        : [11, 29],
        0x10000800                                        : [11, 28],
        0x08000800                                        : [11, 27],
        0x04000800                                        : [11, 26],
        0x02000800                                        : [11, 25],
        0x01000800                                        : [11, 24],
        0x00800800                                        : [11, 23],
        0x00400800                                        : [11, 22],
        0x00200800                                        : [11, 21],
        0x00100800                                        : [11, 20],
        0x00080800                                        : [11, 19],
        0x00040800                                        : [11, 18],
        0x00020800                                        : [11, 17],
        0x00010800                                        : [11, 16],
        0x00008800                                        : [11, 15],
        0x00004800                                        : [11, 14],
        0x00002800                                        : [11, 13],
        0x00001800                                        : [11, 12],
        0x80000400                                        : [10, 31],
        0x40000400                                        : [10, 30],
        0x20000400                                        : [10, 29],
        0x10000400                                        : [10, 28],
        0x08000400                                        : [10, 27],
        0x04000400                                        : [10, 26],
        0x02000400                                        : [10, 25],
        0x01000400                                        : [10, 24],
        0x00800400                                        : [10, 23],
        0x00400400                                        : [10, 22],
        0x00200400                                        : [10, 21],
        0x00100400                                        : [10, 20],
        0x00080400                                        : [10, 19],
        0x00040400                                        : [10, 18],
        0x00020400                                        : [10, 17],
        0x00010400                                        : [10, 16],
        0x00008400                                        : [10, 15],
        0x00004400                                        : [10, 14],
        0x00002400                                        : [10, 13],
        0x00001400                                        : [10, 12],
        0x00000c00                                        : [10, 11],
        0x80000200                                        : [9, 31],
        0x40000200                                        : [9, 30],
        0x20000200                                        : [9, 29],
        0x10000200                                        : [9, 28],
        0x08000200                                        : [9, 27],
        0x04000200                                        : [9, 26],
        0x02000200                                        : [9, 25],
        0x01000200                                        : [9, 24],
        0x00800200                                        : [9, 23],
        0x00400200                                        : [9, 22],
        0x00200200                                        : [9, 21],
        0x00100200                                        : [9, 20],
        0x00080200                                        : [9, 19],
        0x00040200                                        : [9, 18],
        0x00020200                                        : [9, 17],
        0x00010200                                        : [9, 16],
        0x00008200                                        : [9, 15],
        0x00004200                                        : [9, 14],
        0x00002200                                        : [9, 13],
        0x00001200                                        : [9, 12],
        0x00000a00                                        : [9, 11],
        0x00000600                                        : [9, 10],
        0x80000100                                        : [8, 31],
        0x40000100                                        : [8, 30],
        0x20000100                                        : [8, 29],
        0x10000100                                        : [8, 28],
        0x08000100                                        : [8, 27],
        0x04000100                                        : [8, 26],
        0x02000100                                        : [8, 25],
        0x01000100                                        : [8, 24],
        0x00800100                                        : [8, 23],
        0x00400100                                        : [8, 22],
        0x00200100                                        : [8, 21],
        0x00100100                                        : [8, 20],
        0x00080100                                        : [8, 19],
        0x00040100                                        : [8, 18],
        0x00020100                                        : [8, 17],
        0x00010100                                        : [8, 16],
        0x00008100                                        : [8, 15],
        0x00004100                                        : [8, 14],
        0x00002100                                        : [8, 13],
        0x00001100                                        : [8, 12],
        0x00000900                                        : [8, 11],
        0x00000500                                        : [8, 10],
        0x00000300                                        : [8, 9],
        0x80000080                                        : [7, 31],
        0x40000080                                        : [7, 30],
        0x20000080                                        : [7, 29],
        0x10000080                                        : [7, 28],
        0x08000080                                        : [7, 27],
        0x04000080                                        : [7, 26],
        0x02000080                                        : [7, 25],
        0x01000080                                        : [7, 24],
        0x00800080                                        : [7, 23],
        0x00400080                                        : [7, 22],
        0x00200080                                        : [7, 21],
        0x00100080                                        : [7, 20],
        0x00080080                                        : [7, 19],
        0x00040080                                        : [7, 18],
        0x00020080                                        : [7, 17],
        0x00010080                                        : [7, 16],
        0x00008080                                        : [7, 15],
        0x00004080                                        : [7, 14],
        0x00002080                                        : [7, 13],
        0x00001080                                        : [7, 12],
        0x00000880                                        : [7, 11],
        0x00000480                                        : [7, 10],
        0x00000280                                        : [7, 9],
        0x00000180                                        : [7, 8],
        0x80000040                                        : [6, 31],
        0x40000040                                        : [6, 30],
        0x20000040                                        : [6, 29],
        0x10000040                                        : [6, 28],
        0x08000040                                        : [6, 27],
        0x04000040                                        : [6, 26],
        0x02000040                                        : [6, 25],
        0x01000040                                        : [6, 24],
        0x00800040                                        : [6, 23],
        0x00400040                                        : [6, 22],
        0x00200040                                        : [6, 21],
        0x00100040                                        : [6, 20],
        0x00080040                                        : [6, 19],
        0x00040040                                        : [6, 18],
        0x00020040                                        : [6, 17],
        0x00010040                                        : [6, 16],
        0x00008040                                        : [6, 15],
        0x00004040                                        : [6, 14],
        0x00002040                                        : [6, 13],
        0x00001040                                        : [6, 12],
        0x00000840                                        : [6, 11],
        0x00000440                                        : [6, 10],
        0x00000240                                        : [6, 9],
        0x00000140                                        : [6, 8],
        0x000000c0                                        : [6, 7],
        0x80000020                                        : [5, 31],
        0x40000020                                        : [5, 30],
        0x20000020                                        : [5, 29],
        0x10000020                                        : [5, 28],
        0x08000020                                        : [5, 27],
        0x04000020                                        : [5, 26],
        0x02000020                                        : [5, 25],
        0x01000020                                        : [5, 24],
        0x00800020                                        : [5, 23],
        0x00400020                                        : [5, 22],
        0x00200020                                        : [5, 21],
        0x00100020                                        : [5, 20],
        0x00080020                                        : [5, 19],
        0x00040020                                        : [5, 18],
        0x00020020                                        : [5, 17],
        0x00010020                                        : [5, 16],
        0x00008020                                        : [5, 15],
        0x00004020                                        : [5, 14],
        0x00002020                                        : [5, 13],
        0x00001020                                        : [5, 12],
        0x00000820                                        : [5, 11],
        0x00000420                                        : [5, 10],
        0x00000220                                        : [5, 9],
        0x00000120                                        : [5, 8],
        0x000000a0                                        : [5, 7],
        0x00000060                                        : [5, 6],
        0x80000010                                        : [4, 31],
        0x40000010                                        : [4, 30],
        0x20000010                                        : [4, 29],
        0x10000010                                        : [4, 28],
        0x08000010                                        : [4, 27],
        0x04000010                                        : [4, 26],
        0x02000010                                        : [4, 25],
        0x01000010                                        : [4, 24],
        0x00800010                                        : [4, 23],
        0x00400010                                        : [4, 22],
        0x00200010                                        : [4, 21],
        0x00100010                                        : [4, 20],
        0x00080010                                        : [4, 19],
        0x00040010                                        : [4, 18],
        0x00020010                                        : [4, 17],
        0x00010010                                        : [4, 16],
        0x00008010                                        : [4, 15],
        0x00004010                                        : [4, 14],
        0x00002010                                        : [4, 13],
        0x00001010                                        : [4, 12],
        0x00000810                                        : [4, 11],
        0x00000410                                        : [4, 10],
        0x00000210                                        : [4, 9],
        0x00000110                                        : [4, 8],
        0x00000090                                        : [4, 7],
        0x00000050                                        : [4, 6],
        0x00000030                                        : [4, 5],
        0x80000008                                        : [3, 31],
        0x40000008                                        : [3, 30],
        0x20000008                                        : [3, 29],
        0x10000008                                        : [3, 28],
        0x08000008                                        : [3, 27],
        0x04000008                                        : [3, 26],
        0x02000008                                        : [3, 25],
        0x01000008                                        : [3, 24],
        0x00800008                                        : [3, 23],
        0x00400008                                        : [3, 22],
        0x00200008                                        : [3, 21],
        0x00100008                                        : [3, 20],
        0x00080008                                        : [3, 19],
        0x00040008                                        : [3, 18],
        0x00020008                                        : [3, 17],
        0x00010008                                        : [3, 16],
        0x00008008                                        : [3, 15],
        0x00004008                                        : [3, 14],
        0x00002008                                        : [3, 13],
        0x00001008                                        : [3, 12],
        0x00000808                                        : [3, 11],
        0x00000408                                        : [3, 10],
        0x00000208                                        : [3, 9],
        0x00000108                                        : [3, 8],
        0x00000088                                        : [3, 7],
        0x00000048                                        : [3, 6],
        0x00000028                                        : [3, 5],
        0x00000018                                        : [3, 4],
        0x80000004                                        : [2, 31],
        0x40000004                                        : [2, 30],
        0x20000004                                        : [2, 29],
        0x10000004                                        : [2, 28],
        0x08000004                                        : [2, 27],
        0x04000004                                        : [2, 26],
        0x02000004                                        : [2, 25],
        0x01000004                                        : [2, 24],
        0x00800004                                        : [2, 23],
        0x00400004                                        : [2, 22],
        0x00200004                                        : [2, 21],
        0x00100004                                        : [2, 20],
        0x00080004                                        : [2, 19],
        0x00040004                                        : [2, 18],
        0x00020004                                        : [2, 17],
        0x00010004                                        : [2, 16],
        0x00008004                                        : [2, 15],
        0x00004004                                        : [2, 14],
        0x00002004                                        : [2, 13],
        0x00001004                                        : [2, 12],
        0x00000804                                        : [2, 11],
        0x00000404                                        : [2, 10],
        0x00000204                                        : [2, 9],
        0x00000104                                        : [2, 8],
        0x00000084                                        : [2, 7],
        0x00000044                                        : [2, 6],
        0x00000024                                        : [2, 5],
        0x00000014                                        : [2, 4],
        0x0000000c                                        : [2, 3],
        0x80000002                                        : [1, 31],
        0x40000002                                        : [1, 30],
        0x20000002                                        : [1, 29],
        0x10000002                                        : [1, 28],
        0x08000002                                        : [1, 27],
        0x04000002                                        : [1, 26],
        0x02000002                                        : [1, 25],
        0x01000002                                        : [1, 24],
        0x00800002                                        : [1, 23],
        0x00400002                                        : [1, 22],
        0x00200002                                        : [1, 21],
        0x00100002                                        : [1, 20],
        0x00080002                                        : [1, 19],
        0x00040002                                        : [1, 18],
        0x00020002                                        : [1, 17],
        0x00010002                                        : [1, 16],
        0x00008002                                        : [1, 15],
        0x00004002                                        : [1, 14],
        0x00002002                                        : [1, 13],
        0x00001002                                        : [1, 12],
        0x00000802                                        : [1, 11],
        0x00000402                                        : [1, 10],
        0x00000202                                        : [1, 9],
        0x00000102                                        : [1, 8],
        0x00000082                                        : [1, 7],
        0x00000042                                        : [1, 6],
        0x00000022                                        : [1, 5],
        0x00000012                                        : [1, 4],
        0x0000000a                                        : [1, 3],
        0x00000006                                        : [1, 2],
        0x80000001                                        : [0, 31],
        0x40000001                                        : [0, 30],
        0x20000001                                        : [0, 29],
        0x10000001                                        : [0, 28],
        0x08000001                                        : [0, 27],
        0x04000001                                        : [0, 26],
        0x02000001                                        : [0, 25],
        0x01000001                                        : [0, 24],
        0x00800001                                        : [0, 23],
        0x00400001                                        : [0, 22],
        0x00200001                                        : [0, 21],
        0x00100001                                        : [0, 20],
        0x00080001                                        : [0, 19],
        0x00040001                                        : [0, 18],
        0x00020001                                        : [0, 17],
        0x00010001                                        : [0, 16],
        0x00008001                                        : [0, 15],
        0x00004001                                        : [0, 14],
        0x00002001                                        : [0, 13],
        0x00001001                                        : [0, 12],
        0x00000801                                        : [0, 11],
        0x00000401                                        : [0, 10],
        0x00000201                                        : [0, 9],
        0x00000101                                        : [0, 8],
        0x00000081                                        : [0, 7],
        0x00000041                                        : [0, 6],
        0x00000021                                        : [0, 5],
        0x00000011                                        : [0, 4],
        0x00000009                                        : [0, 3],
        0x00000005                                        : [0, 2],
        0x00000003                                        : [0, 1],
        }

    _one_zero = {
        0x7fffffff                                        : 31,
        0xbfffffff                                        : 30,
        0xdfffffff                                        : 29,
        0xefffffff                                        : 28,
        0xf7ffffff                                        : 27,
        0xfbffffff                                        : 26,
        0xfdffffff                                        : 25,
        0xfeffffff                                        : 24,
        0xff7fffff                                        : 23,
        0xffbfffff                                        : 22,
        0xffdfffff                                        : 21,
        0xffefffff                                        : 20,
        0xfff7ffff                                        : 19,
        0xfffbffff                                        : 18,
        0xfffdffff                                        : 17,
        0xfffeffff                                        : 16,
        0xffff7fff                                        : 15,
        0xffffbfff                                        : 14,
        0xffffdfff                                        : 13,
        0xffffefff                                        : 12,
        0xfffff7ff                                        : 11,
        0xfffffbff                                        : 10,
        0xfffffdff                                        : 9,
        0xfffffeff                                        : 8,
        0xffffff7f                                        : 7,
        0xffffffbf                                        : 6,
        0xffffffdf                                        : 5,
        0xffffffef                                        : 4,
        0xfffffff7                                        : 3,
        0xfffffffb                                        : 2,
        0xfffffffd                                        : 1,
        0xfffffffe                                        : 0,
        }

    _two_zero = {
        0x3fffffff                                        : [30, 31],
        0x5fffffff                                        : [29, 31],
        0x9fffffff                                        : [29, 30],
        0x6fffffff                                        : [28, 31],
        0xafffffff                                        : [28, 30],
        0xcfffffff                                        : [28, 29],
        0x77ffffff                                        : [27, 31],
        0xb7ffffff                                        : [27, 30],
        0xd7ffffff                                        : [27, 29],
        0xe7ffffff                                        : [27, 28],
        0x7bffffff                                        : [26, 31],
        0xbbffffff                                        : [26, 30],
        0xdbffffff                                        : [26, 29],
        0xebffffff                                        : [26, 28],
        0xf3ffffff                                        : [26, 27],
        0x7dffffff                                        : [25, 31],
        0xbdffffff                                        : [25, 30],
        0xddffffff                                        : [25, 29],
        0xedffffff                                        : [25, 28],
        0xf5ffffff                                        : [25, 27],
        0xf9ffffff                                        : [25, 26],
        0x7effffff                                        : [24, 31],
        0xbeffffff                                        : [24, 30],
        0xdeffffff                                        : [24, 29],
        0xeeffffff                                        : [24, 28],
        0xf6ffffff                                        : [24, 27],
        0xfaffffff                                        : [24, 26],
        0xfcffffff                                        : [24, 25],
        0x7f7fffff                                        : [23, 31],
        0xbf7fffff                                        : [23, 30],
        0xdf7fffff                                        : [23, 29],
        0xef7fffff                                        : [23, 28],
        0xf77fffff                                        : [23, 27],
        0xfb7fffff                                        : [23, 26],
        0xfd7fffff                                        : [23, 25],
        0xfe7fffff                                        : [23, 24],
        0x7fbfffff                                        : [22, 31],
        0xbfbfffff                                        : [22, 30],
        0xdfbfffff                                        : [22, 29],
        0xefbfffff                                        : [22, 28],
        0xf7bfffff                                        : [22, 27],
        0xfbbfffff                                        : [22, 26],
        0xfdbfffff                                        : [22, 25],
        0xfebfffff                                        : [22, 24],
        0xff3fffff                                        : [22, 23],
        0x7fdfffff                                        : [21, 31],
        0xbfdfffff                                        : [21, 30],
        0xdfdfffff                                        : [21, 29],
        0xefdfffff                                        : [21, 28],
        0xf7dfffff                                        : [21, 27],
        0xfbdfffff                                        : [21, 26],
        0xfddfffff                                        : [21, 25],
        0xfedfffff                                        : [21, 24],
        0xff5fffff                                        : [21, 23],
        0xff9fffff                                        : [21, 22],
        0x7fefffff                                        : [20, 31],
        0xbfefffff                                        : [20, 30],
        0xdfefffff                                        : [20, 29],
        0xefefffff                                        : [20, 28],
        0xf7efffff                                        : [20, 27],
        0xfbefffff                                        : [20, 26],
        0xfdefffff                                        : [20, 25],
        0xfeefffff                                        : [20, 24],
        0xff6fffff                                        : [20, 23],
        0xffafffff                                        : [20, 22],
        0xffcfffff                                        : [20, 21],
        0x7ff7ffff                                        : [19, 31],
        0xbff7ffff                                        : [19, 30],
        0xdff7ffff                                        : [19, 29],
        0xeff7ffff                                        : [19, 28],
        0xf7f7ffff                                        : [19, 27],
        0xfbf7ffff                                        : [19, 26],
        0xfdf7ffff                                        : [19, 25],
        0xfef7ffff                                        : [19, 24],
        0xff77ffff                                        : [19, 23],
        0xffb7ffff                                        : [19, 22],
        0xffd7ffff                                        : [19, 21],
        0xffe7ffff                                        : [19, 20],
        0x7ffbffff                                        : [18, 31],
        0xbffbffff                                        : [18, 30],
        0xdffbffff                                        : [18, 29],
        0xeffbffff                                        : [18, 28],
        0xf7fbffff                                        : [18, 27],
        0xfbfbffff                                        : [18, 26],
        0xfdfbffff                                        : [18, 25],
        0xfefbffff                                        : [18, 24],
        0xff7bffff                                        : [18, 23],
        0xffbbffff                                        : [18, 22],
        0xffdbffff                                        : [18, 21],
        0xffebffff                                        : [18, 20],
        0xfff3ffff                                        : [18, 19],
        0x7ffdffff                                        : [17, 31],
        0xbffdffff                                        : [17, 30],
        0xdffdffff                                        : [17, 29],
        0xeffdffff                                        : [17, 28],
        0xf7fdffff                                        : [17, 27],
        0xfbfdffff                                        : [17, 26],
        0xfdfdffff                                        : [17, 25],
        0xfefdffff                                        : [17, 24],
        0xff7dffff                                        : [17, 23],
        0xffbdffff                                        : [17, 22],
        0xffddffff                                        : [17, 21],
        0xffedffff                                        : [17, 20],
        0xfff5ffff                                        : [17, 19],
        0xfff9ffff                                        : [17, 18],
        0x7ffeffff                                        : [16, 31],
        0xbffeffff                                        : [16, 30],
        0xdffeffff                                        : [16, 29],
        0xeffeffff                                        : [16, 28],
        0xf7feffff                                        : [16, 27],
        0xfbfeffff                                        : [16, 26],
        0xfdfeffff                                        : [16, 25],
        0xfefeffff                                        : [16, 24],
        0xff7effff                                        : [16, 23],
        0xffbeffff                                        : [16, 22],
        0xffdeffff                                        : [16, 21],
        0xffeeffff                                        : [16, 20],
        0xfff6ffff                                        : [16, 19],
        0xfffaffff                                        : [16, 18],
        0xfffcffff                                        : [16, 17],
        0x7fff7fff                                        : [15, 31],
        0xbfff7fff                                        : [15, 30],
        0xdfff7fff                                        : [15, 29],
        0xefff7fff                                        : [15, 28],
        0xf7ff7fff                                        : [15, 27],
        0xfbff7fff                                        : [15, 26],
        0xfdff7fff                                        : [15, 25],
        0xfeff7fff                                        : [15, 24],
        0xff7f7fff                                        : [15, 23],
        0xffbf7fff                                        : [15, 22],
        0xffdf7fff                                        : [15, 21],
        0xffef7fff                                        : [15, 20],
        0xfff77fff                                        : [15, 19],
        0xfffb7fff                                        : [15, 18],
        0xfffd7fff                                        : [15, 17],
        0xfffe7fff                                        : [15, 16],
        0x7fffbfff                                        : [14, 31],
        0xbfffbfff                                        : [14, 30],
        0xdfffbfff                                        : [14, 29],
        0xefffbfff                                        : [14, 28],
        0xf7ffbfff                                        : [14, 27],
        0xfbffbfff                                        : [14, 26],
        0xfdffbfff                                        : [14, 25],
        0xfeffbfff                                        : [14, 24],
        0xff7fbfff                                        : [14, 23],
        0xffbfbfff                                        : [14, 22],
        0xffdfbfff                                        : [14, 21],
        0xffefbfff                                        : [14, 20],
        0xfff7bfff                                        : [14, 19],
        0xfffbbfff                                        : [14, 18],
        0xfffdbfff                                        : [14, 17],
        0xfffebfff                                        : [14, 16],
        0xffff3fff                                        : [14, 15],
        0x7fffdfff                                        : [13, 31],
        0xbfffdfff                                        : [13, 30],
        0xdfffdfff                                        : [13, 29],
        0xefffdfff                                        : [13, 28],
        0xf7ffdfff                                        : [13, 27],
        0xfbffdfff                                        : [13, 26],
        0xfdffdfff                                        : [13, 25],
        0xfeffdfff                                        : [13, 24],
        0xff7fdfff                                        : [13, 23],
        0xffbfdfff                                        : [13, 22],
        0xffdfdfff                                        : [13, 21],
        0xffefdfff                                        : [13, 20],
        0xfff7dfff                                        : [13, 19],
        0xfffbdfff                                        : [13, 18],
        0xfffddfff                                        : [13, 17],
        0xfffedfff                                        : [13, 16],
        0xffff5fff                                        : [13, 15],
        0xffff9fff                                        : [13, 14],
        0x7fffefff                                        : [12, 31],
        0xbfffefff                                        : [12, 30],
        0xdfffefff                                        : [12, 29],
        0xefffefff                                        : [12, 28],
        0xf7ffefff                                        : [12, 27],
        0xfbffefff                                        : [12, 26],
        0xfdffefff                                        : [12, 25],
        0xfeffefff                                        : [12, 24],
        0xff7fefff                                        : [12, 23],
        0xffbfefff                                        : [12, 22],
        0xffdfefff                                        : [12, 21],
        0xffefefff                                        : [12, 20],
        0xfff7efff                                        : [12, 19],
        0xfffbefff                                        : [12, 18],
        0xfffdefff                                        : [12, 17],
        0xfffeefff                                        : [12, 16],
        0xffff6fff                                        : [12, 15],
        0xffffafff                                        : [12, 14],
        0xffffcfff                                        : [12, 13],
        0x7ffff7ff                                        : [11, 31],
        0xbffff7ff                                        : [11, 30],
        0xdffff7ff                                        : [11, 29],
        0xeffff7ff                                        : [11, 28],
        0xf7fff7ff                                        : [11, 27],
        0xfbfff7ff                                        : [11, 26],
        0xfdfff7ff                                        : [11, 25],
        0xfefff7ff                                        : [11, 24],
        0xff7ff7ff                                        : [11, 23],
        0xffbff7ff                                        : [11, 22],
        0xffdff7ff                                        : [11, 21],
        0xffeff7ff                                        : [11, 20],
        0xfff7f7ff                                        : [11, 19],
        0xfffbf7ff                                        : [11, 18],
        0xfffdf7ff                                        : [11, 17],
        0xfffef7ff                                        : [11, 16],
        0xffff77ff                                        : [11, 15],
        0xffffb7ff                                        : [11, 14],
        0xffffd7ff                                        : [11, 13],
        0xffffe7ff                                        : [11, 12],
        0x7ffffbff                                        : [10, 31],
        0xbffffbff                                        : [10, 30],
        0xdffffbff                                        : [10, 29],
        0xeffffbff                                        : [10, 28],
        0xf7fffbff                                        : [10, 27],
        0xfbfffbff                                        : [10, 26],
        0xfdfffbff                                        : [10, 25],
        0xfefffbff                                        : [10, 24],
        0xff7ffbff                                        : [10, 23],
        0xffbffbff                                        : [10, 22],
        0xffdffbff                                        : [10, 21],
        0xffeffbff                                        : [10, 20],
        0xfff7fbff                                        : [10, 19],
        0xfffbfbff                                        : [10, 18],
        0xfffdfbff                                        : [10, 17],
        0xfffefbff                                        : [10, 16],
        0xffff7bff                                        : [10, 15],
        0xffffbbff                                        : [10, 14],
        0xffffdbff                                        : [10, 13],
        0xffffebff                                        : [10, 12],
        0xfffff3ff                                        : [10, 11],
        0x7ffffdff                                        : [9, 31],
        0xbffffdff                                        : [9, 30],
        0xdffffdff                                        : [9, 29],
        0xeffffdff                                        : [9, 28],
        0xf7fffdff                                        : [9, 27],
        0xfbfffdff                                        : [9, 26],
        0xfdfffdff                                        : [9, 25],
        0xfefffdff                                        : [9, 24],
        0xff7ffdff                                        : [9, 23],
        0xffbffdff                                        : [9, 22],
        0xffdffdff                                        : [9, 21],
        0xffeffdff                                        : [9, 20],
        0xfff7fdff                                        : [9, 19],
        0xfffbfdff                                        : [9, 18],
        0xfffdfdff                                        : [9, 17],
        0xfffefdff                                        : [9, 16],
        0xffff7dff                                        : [9, 15],
        0xffffbdff                                        : [9, 14],
        0xffffddff                                        : [9, 13],
        0xffffedff                                        : [9, 12],
        0xfffff5ff                                        : [9, 11],
        0xfffff9ff                                        : [9, 10],
        0x7ffffeff                                        : [8, 31],
        0xbffffeff                                        : [8, 30],
        0xdffffeff                                        : [8, 29],
        0xeffffeff                                        : [8, 28],
        0xf7fffeff                                        : [8, 27],
        0xfbfffeff                                        : [8, 26],
        0xfdfffeff                                        : [8, 25],
        0xfefffeff                                        : [8, 24],
        0xff7ffeff                                        : [8, 23],
        0xffbffeff                                        : [8, 22],
        0xffdffeff                                        : [8, 21],
        0xffeffeff                                        : [8, 20],
        0xfff7feff                                        : [8, 19],
        0xfffbfeff                                        : [8, 18],
        0xfffdfeff                                        : [8, 17],
        0xfffefeff                                        : [8, 16],
        0xffff7eff                                        : [8, 15],
        0xffffbeff                                        : [8, 14],
        0xffffdeff                                        : [8, 13],
        0xffffeeff                                        : [8, 12],
        0xfffff6ff                                        : [8, 11],
        0xfffffaff                                        : [8, 10],
        0xfffffcff                                        : [8, 9],
        0x7fffff7f                                        : [7, 31],
        0xbfffff7f                                        : [7, 30],
        0xdfffff7f                                        : [7, 29],
        0xefffff7f                                        : [7, 28],
        0xf7ffff7f                                        : [7, 27],
        0xfbffff7f                                        : [7, 26],
        0xfdffff7f                                        : [7, 25],
        0xfeffff7f                                        : [7, 24],
        0xff7fff7f                                        : [7, 23],
        0xffbfff7f                                        : [7, 22],
        0xffdfff7f                                        : [7, 21],
        0xffefff7f                                        : [7, 20],
        0xfff7ff7f                                        : [7, 19],
        0xfffbff7f                                        : [7, 18],
        0xfffdff7f                                        : [7, 17],
        0xfffeff7f                                        : [7, 16],
        0xffff7f7f                                        : [7, 15],
        0xffffbf7f                                        : [7, 14],
        0xffffdf7f                                        : [7, 13],
        0xffffef7f                                        : [7, 12],
        0xfffff77f                                        : [7, 11],
        0xfffffb7f                                        : [7, 10],
        0xfffffd7f                                        : [7, 9],
        0xfffffe7f                                        : [7, 8],
        0x7fffffbf                                        : [6, 31],
        0xbfffffbf                                        : [6, 30],
        0xdfffffbf                                        : [6, 29],
        0xefffffbf                                        : [6, 28],
        0xf7ffffbf                                        : [6, 27],
        0xfbffffbf                                        : [6, 26],
        0xfdffffbf                                        : [6, 25],
        0xfeffffbf                                        : [6, 24],
        0xff7fffbf                                        : [6, 23],
        0xffbfffbf                                        : [6, 22],
        0xffdfffbf                                        : [6, 21],
        0xffefffbf                                        : [6, 20],
        0xfff7ffbf                                        : [6, 19],
        0xfffbffbf                                        : [6, 18],
        0xfffdffbf                                        : [6, 17],
        0xfffeffbf                                        : [6, 16],
        0xffff7fbf                                        : [6, 15],
        0xffffbfbf                                        : [6, 14],
        0xffffdfbf                                        : [6, 13],
        0xffffefbf                                        : [6, 12],
        0xfffff7bf                                        : [6, 11],
        0xfffffbbf                                        : [6, 10],
        0xfffffdbf                                        : [6, 9],
        0xfffffebf                                        : [6, 8],
        0xffffff3f                                        : [6, 7],
        0x7fffffdf                                        : [5, 31],
        0xbfffffdf                                        : [5, 30],
        0xdfffffdf                                        : [5, 29],
        0xefffffdf                                        : [5, 28],
        0xf7ffffdf                                        : [5, 27],
        0xfbffffdf                                        : [5, 26],
        0xfdffffdf                                        : [5, 25],
        0xfeffffdf                                        : [5, 24],
        0xff7fffdf                                        : [5, 23],
        0xffbfffdf                                        : [5, 22],
        0xffdfffdf                                        : [5, 21],
        0xffefffdf                                        : [5, 20],
        0xfff7ffdf                                        : [5, 19],
        0xfffbffdf                                        : [5, 18],
        0xfffdffdf                                        : [5, 17],
        0xfffeffdf                                        : [5, 16],
        0xffff7fdf                                        : [5, 15],
        0xffffbfdf                                        : [5, 14],
        0xffffdfdf                                        : [5, 13],
        0xffffefdf                                        : [5, 12],
        0xfffff7df                                        : [5, 11],
        0xfffffbdf                                        : [5, 10],
        0xfffffddf                                        : [5, 9],
        0xfffffedf                                        : [5, 8],
        0xffffff5f                                        : [5, 7],
        0xffffff9f                                        : [5, 6],
        0x7fffffef                                        : [4, 31],
        0xbfffffef                                        : [4, 30],
        0xdfffffef                                        : [4, 29],
        0xefffffef                                        : [4, 28],
        0xf7ffffef                                        : [4, 27],
        0xfbffffef                                        : [4, 26],
        0xfdffffef                                        : [4, 25],
        0xfeffffef                                        : [4, 24],
        0xff7fffef                                        : [4, 23],
        0xffbfffef                                        : [4, 22],
        0xffdfffef                                        : [4, 21],
        0xffefffef                                        : [4, 20],
        0xfff7ffef                                        : [4, 19],
        0xfffbffef                                        : [4, 18],
        0xfffdffef                                        : [4, 17],
        0xfffeffef                                        : [4, 16],
        0xffff7fef                                        : [4, 15],
        0xffffbfef                                        : [4, 14],
        0xffffdfef                                        : [4, 13],
        0xffffefef                                        : [4, 12],
        0xfffff7ef                                        : [4, 11],
        0xfffffbef                                        : [4, 10],
        0xfffffdef                                        : [4, 9],
        0xfffffeef                                        : [4, 8],
        0xffffff6f                                        : [4, 7],
        0xffffffaf                                        : [4, 6],
        0xffffffcf                                        : [4, 5],
        0x7ffffff7                                        : [3, 31],
        0xbffffff7                                        : [3, 30],
        0xdffffff7                                        : [3, 29],
        0xeffffff7                                        : [3, 28],
        0xf7fffff7                                        : [3, 27],
        0xfbfffff7                                        : [3, 26],
        0xfdfffff7                                        : [3, 25],
        0xfefffff7                                        : [3, 24],
        0xff7ffff7                                        : [3, 23],
        0xffbffff7                                        : [3, 22],
        0xffdffff7                                        : [3, 21],
        0xffeffff7                                        : [3, 20],
        0xfff7fff7                                        : [3, 19],
        0xfffbfff7                                        : [3, 18],
        0xfffdfff7                                        : [3, 17],
        0xfffefff7                                        : [3, 16],
        0xffff7ff7                                        : [3, 15],
        0xffffbff7                                        : [3, 14],
        0xffffdff7                                        : [3, 13],
        0xffffeff7                                        : [3, 12],
        0xfffff7f7                                        : [3, 11],
        0xfffffbf7                                        : [3, 10],
        0xfffffdf7                                        : [3, 9],
        0xfffffef7                                        : [3, 8],
        0xffffff77                                        : [3, 7],
        0xffffffb7                                        : [3, 6],
        0xffffffd7                                        : [3, 5],
        0xffffffe7                                        : [3, 4],
        0x7ffffffb                                        : [2, 31],
        0xbffffffb                                        : [2, 30],
        0xdffffffb                                        : [2, 29],
        0xeffffffb                                        : [2, 28],
        0xf7fffffb                                        : [2, 27],
        0xfbfffffb                                        : [2, 26],
        0xfdfffffb                                        : [2, 25],
        0xfefffffb                                        : [2, 24],
        0xff7ffffb                                        : [2, 23],
        0xffbffffb                                        : [2, 22],
        0xffdffffb                                        : [2, 21],
        0xffeffffb                                        : [2, 20],
        0xfff7fffb                                        : [2, 19],
        0xfffbfffb                                        : [2, 18],
        0xfffdfffb                                        : [2, 17],
        0xfffefffb                                        : [2, 16],
        0xffff7ffb                                        : [2, 15],
        0xffffbffb                                        : [2, 14],
        0xffffdffb                                        : [2, 13],
        0xffffeffb                                        : [2, 12],
        0xfffff7fb                                        : [2, 11],
        0xfffffbfb                                        : [2, 10],
        0xfffffdfb                                        : [2, 9],
        0xfffffefb                                        : [2, 8],
        0xffffff7b                                        : [2, 7],
        0xffffffbb                                        : [2, 6],
        0xffffffdb                                        : [2, 5],
        0xffffffeb                                        : [2, 4],
        0xfffffff3                                        : [2, 3],
        0x7ffffffd                                        : [1, 31],
        0xbffffffd                                        : [1, 30],
        0xdffffffd                                        : [1, 29],
        0xeffffffd                                        : [1, 28],
        0xf7fffffd                                        : [1, 27],
        0xfbfffffd                                        : [1, 26],
        0xfdfffffd                                        : [1, 25],
        0xfefffffd                                        : [1, 24],
        0xff7ffffd                                        : [1, 23],
        0xffbffffd                                        : [1, 22],
        0xffdffffd                                        : [1, 21],
        0xffeffffd                                        : [1, 20],
        0xfff7fffd                                        : [1, 19],
        0xfffbfffd                                        : [1, 18],
        0xfffdfffd                                        : [1, 17],
        0xfffefffd                                        : [1, 16],
        0xffff7ffd                                        : [1, 15],
        0xffffbffd                                        : [1, 14],
        0xffffdffd                                        : [1, 13],
        0xffffeffd                                        : [1, 12],
        0xfffff7fd                                        : [1, 11],
        0xfffffbfd                                        : [1, 10],
        0xfffffdfd                                        : [1, 9],
        0xfffffefd                                        : [1, 8],
        0xffffff7d                                        : [1, 7],
        0xffffffbd                                        : [1, 6],
        0xffffffdd                                        : [1, 5],
        0xffffffed                                        : [1, 4],
        0xfffffff5                                        : [1, 3],
        0xfffffff9                                        : [1, 2],
        0x7ffffffe                                        : [0, 31],
        0xbffffffe                                        : [0, 30],
        0xdffffffe                                        : [0, 29],
        0xeffffffe                                        : [0, 28],
        0xf7fffffe                                        : [0, 27],
        0xfbfffffe                                        : [0, 26],
        0xfdfffffe                                        : [0, 25],
        0xfefffffe                                        : [0, 24],
        0xff7ffffe                                        : [0, 23],
        0xffbffffe                                        : [0, 22],
        0xffdffffe                                        : [0, 21],
        0xffeffffe                                        : [0, 20],
        0xfff7fffe                                        : [0, 19],
        0xfffbfffe                                        : [0, 18],
        0xfffdfffe                                        : [0, 17],
        0xfffefffe                                        : [0, 16],
        0xffff7ffe                                        : [0, 15],
        0xffffbffe                                        : [0, 14],
        0xffffdffe                                        : [0, 13],
        0xffffeffe                                        : [0, 12],
        0xfffff7fe                                        : [0, 11],
        0xfffffbfe                                        : [0, 10],
        0xfffffdfe                                        : [0, 9],
        0xfffffefe                                        : [0, 8],
        0xffffff7e                                        : [0, 7],
        0xffffffbe                                        : [0, 6],
        0xffffffde                                        : [0, 5],
        0xffffffee                                        : [0, 4],
        0xfffffff6                                        : [0, 3],
        0xfffffffa                                        : [0, 2],
        0xfffffffc                                        : [0, 1],
        }

    _shl_one = {
        0xfffffffe                                        : 1,
        0xfffffffc                                        : 2,
        0xfffffff8                                        : 3,
        0xfffffff0                                        : 4,
        0xffffffe0                                        : 5,
        0xffffffc0                                        : 6,
        0xffffff80                                        : 7,
        0xffffff00                                        : 8,
        0xfffffe00                                        : 9,
        0xfffffc00                                        : 10,
        0xfffff800                                        : 11,
        0xfffff000                                        : 12,
        0xffffe000                                        : 13,
        0xffffc000                                        : 14,
        0xffff8000                                        : 15,
        0xffff0000                                        : 16,
        0xfffe0000                                        : 17,
        0xfffc0000                                        : 18,
        0xfff80000                                        : 19,
        0xfff00000                                        : 20,
        0xffe00000                                        : 21,
        0xffc00000                                        : 22,
        0xff800000                                        : 23,
        0xff000000                                        : 24,
        0xfe000000                                        : 25,
        0xfc000000                                        : 26,
        0xf8000000                                        : 27,
        0xf0000000                                        : 28,
        0xe0000000                                        : 29,
        0xc0000000                                        : 30,
        0x80000000                                        : 31,
        }

    _shl_one_one = {
        0xfffffffd                                        : [2, 0],
        0xfffffff9                                        : [3, 0],
        0xfffffffa                                        : [3, 1],
        0xfffffff1                                        : [4, 0],
        0xfffffff2                                        : [4, 1],
        0xfffffff4                                        : [4, 2],
        0xffffffe1                                        : [5, 0],
        0xffffffe2                                        : [5, 1],
        0xffffffe4                                        : [5, 2],
        0xffffffe8                                        : [5, 3],
        0xffffffc1                                        : [6, 0],
        0xffffffc2                                        : [6, 1],
        0xffffffc4                                        : [6, 2],
        0xffffffc8                                        : [6, 3],
        0xffffffd0                                        : [6, 4],
        0xffffff81                                        : [7, 0],
        0xffffff82                                        : [7, 1],
        0xffffff84                                        : [7, 2],
        0xffffff88                                        : [7, 3],
        0xffffff90                                        : [7, 4],
        0xffffffa0                                        : [7, 5],
        0xffffff01                                        : [8, 0],
        0xffffff02                                        : [8, 1],
        0xffffff04                                        : [8, 2],
        0xffffff08                                        : [8, 3],
        0xffffff10                                        : [8, 4],
        0xffffff20                                        : [8, 5],
        0xffffff40                                        : [8, 6],
        0xfffffe01                                        : [9, 0],
        0xfffffe02                                        : [9, 1],
        0xfffffe04                                        : [9, 2],
        0xfffffe08                                        : [9, 3],
        0xfffffe10                                        : [9, 4],
        0xfffffe20                                        : [9, 5],
        0xfffffe40                                        : [9, 6],
        0xfffffe80                                        : [9, 7],
        0xfffffc01                                        : [10, 0],
        0xfffffc02                                        : [10, 1],
        0xfffffc04                                        : [10, 2],
        0xfffffc08                                        : [10, 3],
        0xfffffc10                                        : [10, 4],
        0xfffffc20                                        : [10, 5],
        0xfffffc40                                        : [10, 6],
        0xfffffc80                                        : [10, 7],
        0xfffffd00                                        : [10, 8],
        0xfffff801                                        : [11, 0],
        0xfffff802                                        : [11, 1],
        0xfffff804                                        : [11, 2],
        0xfffff808                                        : [11, 3],
        0xfffff810                                        : [11, 4],
        0xfffff820                                        : [11, 5],
        0xfffff840                                        : [11, 6],
        0xfffff880                                        : [11, 7],
        0xfffff900                                        : [11, 8],
        0xfffffa00                                        : [11, 9],
        0xfffff001                                        : [12, 0],
        0xfffff002                                        : [12, 1],
        0xfffff004                                        : [12, 2],
        0xfffff008                                        : [12, 3],
        0xfffff010                                        : [12, 4],
        0xfffff020                                        : [12, 5],
        0xfffff040                                        : [12, 6],
        0xfffff080                                        : [12, 7],
        0xfffff100                                        : [12, 8],
        0xfffff200                                        : [12, 9],
        0xfffff400                                        : [12, 10],
        0xffffe001                                        : [13, 0],
        0xffffe002                                        : [13, 1],
        0xffffe004                                        : [13, 2],
        0xffffe008                                        : [13, 3],
        0xffffe010                                        : [13, 4],
        0xffffe020                                        : [13, 5],
        0xffffe040                                        : [13, 6],
        0xffffe080                                        : [13, 7],
        0xffffe100                                        : [13, 8],
        0xffffe200                                        : [13, 9],
        0xffffe400                                        : [13, 10],
        0xffffe800                                        : [13, 11],
        0xffffc001                                        : [14, 0],
        0xffffc002                                        : [14, 1],
        0xffffc004                                        : [14, 2],
        0xffffc008                                        : [14, 3],
        0xffffc010                                        : [14, 4],
        0xffffc020                                        : [14, 5],
        0xffffc040                                        : [14, 6],
        0xffffc080                                        : [14, 7],
        0xffffc100                                        : [14, 8],
        0xffffc200                                        : [14, 9],
        0xffffc400                                        : [14, 10],
        0xffffc800                                        : [14, 11],
        0xffffd000                                        : [14, 12],
        0xffff8001                                        : [15, 0],
        0xffff8002                                        : [15, 1],
        0xffff8004                                        : [15, 2],
        0xffff8008                                        : [15, 3],
        0xffff8010                                        : [15, 4],
        0xffff8020                                        : [15, 5],
        0xffff8040                                        : [15, 6],
        0xffff8080                                        : [15, 7],
        0xffff8100                                        : [15, 8],
        0xffff8200                                        : [15, 9],
        0xffff8400                                        : [15, 10],
        0xffff8800                                        : [15, 11],
        0xffff9000                                        : [15, 12],
        0xffffa000                                        : [15, 13],
        0xffff0001                                        : [16, 0],
        0xffff0002                                        : [16, 1],
        0xffff0004                                        : [16, 2],
        0xffff0008                                        : [16, 3],
        0xffff0010                                        : [16, 4],
        0xffff0020                                        : [16, 5],
        0xffff0040                                        : [16, 6],
        0xffff0080                                        : [16, 7],
        0xffff0100                                        : [16, 8],
        0xffff0200                                        : [16, 9],
        0xffff0400                                        : [16, 10],
        0xffff0800                                        : [16, 11],
        0xffff1000                                        : [16, 12],
        0xffff2000                                        : [16, 13],
        0xffff4000                                        : [16, 14],
        0xfffe0001                                        : [17, 0],
        0xfffe0002                                        : [17, 1],
        0xfffe0004                                        : [17, 2],
        0xfffe0008                                        : [17, 3],
        0xfffe0010                                        : [17, 4],
        0xfffe0020                                        : [17, 5],
        0xfffe0040                                        : [17, 6],
        0xfffe0080                                        : [17, 7],
        0xfffe0100                                        : [17, 8],
        0xfffe0200                                        : [17, 9],
        0xfffe0400                                        : [17, 10],
        0xfffe0800                                        : [17, 11],
        0xfffe1000                                        : [17, 12],
        0xfffe2000                                        : [17, 13],
        0xfffe4000                                        : [17, 14],
        0xfffe8000                                        : [17, 15],
        0xfffc0001                                        : [18, 0],
        0xfffc0002                                        : [18, 1],
        0xfffc0004                                        : [18, 2],
        0xfffc0008                                        : [18, 3],
        0xfffc0010                                        : [18, 4],
        0xfffc0020                                        : [18, 5],
        0xfffc0040                                        : [18, 6],
        0xfffc0080                                        : [18, 7],
        0xfffc0100                                        : [18, 8],
        0xfffc0200                                        : [18, 9],
        0xfffc0400                                        : [18, 10],
        0xfffc0800                                        : [18, 11],
        0xfffc1000                                        : [18, 12],
        0xfffc2000                                        : [18, 13],
        0xfffc4000                                        : [18, 14],
        0xfffc8000                                        : [18, 15],
        0xfffd0000                                        : [18, 16],
        0xfff80001                                        : [19, 0],
        0xfff80002                                        : [19, 1],
        0xfff80004                                        : [19, 2],
        0xfff80008                                        : [19, 3],
        0xfff80010                                        : [19, 4],
        0xfff80020                                        : [19, 5],
        0xfff80040                                        : [19, 6],
        0xfff80080                                        : [19, 7],
        0xfff80100                                        : [19, 8],
        0xfff80200                                        : [19, 9],
        0xfff80400                                        : [19, 10],
        0xfff80800                                        : [19, 11],
        0xfff81000                                        : [19, 12],
        0xfff82000                                        : [19, 13],
        0xfff84000                                        : [19, 14],
        0xfff88000                                        : [19, 15],
        0xfff90000                                        : [19, 16],
        0xfffa0000                                        : [19, 17],
        0xfff00001                                        : [20, 0],
        0xfff00002                                        : [20, 1],
        0xfff00004                                        : [20, 2],
        0xfff00008                                        : [20, 3],
        0xfff00010                                        : [20, 4],
        0xfff00020                                        : [20, 5],
        0xfff00040                                        : [20, 6],
        0xfff00080                                        : [20, 7],
        0xfff00100                                        : [20, 8],
        0xfff00200                                        : [20, 9],
        0xfff00400                                        : [20, 10],
        0xfff00800                                        : [20, 11],
        0xfff01000                                        : [20, 12],
        0xfff02000                                        : [20, 13],
        0xfff04000                                        : [20, 14],
        0xfff08000                                        : [20, 15],
        0xfff10000                                        : [20, 16],
        0xfff20000                                        : [20, 17],
        0xfff40000                                        : [20, 18],
        0xffe00001                                        : [21, 0],
        0xffe00002                                        : [21, 1],
        0xffe00004                                        : [21, 2],
        0xffe00008                                        : [21, 3],
        0xffe00010                                        : [21, 4],
        0xffe00020                                        : [21, 5],
        0xffe00040                                        : [21, 6],
        0xffe00080                                        : [21, 7],
        0xffe00100                                        : [21, 8],
        0xffe00200                                        : [21, 9],
        0xffe00400                                        : [21, 10],
        0xffe00800                                        : [21, 11],
        0xffe01000                                        : [21, 12],
        0xffe02000                                        : [21, 13],
        0xffe04000                                        : [21, 14],
        0xffe08000                                        : [21, 15],
        0xffe10000                                        : [21, 16],
        0xffe20000                                        : [21, 17],
        0xffe40000                                        : [21, 18],
        0xffe80000                                        : [21, 19],
        0xffc00001                                        : [22, 0],
        0xffc00002                                        : [22, 1],
        0xffc00004                                        : [22, 2],
        0xffc00008                                        : [22, 3],
        0xffc00010                                        : [22, 4],
        0xffc00020                                        : [22, 5],
        0xffc00040                                        : [22, 6],
        0xffc00080                                        : [22, 7],
        0xffc00100                                        : [22, 8],
        0xffc00200                                        : [22, 9],
        0xffc00400                                        : [22, 10],
        0xffc00800                                        : [22, 11],
        0xffc01000                                        : [22, 12],
        0xffc02000                                        : [22, 13],
        0xffc04000                                        : [22, 14],
        0xffc08000                                        : [22, 15],
        0xffc10000                                        : [22, 16],
        0xffc20000                                        : [22, 17],
        0xffc40000                                        : [22, 18],
        0xffc80000                                        : [22, 19],
        0xffd00000                                        : [22, 20],
        0xff800001                                        : [23, 0],
        0xff800002                                        : [23, 1],
        0xff800004                                        : [23, 2],
        0xff800008                                        : [23, 3],
        0xff800010                                        : [23, 4],
        0xff800020                                        : [23, 5],
        0xff800040                                        : [23, 6],
        0xff800080                                        : [23, 7],
        0xff800100                                        : [23, 8],
        0xff800200                                        : [23, 9],
        0xff800400                                        : [23, 10],
        0xff800800                                        : [23, 11],
        0xff801000                                        : [23, 12],
        0xff802000                                        : [23, 13],
        0xff804000                                        : [23, 14],
        0xff808000                                        : [23, 15],
        0xff810000                                        : [23, 16],
        0xff820000                                        : [23, 17],
        0xff840000                                        : [23, 18],
        0xff880000                                        : [23, 19],
        0xff900000                                        : [23, 20],
        0xffa00000                                        : [23, 21],
        0xff000001                                        : [24, 0],
        0xff000002                                        : [24, 1],
        0xff000004                                        : [24, 2],
        0xff000008                                        : [24, 3],
        0xff000010                                        : [24, 4],
        0xff000020                                        : [24, 5],
        0xff000040                                        : [24, 6],
        0xff000080                                        : [24, 7],
        0xff000100                                        : [24, 8],
        0xff000200                                        : [24, 9],
        0xff000400                                        : [24, 10],
        0xff000800                                        : [24, 11],
        0xff001000                                        : [24, 12],
        0xff002000                                        : [24, 13],
        0xff004000                                        : [24, 14],
        0xff008000                                        : [24, 15],
        0xff010000                                        : [24, 16],
        0xff020000                                        : [24, 17],
        0xff040000                                        : [24, 18],
        0xff080000                                        : [24, 19],
        0xff100000                                        : [24, 20],
        0xff200000                                        : [24, 21],
        0xff400000                                        : [24, 22],
        0xfe000001                                        : [25, 0],
        0xfe000002                                        : [25, 1],
        0xfe000004                                        : [25, 2],
        0xfe000008                                        : [25, 3],
        0xfe000010                                        : [25, 4],
        0xfe000020                                        : [25, 5],
        0xfe000040                                        : [25, 6],
        0xfe000080                                        : [25, 7],
        0xfe000100                                        : [25, 8],
        0xfe000200                                        : [25, 9],
        0xfe000400                                        : [25, 10],
        0xfe000800                                        : [25, 11],
        0xfe001000                                        : [25, 12],
        0xfe002000                                        : [25, 13],
        0xfe004000                                        : [25, 14],
        0xfe008000                                        : [25, 15],
        0xfe010000                                        : [25, 16],
        0xfe020000                                        : [25, 17],
        0xfe040000                                        : [25, 18],
        0xfe080000                                        : [25, 19],
        0xfe100000                                        : [25, 20],
        0xfe200000                                        : [25, 21],
        0xfe400000                                        : [25, 22],
        0xfe800000                                        : [25, 23],
        0xfc000001                                        : [26, 0],
        0xfc000002                                        : [26, 1],
        0xfc000004                                        : [26, 2],
        0xfc000008                                        : [26, 3],
        0xfc000010                                        : [26, 4],
        0xfc000020                                        : [26, 5],
        0xfc000040                                        : [26, 6],
        0xfc000080                                        : [26, 7],
        0xfc000100                                        : [26, 8],
        0xfc000200                                        : [26, 9],
        0xfc000400                                        : [26, 10],
        0xfc000800                                        : [26, 11],
        0xfc001000                                        : [26, 12],
        0xfc002000                                        : [26, 13],
        0xfc004000                                        : [26, 14],
        0xfc008000                                        : [26, 15],
        0xfc010000                                        : [26, 16],
        0xfc020000                                        : [26, 17],
        0xfc040000                                        : [26, 18],
        0xfc080000                                        : [26, 19],
        0xfc100000                                        : [26, 20],
        0xfc200000                                        : [26, 21],
        0xfc400000                                        : [26, 22],
        0xfc800000                                        : [26, 23],
        0xfd000000                                        : [26, 24],
        0xf8000001                                        : [27, 0],
        0xf8000002                                        : [27, 1],
        0xf8000004                                        : [27, 2],
        0xf8000008                                        : [27, 3],
        0xf8000010                                        : [27, 4],
        0xf8000020                                        : [27, 5],
        0xf8000040                                        : [27, 6],
        0xf8000080                                        : [27, 7],
        0xf8000100                                        : [27, 8],
        0xf8000200                                        : [27, 9],
        0xf8000400                                        : [27, 10],
        0xf8000800                                        : [27, 11],
        0xf8001000                                        : [27, 12],
        0xf8002000                                        : [27, 13],
        0xf8004000                                        : [27, 14],
        0xf8008000                                        : [27, 15],
        0xf8010000                                        : [27, 16],
        0xf8020000                                        : [27, 17],
        0xf8040000                                        : [27, 18],
        0xf8080000                                        : [27, 19],
        0xf8100000                                        : [27, 20],
        0xf8200000                                        : [27, 21],
        0xf8400000                                        : [27, 22],
        0xf8800000                                        : [27, 23],
        0xf9000000                                        : [27, 24],
        0xfa000000                                        : [27, 25],
        0xf0000001                                        : [28, 0],
        0xf0000002                                        : [28, 1],
        0xf0000004                                        : [28, 2],
        0xf0000008                                        : [28, 3],
        0xf0000010                                        : [28, 4],
        0xf0000020                                        : [28, 5],
        0xf0000040                                        : [28, 6],
        0xf0000080                                        : [28, 7],
        0xf0000100                                        : [28, 8],
        0xf0000200                                        : [28, 9],
        0xf0000400                                        : [28, 10],
        0xf0000800                                        : [28, 11],
        0xf0001000                                        : [28, 12],
        0xf0002000                                        : [28, 13],
        0xf0004000                                        : [28, 14],
        0xf0008000                                        : [28, 15],
        0xf0010000                                        : [28, 16],
        0xf0020000                                        : [28, 17],
        0xf0040000                                        : [28, 18],
        0xf0080000                                        : [28, 19],
        0xf0100000                                        : [28, 20],
        0xf0200000                                        : [28, 21],
        0xf0400000                                        : [28, 22],
        0xf0800000                                        : [28, 23],
        0xf1000000                                        : [28, 24],
        0xf2000000                                        : [28, 25],
        0xf4000000                                        : [28, 26],
        0xe0000001                                        : [29, 0],
        0xe0000002                                        : [29, 1],
        0xe0000004                                        : [29, 2],
        0xe0000008                                        : [29, 3],
        0xe0000010                                        : [29, 4],
        0xe0000020                                        : [29, 5],
        0xe0000040                                        : [29, 6],
        0xe0000080                                        : [29, 7],
        0xe0000100                                        : [29, 8],
        0xe0000200                                        : [29, 9],
        0xe0000400                                        : [29, 10],
        0xe0000800                                        : [29, 11],
        0xe0001000                                        : [29, 12],
        0xe0002000                                        : [29, 13],
        0xe0004000                                        : [29, 14],
        0xe0008000                                        : [29, 15],
        0xe0010000                                        : [29, 16],
        0xe0020000                                        : [29, 17],
        0xe0040000                                        : [29, 18],
        0xe0080000                                        : [29, 19],
        0xe0100000                                        : [29, 20],
        0xe0200000                                        : [29, 21],
        0xe0400000                                        : [29, 22],
        0xe0800000                                        : [29, 23],
        0xe1000000                                        : [29, 24],
        0xe2000000                                        : [29, 25],
        0xe4000000                                        : [29, 26],
        0xe8000000                                        : [29, 27],
        0xc0000001                                        : [30, 0],
        0xc0000002                                        : [30, 1],
        0xc0000004                                        : [30, 2],
        0xc0000008                                        : [30, 3],
        0xc0000010                                        : [30, 4],
        0xc0000020                                        : [30, 5],
        0xc0000040                                        : [30, 6],
        0xc0000080                                        : [30, 7],
        0xc0000100                                        : [30, 8],
        0xc0000200                                        : [30, 9],
        0xc0000400                                        : [30, 10],
        0xc0000800                                        : [30, 11],
        0xc0001000                                        : [30, 12],
        0xc0002000                                        : [30, 13],
        0xc0004000                                        : [30, 14],
        0xc0008000                                        : [30, 15],
        0xc0010000                                        : [30, 16],
        0xc0020000                                        : [30, 17],
        0xc0040000                                        : [30, 18],
        0xc0080000                                        : [30, 19],
        0xc0100000                                        : [30, 20],
        0xc0200000                                        : [30, 21],
        0xc0400000                                        : [30, 22],
        0xc0800000                                        : [30, 23],
        0xc1000000                                        : [30, 24],
        0xc2000000                                        : [30, 25],
        0xc4000000                                        : [30, 26],
        0xc8000000                                        : [30, 27],
        0xd0000000                                        : [30, 28],
        0x80000001                                        : [31, 0],
        0x80000002                                        : [31, 1],
        0x80000004                                        : [31, 2],
        0x80000008                                        : [31, 3],
        0x80000010                                        : [31, 4],
        0x80000020                                        : [31, 5],
        0x80000040                                        : [31, 6],
        0x80000080                                        : [31, 7],
        0x80000100                                        : [31, 8],
        0x80000200                                        : [31, 9],
        0x80000400                                        : [31, 10],
        0x80000800                                        : [31, 11],
        0x80001000                                        : [31, 12],
        0x80002000                                        : [31, 13],
        0x80004000                                        : [31, 14],
        0x80008000                                        : [31, 15],
        0x80010000                                        : [31, 16],
        0x80020000                                        : [31, 17],
        0x80040000                                        : [31, 18],
        0x80080000                                        : [31, 19],
        0x80100000                                        : [31, 20],
        0x80200000                                        : [31, 21],
        0x80400000                                        : [31, 22],
        0x80800000                                        : [31, 23],
        0x81000000                                        : [31, 24],
        0x82000000                                        : [31, 25],
        0x84000000                                        : [31, 26],
        0x88000000                                        : [31, 27],
        0x90000000                                        : [31, 28],
        0xa0000000                                        : [31, 29],
        }

    _shl_one_zero = {
        0xfffffff6                                        : [1, 3],
        0xffffffee                                        : [1, 4],
        0xffffffde                                        : [1, 5],
        0xffffffbe                                        : [1, 6],
        0xffffff7e                                        : [1, 7],
        0xfffffefe                                        : [1, 8],
        0xfffffdfe                                        : [1, 9],
        0xfffffbfe                                        : [1, 10],
        0xfffff7fe                                        : [1, 11],
        0xffffeffe                                        : [1, 12],
        0xffffdffe                                        : [1, 13],
        0xffffbffe                                        : [1, 14],
        0xffff7ffe                                        : [1, 15],
        0xfffefffe                                        : [1, 16],
        0xfffdfffe                                        : [1, 17],
        0xfffbfffe                                        : [1, 18],
        0xfff7fffe                                        : [1, 19],
        0xffeffffe                                        : [1, 20],
        0xffdffffe                                        : [1, 21],
        0xffbffffe                                        : [1, 22],
        0xff7ffffe                                        : [1, 23],
        0xfefffffe                                        : [1, 24],
        0xfdfffffe                                        : [1, 25],
        0xfbfffffe                                        : [1, 26],
        0xf7fffffe                                        : [1, 27],
        0xeffffffe                                        : [1, 28],
        0xdffffffe                                        : [1, 29],
        0xbffffffe                                        : [1, 30],
        0x7ffffffe                                        : [1, 31],
        0xffffffec                                        : [2, 4],
        0xffffffdc                                        : [2, 5],
        0xffffffbc                                        : [2, 6],
        0xffffff7c                                        : [2, 7],
        0xfffffefc                                        : [2, 8],
        0xfffffdfc                                        : [2, 9],
        0xfffffbfc                                        : [2, 10],
        0xfffff7fc                                        : [2, 11],
        0xffffeffc                                        : [2, 12],
        0xffffdffc                                        : [2, 13],
        0xffffbffc                                        : [2, 14],
        0xffff7ffc                                        : [2, 15],
        0xfffefffc                                        : [2, 16],
        0xfffdfffc                                        : [2, 17],
        0xfffbfffc                                        : [2, 18],
        0xfff7fffc                                        : [2, 19],
        0xffeffffc                                        : [2, 20],
        0xffdffffc                                        : [2, 21],
        0xffbffffc                                        : [2, 22],
        0xff7ffffc                                        : [2, 23],
        0xfefffffc                                        : [2, 24],
        0xfdfffffc                                        : [2, 25],
        0xfbfffffc                                        : [2, 26],
        0xf7fffffc                                        : [2, 27],
        0xeffffffc                                        : [2, 28],
        0xdffffffc                                        : [2, 29],
        0xbffffffc                                        : [2, 30],
        0x7ffffffc                                        : [2, 31],
        0xffffffd8                                        : [3, 5],
        0xffffffb8                                        : [3, 6],
        0xffffff78                                        : [3, 7],
        0xfffffef8                                        : [3, 8],
        0xfffffdf8                                        : [3, 9],
        0xfffffbf8                                        : [3, 10],
        0xfffff7f8                                        : [3, 11],
        0xffffeff8                                        : [3, 12],
        0xffffdff8                                        : [3, 13],
        0xffffbff8                                        : [3, 14],
        0xffff7ff8                                        : [3, 15],
        0xfffefff8                                        : [3, 16],
        0xfffdfff8                                        : [3, 17],
        0xfffbfff8                                        : [3, 18],
        0xfff7fff8                                        : [3, 19],
        0xffeffff8                                        : [3, 20],
        0xffdffff8                                        : [3, 21],
        0xffbffff8                                        : [3, 22],
        0xff7ffff8                                        : [3, 23],
        0xfefffff8                                        : [3, 24],
        0xfdfffff8                                        : [3, 25],
        0xfbfffff8                                        : [3, 26],
        0xf7fffff8                                        : [3, 27],
        0xeffffff8                                        : [3, 28],
        0xdffffff8                                        : [3, 29],
        0xbffffff8                                        : [3, 30],
        0x7ffffff8                                        : [3, 31],
        0xffffffb0                                        : [4, 6],
        0xffffff70                                        : [4, 7],
        0xfffffef0                                        : [4, 8],
        0xfffffdf0                                        : [4, 9],
        0xfffffbf0                                        : [4, 10],
        0xfffff7f0                                        : [4, 11],
        0xffffeff0                                        : [4, 12],
        0xffffdff0                                        : [4, 13],
        0xffffbff0                                        : [4, 14],
        0xffff7ff0                                        : [4, 15],
        0xfffefff0                                        : [4, 16],
        0xfffdfff0                                        : [4, 17],
        0xfffbfff0                                        : [4, 18],
        0xfff7fff0                                        : [4, 19],
        0xffeffff0                                        : [4, 20],
        0xffdffff0                                        : [4, 21],
        0xffbffff0                                        : [4, 22],
        0xff7ffff0                                        : [4, 23],
        0xfefffff0                                        : [4, 24],
        0xfdfffff0                                        : [4, 25],
        0xfbfffff0                                        : [4, 26],
        0xf7fffff0                                        : [4, 27],
        0xeffffff0                                        : [4, 28],
        0xdffffff0                                        : [4, 29],
        0xbffffff0                                        : [4, 30],
        0x7ffffff0                                        : [4, 31],
        0xffffff60                                        : [5, 7],
        0xfffffee0                                        : [5, 8],
        0xfffffde0                                        : [5, 9],
        0xfffffbe0                                        : [5, 10],
        0xfffff7e0                                        : [5, 11],
        0xffffefe0                                        : [5, 12],
        0xffffdfe0                                        : [5, 13],
        0xffffbfe0                                        : [5, 14],
        0xffff7fe0                                        : [5, 15],
        0xfffeffe0                                        : [5, 16],
        0xfffdffe0                                        : [5, 17],
        0xfffbffe0                                        : [5, 18],
        0xfff7ffe0                                        : [5, 19],
        0xffefffe0                                        : [5, 20],
        0xffdfffe0                                        : [5, 21],
        0xffbfffe0                                        : [5, 22],
        0xff7fffe0                                        : [5, 23],
        0xfeffffe0                                        : [5, 24],
        0xfdffffe0                                        : [5, 25],
        0xfbffffe0                                        : [5, 26],
        0xf7ffffe0                                        : [5, 27],
        0xefffffe0                                        : [5, 28],
        0xdfffffe0                                        : [5, 29],
        0xbfffffe0                                        : [5, 30],
        0x7fffffe0                                        : [5, 31],
        0xfffffec0                                        : [6, 8],
        0xfffffdc0                                        : [6, 9],
        0xfffffbc0                                        : [6, 10],
        0xfffff7c0                                        : [6, 11],
        0xffffefc0                                        : [6, 12],
        0xffffdfc0                                        : [6, 13],
        0xffffbfc0                                        : [6, 14],
        0xffff7fc0                                        : [6, 15],
        0xfffeffc0                                        : [6, 16],
        0xfffdffc0                                        : [6, 17],
        0xfffbffc0                                        : [6, 18],
        0xfff7ffc0                                        : [6, 19],
        0xffefffc0                                        : [6, 20],
        0xffdfffc0                                        : [6, 21],
        0xffbfffc0                                        : [6, 22],
        0xff7fffc0                                        : [6, 23],
        0xfeffffc0                                        : [6, 24],
        0xfdffffc0                                        : [6, 25],
        0xfbffffc0                                        : [6, 26],
        0xf7ffffc0                                        : [6, 27],
        0xefffffc0                                        : [6, 28],
        0xdfffffc0                                        : [6, 29],
        0xbfffffc0                                        : [6, 30],
        0x7fffffc0                                        : [6, 31],
        0xfffffd80                                        : [7, 9],
        0xfffffb80                                        : [7, 10],
        0xfffff780                                        : [7, 11],
        0xffffef80                                        : [7, 12],
        0xffffdf80                                        : [7, 13],
        0xffffbf80                                        : [7, 14],
        0xffff7f80                                        : [7, 15],
        0xfffeff80                                        : [7, 16],
        0xfffdff80                                        : [7, 17],
        0xfffbff80                                        : [7, 18],
        0xfff7ff80                                        : [7, 19],
        0xffefff80                                        : [7, 20],
        0xffdfff80                                        : [7, 21],
        0xffbfff80                                        : [7, 22],
        0xff7fff80                                        : [7, 23],
        0xfeffff80                                        : [7, 24],
        0xfdffff80                                        : [7, 25],
        0xfbffff80                                        : [7, 26],
        0xf7ffff80                                        : [7, 27],
        0xefffff80                                        : [7, 28],
        0xdfffff80                                        : [7, 29],
        0xbfffff80                                        : [7, 30],
        0x7fffff80                                        : [7, 31],
        0xfffffb00                                        : [8, 10],
        0xfffff700                                        : [8, 11],
        0xffffef00                                        : [8, 12],
        0xffffdf00                                        : [8, 13],
        0xffffbf00                                        : [8, 14],
        0xffff7f00                                        : [8, 15],
        0xfffeff00                                        : [8, 16],
        0xfffdff00                                        : [8, 17],
        0xfffbff00                                        : [8, 18],
        0xfff7ff00                                        : [8, 19],
        0xffefff00                                        : [8, 20],
        0xffdfff00                                        : [8, 21],
        0xffbfff00                                        : [8, 22],
        0xff7fff00                                        : [8, 23],
        0xfeffff00                                        : [8, 24],
        0xfdffff00                                        : [8, 25],
        0xfbffff00                                        : [8, 26],
        0xf7ffff00                                        : [8, 27],
        0xefffff00                                        : [8, 28],
        0xdfffff00                                        : [8, 29],
        0xbfffff00                                        : [8, 30],
        0x7fffff00                                        : [8, 31],
        0xfffff600                                        : [9, 11],
        0xffffee00                                        : [9, 12],
        0xffffde00                                        : [9, 13],
        0xffffbe00                                        : [9, 14],
        0xffff7e00                                        : [9, 15],
        0xfffefe00                                        : [9, 16],
        0xfffdfe00                                        : [9, 17],
        0xfffbfe00                                        : [9, 18],
        0xfff7fe00                                        : [9, 19],
        0xffeffe00                                        : [9, 20],
        0xffdffe00                                        : [9, 21],
        0xffbffe00                                        : [9, 22],
        0xff7ffe00                                        : [9, 23],
        0xfefffe00                                        : [9, 24],
        0xfdfffe00                                        : [9, 25],
        0xfbfffe00                                        : [9, 26],
        0xf7fffe00                                        : [9, 27],
        0xeffffe00                                        : [9, 28],
        0xdffffe00                                        : [9, 29],
        0xbffffe00                                        : [9, 30],
        0x7ffffe00                                        : [9, 31],
        0xffffec00                                        : [10, 12],
        0xffffdc00                                        : [10, 13],
        0xffffbc00                                        : [10, 14],
        0xffff7c00                                        : [10, 15],
        0xfffefc00                                        : [10, 16],
        0xfffdfc00                                        : [10, 17],
        0xfffbfc00                                        : [10, 18],
        0xfff7fc00                                        : [10, 19],
        0xffeffc00                                        : [10, 20],
        0xffdffc00                                        : [10, 21],
        0xffbffc00                                        : [10, 22],
        0xff7ffc00                                        : [10, 23],
        0xfefffc00                                        : [10, 24],
        0xfdfffc00                                        : [10, 25],
        0xfbfffc00                                        : [10, 26],
        0xf7fffc00                                        : [10, 27],
        0xeffffc00                                        : [10, 28],
        0xdffffc00                                        : [10, 29],
        0xbffffc00                                        : [10, 30],
        0x7ffffc00                                        : [10, 31],
        0xffffd800                                        : [11, 13],
        0xffffb800                                        : [11, 14],
        0xffff7800                                        : [11, 15],
        0xfffef800                                        : [11, 16],
        0xfffdf800                                        : [11, 17],
        0xfffbf800                                        : [11, 18],
        0xfff7f800                                        : [11, 19],
        0xffeff800                                        : [11, 20],
        0xffdff800                                        : [11, 21],
        0xffbff800                                        : [11, 22],
        0xff7ff800                                        : [11, 23],
        0xfefff800                                        : [11, 24],
        0xfdfff800                                        : [11, 25],
        0xfbfff800                                        : [11, 26],
        0xf7fff800                                        : [11, 27],
        0xeffff800                                        : [11, 28],
        0xdffff800                                        : [11, 29],
        0xbffff800                                        : [11, 30],
        0x7ffff800                                        : [11, 31],
        0xffffb000                                        : [12, 14],
        0xffff7000                                        : [12, 15],
        0xfffef000                                        : [12, 16],
        0xfffdf000                                        : [12, 17],
        0xfffbf000                                        : [12, 18],
        0xfff7f000                                        : [12, 19],
        0xffeff000                                        : [12, 20],
        0xffdff000                                        : [12, 21],
        0xffbff000                                        : [12, 22],
        0xff7ff000                                        : [12, 23],
        0xfefff000                                        : [12, 24],
        0xfdfff000                                        : [12, 25],
        0xfbfff000                                        : [12, 26],
        0xf7fff000                                        : [12, 27],
        0xeffff000                                        : [12, 28],
        0xdffff000                                        : [12, 29],
        0xbffff000                                        : [12, 30],
        0x7ffff000                                        : [12, 31],
        0xffff6000                                        : [13, 15],
        0xfffee000                                        : [13, 16],
        0xfffde000                                        : [13, 17],
        0xfffbe000                                        : [13, 18],
        0xfff7e000                                        : [13, 19],
        0xffefe000                                        : [13, 20],
        0xffdfe000                                        : [13, 21],
        0xffbfe000                                        : [13, 22],
        0xff7fe000                                        : [13, 23],
        0xfeffe000                                        : [13, 24],
        0xfdffe000                                        : [13, 25],
        0xfbffe000                                        : [13, 26],
        0xf7ffe000                                        : [13, 27],
        0xefffe000                                        : [13, 28],
        0xdfffe000                                        : [13, 29],
        0xbfffe000                                        : [13, 30],
        0x7fffe000                                        : [13, 31],
        0xfffec000                                        : [14, 16],
        0xfffdc000                                        : [14, 17],
        0xfffbc000                                        : [14, 18],
        0xfff7c000                                        : [14, 19],
        0xffefc000                                        : [14, 20],
        0xffdfc000                                        : [14, 21],
        0xffbfc000                                        : [14, 22],
        0xff7fc000                                        : [14, 23],
        0xfeffc000                                        : [14, 24],
        0xfdffc000                                        : [14, 25],
        0xfbffc000                                        : [14, 26],
        0xf7ffc000                                        : [14, 27],
        0xefffc000                                        : [14, 28],
        0xdfffc000                                        : [14, 29],
        0xbfffc000                                        : [14, 30],
        0x7fffc000                                        : [14, 31],
        0xfffd8000                                        : [15, 17],
        0xfffb8000                                        : [15, 18],
        0xfff78000                                        : [15, 19],
        0xffef8000                                        : [15, 20],
        0xffdf8000                                        : [15, 21],
        0xffbf8000                                        : [15, 22],
        0xff7f8000                                        : [15, 23],
        0xfeff8000                                        : [15, 24],
        0xfdff8000                                        : [15, 25],
        0xfbff8000                                        : [15, 26],
        0xf7ff8000                                        : [15, 27],
        0xefff8000                                        : [15, 28],
        0xdfff8000                                        : [15, 29],
        0xbfff8000                                        : [15, 30],
        0x7fff8000                                        : [15, 31],
        0xfffb0000                                        : [16, 18],
        0xfff70000                                        : [16, 19],
        0xffef0000                                        : [16, 20],
        0xffdf0000                                        : [16, 21],
        0xffbf0000                                        : [16, 22],
        0xff7f0000                                        : [16, 23],
        0xfeff0000                                        : [16, 24],
        0xfdff0000                                        : [16, 25],
        0xfbff0000                                        : [16, 26],
        0xf7ff0000                                        : [16, 27],
        0xefff0000                                        : [16, 28],
        0xdfff0000                                        : [16, 29],
        0xbfff0000                                        : [16, 30],
        0x7fff0000                                        : [16, 31],
        0xfff60000                                        : [17, 19],
        0xffee0000                                        : [17, 20],
        0xffde0000                                        : [17, 21],
        0xffbe0000                                        : [17, 22],
        0xff7e0000                                        : [17, 23],
        0xfefe0000                                        : [17, 24],
        0xfdfe0000                                        : [17, 25],
        0xfbfe0000                                        : [17, 26],
        0xf7fe0000                                        : [17, 27],
        0xeffe0000                                        : [17, 28],
        0xdffe0000                                        : [17, 29],
        0xbffe0000                                        : [17, 30],
        0x7ffe0000                                        : [17, 31],
        0xffec0000                                        : [18, 20],
        0xffdc0000                                        : [18, 21],
        0xffbc0000                                        : [18, 22],
        0xff7c0000                                        : [18, 23],
        0xfefc0000                                        : [18, 24],
        0xfdfc0000                                        : [18, 25],
        0xfbfc0000                                        : [18, 26],
        0xf7fc0000                                        : [18, 27],
        0xeffc0000                                        : [18, 28],
        0xdffc0000                                        : [18, 29],
        0xbffc0000                                        : [18, 30],
        0x7ffc0000                                        : [18, 31],
        0xffd80000                                        : [19, 21],
        0xffb80000                                        : [19, 22],
        0xff780000                                        : [19, 23],
        0xfef80000                                        : [19, 24],
        0xfdf80000                                        : [19, 25],
        0xfbf80000                                        : [19, 26],
        0xf7f80000                                        : [19, 27],
        0xeff80000                                        : [19, 28],
        0xdff80000                                        : [19, 29],
        0xbff80000                                        : [19, 30],
        0x7ff80000                                        : [19, 31],
        0xffb00000                                        : [20, 22],
        0xff700000                                        : [20, 23],
        0xfef00000                                        : [20, 24],
        0xfdf00000                                        : [20, 25],
        0xfbf00000                                        : [20, 26],
        0xf7f00000                                        : [20, 27],
        0xeff00000                                        : [20, 28],
        0xdff00000                                        : [20, 29],
        0xbff00000                                        : [20, 30],
        0x7ff00000                                        : [20, 31],
        0xff600000                                        : [21, 23],
        0xfee00000                                        : [21, 24],
        0xfde00000                                        : [21, 25],
        0xfbe00000                                        : [21, 26],
        0xf7e00000                                        : [21, 27],
        0xefe00000                                        : [21, 28],
        0xdfe00000                                        : [21, 29],
        0xbfe00000                                        : [21, 30],
        0x7fe00000                                        : [21, 31],
        0xfec00000                                        : [22, 24],
        0xfdc00000                                        : [22, 25],
        0xfbc00000                                        : [22, 26],
        0xf7c00000                                        : [22, 27],
        0xefc00000                                        : [22, 28],
        0xdfc00000                                        : [22, 29],
        0xbfc00000                                        : [22, 30],
        0x7fc00000                                        : [22, 31],
        0xfd800000                                        : [23, 25],
        0xfb800000                                        : [23, 26],
        0xf7800000                                        : [23, 27],
        0xef800000                                        : [23, 28],
        0xdf800000                                        : [23, 29],
        0xbf800000                                        : [23, 30],
        0x7f800000                                        : [23, 31],
        0xfb000000                                        : [24, 26],
        0xf7000000                                        : [24, 27],
        0xef000000                                        : [24, 28],
        0xdf000000                                        : [24, 29],
        0xbf000000                                        : [24, 30],
        0x7f000000                                        : [24, 31],
        0xf6000000                                        : [25, 27],
        0xee000000                                        : [25, 28],
        0xde000000                                        : [25, 29],
        0xbe000000                                        : [25, 30],
        0x7e000000                                        : [25, 31],
        0xec000000                                        : [26, 28],
        0xdc000000                                        : [26, 29],
        0xbc000000                                        : [26, 30],
        0x7c000000                                        : [26, 31],
        0xd8000000                                        : [27, 29],
        0xb8000000                                        : [27, 30],
        0x78000000                                        : [27, 31],
        0xb0000000                                        : [28, 30],
        0x70000000                                        : [28, 31],
        0x60000000                                        : [29, 31],
        }

    _shr_one = {
        0x7fffffff                                        : 1,
        0x3fffffff                                        : 2,
        0x1fffffff                                        : 3,
        0x0fffffff                                        : 4,
        0x07ffffff                                        : 5,
        0x03ffffff                                        : 6,
        0x01ffffff                                        : 7,
        0x00ffffff                                        : 8,
        0x007fffff                                        : 9,
        0x003fffff                                        : 10,
        0x001fffff                                        : 11,
        0x000fffff                                        : 12,
        0x0007ffff                                        : 13,
        0x0003ffff                                        : 14,
        0x0001ffff                                        : 15,
        0x0000ffff                                        : 16,
        0x00007fff                                        : 17,
        0x00003fff                                        : 18,
        0x00001fff                                        : 19,
        0x00000fff                                        : 20,
        0x000007ff                                        : 21,
        0x000003ff                                        : 22,
        0x000001ff                                        : 23,
        0x000000ff                                        : 24,
        0x0000007f                                        : 25,
        0x0000003f                                        : 26,
        0x0000001f                                        : 27,
        0x0000000f                                        : 28,
        0x00000007                                        : 29,
        0x00000003                                        : 30,
        0x00000001                                        : 31,
        }

    _shr_one_one = {
        0xbfffffff                                        : [2, 31],
        0x9fffffff                                        : [3, 31],
        0x5fffffff                                        : [3, 30],
        0x8fffffff                                        : [4, 31],
        0x4fffffff                                        : [4, 30],
        0x2fffffff                                        : [4, 29],
        0x87ffffff                                        : [5, 31],
        0x47ffffff                                        : [5, 30],
        0x27ffffff                                        : [5, 29],
        0x17ffffff                                        : [5, 28],
        0x83ffffff                                        : [6, 31],
        0x43ffffff                                        : [6, 30],
        0x23ffffff                                        : [6, 29],
        0x13ffffff                                        : [6, 28],
        0x0bffffff                                        : [6, 27],
        0x81ffffff                                        : [7, 31],
        0x41ffffff                                        : [7, 30],
        0x21ffffff                                        : [7, 29],
        0x11ffffff                                        : [7, 28],
        0x09ffffff                                        : [7, 27],
        0x05ffffff                                        : [7, 26],
        0x80ffffff                                        : [8, 31],
        0x40ffffff                                        : [8, 30],
        0x20ffffff                                        : [8, 29],
        0x10ffffff                                        : [8, 28],
        0x08ffffff                                        : [8, 27],
        0x04ffffff                                        : [8, 26],
        0x02ffffff                                        : [8, 25],
        0x807fffff                                        : [9, 31],
        0x407fffff                                        : [9, 30],
        0x207fffff                                        : [9, 29],
        0x107fffff                                        : [9, 28],
        0x087fffff                                        : [9, 27],
        0x047fffff                                        : [9, 26],
        0x027fffff                                        : [9, 25],
        0x017fffff                                        : [9, 24],
        0x803fffff                                        : [10, 31],
        0x403fffff                                        : [10, 30],
        0x203fffff                                        : [10, 29],
        0x103fffff                                        : [10, 28],
        0x083fffff                                        : [10, 27],
        0x043fffff                                        : [10, 26],
        0x023fffff                                        : [10, 25],
        0x013fffff                                        : [10, 24],
        0x00bfffff                                        : [10, 23],
        0x801fffff                                        : [11, 31],
        0x401fffff                                        : [11, 30],
        0x201fffff                                        : [11, 29],
        0x101fffff                                        : [11, 28],
        0x081fffff                                        : [11, 27],
        0x041fffff                                        : [11, 26],
        0x021fffff                                        : [11, 25],
        0x011fffff                                        : [11, 24],
        0x009fffff                                        : [11, 23],
        0x005fffff                                        : [11, 22],
        0x800fffff                                        : [12, 31],
        0x400fffff                                        : [12, 30],
        0x200fffff                                        : [12, 29],
        0x100fffff                                        : [12, 28],
        0x080fffff                                        : [12, 27],
        0x040fffff                                        : [12, 26],
        0x020fffff                                        : [12, 25],
        0x010fffff                                        : [12, 24],
        0x008fffff                                        : [12, 23],
        0x004fffff                                        : [12, 22],
        0x002fffff                                        : [12, 21],
        0x8007ffff                                        : [13, 31],
        0x4007ffff                                        : [13, 30],
        0x2007ffff                                        : [13, 29],
        0x1007ffff                                        : [13, 28],
        0x0807ffff                                        : [13, 27],
        0x0407ffff                                        : [13, 26],
        0x0207ffff                                        : [13, 25],
        0x0107ffff                                        : [13, 24],
        0x0087ffff                                        : [13, 23],
        0x0047ffff                                        : [13, 22],
        0x0027ffff                                        : [13, 21],
        0x0017ffff                                        : [13, 20],
        0x8003ffff                                        : [14, 31],
        0x4003ffff                                        : [14, 30],
        0x2003ffff                                        : [14, 29],
        0x1003ffff                                        : [14, 28],
        0x0803ffff                                        : [14, 27],
        0x0403ffff                                        : [14, 26],
        0x0203ffff                                        : [14, 25],
        0x0103ffff                                        : [14, 24],
        0x0083ffff                                        : [14, 23],
        0x0043ffff                                        : [14, 22],
        0x0023ffff                                        : [14, 21],
        0x0013ffff                                        : [14, 20],
        0x000bffff                                        : [14, 19],
        0x8001ffff                                        : [15, 31],
        0x4001ffff                                        : [15, 30],
        0x2001ffff                                        : [15, 29],
        0x1001ffff                                        : [15, 28],
        0x0801ffff                                        : [15, 27],
        0x0401ffff                                        : [15, 26],
        0x0201ffff                                        : [15, 25],
        0x0101ffff                                        : [15, 24],
        0x0081ffff                                        : [15, 23],
        0x0041ffff                                        : [15, 22],
        0x0021ffff                                        : [15, 21],
        0x0011ffff                                        : [15, 20],
        0x0009ffff                                        : [15, 19],
        0x0005ffff                                        : [15, 18],
        0x8000ffff                                        : [16, 31],
        0x4000ffff                                        : [16, 30],
        0x2000ffff                                        : [16, 29],
        0x1000ffff                                        : [16, 28],
        0x0800ffff                                        : [16, 27],
        0x0400ffff                                        : [16, 26],
        0x0200ffff                                        : [16, 25],
        0x0100ffff                                        : [16, 24],
        0x0080ffff                                        : [16, 23],
        0x0040ffff                                        : [16, 22],
        0x0020ffff                                        : [16, 21],
        0x0010ffff                                        : [16, 20],
        0x0008ffff                                        : [16, 19],
        0x0004ffff                                        : [16, 18],
        0x0002ffff                                        : [16, 17],
        0x80007fff                                        : [17, 31],
        0x40007fff                                        : [17, 30],
        0x20007fff                                        : [17, 29],
        0x10007fff                                        : [17, 28],
        0x08007fff                                        : [17, 27],
        0x04007fff                                        : [17, 26],
        0x02007fff                                        : [17, 25],
        0x01007fff                                        : [17, 24],
        0x00807fff                                        : [17, 23],
        0x00407fff                                        : [17, 22],
        0x00207fff                                        : [17, 21],
        0x00107fff                                        : [17, 20],
        0x00087fff                                        : [17, 19],
        0x00047fff                                        : [17, 18],
        0x00027fff                                        : [17, 17],
        0x00017fff                                        : [17, 16],
        0x80003fff                                        : [18, 31],
        0x40003fff                                        : [18, 30],
        0x20003fff                                        : [18, 29],
        0x10003fff                                        : [18, 28],
        0x08003fff                                        : [18, 27],
        0x04003fff                                        : [18, 26],
        0x02003fff                                        : [18, 25],
        0x01003fff                                        : [18, 24],
        0x00803fff                                        : [18, 23],
        0x00403fff                                        : [18, 22],
        0x00203fff                                        : [18, 21],
        0x00103fff                                        : [18, 20],
        0x00083fff                                        : [18, 19],
        0x00043fff                                        : [18, 18],
        0x00023fff                                        : [18, 17],
        0x00013fff                                        : [18, 16],
        0x0000bfff                                        : [18, 15],
        0x80001fff                                        : [19, 31],
        0x40001fff                                        : [19, 30],
        0x20001fff                                        : [19, 29],
        0x10001fff                                        : [19, 28],
        0x08001fff                                        : [19, 27],
        0x04001fff                                        : [19, 26],
        0x02001fff                                        : [19, 25],
        0x01001fff                                        : [19, 24],
        0x00801fff                                        : [19, 23],
        0x00401fff                                        : [19, 22],
        0x00201fff                                        : [19, 21],
        0x00101fff                                        : [19, 20],
        0x00081fff                                        : [19, 19],
        0x00041fff                                        : [19, 18],
        0x00021fff                                        : [19, 17],
        0x00011fff                                        : [19, 16],
        0x00009fff                                        : [19, 15],
        0x00005fff                                        : [19, 14],
        0x80000fff                                        : [20, 31],
        0x40000fff                                        : [20, 30],
        0x20000fff                                        : [20, 29],
        0x10000fff                                        : [20, 28],
        0x08000fff                                        : [20, 27],
        0x04000fff                                        : [20, 26],
        0x02000fff                                        : [20, 25],
        0x01000fff                                        : [20, 24],
        0x00800fff                                        : [20, 23],
        0x00400fff                                        : [20, 22],
        0x00200fff                                        : [20, 21],
        0x00100fff                                        : [20, 20],
        0x00080fff                                        : [20, 19],
        0x00040fff                                        : [20, 18],
        0x00020fff                                        : [20, 17],
        0x00010fff                                        : [20, 16],
        0x00008fff                                        : [20, 15],
        0x00004fff                                        : [20, 14],
        0x00002fff                                        : [20, 13],
        0x800007ff                                        : [21, 31],
        0x400007ff                                        : [21, 30],
        0x200007ff                                        : [21, 29],
        0x100007ff                                        : [21, 28],
        0x080007ff                                        : [21, 27],
        0x040007ff                                        : [21, 26],
        0x020007ff                                        : [21, 25],
        0x010007ff                                        : [21, 24],
        0x008007ff                                        : [21, 23],
        0x004007ff                                        : [21, 22],
        0x002007ff                                        : [21, 21],
        0x001007ff                                        : [21, 20],
        0x000807ff                                        : [21, 19],
        0x000407ff                                        : [21, 18],
        0x000207ff                                        : [21, 17],
        0x000107ff                                        : [21, 16],
        0x000087ff                                        : [21, 15],
        0x000047ff                                        : [21, 14],
        0x000027ff                                        : [21, 13],
        0x000017ff                                        : [21, 12],
        0x800003ff                                        : [22, 31],
        0x400003ff                                        : [22, 30],
        0x200003ff                                        : [22, 29],
        0x100003ff                                        : [22, 28],
        0x080003ff                                        : [22, 27],
        0x040003ff                                        : [22, 26],
        0x020003ff                                        : [22, 25],
        0x010003ff                                        : [22, 24],
        0x008003ff                                        : [22, 23],
        0x004003ff                                        : [22, 22],
        0x002003ff                                        : [22, 21],
        0x001003ff                                        : [22, 20],
        0x000803ff                                        : [22, 19],
        0x000403ff                                        : [22, 18],
        0x000203ff                                        : [22, 17],
        0x000103ff                                        : [22, 16],
        0x000083ff                                        : [22, 15],
        0x000043ff                                        : [22, 14],
        0x000023ff                                        : [22, 13],
        0x000013ff                                        : [22, 12],
        0x00000bff                                        : [22, 11],
        0x800001ff                                        : [23, 31],
        0x400001ff                                        : [23, 30],
        0x200001ff                                        : [23, 29],
        0x100001ff                                        : [23, 28],
        0x080001ff                                        : [23, 27],
        0x040001ff                                        : [23, 26],
        0x020001ff                                        : [23, 25],
        0x010001ff                                        : [23, 24],
        0x008001ff                                        : [23, 23],
        0x004001ff                                        : [23, 22],
        0x002001ff                                        : [23, 21],
        0x001001ff                                        : [23, 20],
        0x000801ff                                        : [23, 19],
        0x000401ff                                        : [23, 18],
        0x000201ff                                        : [23, 17],
        0x000101ff                                        : [23, 16],
        0x000081ff                                        : [23, 15],
        0x000041ff                                        : [23, 14],
        0x000021ff                                        : [23, 13],
        0x000011ff                                        : [23, 12],
        0x000009ff                                        : [23, 11],
        0x000005ff                                        : [23, 10],
        0x800000ff                                        : [24, 31],
        0x400000ff                                        : [24, 30],
        0x200000ff                                        : [24, 29],
        0x100000ff                                        : [24, 28],
        0x080000ff                                        : [24, 27],
        0x040000ff                                        : [24, 26],
        0x020000ff                                        : [24, 25],
        0x010000ff                                        : [24, 24],
        0x008000ff                                        : [24, 23],
        0x004000ff                                        : [24, 22],
        0x002000ff                                        : [24, 21],
        0x001000ff                                        : [24, 20],
        0x000800ff                                        : [24, 19],
        0x000400ff                                        : [24, 18],
        0x000200ff                                        : [24, 17],
        0x000100ff                                        : [24, 16],
        0x000080ff                                        : [24, 15],
        0x000040ff                                        : [24, 14],
        0x000020ff                                        : [24, 13],
        0x000010ff                                        : [24, 12],
        0x000008ff                                        : [24, 11],
        0x000004ff                                        : [24, 10],
        0x000002ff                                        : [24, 9],
        0x8000007f                                        : [25, 31],
        0x4000007f                                        : [25, 30],
        0x2000007f                                        : [25, 29],
        0x1000007f                                        : [25, 28],
        0x0800007f                                        : [25, 27],
        0x0400007f                                        : [25, 26],
        0x0200007f                                        : [25, 25],
        0x0100007f                                        : [25, 24],
        0x0080007f                                        : [25, 23],
        0x0040007f                                        : [25, 22],
        0x0020007f                                        : [25, 21],
        0x0010007f                                        : [25, 20],
        0x0008007f                                        : [25, 19],
        0x0004007f                                        : [25, 18],
        0x0002007f                                        : [25, 17],
        0x0001007f                                        : [25, 16],
        0x0000807f                                        : [25, 15],
        0x0000407f                                        : [25, 14],
        0x0000207f                                        : [25, 13],
        0x0000107f                                        : [25, 12],
        0x0000087f                                        : [25, 11],
        0x0000047f                                        : [25, 10],
        0x0000027f                                        : [25, 9],
        0x0000017f                                        : [25, 8],
        0x8000003f                                        : [26, 31],
        0x4000003f                                        : [26, 30],
        0x2000003f                                        : [26, 29],
        0x1000003f                                        : [26, 28],
        0x0800003f                                        : [26, 27],
        0x0400003f                                        : [26, 26],
        0x0200003f                                        : [26, 25],
        0x0100003f                                        : [26, 24],
        0x0080003f                                        : [26, 23],
        0x0040003f                                        : [26, 22],
        0x0020003f                                        : [26, 21],
        0x0010003f                                        : [26, 20],
        0x0008003f                                        : [26, 19],
        0x0004003f                                        : [26, 18],
        0x0002003f                                        : [26, 17],
        0x0001003f                                        : [26, 16],
        0x0000803f                                        : [26, 15],
        0x0000403f                                        : [26, 14],
        0x0000203f                                        : [26, 13],
        0x0000103f                                        : [26, 12],
        0x0000083f                                        : [26, 11],
        0x0000043f                                        : [26, 10],
        0x0000023f                                        : [26, 9],
        0x0000013f                                        : [26, 8],
        0x000000bf                                        : [26, 7],
        0x8000001f                                        : [27, 31],
        0x4000001f                                        : [27, 30],
        0x2000001f                                        : [27, 29],
        0x1000001f                                        : [27, 28],
        0x0800001f                                        : [27, 27],
        0x0400001f                                        : [27, 26],
        0x0200001f                                        : [27, 25],
        0x0100001f                                        : [27, 24],
        0x0080001f                                        : [27, 23],
        0x0040001f                                        : [27, 22],
        0x0020001f                                        : [27, 21],
        0x0010001f                                        : [27, 20],
        0x0008001f                                        : [27, 19],
        0x0004001f                                        : [27, 18],
        0x0002001f                                        : [27, 17],
        0x0001001f                                        : [27, 16],
        0x0000801f                                        : [27, 15],
        0x0000401f                                        : [27, 14],
        0x0000201f                                        : [27, 13],
        0x0000101f                                        : [27, 12],
        0x0000081f                                        : [27, 11],
        0x0000041f                                        : [27, 10],
        0x0000021f                                        : [27, 9],
        0x0000011f                                        : [27, 8],
        0x0000009f                                        : [27, 7],
        0x0000005f                                        : [27, 6],
        0x8000000f                                        : [28, 31],
        0x4000000f                                        : [28, 30],
        0x2000000f                                        : [28, 29],
        0x1000000f                                        : [28, 28],
        0x0800000f                                        : [28, 27],
        0x0400000f                                        : [28, 26],
        0x0200000f                                        : [28, 25],
        0x0100000f                                        : [28, 24],
        0x0080000f                                        : [28, 23],
        0x0040000f                                        : [28, 22],
        0x0020000f                                        : [28, 21],
        0x0010000f                                        : [28, 20],
        0x0008000f                                        : [28, 19],
        0x0004000f                                        : [28, 18],
        0x0002000f                                        : [28, 17],
        0x0001000f                                        : [28, 16],
        0x0000800f                                        : [28, 15],
        0x0000400f                                        : [28, 14],
        0x0000200f                                        : [28, 13],
        0x0000100f                                        : [28, 12],
        0x0000080f                                        : [28, 11],
        0x0000040f                                        : [28, 10],
        0x0000020f                                        : [28, 9],
        0x0000010f                                        : [28, 8],
        0x0000008f                                        : [28, 7],
        0x0000004f                                        : [28, 6],
        0x0000002f                                        : [28, 5],
        0x80000007                                        : [29, 31],
        0x40000007                                        : [29, 30],
        0x20000007                                        : [29, 29],
        0x10000007                                        : [29, 28],
        0x08000007                                        : [29, 27],
        0x04000007                                        : [29, 26],
        0x02000007                                        : [29, 25],
        0x01000007                                        : [29, 24],
        0x00800007                                        : [29, 23],
        0x00400007                                        : [29, 22],
        0x00200007                                        : [29, 21],
        0x00100007                                        : [29, 20],
        0x00080007                                        : [29, 19],
        0x00040007                                        : [29, 18],
        0x00020007                                        : [29, 17],
        0x00010007                                        : [29, 16],
        0x00008007                                        : [29, 15],
        0x00004007                                        : [29, 14],
        0x00002007                                        : [29, 13],
        0x00001007                                        : [29, 12],
        0x00000807                                        : [29, 11],
        0x00000407                                        : [29, 10],
        0x00000207                                        : [29, 9],
        0x00000107                                        : [29, 8],
        0x00000087                                        : [29, 7],
        0x00000047                                        : [29, 6],
        0x00000027                                        : [29, 5],
        0x00000017                                        : [29, 4],
        0x80000003                                        : [30, 31],
        0x40000003                                        : [30, 30],
        0x20000003                                        : [30, 29],
        0x10000003                                        : [30, 28],
        0x08000003                                        : [30, 27],
        0x04000003                                        : [30, 26],
        0x02000003                                        : [30, 25],
        0x01000003                                        : [30, 24],
        0x00800003                                        : [30, 23],
        0x00400003                                        : [30, 22],
        0x00200003                                        : [30, 21],
        0x00100003                                        : [30, 20],
        0x00080003                                        : [30, 19],
        0x00040003                                        : [30, 18],
        0x00020003                                        : [30, 17],
        0x00010003                                        : [30, 16],
        0x00008003                                        : [30, 15],
        0x00004003                                        : [30, 14],
        0x00002003                                        : [30, 13],
        0x00001003                                        : [30, 12],
        0x00000803                                        : [30, 11],
        0x00000403                                        : [30, 10],
        0x00000203                                        : [30, 9],
        0x00000103                                        : [30, 8],
        0x00000083                                        : [30, 7],
        0x00000043                                        : [30, 6],
        0x00000023                                        : [30, 5],
        0x00000013                                        : [30, 4],
        0x0000000b                                        : [30, 3],
        0x80000001                                        : [31, 31],
        0x40000001                                        : [31, 30],
        0x20000001                                        : [31, 29],
        0x10000001                                        : [31, 28],
        0x08000001                                        : [31, 27],
        0x04000001                                        : [31, 26],
        0x02000001                                        : [31, 25],
        0x01000001                                        : [31, 24],
        0x00800001                                        : [31, 23],
        0x00400001                                        : [31, 22],
        0x00200001                                        : [31, 21],
        0x00100001                                        : [31, 20],
        0x00080001                                        : [31, 19],
        0x00040001                                        : [31, 18],
        0x00020001                                        : [31, 17],
        0x00010001                                        : [31, 16],
        0x00008001                                        : [31, 15],
        0x00004001                                        : [31, 14],
        0x00002001                                        : [31, 13],
        0x00001001                                        : [31, 12],
        0x00000801                                        : [31, 11],
        0x00000401                                        : [31, 10],
        0x00000201                                        : [31, 9],
        0x00000101                                        : [31, 8],
        0x00000081                                        : [31, 7],
        0x00000041                                        : [31, 6],
        0x00000021                                        : [31, 5],
        0x00000011                                        : [31, 4],
        0x00000009                                        : [31, 3],
        0x00000005                                        : [31, 2],
        }

    _shr_one_zero = {
        0x6fffffff                                        : [1, 28],
        0x77ffffff                                        : [1, 27],
        0x7bffffff                                        : [1, 26],
        0x7dffffff                                        : [1, 25],
        0x7effffff                                        : [1, 24],
        0x7f7fffff                                        : [1, 23],
        0x7fbfffff                                        : [1, 22],
        0x7fdfffff                                        : [1, 21],
        0x7fefffff                                        : [1, 20],
        0x7ff7ffff                                        : [1, 19],
        0x7ffbffff                                        : [1, 18],
        0x7ffdffff                                        : [1, 17],
        0x7ffeffff                                        : [1, 16],
        0x7fff7fff                                        : [1, 15],
        0x7fffbfff                                        : [1, 14],
        0x7fffdfff                                        : [1, 13],
        0x7fffefff                                        : [1, 12],
        0x7ffff7ff                                        : [1, 11],
        0x7ffffbff                                        : [1, 10],
        0x7ffffdff                                        : [1, 9],
        0x7ffffeff                                        : [1, 8],
        0x7fffff7f                                        : [1, 7],
        0x7fffffbf                                        : [1, 6],
        0x7fffffdf                                        : [1, 5],
        0x7fffffef                                        : [1, 4],
        0x7ffffff7                                        : [1, 3],
        0x7ffffffb                                        : [1, 2],
        0x7ffffffd                                        : [1, 1],
        0x7ffffffe                                        : [1, 0],
        0x37ffffff                                        : [2, 27],
        0x3bffffff                                        : [2, 26],
        0x3dffffff                                        : [2, 25],
        0x3effffff                                        : [2, 24],
        0x3f7fffff                                        : [2, 23],
        0x3fbfffff                                        : [2, 22],
        0x3fdfffff                                        : [2, 21],
        0x3fefffff                                        : [2, 20],
        0x3ff7ffff                                        : [2, 19],
        0x3ffbffff                                        : [2, 18],
        0x3ffdffff                                        : [2, 17],
        0x3ffeffff                                        : [2, 16],
        0x3fff7fff                                        : [2, 15],
        0x3fffbfff                                        : [2, 14],
        0x3fffdfff                                        : [2, 13],
        0x3fffefff                                        : [2, 12],
        0x3ffff7ff                                        : [2, 11],
        0x3ffffbff                                        : [2, 10],
        0x3ffffdff                                        : [2, 9],
        0x3ffffeff                                        : [2, 8],
        0x3fffff7f                                        : [2, 7],
        0x3fffffbf                                        : [2, 6],
        0x3fffffdf                                        : [2, 5],
        0x3fffffef                                        : [2, 4],
        0x3ffffff7                                        : [2, 3],
        0x3ffffffb                                        : [2, 2],
        0x3ffffffd                                        : [2, 1],
        0x3ffffffe                                        : [2, 0],
        0x1bffffff                                        : [3, 26],
        0x1dffffff                                        : [3, 25],
        0x1effffff                                        : [3, 24],
        0x1f7fffff                                        : [3, 23],
        0x1fbfffff                                        : [3, 22],
        0x1fdfffff                                        : [3, 21],
        0x1fefffff                                        : [3, 20],
        0x1ff7ffff                                        : [3, 19],
        0x1ffbffff                                        : [3, 18],
        0x1ffdffff                                        : [3, 17],
        0x1ffeffff                                        : [3, 16],
        0x1fff7fff                                        : [3, 15],
        0x1fffbfff                                        : [3, 14],
        0x1fffdfff                                        : [3, 13],
        0x1fffefff                                        : [3, 12],
        0x1ffff7ff                                        : [3, 11],
        0x1ffffbff                                        : [3, 10],
        0x1ffffdff                                        : [3, 9],
        0x1ffffeff                                        : [3, 8],
        0x1fffff7f                                        : [3, 7],
        0x1fffffbf                                        : [3, 6],
        0x1fffffdf                                        : [3, 5],
        0x1fffffef                                        : [3, 4],
        0x1ffffff7                                        : [3, 3],
        0x1ffffffb                                        : [3, 2],
        0x1ffffffd                                        : [3, 1],
        0x1ffffffe                                        : [3, 0],
        0x0dffffff                                        : [4, 25],
        0x0effffff                                        : [4, 24],
        0x0f7fffff                                        : [4, 23],
        0x0fbfffff                                        : [4, 22],
        0x0fdfffff                                        : [4, 21],
        0x0fefffff                                        : [4, 20],
        0x0ff7ffff                                        : [4, 19],
        0x0ffbffff                                        : [4, 18],
        0x0ffdffff                                        : [4, 17],
        0x0ffeffff                                        : [4, 16],
        0x0fff7fff                                        : [4, 15],
        0x0fffbfff                                        : [4, 14],
        0x0fffdfff                                        : [4, 13],
        0x0fffefff                                        : [4, 12],
        0x0ffff7ff                                        : [4, 11],
        0x0ffffbff                                        : [4, 10],
        0x0ffffdff                                        : [4, 9],
        0x0ffffeff                                        : [4, 8],
        0x0fffff7f                                        : [4, 7],
        0x0fffffbf                                        : [4, 6],
        0x0fffffdf                                        : [4, 5],
        0x0fffffef                                        : [4, 4],
        0x0ffffff7                                        : [4, 3],
        0x0ffffffb                                        : [4, 2],
        0x0ffffffd                                        : [4, 1],
        0x0ffffffe                                        : [4, 0],
        0x06ffffff                                        : [5, 24],
        0x077fffff                                        : [5, 23],
        0x07bfffff                                        : [5, 22],
        0x07dfffff                                        : [5, 21],
        0x07efffff                                        : [5, 20],
        0x07f7ffff                                        : [5, 19],
        0x07fbffff                                        : [5, 18],
        0x07fdffff                                        : [5, 17],
        0x07feffff                                        : [5, 16],
        0x07ff7fff                                        : [5, 15],
        0x07ffbfff                                        : [5, 14],
        0x07ffdfff                                        : [5, 13],
        0x07ffefff                                        : [5, 12],
        0x07fff7ff                                        : [5, 11],
        0x07fffbff                                        : [5, 10],
        0x07fffdff                                        : [5, 9],
        0x07fffeff                                        : [5, 8],
        0x07ffff7f                                        : [5, 7],
        0x07ffffbf                                        : [5, 6],
        0x07ffffdf                                        : [5, 5],
        0x07ffffef                                        : [5, 4],
        0x07fffff7                                        : [5, 3],
        0x07fffffb                                        : [5, 2],
        0x07fffffd                                        : [5, 1],
        0x07fffffe                                        : [5, 0],
        0x037fffff                                        : [6, 23],
        0x03bfffff                                        : [6, 22],
        0x03dfffff                                        : [6, 21],
        0x03efffff                                        : [6, 20],
        0x03f7ffff                                        : [6, 19],
        0x03fbffff                                        : [6, 18],
        0x03fdffff                                        : [6, 17],
        0x03feffff                                        : [6, 16],
        0x03ff7fff                                        : [6, 15],
        0x03ffbfff                                        : [6, 14],
        0x03ffdfff                                        : [6, 13],
        0x03ffefff                                        : [6, 12],
        0x03fff7ff                                        : [6, 11],
        0x03fffbff                                        : [6, 10],
        0x03fffdff                                        : [6, 9],
        0x03fffeff                                        : [6, 8],
        0x03ffff7f                                        : [6, 7],
        0x03ffffbf                                        : [6, 6],
        0x03ffffdf                                        : [6, 5],
        0x03ffffef                                        : [6, 4],
        0x03fffff7                                        : [6, 3],
        0x03fffffb                                        : [6, 2],
        0x03fffffd                                        : [6, 1],
        0x03fffffe                                        : [6, 0],
        0x01bfffff                                        : [7, 22],
        0x01dfffff                                        : [7, 21],
        0x01efffff                                        : [7, 20],
        0x01f7ffff                                        : [7, 19],
        0x01fbffff                                        : [7, 18],
        0x01fdffff                                        : [7, 17],
        0x01feffff                                        : [7, 16],
        0x01ff7fff                                        : [7, 15],
        0x01ffbfff                                        : [7, 14],
        0x01ffdfff                                        : [7, 13],
        0x01ffefff                                        : [7, 12],
        0x01fff7ff                                        : [7, 11],
        0x01fffbff                                        : [7, 10],
        0x01fffdff                                        : [7, 9],
        0x01fffeff                                        : [7, 8],
        0x01ffff7f                                        : [7, 7],
        0x01ffffbf                                        : [7, 6],
        0x01ffffdf                                        : [7, 5],
        0x01ffffef                                        : [7, 4],
        0x01fffff7                                        : [7, 3],
        0x01fffffb                                        : [7, 2],
        0x01fffffd                                        : [7, 1],
        0x01fffffe                                        : [7, 0],
        0x00dfffff                                        : [8, 21],
        0x00efffff                                        : [8, 20],
        0x00f7ffff                                        : [8, 19],
        0x00fbffff                                        : [8, 18],
        0x00fdffff                                        : [8, 17],
        0x00feffff                                        : [8, 16],
        0x00ff7fff                                        : [8, 15],
        0x00ffbfff                                        : [8, 14],
        0x00ffdfff                                        : [8, 13],
        0x00ffefff                                        : [8, 12],
        0x00fff7ff                                        : [8, 11],
        0x00fffbff                                        : [8, 10],
        0x00fffdff                                        : [8, 9],
        0x00fffeff                                        : [8, 8],
        0x00ffff7f                                        : [8, 7],
        0x00ffffbf                                        : [8, 6],
        0x00ffffdf                                        : [8, 5],
        0x00ffffef                                        : [8, 4],
        0x00fffff7                                        : [8, 3],
        0x00fffffb                                        : [8, 2],
        0x00fffffd                                        : [8, 1],
        0x00fffffe                                        : [8, 0],
        0x006fffff                                        : [9, 20],
        0x0077ffff                                        : [9, 19],
        0x007bffff                                        : [9, 18],
        0x007dffff                                        : [9, 17],
        0x007effff                                        : [9, 16],
        0x007f7fff                                        : [9, 15],
        0x007fbfff                                        : [9, 14],
        0x007fdfff                                        : [9, 13],
        0x007fefff                                        : [9, 12],
        0x007ff7ff                                        : [9, 11],
        0x007ffbff                                        : [9, 10],
        0x007ffdff                                        : [9, 9],
        0x007ffeff                                        : [9, 8],
        0x007fff7f                                        : [9, 7],
        0x007fffbf                                        : [9, 6],
        0x007fffdf                                        : [9, 5],
        0x007fffef                                        : [9, 4],
        0x007ffff7                                        : [9, 3],
        0x007ffffb                                        : [9, 2],
        0x007ffffd                                        : [9, 1],
        0x007ffffe                                        : [9, 0],
        0x0037ffff                                        : [10, 19],
        0x003bffff                                        : [10, 18],
        0x003dffff                                        : [10, 17],
        0x003effff                                        : [10, 16],
        0x003f7fff                                        : [10, 15],
        0x003fbfff                                        : [10, 14],
        0x003fdfff                                        : [10, 13],
        0x003fefff                                        : [10, 12],
        0x003ff7ff                                        : [10, 11],
        0x003ffbff                                        : [10, 10],
        0x003ffdff                                        : [10, 9],
        0x003ffeff                                        : [10, 8],
        0x003fff7f                                        : [10, 7],
        0x003fffbf                                        : [10, 6],
        0x003fffdf                                        : [10, 5],
        0x003fffef                                        : [10, 4],
        0x003ffff7                                        : [10, 3],
        0x003ffffb                                        : [10, 2],
        0x003ffffd                                        : [10, 1],
        0x003ffffe                                        : [10, 0],
        0x001bffff                                        : [11, 18],
        0x001dffff                                        : [11, 17],
        0x001effff                                        : [11, 16],
        0x001f7fff                                        : [11, 15],
        0x001fbfff                                        : [11, 14],
        0x001fdfff                                        : [11, 13],
        0x001fefff                                        : [11, 12],
        0x001ff7ff                                        : [11, 11],
        0x001ffbff                                        : [11, 10],
        0x001ffdff                                        : [11, 9],
        0x001ffeff                                        : [11, 8],
        0x001fff7f                                        : [11, 7],
        0x001fffbf                                        : [11, 6],
        0x001fffdf                                        : [11, 5],
        0x001fffef                                        : [11, 4],
        0x001ffff7                                        : [11, 3],
        0x001ffffb                                        : [11, 2],
        0x001ffffd                                        : [11, 1],
        0x001ffffe                                        : [11, 0],
        0x000dffff                                        : [12, 17],
        0x000effff                                        : [12, 16],
        0x000f7fff                                        : [12, 15],
        0x000fbfff                                        : [12, 14],
        0x000fdfff                                        : [12, 13],
        0x000fefff                                        : [12, 12],
        0x000ff7ff                                        : [12, 11],
        0x000ffbff                                        : [12, 10],
        0x000ffdff                                        : [12, 9],
        0x000ffeff                                        : [12, 8],
        0x000fff7f                                        : [12, 7],
        0x000fffbf                                        : [12, 6],
        0x000fffdf                                        : [12, 5],
        0x000fffef                                        : [12, 4],
        0x000ffff7                                        : [12, 3],
        0x000ffffb                                        : [12, 2],
        0x000ffffd                                        : [12, 1],
        0x000ffffe                                        : [12, 0],
        0x0006ffff                                        : [13, 16],
        0x00077fff                                        : [13, 15],
        0x0007bfff                                        : [13, 14],
        0x0007dfff                                        : [13, 13],
        0x0007efff                                        : [13, 12],
        0x0007f7ff                                        : [13, 11],
        0x0007fbff                                        : [13, 10],
        0x0007fdff                                        : [13, 9],
        0x0007feff                                        : [13, 8],
        0x0007ff7f                                        : [13, 7],
        0x0007ffbf                                        : [13, 6],
        0x0007ffdf                                        : [13, 5],
        0x0007ffef                                        : [13, 4],
        0x0007fff7                                        : [13, 3],
        0x0007fffb                                        : [13, 2],
        0x0007fffd                                        : [13, 1],
        0x0007fffe                                        : [13, 0],
        0x00037fff                                        : [14, 15],
        0x0003bfff                                        : [14, 14],
        0x0003dfff                                        : [14, 13],
        0x0003efff                                        : [14, 12],
        0x0003f7ff                                        : [14, 11],
        0x0003fbff                                        : [14, 10],
        0x0003fdff                                        : [14, 9],
        0x0003feff                                        : [14, 8],
        0x0003ff7f                                        : [14, 7],
        0x0003ffbf                                        : [14, 6],
        0x0003ffdf                                        : [14, 5],
        0x0003ffef                                        : [14, 4],
        0x0003fff7                                        : [14, 3],
        0x0003fffb                                        : [14, 2],
        0x0003fffd                                        : [14, 1],
        0x0003fffe                                        : [14, 0],
        0x0001bfff                                        : [15, 14],
        0x0001dfff                                        : [15, 13],
        0x0001efff                                        : [15, 12],
        0x0001f7ff                                        : [15, 11],
        0x0001fbff                                        : [15, 10],
        0x0001fdff                                        : [15, 9],
        0x0001feff                                        : [15, 8],
        0x0001ff7f                                        : [15, 7],
        0x0001ffbf                                        : [15, 6],
        0x0001ffdf                                        : [15, 5],
        0x0001ffef                                        : [15, 4],
        0x0001fff7                                        : [15, 3],
        0x0001fffb                                        : [15, 2],
        0x0001fffd                                        : [15, 1],
        0x0001fffe                                        : [15, 0],
        0x0000dfff                                        : [16, 13],
        0x0000efff                                        : [16, 12],
        0x0000f7ff                                        : [16, 11],
        0x0000fbff                                        : [16, 10],
        0x0000fdff                                        : [16, 9],
        0x0000feff                                        : [16, 8],
        0x0000ff7f                                        : [16, 7],
        0x0000ffbf                                        : [16, 6],
        0x0000ffdf                                        : [16, 5],
        0x0000ffef                                        : [16, 4],
        0x0000fff7                                        : [16, 3],
        0x0000fffb                                        : [16, 2],
        0x0000fffd                                        : [16, 1],
        0x0000fffe                                        : [16, 0],
        0x00006fff                                        : [17, 12],
        0x000077ff                                        : [17, 11],
        0x00007bff                                        : [17, 10],
        0x00007dff                                        : [17, 9],
        0x00007eff                                        : [17, 8],
        0x00007f7f                                        : [17, 7],
        0x00007fbf                                        : [17, 6],
        0x00007fdf                                        : [17, 5],
        0x00007fef                                        : [17, 4],
        0x00007ff7                                        : [17, 3],
        0x00007ffb                                        : [17, 2],
        0x00007ffd                                        : [17, 1],
        0x00007ffe                                        : [17, 0],
        0x000037ff                                        : [18, 11],
        0x00003bff                                        : [18, 10],
        0x00003dff                                        : [18, 9],
        0x00003eff                                        : [18, 8],
        0x00003f7f                                        : [18, 7],
        0x00003fbf                                        : [18, 6],
        0x00003fdf                                        : [18, 5],
        0x00003fef                                        : [18, 4],
        0x00003ff7                                        : [18, 3],
        0x00003ffb                                        : [18, 2],
        0x00003ffd                                        : [18, 1],
        0x00003ffe                                        : [18, 0],
        0x00001bff                                        : [19, 10],
        0x00001dff                                        : [19, 9],
        0x00001eff                                        : [19, 8],
        0x00001f7f                                        : [19, 7],
        0x00001fbf                                        : [19, 6],
        0x00001fdf                                        : [19, 5],
        0x00001fef                                        : [19, 4],
        0x00001ff7                                        : [19, 3],
        0x00001ffb                                        : [19, 2],
        0x00001ffd                                        : [19, 1],
        0x00001ffe                                        : [19, 0],
        0x00000dff                                        : [20, 9],
        0x00000eff                                        : [20, 8],
        0x00000f7f                                        : [20, 7],
        0x00000fbf                                        : [20, 6],
        0x00000fdf                                        : [20, 5],
        0x00000fef                                        : [20, 4],
        0x00000ff7                                        : [20, 3],
        0x00000ffb                                        : [20, 2],
        0x00000ffd                                        : [20, 1],
        0x00000ffe                                        : [20, 0],
        0x000006ff                                        : [21, 8],
        0x0000077f                                        : [21, 7],
        0x000007bf                                        : [21, 6],
        0x000007df                                        : [21, 5],
        0x000007ef                                        : [21, 4],
        0x000007f7                                        : [21, 3],
        0x000007fb                                        : [21, 2],
        0x000007fd                                        : [21, 1],
        0x000007fe                                        : [21, 0],
        0x0000037f                                        : [22, 7],
        0x000003bf                                        : [22, 6],
        0x000003df                                        : [22, 5],
        0x000003ef                                        : [22, 4],
        0x000003f7                                        : [22, 3],
        0x000003fb                                        : [22, 2],
        0x000003fd                                        : [22, 1],
        0x000003fe                                        : [22, 0],
        0x000001bf                                        : [23, 6],
        0x000001df                                        : [23, 5],
        0x000001ef                                        : [23, 4],
        0x000001f7                                        : [23, 3],
        0x000001fb                                        : [23, 2],
        0x000001fd                                        : [23, 1],
        0x000001fe                                        : [23, 0],
        0x000000df                                        : [24, 5],
        0x000000ef                                        : [24, 4],
        0x000000f7                                        : [24, 3],
        0x000000fb                                        : [24, 2],
        0x000000fd                                        : [24, 1],
        0x000000fe                                        : [24, 0],
        0x0000006f                                        : [25, 4],
        0x00000077                                        : [25, 3],
        0x0000007b                                        : [25, 2],
        0x0000007d                                        : [25, 1],
        0x0000007e                                        : [25, 0],
        0x00000037                                        : [26, 3],
        0x0000003b                                        : [26, 2],
        0x0000003d                                        : [26, 1],
        0x0000003e                                        : [26, 0],
        0x0000001b                                        : [27, 2],
        0x0000001d                                        : [27, 1],
        0x0000001e                                        : [27, 0],
        0x0000000d                                        : [28, 1],
        0x0000000e                                        : [28, 0],
        0x00000006                                        : [29, 0],
        }

    _shr_shl = {
        0x3ffffffe                                        : [3, 1],
        0x1ffffffe                                        : [4, 1],
        0x3ffffffc                                        : [4, 2],
        0x0ffffffe                                        : [5, 1],
        0x1ffffffc                                        : [5, 2],
        0x3ffffff8                                        : [5, 3],
        0x07fffffe                                        : [6, 1],
        0x0ffffffc                                        : [6, 2],
        0x1ffffff8                                        : [6, 3],
        0x3ffffff0                                        : [6, 4],
        0x03fffffe                                        : [7, 1],
        0x07fffffc                                        : [7, 2],
        0x0ffffff8                                        : [7, 3],
        0x1ffffff0                                        : [7, 4],
        0x3fffffe0                                        : [7, 5],
        0x01fffffe                                        : [8, 1],
        0x03fffffc                                        : [8, 2],
        0x07fffff8                                        : [8, 3],
        0x0ffffff0                                        : [8, 4],
        0x1fffffe0                                        : [8, 5],
        0x3fffffc0                                        : [8, 6],
        0x00fffffe                                        : [9, 1],
        0x01fffffc                                        : [9, 2],
        0x03fffff8                                        : [9, 3],
        0x07fffff0                                        : [9, 4],
        0x0fffffe0                                        : [9, 5],
        0x1fffffc0                                        : [9, 6],
        0x3fffff80                                        : [9, 7],
        0x007ffffe                                        : [10, 1],
        0x00fffffc                                        : [10, 2],
        0x01fffff8                                        : [10, 3],
        0x03fffff0                                        : [10, 4],
        0x07ffffe0                                        : [10, 5],
        0x0fffffc0                                        : [10, 6],
        0x1fffff80                                        : [10, 7],
        0x3fffff00                                        : [10, 8],
        0x003ffffe                                        : [11, 1],
        0x007ffffc                                        : [11, 2],
        0x00fffff8                                        : [11, 3],
        0x01fffff0                                        : [11, 4],
        0x03ffffe0                                        : [11, 5],
        0x07ffffc0                                        : [11, 6],
        0x0fffff80                                        : [11, 7],
        0x1fffff00                                        : [11, 8],
        0x3ffffe00                                        : [11, 9],
        0x001ffffe                                        : [12, 1],
        0x003ffffc                                        : [12, 2],
        0x007ffff8                                        : [12, 3],
        0x00fffff0                                        : [12, 4],
        0x01ffffe0                                        : [12, 5],
        0x03ffffc0                                        : [12, 6],
        0x07ffff80                                        : [12, 7],
        0x0fffff00                                        : [12, 8],
        0x1ffffe00                                        : [12, 9],
        0x3ffffc00                                        : [12, 10],
        0x000ffffe                                        : [13, 1],
        0x001ffffc                                        : [13, 2],
        0x003ffff8                                        : [13, 3],
        0x007ffff0                                        : [13, 4],
        0x00ffffe0                                        : [13, 5],
        0x01ffffc0                                        : [13, 6],
        0x03ffff80                                        : [13, 7],
        0x07ffff00                                        : [13, 8],
        0x0ffffe00                                        : [13, 9],
        0x1ffffc00                                        : [13, 10],
        0x3ffff800                                        : [13, 11],
        0x0007fffe                                        : [14, 1],
        0x000ffffc                                        : [14, 2],
        0x001ffff8                                        : [14, 3],
        0x003ffff0                                        : [14, 4],
        0x007fffe0                                        : [14, 5],
        0x00ffffc0                                        : [14, 6],
        0x01ffff80                                        : [14, 7],
        0x03ffff00                                        : [14, 8],
        0x07fffe00                                        : [14, 9],
        0x0ffffc00                                        : [14, 10],
        0x1ffff800                                        : [14, 11],
        0x3ffff000                                        : [14, 12],
        0x0003fffe                                        : [15, 1],
        0x0007fffc                                        : [15, 2],
        0x000ffff8                                        : [15, 3],
        0x001ffff0                                        : [15, 4],
        0x003fffe0                                        : [15, 5],
        0x007fffc0                                        : [15, 6],
        0x00ffff80                                        : [15, 7],
        0x01ffff00                                        : [15, 8],
        0x03fffe00                                        : [15, 9],
        0x07fffc00                                        : [15, 10],
        0x0ffff800                                        : [15, 11],
        0x1ffff000                                        : [15, 12],
        0x3fffe000                                        : [15, 13],
        }

    @staticmethod
    def _msb_one(num) :
        for lv in range(32) :
            if num & (1 << (31 - lv)) :
                return 31 - lv

    @staticmethod
    def _lsb_one(num) :
        for lv in range(32) :
            if num & (1 << lv) :
                return lv + 0

    @staticmethod
    def _is_a_huge_number(num) :
        return (num & 0xffff0000) and\
               (num != 0xffffffff) and\
               not CmmAssgnInstruction._one_one.has_key(num) and\
               not CmmAssgnInstruction._one_zero.has_key(num) and\
               not CmmAssgnInstruction._shl_one.has_key(num) and\
               not CmmAssgnInstruction._shr_one.has_key(num) and\
               not CmmAssgnInstruction._two_one.has_key(num) and\
               not CmmAssgnInstruction._two_zero.has_key(num) and\
               not CmmAssgnInstruction._shl_one_one.has_key(num) and\
               not CmmAssgnInstruction._shr_one_one.has_key(num) and\
               not CmmAssgnInstruction._shl_one_zero.has_key(num) and\
               not CmmAssgnInstruction._shr_one_zero.has_key(num) and\
               not CmmAssgnInstruction._shr_shl.has_key(num) and\
               (CmmAssgnInstruction._msb_one(num) - CmmAssgnInstruction._lsb_one(num) >= 16)

    def is_a_huge_number_assignment(self) :
        if self.body[0] == 'mviw' and self.body[2].is_a_constant() :
            num = self.body[2]

            upper_32 = (num >> 32) & 0xffffffff
            lower_32 = num & 0xffffffff

            return CmmAssgnInstruction._is_a_huge_number(upper_32) or\
                   CmmAssgnInstruction._is_a_huge_number(lower_32)

    def _assign_imm_32(self, reg, imm) :
        imm &= 0xffffffff

        if imm == 0 :
            return [
                CmmAssgnInstruction([ 'move', reg, CmmSR('zero', CmmType('u', 32)) ])
                ]
        elif imm == 0xffffffff :
            return [
                CmmAssgnInstruction([ 'move', reg, CmmSR('one', CmmType('u', 32)) ])
                ]
        elif imm & 0xffff0000 == 0 :
            return [
                CmmAssgnInstruction([ 'mviw', reg, imm ])
                ]
        elif CmmAssgnInstruction._one_one.has_key(imm) :
            return [
                CmmOrigAsmInstruction([ 'bitci', reg, CmmSR('zero', CmmType('u', 32)), CmmConst(CmmAssgnInstruction._one_one[imm]), CmmConst(1) ])
                ]
        elif CmmAssgnInstruction._one_zero.has_key(imm) :
            return [
                CmmOrigAsmInstruction([ 'bitci', reg, CmmSR('one', CmmType('u', 32)), CmmConst(CmmAssgnInstruction._one_zero[imm]), CmmConst(0) ])
                ]
        elif CmmAssgnInstruction._shl_one.has_key(imm) :
            return [
                CmmBinAsInstruction([ 'shli', reg, CmmSR('one', CmmType('u', 32)), CmmConst(CmmAssgnInstruction._shl_one[imm]) ])
                ]
        elif CmmAssgnInstruction._shr_one.has_key(imm) :
            return [
                CmmBinAsInstruction([ 'shri', reg, CmmSR('one', CmmType('u', 32)), CmmConst(CmmAssgnInstruction._shr_one[imm]) ])
                ]
        elif reg.is_a_rf_reg() :
            [reg_1, reg_0] = reg.splt_into_2()

            return [
                CmmAssgnInstruction([ 'mviw', reg_0, imm & 0xffff ] ),
                CmmAssgnInstruction([ 'mviw', reg_1, (imm >> 16) & 0xffff ] ),
                ]
        elif CmmAssgnInstruction._two_one.has_key(imm) :
            [fst, sec] = [ CmmConst(x) for x in CmmAssgnInstruction._two_one[imm] ]

            return [
                CmmOrigAsmInstruction([ 'bitci', reg, CmmSR('zero', CmmType('u', 32)), fst, CmmConst(1) ]),
                CmmOrigAsmInstruction([ 'bitci', reg, reg, sec, CmmConst(1) ]),
                ]
        elif CmmAssgnInstruction._two_zero.has_key(imm) :
            [fst, sec] = [ CmmConst(x) for x in CmmAssgnInstruction._two_zero[imm] ]

            return [
                CmmOrigAsmInstruction([ 'bitci', reg, CmmSR('one', CmmType('u', 32)), fst, CmmConst(0) ]),
                CmmOrigAsmInstruction([ 'bitci', reg, reg, sec, CmmConst(0) ]),
                ]
        elif CmmAssgnInstruction._shl_one_one.has_key(imm) :
            [fst, sec] = [ CmmConst(x) for x in CmmAssgnInstruction._shl_one_one[imm] ]

            return [
                CmmBinAsInstruction([ 'shli', reg, CmmSR('one', CmmType('u', 32)), fst ]),
                CmmOrigAsmInstruction([ 'bitci', reg, reg, sec, CmmConst(1) ]),
                ]
        elif CmmAssgnInstruction._shr_one_one.has_key(imm) :
            [fst, sec] = [ CmmConst(x) for x in CmmAssgnInstruction._shr_one_one[imm] ]

            return [
                CmmBinAsInstruction([ 'shri', reg, CmmSR('one', CmmType('u', 32)), fst ]),
                CmmOrigAsmInstruction([ 'bitci', reg, reg, sec, CmmConst(1) ]),
                ]
        elif CmmAssgnInstruction._shl_one_zero.has_key(imm) :
            [fst, sec] = [ CmmConst(x) for x in CmmAssgnInstruction._shl_one_zero[imm] ]

            return [
                CmmBinAsInstruction([ 'shli', reg, CmmSR('one', CmmType('u', 32)), fst ]),
                CmmOrigAsmInstruction([ 'bitci', reg, reg, sec, CmmConst(0) ]),
                ]
        elif CmmAssgnInstruction._shr_one_zero.has_key(imm) :
            [fst, sec] = [ CmmConst(x) for x in CmmAssgnInstruction._shr_one_zero[imm] ]

            return [
                CmmBinAsInstruction([ 'shri', reg, CmmSR('one', CmmType('u', 32)), fst ]),
                CmmOrigAsmInstruction([ 'bitci', reg, reg, sec, CmmConst(0) ]),
                ]
        elif CmmAssgnInstruction._shr_shl.has_key(imm) :
            [fst, sec] = [ CmmConst(x) for x in CmmAssgnInstruction._shr_shl[imm] ]

            return [
                CmmBinAsInstruction([ 'shri', reg, CmmSR('one', CmmType('u', 32)), fst ]),
                CmmBinAsInstruction([ 'shli', reg, reg, sec ]),
                ]
        else :
            lsb = CmmConst(CmmAssgnInstruction._lsb_one(imm))

            # it is garanteed that $msb - $lsb < 16, other wise register would be from register file

            return [
                CmmAssgnInstruction([ 'mviw', reg, imm >> lsb ]),
                CmmBinAsInstruction([ 'shli', reg, reg, lsb ]),
                ]

    mp_spec_values_zero = {
        0x8000	: CmmConst(31),
        0x4000	: CmmConst(30),
        0x2000	: CmmConst(29),
        0x1000	: CmmConst(28),
        0x0800	: CmmConst(27),
        0x0400	: CmmConst(26),
        0x0200	: CmmConst(25),
        0x0100	: CmmConst(24),
        0x0080	: CmmConst(23),
        0x0040	: CmmConst(22),
        0x0020	: CmmConst(21),
        0x0010	: CmmConst(20),
        0x0008	: CmmConst(19),
        0x0004	: CmmConst(18),
        0x0002	: CmmConst(17),
        0x0001	: CmmConst(16),
    }

    mp_spec_values_one = {
        0x8000	: CmmConst(31),
        0xc000	: CmmConst(30),
        0xe000	: CmmConst(29),
        0xf000	: CmmConst(28),
        0xf800	: CmmConst(27),
        0xfc00	: CmmConst(26),
        0xfe00	: CmmConst(25),
        0xff00	: CmmConst(24),
        0xff80	: CmmConst(23),
        0xffc0	: CmmConst(22),
        0xffe0	: CmmConst(21),
        0xfff0	: CmmConst(20),
        0xfff8	: CmmConst(19),
        0xfffc	: CmmConst(18),
        0xfffe	: CmmConst(17),
        0xffff	: CmmConst(16),
    }

    def _fix_mvi_constant(self) :
        [reg, num] = [ x.get_reg() if x.is_a_var() else x for x in self.body[1 : 3] ]

        if reg.is_a_mp_reg() :
            key = num >> 16
            tmp = reg.new()

            tmp.set_size(16)

            if CmmAssgnInstruction.mp_spec_values_zero.has_key(key) :
                return [
                    CmmOrigAsmInstruction([ 'bitci', reg, CmmSR('zero', CmmType('u', 32)), CmmAssgnInstruction.mp_spec_values_zero[key], CmmConst(1) ])
                    ] +\
                    [ CmmAssgnInstruction([ 'mviw', tmp, num & 0xffff ]) ] if num & 0xffff else []
            elif CmmAssgnInstruction.mp_spec_values_one.has_key(key) :
                return [
                    CmmBinAsInstruction([ 'shli', reg, CmmSR('one', CmmType('u', 32)), CmmAssgnInstruction.mp_spec_values_one[key] ])
                    ] +\
                    [ CmmAssgnInstruction([ 'mviw', tmp, num & 0xffff ]) ] if num & 0xffff else []
            else :
                if key :
                    return [
                        CmmAssgnInstruction([ 'mviw', tmp, key ]),
                        CmmBinAsInstruction([ 'shli', reg, reg, CmmConst(16) ])
                        ] +\
                        [ CmmAssgnInstruction([ 'mviw', tmp, num & 0xffff ]) ] if num & 0xffff else []
                else :
                    return [
                        CmmAssgnInstruction([ 'mviw', reg, num ])
                        ]
        else :
            reg_size = reg.get_size()

            [reg_1, reg_0] = reg.splt_into_2()

            if self.is_a_huge_number_assignment() :
                print ("_fix_mvi_constant 0 reg = " + str(reg) + ", reg1 = " + str(reg_1) + ", reg0 = " + str(reg_0) + ", num = " + str(num) + "\n")
                if reg_size == 64 :
                    return self._assign_imm_32(reg_0, num) + self._assign_imm_32(reg_1, num >> 32)
                elif reg_size == 32 :
                    return self._assign_imm_32(reg, num)
            elif reg_size == 64 :
                if num == 0 :
                    return [
                        CmmAssgnInstruction([ 'move', reg, CmmSR('zero', CmmType('u', 64)) ])
                        ]
                elif (num & 0xffffffffffffffffL) == 0xffffffffffffffffL :
                    return [
                        CmmAssgnInstruction([ 'move', reg, CmmSR('one', CmmType('u', 64)) ])
                        ]
                else :
                    return self._assign_imm_32(reg_0, num) + self._assign_imm_32(reg_1, num >> 32)
            elif reg_size == 32 :
                return self._assign_imm_32(reg, num)
            elif reg_size == 16 :
                return [
                    CmmAssgnInstruction([ 'mviw', reg, num ])        
                    ]
            else : # reg_size == 8
                return [
                    CmmAssgnInstruction([ 'mvib', reg, num & 0xff ])
                    ]

    def _fix_mvi_label(self) :
        [opC, opA] = [ x.get_reg() if x.is_a_var() else x for x in self.body[1 : 3] ]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if opC_size == 64 :
            if (opC == 'cmd') :
                # bug, should be fixed
                return [ self.cwl() ]
            else :
                [opC_1, opC_0] = opC.splt_into_2()

                return [
                    CmmAssgnInstruction([ 'mviw', opC_0, opA] ) ,
                    CmmAssgnInstruction([ 'move', opC_1, CmmSR('zero', CmmType('u', 32)) ])
                    ]
        else :
            return [ self.cwl() ]

    def fix_64__ss_8_8(self) :
        return []

    def fix_64__ss_8_16(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        opC_size = opC.get_size()

        r = opA

        if opA.is_a_rf_reg() :
            while r.get_size() > opC_size :
                [_, r] = r.splt_into_2()
        else :
            if r.get_size() > 32 :
                [_, r] = r.splt_into_2()

            r.set_size(opC_size)

        return [
            CmmAssgnInstruction([ 'move', opC, r ]),
            ]

    def fix_64__ss_16_8(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        opA_size = opA.get_size()
        opC_size = opC.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneC = CmmSR('one', CmmType('u', opC_size))

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', opC, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', opC, opC, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', opC, opC, opA ]),
            ]

    def fix_64__ss_32_64(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        [opA_1, opA_0] = opA.splt_into_2()

        return [
            CmmAssgnInstruction([ 'move', opC, opA_0 ]),
            ]

    def fix_64__ss_64_8(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA_size = opA.get_size()
        opC_0_size = opC_0.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', opC_1, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ 'shli', opC_0, opC_1, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', opC_0, opC_0, opA ]),
            ]

    def fix_64__ss_64_32(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA_size = opA.get_size()
        opC_0_size = opC_0.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))

        return [
            CmmAssgnInstruction([ 'move', opC_0, opA ]),
            CmmBinAsInstruction([ 'add', zeroA, opC_0, opC_0 ]),
            CmmBinAsInstruction([ 'addc', opC_1, oneC_0, oneC_0 ]),
            ]

    def fix_64__ss_64_64(self) :
        return []

    def fix_64__uu_8_8(self) :
        return []

    def fix_64__uu_8_16(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        opC_size = opC.get_size()

        r = opA

        if opA.is_a_rf_reg() :
            while r.get_size() > opC_size :
                [_, r] = r.splt_into_2()
        else :
            if r.get_size() > 32 :
                [_, r] = r.splt_into_2()

            r.set_size(opC_size)

        return [
            CmmAssgnInstruction([ 'move', opC, r ]),
            ]

    def fix_64__uu_16_8(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        opA_size = opA.get_size()
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        return [
            CmmBinAsInstruction([ 'or', opC, zeroA, opA ]),
            ]

    def fix_64__uu_32_64(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        [opA_1, opA_0] = opA.splt_into_2()

        return [
            CmmAssgnInstruction([ 'move', opC, opA_0 ]),
            ]

    def fix_64__uu_64_8(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()

        return [
            CmmBinAsInstruction([ 'or', opC_0, CmmSR('zero', CmmType('u', 32)), opA ]),
            CmmAssgnInstruction([ 'move', opC_1, CmmSR('zero', CmmType('u', 32)) ]),
            ]

    def fix_64__uu_64_32(self) :
        opA = self.get_opA()
        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()

        return [
            CmmAssgnInstruction([ 'move', opC_0, opA]) ,
            CmmAssgnInstruction([ 'move', opC_1, CmmSR('zero', CmmType('u', 32)) ]),
            ]

    def fix_64__uu_64_64(self) :
        return []

    fix_64 = {
        'ss' : {
            8  : {
                8  : fix_64__ss_8_8,
                16 : fix_64__ss_8_16,
                32 : fix_64__ss_8_16,
                64 : fix_64__ss_8_16,
            },
            16 : {
                8  : fix_64__ss_16_8,
                16 : fix_64__ss_8_8,
                32 : fix_64__ss_8_16,
                64 : fix_64__ss_8_16,
            },
            32 : {
                8  : fix_64__ss_16_8,
                16 : fix_64__ss_16_8,
                32 : fix_64__ss_8_8,
                64 : fix_64__ss_32_64,
            },
            64 : {
                8  : fix_64__ss_64_8,
                16 : fix_64__ss_64_8,
                32 : fix_64__ss_64_32,
                64 : fix_64__ss_64_64,
            },
        },
        'uu' : {
            8  : {
                8  : fix_64__uu_8_8,
                16 : fix_64__uu_8_16,
                32 : fix_64__uu_8_16,
                64 : fix_64__uu_8_16,
            },
            16 : {
                8  : fix_64__uu_16_8,
                16 : fix_64__uu_8_8,
                32 : fix_64__uu_8_16,
                64 : fix_64__uu_8_16,
            },
            32 : {
                8  : fix_64__uu_16_8,
                16 : fix_64__uu_16_8,
                32 : fix_64__uu_8_8,
                64 : fix_64__uu_32_64,
            },
            64 : {
                8  : fix_64__uu_64_8,
                16 : fix_64__uu_64_8,
                32 : fix_64__uu_64_32,
                64 : fix_64__uu_64_64,
            },
        },
    }

    def _fix_move(self) :
        opA = self.get_opA()
        opC = self.get_opC()

        opA_size = opA.get_size()
        opC_size = opC.get_size()

        opA_size = self.get_normalized_size(opA_size)
        opC_size = self.get_normalized_size(opC_size)

        [opC, opA] = [ x.get_reg() if x.is_a_var() else x for x in [opC, opA] ]

        self.set_opC(opC)
        self.set_opA(opA)

        if opA.is_signed() and opC.is_signed() :
            return CmmAssgnInstruction.fix_64['ss'][opC_size][opA_size](self)
        elif opA.is_signed() and opC.is_unsigned() :
            return CmmAssgnInstruction.fix_64['uu'][opC_size][opA_size](self)
        elif opA.is_unsigned() and opC.is_signed() :
            return CmmAssgnInstruction.fix_64['ss'][opC_size][opA_size](self)
        else :
            return CmmAssgnInstruction.fix_64['uu'][opC_size][opA_size](self)

    def fix_64_bit_arithmetic(self) :
        instr = []

        if self.is_mviw() :
            if self.body[2].is_a_constant() :
                instr = self._fix_mvi_constant()
            elif self.body[2].is_a_label() :
                instr = self._fix_mvi_label()
        elif self.is_move() :
            instr = self._fix_move()

        self.set_final_asm_form(instr)

    def _load_fix_size(self) :
        [mp] = [ x.get_reg() if x.is_a_var() else x for x in [self.body[3]] ]
        mp_index = mp.get_index()

        self.body.pop()

        self.body += [mp_index] + [CmmConst(0)] * (6 if CmmMisc.get_chip() == 'nP7510' else 4)

    def _store_fix_size(self) :
        [mp] = [ x.get_reg() if x.is_a_var() else x for x in [self.body[3]] ]
        mp_index = mp.get_index()

        self.body.pop()

        self.body += [mp_index] + [CmmConst(0)] * (4 if CmmMisc.get_chip() == 'nP7510' else 5)

        self.body[-2] = CmmConst(0xf)

    def _move_fix_size(self) :
        opA = self.body[2]
        opA_size = opA.get_size()

        self.body += [self.get_asm_operand_size(opA_size, opA_size, '_')]

    def _mvi_fix_size(self) :
        opC = self.body[1]
        opC_size = opC.get_size()

        print("_mvi_fix_size " + str(opC) + " " + str(opC_size) + "\n")
        self.body += [self.get_asm_operand_size(opC_size, '_', '_')]

    def fix_size_operand(self) :
        instr = self.get_final_asm_form()

        if instr :
            for lv in instr :
                return lv.fix_size_operand()
        else :
            if self.is_move() :
                return self._move_fix_size()
            elif self.is_mviw() or self.is_mvib() :
                return self._mvi_fix_size()
            elif self.is_load() :
                return self._load_fix_size()
            else :
                return self._store_fix_size()

    def compute_left(self) :
        op = self.body[0]

        if op == '=' or op == 'move' or op == 'mviw' or op == 'mvib' :
            return [self.body[1]]
        elif op == 'load' :
            return [self.body[2]]
        else :
            return []

    def is_goto(self) :
        self.body[1].is_pc()

    def is_move(self) :
        return self.body[0] == 'move'

    def is_mviw(self) :
        return self.body[0] == 'mviw'

    def is_mvib(self) :
        return self.body[0] == 'mvib'

    def is_load(self) :
        return self.body[0] == 'load'

    def is_store(self) :
        return self.body[0] == 'store'

    def evaluate(self) :
        print("evaluate " + str(self) + " " + str(self.body) + "\n")

        [left] = self.get_left()
        [right] = self.get_operands()

        if left : # there is no 'left' in store
            left.set_value(right.new() if right.is_a_constant() else right.get_value())

    def get_asm_code(self) :
        instr = self.get_final_asm_form()

        if instr :
            return [ x.get_asm_code() for x in instr ]
        else :
            code = ''

            body = [ x.get_reg() if hasattr(x, 'is_a_var') and x.is_a_var() else x for x in self.body ]

            print "get_asm_code 0 body\n"

            code = "\t" + str(body.pop(0)) + "\t"

            if len(body) < 5 :
                print("get_asm_code 1 " + str(body) + "\n")
                print("body is " + str(self.body) + "\n")

                if self.body[0] == "mviw" :
                    const = "0x%04x" % body[1].get_value() if body[1].is_a_constant() else str(body[1])
                    code += str(body.pop(0)) + ",\t" + const + ",\t" + str(body.pop())
                elif self.body[0] == "mvib" :
                    const = "0x%02x" % body[1].get_value()
                    code += str(body.pop(0)) + ",\t" + const + ",\t" + str(body.pop())
                else :
                    if body[0] == body[1] :
                        print("get_asm_code 2 " + str(body) + "\n")
                        return []

                    code += ",\t".join([ str(x) for x in body ])
            elif self.body[0] == 'load' :
                print ("Load " + str(self.body[2]) + " " + str(self.body[2].get_size()) + "\n")
                z = body[0:4]
                body[0:4] = []
                code += ",\t".join([ str(x) for x in z ])
                code += ",  " + ",  ".join([ str(x) for x in  body])
            else :
                body[-2] = "0x%1x" % body[-2].get_value()
                z = body[0:4]
                body[0:4] = []
                code += ",\t".join([ str(x) for x in z ])
                code += ",  " + ",  ".join([ str(x) for x in  body])

            return [ code ]
