from CmmBinAsDummyInstruction import CmmBinAsDummyInstruction
from CmmAssgnDummyInstruction import CmmAssgnDummyInstruction
from CmmLabel import CmmLabel
from CmmShlLabel import CmmShlLabel
from CmmShrLabel import CmmShrLabel
from CmmValueLattice import CmmValueLattice

class CmmBinAsInstruction(CmmInstruction) :
    def swap_opA_opB(self) :
        opA = self.get_opA()
        opB = self.get_opB()

        self.set_opA(opB)
        self.set_opB(opA)

    def get_op(self) :
        return self.body[0]

    def get_opC(self) :
        return self.body[1]

    def set_opC(self, opC) :
        self.body[1] = opC

    def get_opA(self) :
        return self.body[2]

    def set_opA(self, opA) :
        self.body[2] = opA

    def get_opB(self) :
        return self.body[3]

    def set_opB(self, opB) :
        self.body[3] = opB

    def _get_auxiliary_var(self) :
        self.body.pop()

    def _set_auxiliary_var(self, var) :
        self.body += [var]

    @staticmethod
    def _is_not_a_commutative_op(op) :
        return op == '-' or op == '<<' or op == '>>'

    @staticmethod
    def _is_a_commutative_op(op) :
        return op == '+' or op == '&' or op == '|' or op == '^'

    _assignment = {
        '+'                                                 : 0 ,
        '-'                                                 : 0 ,
        '&'                                                 : 0 ,
        '|'                                                 : 0 ,
        '^'                                                 : 0 ,
        '<<'                                                : 0 ,
        '>>'                                                : 0 ,
        }

    def is_set_E_instruction(self) :
        return CmmBinAsInstruction._assignment.has_key(self.body[0])

    def fix_bitfield_def_btws_opC_opA_opB_are_the_same(self) :
        op = self.get_op()
        opC = self.get_opC()

        zero = opC.get_zero()

        if op == '&' or  op == '|' :
            self.set_dead()
        else : # ^
            self.get_opC().set_envelope()
            self.set_op('&')
            self.get_opA().set_envelope()
            self.set_opB(zero)

        return [ self ]

    def fix_bitfield_def_btws_opC_opA_are_the_same(self) :
        op = self.get_op()
        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        instr = []

        mask = opC.get_mask()
        zero = opC.get_zero()
        shft = opC.get_shift_correction_value()

        if opB.is_a_constant() :
            if op == '|' :
                if opB :
                    opC.set_envelope()
                    opA.set_envelope()
                    self.set_opB((opB & mask) << shft)
                else :
                    self.set_dead()
            elif op == '&' :
                if opB & mask != mask :
                    opC.set_envelope()
                    opA.set_envelope()
                    self.set_opB(((opB & mask) << shft) | zero)
                else :
                    self.set_dead()
            elif op == '^' :
                opC.set_envelope()
                opA.set_envelope()
                self.set_opB((((opB & mask) ^ mask) << shft) | zero)

                tmp_C_0 = opC.new()
                tmp_C_1 = tmp_C_0.new()

                self.set_opC(tmp_C_0)

                instr += [CmmBinAsInstruction([ '^', opC, tmp_C_1, CmmSR('one', CmmType('u', opC.get_size())) ])]
        else :
            opC.set_envelope()

            tmp_C_0 = CmmSpecVar('','tmp', CmmType('u', opC.get_envelope_size()))

            self.set_op('&')
            self.set_opA(opB)
            self.set_opB(mask)
            self.set_opC(tmp_C_0)

            if op == '|' :
                tmp_C_1 = tmp_C_0.new()
                tmp_C_2 = tmp_C_0.new()
                tmp_C_3 = tmp_C_0.new()

                if shft :
                    instr += [CmmBinAsInstruction([ '<<', tmp_C_2, tmp_C_1, shft ])]
                instr += [CmmBinAsInstruction([ '|', opC,  opA, tmp_C_3 ])]
            elif op == '&' :
                tmp_C_1 = tmp_C_0.new()
                tmp_C_2 = tmp_C_0.new()
                tmp_C_3 = tmp_C_0.new()
                tmp_C_4 = tmp_C_0.new()
                tmp_C_5 = tmp_C_0.new()

                if shft :
                    instr += [CmmBinAsInstruction([ '<<', tmp_C_2, tmp_C_1, shft ])]
                instr += [
                          CmmBinAsInstruction([ '|', tmp_C_4,  tmp_C_3, zero ]),
                          CmmBinAsInstruction([ '&', opC,  opA, tmp_C_5 ])
                         ]
            elif op == '^' :
                tmp_C_1 = tmp_C_0.new()
                tmp_C_2 = tmp_C_0.new()
                tmp_C_3 = tmp_C_0.new()
                tmp_C_4 = tmp_C_0.new()
                tmp_C_5 = tmp_C_0.new()
                tmp_C_6 = tmp_C_0.new()
                tmp_C_7 = tmp_C_0.new()
                tmp_C_8 = opC.new()
                tmp_C_9 = tmp_C_8.new()

                instr += [CmmBinAsInstruction([ '^', tmp_C_2,  tmp_C_1, mask ])]
                if shft :
                    instr += [CmmBinAsInstruction([ '<<', tmp_C_4, tmp_C_3, shft ])]
                instr += [
                          CmmBinAsInstruction([ '|', tmp_C_6,  tmp_C_5, zero ]),
                          CmmBinAsInstruction([ '^', tmp_C_8,  opA, tmp_C_7 ]),
                          CmmBinAsInstruction([ '^', opC,  tmp_C_9, ((CmmConst(1) << opC.get_envelope_size()) - 1) ])
                         ]

        instr[0:0] = CmmBinAsInstruction( self.body )

        return instr

    def fix_bitfield_def_btws_opA_opB_are_the_same(self) :
        op = self.get_op()
        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        instr = []

        mask = opC.get_mask()
        zero = opC.get_zero()
        shft = opC.get_shift_correction_value()

        if op == '&' or  op == '|' :
            opC.set_envelope()

            tmp_C_0 = opC.new()
            tmp_C_1 = tmp_C_0.new()
            tmp_C_2 = tmp_C_0.new()

            self.set_op('&')
            self.set_opC(tmp_C_1)
            self.set_opA(tmp_C_0)
            self.set_opB(zero)

            tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', opC.get_envelope_size()))
            tmp_A_1 = tmp_A_0.new()
            tmp_A_2 = tmp_A_0.new()
            tmp_A_3 = tmp_A_0.new()

            instr += [CmmBinAsInstruction([ '&', tmp_A_0, opA, mask ])]
            if shft :
                instr += [CmmBinAsInstruction([ '<<', tmp_A_2, tmp_A_1, shft ])]
            instr += [CmmBinAsInstruction([ '|', opC, tmp_C_2, tmp_A_3 ])]
        else : # ^
            self.set_opC('&')
            self.get_opC().set_envelope()
            self.set_opA(self.get_opC().new())
            self.set_opB(zero)

        instr[0:0] = CmmBinAsInstruction( self.body )

        return instr

    def fix_bitfield_def_btws(self) :
        op = self.get_op()
        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        instr = []

        mask = opC.get_mask()
        zero = opC.get_zero()
        shft = opC.get_shift_correction_value()

        opC.set_envelope()

        tmp_C_0 = opC.new()
        tmp_C_1 = tmp_C_0.new()
        tmp_C_2 = tmp_C_0.new()

        self.set_op('&')
        self.set_opC(tmp_C_1)
        self.set_opA(tmp_C_0)
        self.set_opB(zero)

        tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', opC.get_envelope_size()))
        tmp_A_1 = tmp_A_0.new()
        tmp_A_2 = tmp_A_0.new()
        tmp_A_3 = tmp_A_0.new()
        tmp_A_4 = tmp_A_0.new()
        tmp_A_5 = tmp_A_0.new()

        instr += [
                  CmmBinAsInstruction([ op, tmp_A_0, opA, opB ]),
                  CmmBinAsInstruction([ '&', tmp_A_2, tmp_A_1, mask])
                 ]
        if shft :
            instr += [CmmBinAsInstruction([ '<<', tmp_A_4, tmp_A_3, shft ])]
        instr += [CmmBinAsInstruction([ '|', opC, tmp_C_2, tmp_A_5 ])]

        instr[0:0] = CmmBinAsInstruction( self.body )

        return instr

    def fix_bitfield_def_non_btws(self) :
        return self.fix_bitfield_def_btws()

    @staticmethod
    def _is_bitwise(op) :
        return op == '|' or op == '&' or op == '^'

    def fix_bitfield_def(self) :
        op = self.get_op()
        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        instr = []

        if opC.is_a_var() and opC.is_nonalligned() :
            if CmmBinAsInstruction._is_bitwise(op) :
                opC_id = opC.get_id()
                opA_id = opA.get_id()

                if opB.is_a_var() :
                    opB_id = opB.get_id()

                    if opC_id == opA_id and opC_id == opB_id :     # a.x = a.x  a.x ( is a bitwise operation)
                        return self.fix_bitfield_def_btws_opC_opA_opB_are_the_same()
                    elif opC_id == opA_id or opC_id == opB_id :    # a.x = a.x  var
                        if opC_id == opB_id :
                            self.set_opA(opB)
                            self.set_opB(opA)

                        return self.fix_bitfield_def_btws_opC_opA_are_the_same()
                    elif opA_id == opB_id :                        # a.x = var  var
                        return self.fix_bitfield_def_btws_opA_opB_are_the_same()
                    else :                                         # a.x = var0  var1
                        return self.fix_bitfield_def_btws()
                else :
                    if opC_id == opA_id :
                        return self.fix_bitfield_def_btws_opC_opA_are_the_same()
                    else :
                        return self.fix_bitfield_def_btws()
            else :
                return self.fix_bitfield_def_non_btws()
        else :
            return [ self ]

    def process_64_bit_spec_reg(self) :
        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        pre = []
        post = []

        if opC.is_a_spec_reg() and opC_size == 64 :
            tmp_C_0 = CmmSpecVar('','tmp', CmmType('u', opC_size))
            tmp_C_1 = tmp_C_0.new()

            self.set_opC(tmp_C_0)

            post += [CmmAssgnInstruction([ '=', opC, tmp_C_1 ])]

        if opA.is_a_spec_reg() and opA_size == 64 :
            tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', opA_size))
            tmp_A_1 = tmp_A_0.new()

            pre += [CmmAssgnInstruction([ '=', tmp_A_0, opA ])]

            self.set_opA(tmp_A_1)

        if opB.is_a_spec_reg() and opB_size == 64 :
            tmp_B_0 = CmmSpecVar('','tmp', CmmType('u', opB_size))
            tmp_B_1 = tmp_B_0.new()

            pre += [CmmAssgnInstruction([ '=', tmp_B_0, opB ])]

            self.set_opB(tmp_B_1)

        return pre + self.cwl() + post

    def fix_use(self) :
        op = self.get_op()
        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        pre = []
        post = []

        print "start fix_use\n"
        if opA.is_a_var() :
            new_size = self.get_normalized_size(max(opA_size, opC_size))

            if opA.is_nonalligned() :
                pass
        elif opA.is_a_constant() :
            if op != '<<' and op != '>>' :
                if opA_size <= 8 and op != '-' :
                    self.set_opA(opB)
                    self.set_opB(opA)
                else :
                    new_size = self.get_normalized_size(opC_size)

                    if new_size == 64 and opA_size <= 32 :
                        new_size = 32

                    tmp_A_0 = CmmSpecVar('', 'tmp', CmmType('u', new_size))
                    tmp_A_1 = tmp_A_0.new()

                    pre += [CmmAssgnInstruction([ '=', tmp_A_0, opA & ((1 << new_size) - 1)])]

                    self.set_opA(tmp_A_1)
            else :
                tmp_A_0 = CmmSpecVar('', 'tmp', CmmType('u', opA_size))
                tmp_A_1 = tmp_A_0.new()

                self.set_opA(tmp_A_1)

                pre += [CmmAssgnInstruction([ '=', tmp_A_0, opA ])]

        print "fix_use 0\n"
        if opB.is_a_var() :
            print "fix_use 1\n"
            new_size = self.get_normalized_size(max(opB_size, opC_size))

            if opB.is_nonalligned() :
                tmp_B_0 = CmmSpecVar('', 'tmp', CmmType('u', new_size))
                tmp_B_1 = tmp_B_0.new()
                tmp_B_2 = CmmSpecVar('', 'tmp', CmmType('u', new_size))

                pre += [CmmAssgnDummyInstruction([ '=', tmp_B_0, CmmConst(0) ])]
                pre += [CmmBinAsInstruction([ "ext_" + str(new_size) + "_" + str(opB_size), tmp_B_2, tmp_B_1, opB ])]

                self.set_opB(tmp_B_2)
        elif opB.is_a_constant() :
            print "fix_use 2\n"
            if op == '<<' or op == '>>' :
                print "fix_use 3\n"
                if opB_size > 8 : # there is no simple assignment here to preserve flags setting
                    self.set_opB(CmmConst(0xff))
            elif opB_size > 8 :
                print "fix_use 4\n"
                # bug : you should check if RF register will be allocated for opA and if it does then only if size >= 32 and != 40 this changes should be done
                tmp_B_0 = CmmSpecVar('', 'tmp', CmmType('u', opB_size))
                tmp_B_1 = tmp_B_0.new()

                pre += [CmmAssgnInstruction([ '=', tmp_B_0, opB ])]

                self.set_opB(tmp_B_1)

        # push @pre, $self->cwl() ;
        ### 02.11.03 push @pre, CmmBinAsInstruction->new( $self->{body} ) ;

        print "finish fix_use\n"

        return pre + self.cwl() + post

    def fix_memory_usage_0(self) :
        print("fix_memory_usage_and_order_of_args " + str(self.body) + "\n")

        opA = self.get_opA()
        opB = self.get_opB()
        instr = []

        print("opA = " + str(opA) + ", opB = " + str(opB) + "\n")
        if opA.is_a_mem_ref() :
            tmp_A_0 = CmmSpecVar('','mem', CmmType('u', 64))
            tmp_A_1 = tmp_A_0.new()

            self.set_opA(tmp_A_1)

            instr += [CmmAssgnInstruction([ '=', tmp_A_0, opA ])]

        if opB.is_a_mem_ref() :
            tmp_B_0 = CmmSpecVar('','mem', CmmType('u', 64))
            tmp_B_1 = tmp_B_0.new()

            self.set_opB(tmp_B_1)

            instr += [CmmAssgnInstruction([ '=', tmp_B_0, opB ])]

        instr += self.cwl()
        ### 02.11.03 push @instr, CmmBinAsInstruction->new( $self->{body} ) ;

        return instr

    def fix_memory_usage_1(self) :
        instr = []

        op = self.get_op()
        opC = self.get_opC()
        opC_size = opC.get_size()

        if opC.is_a_mem_ref() and opC_size == 64 :
            tmp_C_0 = CmmSpecVar('','tmp', CmmType('u', opC_size))
            tmp_C_1 = tmp_C_0.new()

            self.set_opC(tmp_C_0)

            instr += [CmmAssgnInstruction([ '=', opC, tmp_C_1 ])]

        instr[0:0] = self.cwl()
        ### 02.11.03 unshift @instr, CmmBinAsInstruction->new( $self->{body} ) ;

        return instr

    def put_constant_at_opB(self) :
        opA = self.get_opA()
        opB = self.get_opB()
        instr = []

        if opA.is_a_constant() :
            if CmmBinAsInstruction.CmmBinAsInstruction._is_a_commutative_op(self.get_op()) :
                self.set_opA(opB)
                self.set_opB(opA)

        instr += self.cwl()
        ### 02.11.03 push @instr, CmmBinAsInstruction->new( $self->{body} ) ;

        return instr

    def fix_instr_with_opA_smaller_than_opB(self) :
        instr = []

        op = self.get_op()
        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        bad_diff = {}

        if opA_size and opB_size and opA_size < opB_size :
            if opC_size and opA_size and opC_size < opA_size and op != '>>' : # do nothing
                pass
            else :
                if opC_size == 32 or opC_size == 16 :
                    size = max(opC_size, opB_size)
                else :
                    size = opB_size

                bad_diff[24] = bad_diff[40] = bad_diff[48] = bad_diff[56] = 1

                if op == '-' :
                    if bad_diff.has_key(opB_size - opA_size) :
                        tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', size))
                        tmp_A_1 = tmp_A_0.new()

                        self.set_opA(tmp_A_1)

                        instr += [CmmBinAsInstruction([ '|', tmp_A_0, CmmSR('zero', CmmType('u', size)), opA ])]
                else :
                    self.set_opA(opB)
                    self.set_opB(opA)

        instr += [CmmBinAsInstruction( self.body )]

        return instr

    def fix_instr_with_opC_smaller_than_opA(self) :
        instr = []

        op = self.get_op()
        opC = self.get_opC()
        opA = self.get_opA()
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if opC_size and opA_size and opC_size < opA_size :
            if op == '>>' :
                ## my $tmp_C_0 = CmmSpecVar->new('','tmp', CmmType->new('u', $opA_size)) ;
                ## my $tmp_C_1 = $tmp_C_0->new() ;

                ## $self->get_opC($tmp_C_0) ;

                ## push @instr, CmmAssgnInstruction->new([ '=', $opC, $tmp_C_1 ]) ;

                ## $tmp_C_0->set_use($tmp_C_1) ;
                ## $tmp_C_1->set_def($tmp_C_0) ;
                pass
            else :
                self.get_opA().set_size(opC_size)
                self.get_opB().set_size(opC_size)

        instr[0:0] = [CmmBinAsInstruction( self.body )]

        return instr

    def fix_assignment_with_small_right_side_0(self) :
        instr = []

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        desirable_size = 32 if opC_size == 64 else opC_size

        if opA_size < desirable_size :
            # bug should be rewritten
            tmp_A_0 = CmmSpecVar('','tmp', CmmType('u', desirable_size))
            tmp_A_1 = tmp_A_0.new()

            self.set_opA(tmp_A_1)

            instr += [CmmBinAsInstruction([ '|', tmp_A_0, CmmSR('zero', CmmType('u', desirable_size)), opA ])]

        instr += [CmmBinAsInstruction( self.body )]

        return instr

    def fix_shift_operation(self) :
        print ("I am here !!!!!!!!!!!!! 0 " + str(self.body) + "\n")
        op = self.get_op()
        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        pre_instr = []
        post_instr = []

        print ("I am here !!!!!!!!!!!!! 0 " + str(self.body) + "\n")
        if op == '>>' or op == '<<' :
            if opB.is_a_constant() :
                if opA_size == 64 :
                    if opB & 0x20 :
                        post_instr += [CmmBinAsDummyInstruction([ '|', opA, opA, CmmConst(0xf) ])]
                    else :
                        self.body[5] = CmmSpecVar('', 'shift', CmmType('u', 32))
                        pre_instr += [CmmAssgnDummyInstruction([ '=', self.body[5].new(), CmmConst(0) ])]
                        post_instr += [CmmBinAsDummyInstruction([ '|', self.body[5].new(), self.body[5].new(), opA ])]
                elif opC_size == 64 :
                    if opB & 0x20 :
                        pass
                    else :
                        post_instr += [CmmBinAsDummyInstruction([ '|', opA, opA, CmmConst(0xf) ])]
            else :
                if opA_size == 64 or opB_size == 64 :
                    print ("I am here !!!!!!!!!!!!! 1 " + str(self.body) + "\n")
                    self.body[5] = CmmSpecVar('', 'shift', CmmType('u', 32))
                    pre_instr += [CmmAssgnDummyInstruction([ '=', self.body[5].new(), CmmConst(0) ])]
                    post_instr += [CmmBinAsDummyInstruction([ '|', self.body[5].new(), self.body[5].new(), opA ])]
                    post_instr += [CmmBinAsDummyInstruction([ '|', self.body[5].new(), self.body[5].new(), opB ])]
                elif opC_size == 64 :
                    if op == '<<' :
                        self.body[5] = CmmSpecVar('', 'shift', CmmType('u', opB_size))
                        pre_instr += [CmmAssgnDummyInstruction([ '=', self.body[5].new(), CmmConst(0) ])]
                        post_instr += [CmmBinAsDummyInstruction([ '|', self.body[5].new(), self.body[5].new(), opA ])]
                        post_instr += [CmmBinAsDummyInstruction([ '|', self.body[5].new(), self.body[5].new(), opB ])]

        return pre_instr + [CmmBinAsInstruction( self.body )] + post_instr

    def enlarge_opA(self, size) :
        opA = self.get_opA()
        opA_size = opA.get_size()

        tmp_A_0 = CmmSpecVar('','tmp', CmmType(opA.get_sign(), size))
        tmp_A_1 = tmp_A_0.new()

        self.set_opA(tmp_A_1)

        return CmmAssgnInstruction([ "ext_" + str(size} + "_" + str(opA_size), tmp_A_0, opA ])

    def enlarge_opB(self, size) :
        opB = self.get_opB()
        opB_size = opB.get_size()

        tmp_B_0 = CmmSpecVar('','tmp', CmmType(opB.get_sign(), size))
        tmp_B_1 = tmp_B_0.new()

        self.set_opB(tmp_B_1)

        return CmmBssgnInstruction([ "ext_" + str(size} + "_" + str(opB_size), tmp_B_0, opB ])

    def try_to_put_one_in_RF(self, var, opA, new_opA_size, opB, new_opB_size) :
        instr = []

        if var.can_be_put_in_RF() :
            var.put_in_RF()
        else :
            if opA :
                instr += [self.enlarge_opA(new_opA_size)]

            if opB :
                instr += [self.enlarge_opB(new_opB_size)]

        return instr

    def try_to_put_two_in_RF(self, var0, var1, opA, new_opA_size, opB, new_opB_size, self) :
        instr = []

        if var0.can_be_put_in_RF() and var1.can_be_put_in_RF() :
            var0.put_in_RF()
            var1.put_in_RF()
        else :
            if opA :
                instr += [self.enlarge_opA(new_opA_size)]

            if opB :
                instr += [self.enlarge_opB(new_opB_size)]

        return instr

    def customize_size__8_8_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_8_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_8_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_8_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_16_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_16_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_16_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_16_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_32_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_32_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_32_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_32_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_64_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_64_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_64_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__8_64_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(8)

            opB = self.get_opB()
            opB.set_size(8)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_8_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opC = self.get_opC()

            pre_instr += [self.try_to_put_one_in_RF(opC, self.get_opA(), 16, None, 0)]
        elif op == '-' :
            opC = self.get_opC()

            pre_instr += [self.try_to_put_one_in_RF(opC, self.get_opA(), 16, None, 0)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_8_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()
        elif op == '-' :
            opC = self.get_opC()
            opA = self.get_opA()
            opB = self.get_opB()

            pre_instr += [self.try_to_put_two_in_RF(opC, opB, self.get_opA(), 16, None, 0)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_8_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()

            opA = self.get_opA()
            opA.set_size(16)
        elif op == '-' :
            opC = self.get_opC()
            opB = self.get_opB()

            pre_instr += [self.try_to_put_two_in_RF(opC, opB, self.get_opA(), 16, None, 0)]
            opB.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_8_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()

            opA = self.get_opA()
            opA.set_size(16)
        elif op == '-' :
            opC = self.get_opC()
            opB = self.get_opB()

            pre_instr += [self.try_to_put_two_in_RF(opC, opB, self.get_opA(), 16, None, 0)]
            opB.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_16_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_16_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_16_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opB = self.get_opB()
            opB.set_size(16)
        elif op == '-' :
            opB = self.get_opB()
            opB.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_16_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opB = self.get_opB()
            opB.set_size(16)
        elif op == '-' :
            opB = self.get_opB()
            opB.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_32_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(16)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_32_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(16)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_32_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(16)

            opB = self.get_opB()
            opB.set_size(16)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(16)

            opB = self.get_opB()
            opB.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_32_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(16)

            opB = self.get_opB()
            opB.set_size(16)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(16)

            opB = self.get_opB()
            opB.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_64_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(16)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_64_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(16)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_64_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(16)

            opB = self.get_opB()
            opB.set_size(16)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(16)

            opB = self.get_opB()
            opB.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__16_64_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(16)

            opB = self.get_opB()
            opB.set_size(16)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(16)

            opB = self.get_opB()
            opB.set_size(16)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_8_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pre_instr += [self.enlarge_opA(32)]
        elif op == '-' :
            pre_instr += [self.enlarge_opA(32)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_8_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()

            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '-' :
            pre_instr += [self.enlarge_opA(32)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_8_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()
        elif op == '-' :
            pre_instr += [self.enlarge_opA(32)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_8_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()

            opA = self.get_opA()
            opA.set_size(32)
        elif op == '-' :
            opB = self.get_opB()
            opB.set_size(32)

            pre_instr += [self.enlarge_opA(32)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_16_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '-' :
            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_16_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '-' :
            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_16_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()
        elif op == '-' :
            opC = self.get_opC()
            opA = self.get_opA()
            opB = self.get_opB()

            pre_instr += [self.try_to_put_two_in_RF(opC, opB, opA, 32, None, 0)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_16_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()

            opA = self.get_opA()
            opA.set_size(32)
        elif op == '-' :
            opC = self.get_opC()
            opA = self.get_opA()
            opB = self.get_opB()

            pre_instr += [self.try_to_put_two_in_RF(opC, opB, opA, 32, None, 0)]

            opB.set_size(32)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_32_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_32_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_32_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_32_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opB = self.get_opB()
            opB.set_size(32)
        elif op == '-' :
            opB = self.get_opB()
            opB.set_size(32)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_64_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(32)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(32)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_64_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(32)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(32)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_64_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(32)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(32)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__32_64_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opA = self.get_opA()
            opA.set_size(32)

            opB = self.get_opB()
            opB.set_size(32)
        elif op == '-' :
            opA = self.get_opA()
            opA.set_size(32)

            opB = self.get_opB()
            opB.set_size(32)
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_8_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pre_instr += [self.enlarge_opA(32)]
        elif op == '-' :
            pre_instr += [self.enlarge_opA(32)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_8_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()

            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '-' :
            pre_instr += [self.enlarge_opA(32)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_8_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()
        elif op == '-' :
            pre_instr += [self.enlarge_opA(32)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_8_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()
        elif op == '-' :
            pre_instr += [self.enlarge_opA(32)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_16_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '-' :
            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_16_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '-' :
            opC = self.get_opC()
            opA = self.get_opA()

            pre_instr += [self.try_to_put_one_in_RF(opC, opA, 32, None, 0)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_16_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()
        elif op == '-' :
            opC = self.get_opC()
            opA = self.get_opA()
            opB = self.get_opB()

            pre_instr += [self.try_to_put_two_in_RF(opC, opB, opA, 32, None, 0)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_16_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            self.swap_opA_opB()
        elif op == '-' :
            opC = self.get_opC()
            opA = self.get_opA()
            opB = self.get_opB()

            pre_instr += [self.try_to_put_two_in_RF(opC, opB, opA, 32, None, 0)]
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_32_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_32_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_32_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_32_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_64_8(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_64_16(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_64_32(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    def customize_size__64_64_64(self) :
        op = self.get_op()

        pre_instr = []
        post_instr = []

        if CmmBinAsInstruction._is_a_commutative_op(op) :
            pass
        elif op == '-' :
            pass
        elif op == '<<' :
            pass
        elif op == '>>' :
            pass

        return pre_instr + [self] + post_instr

    customize_size = {
        8  : {
            8  : {
                8  : CmmBinAsInstruction.customize_size__8_8_8,
                16 : CmmBinAsInstruction.customize_size__8_8_16,
                32 : CmmBinAsInstruction.customize_size__8_8_32,
                64 : CmmBinAsInstruction.customize_size__8_8_64,
            },
            16 : {
                8  : CmmBinAsInstruction.customize_size__8_16_8,
                16 : CmmBinAsInstruction.customize_size__8_16_16,
                32 : CmmBinAsInstruction.customize_size__8_16_32,
                64 : CmmBinAsInstruction.customize_size__8_16_64,
            },
            32 : {
                8  : CmmBinAsInstruction.customize_size__8_32_8,
                16 : CmmBinAsInstruction.customize_size__8_32_16,
                32 : CmmBinAsInstruction.customize_size__8_32_32,
                64 : CmmBinAsInstruction.customize_size__8_32_64,
            },
            64 : {
                8  : CmmBinAsInstruction.customize_size__8_64_8,
                16 : CmmBinAsInstruction.customize_size__8_64_16,
                32 : CmmBinAsInstruction.customize_size__8_64_32,
                64 : CmmBinAsInstruction.customize_size__8_64_64,
            },
        },
        16 : {
            8  : {
                8  : CmmBinAsInstruction.customize_size__16_8_8,
                16 : CmmBinAsInstruction.customize_size__16_8_16,
                32 : CmmBinAsInstruction.customize_size__16_8_32,
                64 : CmmBinAsInstruction.customize_size__16_8_64,
            },
            16 : {
                8  : CmmBinAsInstruction.customize_size__16_16_8,
                16 : CmmBinAsInstruction.customize_size__16_16_16,
                32 : CmmBinAsInstruction.customize_size__16_16_32,
                64 : CmmBinAsInstruction.customize_size__16_16_64,
            },
            32 : {
                8  : CmmBinAsInstruction.customize_size__16_32_8,
                16 : CmmBinAsInstruction.customize_size__16_32_16,
                32 : CmmBinAsInstruction.customize_size__16_32_32,
                64 : CmmBinAsInstruction.customize_size__16_32_64,
            },
            64 : {
                8  : CmmBinAsInstruction.customize_size__16_64_8,
                16 : CmmBinAsInstruction.customize_size__16_64_16,
                32 : CmmBinAsInstruction.customize_size__16_64_32,
                64 : CmmBinAsInstruction.customize_size__16_64_64,
            },
        },
        32 : {
            8  : {
                8  : CmmBinAsInstruction.customize_size__32_8_8,
                16 : CmmBinAsInstruction.customize_size__32_8_16,
                32 : CmmBinAsInstruction.customize_size__32_8_32,
                64 : CmmBinAsInstruction.customize_size__32_8_64,
            },
            16 : {
                8  : CmmBinAsInstruction.customize_size__32_16_8,
                16 : CmmBinAsInstruction.customize_size__32_16_16,
                32 : CmmBinAsInstruction.customize_size__32_16_32,
                64 : CmmBinAsInstruction.customize_size__32_16_64,
            },
            32 : {
                8  : CmmBinAsInstruction.customize_size__32_32_8,
                16 : CmmBinAsInstruction.customize_size__32_32_16,
                32 : CmmBinAsInstruction.customize_size__32_32_32,
                64 : CmmBinAsInstruction.customize_size__32_32_64,
            },
            64 : {
                8  : CmmBinAsInstruction.customize_size__32_64_8,
                16 : CmmBinAsInstruction.customize_size__32_64_16,
                32 : CmmBinAsInstruction.customize_size__32_64_32,
                64 : CmmBinAsInstruction.customize_size__32_64_64,
            },
        },
        64 : {
            8  : {
                8  : CmmBinAsInstruction.customize_size__64_8_8,
                16 : CmmBinAsInstruction.customize_size__64_8_16,
                32 : CmmBinAsInstruction.customize_size__64_8_32,
                64 : CmmBinAsInstruction.customize_size__64_8_64,
            },
            16 : {
                8  : CmmBinAsInstruction.customize_size__64_16_8,
                16 : CmmBinAsInstruction.customize_size__64_16_16,
                32 : CmmBinAsInstruction.customize_size__64_16_32,
                64 : CmmBinAsInstruction.customize_size__64_16_64,
            },
            32 : {
                8  : CmmBinAsInstruction.customize_size__64_32_8,
                16 : CmmBinAsInstruction.customize_size__64_32_16,
                32 : CmmBinAsInstruction.customize_size__64_32_32,
                64 : CmmBinAsInstruction.customize_size__64_32_64,
            },
            64 : {
                8  : CmmBinAsInstruction.customize_size__64_64_8,
                16 : CmmBinAsInstruction.customize_size__64_64_16,
                32 : CmmBinAsInstruction.customize_size__64_64_32,
                64 : CmmBinAsInstruction.customize_size__64_64_64,
            },
        },
    }

    def customize_in_accordance_with_sizes(self) :
        instr = []

        op = self.get_op()
        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        opC_size = self.get_normalized_size(opC_size)
        opA_size = self.get_normalized_size(opA_size)
        opB_size = self.get_normalized_size(opB_size)

        print ("customize_size " + str(self.body) + " || " + str(opC_size) + " " + str(opA_size) + " " + str(opB_size) + "\n")

        if 0 :
            return CmmBinAsInstruction.customize_size[opC_size][opA_size][opB_size](self)
        else :
            var0 = CmmSpecVar('','tmp', CmmType('u', 32))
            var1 = CmmSpecVar('','tmp', CmmType('u', 32))
            instr = []

            self._set_auxiliary_var(var0.new())
            self._set_auxiliary_var(var1.new())

            if self.get_op() == '<<' :
                if opC.get_size() == 64 :
                    var = CmmSpecVar('','tmp', CmmType('u', 32))

                    self._set_auxiliary_var(var.new())
                    instr += [CmmAssgnDummyInstruction([ 'mviw', var, CmmConst(0) ])]

                    if opA.is_signed() and opA.get_size() < 64 :
                        var = CmmSpecVar('','tmp', CmmType('u', 32))

                        self._set_auxiliary_var(var.new())
                        instr += [CmmAssgnDummyInstruction([ 'mviw', var, CmmConst(0) ])]
            elif self.get_op() == '>>' :
                if opA.is_signed() :
                    if opC.get_size() == 64 :
                        var = CmmSpecVar('','tmp', CmmType('u', 32))

                        self._set_auxiliary_var(var.new())
                        instr += [CmmAssgnDummyInstruction([ 'mviw', var, CmmConst(0) ])]

                    if opA.get_size() == 64 :
                        var = CmmSpecVar('','tmp', CmmType('u', 32))

                        self._set_auxiliary_var(var.new())
                        instr += [CmmAssgnDummyInstruction([ 'mviw', var, CmmConst(0) ])]
            else :
                if opB.is_signed() :
                    opC_size = self.get_opC().get_size()
                    opA_size = self.get_opA().get_size()
                    opB_size = self.get_opB().get_size()

                    if opC_size == 64 :
                        if opA_size < 32 and opB_size < 64 :
                            var = CmmSpecVar('','tmp', CmmType('u', 32))

                            self._set_auxiliary_var(var.new())
                            instr += [CmmAssgnDummyInstruction([ 'mviw', var, CmmConst(0) ])]

                        if opB_size < 32 and opA_size < 64 :
                            var = CmmSpecVar('','tmp', CmmType('u', 32))

                            self._set_auxiliary_var(var.new())
                            instr += [CmmAssgnDummyInstruction([ 'mviw', var, CmmConst(0) ])]

            return [
                CmmAssgnDummyInstruction([ 'mviw', var0, CmmConst(0) ]),
                CmmAssgnDummyInstruction([ 'mviw', var1, CmmConst(0) ]),
                ] + instr + [ self ]

    alu_reg_reg_reg_asm = {
    	'+' : 'add'  ,
    	'-' : 'sub'  ,
    	'|' : 'or'   ,
    	'&' : 'and'  ,
    	'^' : 'xor'  ,
    	'<<': 'shl'  ,
    	'>>': 'shr'  ,
    }

    def _alu_reg_reg_reg_asm_form(self) :
        self.body = [ CmmBinAsInstruction.alu_reg_reg_reg_asm[self.get_op()], self.get_opC(), self.get_opA(), self.get_opB() ]

    alu_mem_reg_reg_asm = {
    	'+' : 'addm'  ,
    	'-' : 'subm'  ,
    	'|' : 'orm'   ,
    	'&' : 'andm'  ,
    	'^' : 'xorm'  ,
    	'<<': 'shlm'  ,
    	'>>': 'shrm'  ,
    }

    def _alu_mem_reg_reg_asm_form(self) :
        self.body = [ CmmBinAsInstruction.alu_mem_reg_reg_asm[self.get_op()], CmmConst(0), self.get_opA(), self.get_opB(), self.get_opC().get_mem_address_var() ]

    alu_reg_reg_imm_asm = {
    	'+' : 'adi'  ,
    	'-' : 'sbi'  ,
    	'|' : 'ori'  ,
    	'&' : 'andi' ,
    	'^' : 'xori' ,
    	'<<': 'shli' ,
    	'>>': 'shri' ,
    }

    def _alu_reg_reg_imm_asm_form(self) :
        self.body = [ CmmBinAsInstruction.alu_reg_reg_imm_asm[self.get_op()], self.get_opC(), self.get_opA(), self.get_opB() ]

    alu_mem_reg_imm_asm = {
    	'+' : 'adim'  ,
    	'-' : 'sbim'  ,
    	'|' : 'orim'  ,
    	'&' : 'andim' ,
    	'^' : 'xorim' ,
    	'<<': 'shlim' ,
    	'>>': 'shrim' ,
    }

    def _alu_mem_reg_imm_asm_form(self) :
        self.body = [ CmmBinAsInstruction.alu_mem_reg_imm_asm[self.get_op()], CmmConst(0), self.get_opA(), self.get_opB(), self.get_opC().get_mem_address_var() ]

    def transform_to_pseudo_asm_form(self) :
        opC = self.get_opC()
        opB = self.get_opB()

        print(str(self.body) + "\n")
        aux_data = self.body[4 : ]
        self.body[4 : ] = []

        if opB.is_a_var() or opB.is_a_reg() :
            if opC.is_a_var() or opC.is_a_reg() :
                self._alu_reg_reg_reg_asm_form()
            else :
                self._alu_mem_reg_reg_asm_form()
        else :
            if opC.is_a_var() or opC.is_a_reg() :
                self._alu_reg_reg_imm_asm_form()
            else :
                self._alu_mem_reg_imm_asm_form()

        self.body += aux_data

    # bug and what if the destination is memory ???

    def _fix_add(self) :
        instr = []
        [opC, opA, opB] = [ x.get_reg() if x.is_a_var() else x for x in self.body[1 : 4]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        if opC_size == 64 :
            [opC_1, opC_0] = opC.splt_into_2()

            if opA_size == 64 :
                [opA_1, opA_0] = opA.splt_into_2()

                if opB_size == 64 :
                    [opB_1, opB_0] = opB.splt_into_2()

                    instr = [
                             CmmBinAsInstruction([ 'add',  opC_0, opA_0, opB_0 ]) ,
                             CmmBinAsInstruction([ 'addc', opC_1, opA_1, opB_1 ])
                            ]
                else :
                    instr = [
                             CmmBinAsInstruction([ 'add',  opC_0, opA_0, opB ]) ,
                             CmmBinAsInstruction([ 'addc', opC_1, opA_1, CmmSR('zero', CmmType('u', 32)) ])
                            ]
            elif opA_size == 32 :
                instr = [
                         CmmBinAsInstruction([ 'add',  opC_0, opA, opB ]) ,
                         CmmBinAsInstruction([ 'addc', opC_1, CmmSR('zero', CmmType('u', 32)), CmmSR('zero', CmmType('u', 32)) ])
                        ]

        return instr

    def _fix_adi(self) :
        instr = []
        (opC, opA, imm) = [ (x.get_reg()) if x.is_a_var() else (x)  for x in self.body[1 : 4]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if (opC_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            if (opA_size == 64) :
                (opA_1, opA_0) = opA.splt_into_2()

                instr = (        CmmBinAsInstruction([ 'adi',  opC_0, opA_0, imm ] ) ,
                    CmmBinAsInstruction([ 'addc', opC_1, opA_1, CmmSR('zero', CmmType('u', 32)) ]) )
            elif (opA_size == 32) :
                instr = (        CmmBinAsInstruction([ 'adi',  opC_0, opA, imm ] ) ,
                    CmmBinAsInstruction([ 'addc', opC_1, CmmSR('zero', CmmType('u', 32)), CmmSR('zero', CmmType('u', 32)) ]) )

        return instr

    def _fix_sub(self) :
        instr = []
        (opC, opA, opB) = [ (x.get_reg()) if x.is_a_var() else (x)  for x in self.body[1 : 4]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        if (opC_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            if (opA_size == 64) :
                (opA_1, opA_0) = opA.splt_into_2()

                if (opB_size == 64) :
                    (opB_1, opB_0) = opB.splt_into_2()

                    instr = (CmmBinAsInstruction([ 'sub',  opC_0, opA_0, opB_0 ] ) ,
                        CmmBinAsInstruction([ 'subb', opC_1, opA_1, opB_1 ] ) )
                else :
                    instr = (CmmBinAsInstruction([ 'sub',  opC_0, opA_0, opB ] ) ,
                        CmmBinAsInstruction([ 'subb', opC_1, opA_1, CmmSR('zero', CmmType('u', 32)) ]) )
            elif (opA_size == 32) :
                instr = (        CmmBinAsInstruction([ 'sub',  opC_0, opA, opB ] ) ,
                    CmmBinAsInstruction([ 'subb', opC_1, CmmSR('zero', CmmType('u', 32)), CmmSR('zero', CmmType('u', 32)) ] ) )

        return instr

    def _fix_sbi(self) :
        instr = []
        (opC, opA, imm) = [ (x.get_reg()) if x.is_a_var() else (x)  for x in self.body[1 : 4]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if (opC_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            if (opA_size == 64) :
                (opA_1, opA_0) = opA.splt_into_2()

                instr = (        CmmBinAsInstruction([ 'sbi',  opC_0, opA_0, imm ] ) ,
                    CmmBinAsInstruction([ 'subb', opC_1, opA_1, CmmSR('zero', CmmType('u', 32)) ]) )
            elif (opA_size == 32) :
                instr = (        CmmBinAsInstruction([ 'sbi',  opC_0, opA, imm ] ) ,
                    CmmBinAsInstruction([ 'subb', opC_1, CmmSR('zero', CmmType('u', 32)), CmmSR('zero', CmmType('u', 32)) ]) )

        return instr

    def _fix_bw_log_reg_reg(self, op) :
        instr = []
        (opC, opA, opB) = [ (x.get_reg()) if x.is_a_var() else (x)  for x in self.body[1 : 4]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        if (opC_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            if (opA_size == 64) :
                (opA_1, opA_0) = opA.splt_into_2()

                if (opB_size == 64) :
                    (opB_1, opB_0) = opB.splt_into_2()

                    instr = (CmmBinAsInstruction([ op, opC_0, opA_0, opB_0 ]) ,
                        CmmBinAsInstruction([ op, opC_1, opA_1, opB_1 ]) )
                else :
                    instr = (CmmBinAsInstruction([ op, opC_0, opA_0, opB ]) ,
                        CmmBinAsInstruction([ op, opC_1, opA_1, CmmSR('zero', CmmType('u', 32)) ]) )
                # else {
                # @instr = (CmmBinAsInstruction->new([ $op, $opC_0, $opA_0, $self->{body}[3] ]) ) ;
                # }
            elif (opA_size == 32) :
                instr = (        CmmBinAsInstruction([ op, opC_0, opA, opB ]) ,
                    CmmAssgnInstruction([ 'move', opC_1, CmmSR( (('one') if (op == 'xor') else ('zero')), CmmType('u', 32)) ]) )
            # else {
            # @instr = ()
            # }

        return instr

    def _fix_bw_log_reg_imm(self, op) :
        instr = []
        (opC, opA, imm) = [ (x.get_reg()) if x.is_a_var() else (x)  for x in self.body[1 : 4]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        print ("_fix_bw_log_reg_imm " + str(self.body) + "\n")

        if (opC_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            if (opA_size == 64) :
                (opA_1, opA_0) = opA.splt_into_2()

                instr = (        CmmBinAsInstruction([ op,  opC_0, opA_0, imm ]) ,
                    CmmBinAsInstruction([ op,  opC_1, opA_1, CmmConst(0) ]) )
                # CmmBinAsInstruction->new([ $op,  $opC_1, $opA_1, CmmSR->new('zero', CmmType->new('u', 32)) ]) ) ;
            elif (opA_size == 32) :
                instr = (        CmmBinAsInstruction([ op, opC_0, opA, imm ]) ,
                    CmmAssgnInstruction([ 'move', opC_1, CmmSR( (('one') if (op == 'xori') else ('zero')), CmmType('u', 32)) ]) )
        elif (opC_size == 32) :
            instr = []
        elif (opC_size == 16) :
            instr = []

        return instr

    def _fix_and(self) :
        return self._fix_bw_log_reg_reg('and')

    def _fix_andi(self) :
        return self._fix_bw_log_reg_imm('andi')

    def _fix_or(self) :
        return self._fix_bw_log_reg_reg('or')

    def _fix_ori(self) :
        return self._fix_bw_log_reg_imm('ori')

    def _fix_xor(self) :
        return self._fix_bw_log_reg_reg('xor')

    def _fix_xori(self) :
        return self._fix_bw_log_reg_imm('xori')

    def _fix_shl(self) :
        instr = []
        (opC, opA, opB, tmp) = [ (x.get_reg()) if x.is_a_var() else (x)  for x in self.body[1 : 5]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        if (opB_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            (opA_1, opA_0) = opA.splt_into_2()
            (opB_1, opB_0) = opB.splt_into_2()
            nop = None
            label = None

            print ("opB_1_size = " + str(opB_1.get_size()) + "\n")
            print "_fix_shli !!!\n"
            print (opC_1 + ", " + str(opC_0) + ", " + str(opA_1) + ", " + str(opA_0) + "\n")

            instr = (
                CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]) ,
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(1) ]) ,
                CmmAssgnInstruction([ 'move', opC, CmmSR('zero',CmmType('u', 64)) ]) ,
                CmmAssgnInstruction([ 'mviw', CmmSR('pc'), label = CmmShlLabel('','') ]) ,
                CmmBinAsInstruction([ 'shl', opC_0, opA_0, opB_0 ]) ,
                CmmBinAsInstruction([ 'shl', opC_1, opA_1, opB_0 ]) ,
                CmmAssgnInstruction([ 'mviw', tmp, CmmConst(32) ]) ,
                CmmBinAsInstruction([ 'sub', tmp, tmp, opB_0 ]) ,
                CmmBinAsInstruction([ 'shr', tmp, opA_0, tmp ]) ,
                CmmBinAsInstruction([ 'or', opC_1, opC_1, tmp ]) ,
                nop = CmmNullInstruction() )

            nop.add_label(label)
        elif (opA_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()
            (opA_1, opA_0) = opA.splt_into_2()

            instr = (
                CmmBinAsInstruction([ 'shl', opC_0, opA_0, opB ]) ,
                CmmBinAsInstruction([ 'shl', opC_1, opA_1, opB ]) ,
                CmmAssgnInstruction([ 'mviw', tmp, CmmConst(32) ]) ,
                CmmBinAsInstruction([ 'sub', tmp, tmp, opB ]) ,
                CmmBinAsInstruction([ 'shr', tmp, opA_0, tmp ]) ,
                CmmBinAsInstruction([ 'or', opC_1, opC_1, tmp ]) )
        elif (opC_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            instr = (
                CmmBinAsInstruction([ 'shl', opC_0, opA, opB ]) ,
                CmmAssgnInstruction([ 'mviw', tmp, CmmConst(32) ]) ,
                CmmBinAsInstruction([ 'sub', tmp, tmp, opB ]) ,
                CmmBinAsInstruction([ 'shr', opC_1, opA, tmp ]) )

        return instr

    def _fix_shli(self) :
        instr = []
        (opC, opA, imm, tmp) = [ (x.get_reg()) if x.is_a_var() else (x)  for x in self.body[1 : 5]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if (opA_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()
            (opA_1, opA_0) = opA.splt_into_2()

            if (imm & 0x20) :
                instr = (
                    CmmBinAsInstruction([ 'shli', opC_1, opA_0, imm & 0x1f ]) ,
                    CmmAssgnInstruction([ 'move', opC_0, CmmSR('zero', CmmType('u', 32)) ]) )
            else :
                instr = (
                    CmmBinAsInstruction([ 'shli', opC_0, opA_0, imm ]) ,
                    CmmBinAsInstruction([ 'shli', opC_1, opA_1, imm ]) ,
                    CmmBinAsInstruction([ 'shri', tmp, opA_0, 32 - imm ]) ,
                    CmmBinAsInstruction([ 'or', opC_1, opC_1, tmp ]) )
        elif (opC_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            if (imm & 0x20) :
                instr = (
                    CmmBinAsInstruction([ 'shli', opC_1, opA, imm & 0x1f ]) ,
                    CmmAssgnInstruction([ 'move', opC_0, CmmSR('zero', CmmType('u', 32)) ]) )
            else :
                instr = (
                    CmmBinAsInstruction([ 'shli', opC_0, opA, imm ]) ,
                    CmmBinAsInstruction([ 'shri', opC_1, opA, 32 - imm ]) )
        elif (opC_size == 32) :
            if (imm & 0x20) :
                instr = ( CmmAssgnInstruction([ 'move', opC, CmmSR('zero', CmmType('u', 32)) ]) )

        return instr

    def _fix_shr(self) :
        instr = []
        (opC, opA, opB, tmp) = [ (x.get_reg()) if x.is_a_var() else (x)  for x in self.body[1 : 5]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        if (opB_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()
            (opA_1, opA_0) = opA.splt_into_2()
            (opB_1, opB_0) = opB.splt_into_2()
            nop = None
            label = None

            instr = (
                CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]) ,
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(1) ]) ,
                CmmAssgnInstruction([ 'move', opC, CmmSR('zero',CmmType('u', 64)) ]) ,
                CmmAssgnInstruction([ 'mviw', CmmSR('pc'), label = CmmShrLabel('','') ]) ,
                CmmBinAsInstruction([ 'shr', opC_0, opA_0, opB_0 ]) ,
                CmmBinAsInstruction([ 'shr', opC_1, opA_1, opB_0 ]) ,
                CmmAssgnInstruction([ 'mviw', tmp, CmmConst(32) ]) ,
                CmmBinAsInstruction([ 'sub', tmp, tmp, opB_0 ]) ,
                CmmBinAsInstruction([ 'shl', tmp, opA_1, tmp ]) ,
                CmmBinAsInstruction([ 'or', opC_0, opC_0, tmp ]) ,
                nop = CmmNullInstruction() )

            nop.add_label(label)
        elif (opA_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()
            (opA_1, opA_0) = opA.splt_into_2()

            instr = (
                CmmBinAsInstruction([ 'shr', opC_0, opA_0, opB ]) ,
                CmmBinAsInstruction([ 'shr', opC_1, opA_1, opB ]) ,
                CmmAssgnInstruction([ 'mviw', tmp, CmmConst(32) ]) ,
                CmmBinAsInstruction([ 'sub', tmp, tmp, opB ]) ,
                CmmBinAsInstruction([ 'shl', tmp, opA_1, tmp ]) ,
                CmmBinAsInstruction([ 'or', opC_0, opC_0, tmp ]) )
        elif (opC_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            instr = (
                CmmBinAsInstruction([ 'shr', opC_0, opA, opB ]) ,
                CmmAssgnInstruction([ 'move', opC_1, CmmSR('zero') ]) )

        return instr

    def _fix_shri(self) :
        instr = []
        (opC, opA, imm, tmp) = [ (x.get_reg()) if x.is_a_var() else (x)  for x in self.body[1 : 5]]
        opC_size = opC.get_size()
        opA_size = opA.get_size()

        if (opA_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()
            (opA_1, opA_0) = opA.splt_into_2()

            if (imm & 0x20) :
                instr = (
                    CmmBinAsInstruction([ 'shri', opC_0, opA_1, imm & 0x1f ]) ,
                    CmmAssgnInstruction([ 'move', opC_1, CmmSR('zero', CmmType('u', 32)) ]) )
            else :
                instr = (
                    CmmBinAsInstruction([ 'shri', opC_0, opA_0, imm ]) ,
                    CmmBinAsInstruction([ 'shri', opC_1, opA_1, imm ]) ,
                    CmmBinAsInstruction([ 'shli', tmp, opA_1, 32 - imm ]) ,
                    CmmBinAsInstruction([ 'or', opC_0, opC_0, tmp ]) )
        elif (opC_size == 64) :
            (opC_1, opC_0) = opC.splt_into_2()

            if (imm & 0x20) :
                instr = ( CmmAssgnInstruction([ 'move', opC, CmmSR('zero', CmmType('u', 64)) ]) )
            else :
                instr = (
                    CmmBinAsInstruction([ 'shli', opC_0, opA, 32 - imm ]) ,
                    CmmBinAsInstruction([ 'shri', opC_1, opA, imm ]) )
        elif (opC_size == 32) :
            if (imm & 0x20) :
                instr = ( CmmAssgnInstruction([ 'move', opC, CmmSR('zero', CmmType('u', 32)) ]) )

        return instr

    second_op = {
    	'add' :	'addc',
    	'adi' :	'addc',
    	'sub' :	'subb',
    	'sbi' :	'subb',
    	'and' :	'and',
    	'andi':	'and',
    	'or'  :	'or',
    	'ori' :	'or',
    	'xor' :	'xor',
    	'xori':	'xor',
    }

    def fix_64__alu_ss_8_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        return [
            ]

    def fix_64__alu_ss_8_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_ss_8_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_ss_8_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()

        opC_size = opC.get_size()

        opA.set_size(opC_size)

        return [
            ]

    def fix_64__alu_ss_8_16_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opA.set_size(opC_size)
        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_ss_8_16_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opA.set_size(opC_size)
        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_ss_8_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)

        return [
            ]

    def fix_64__alu_ss_8_64_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)
        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_ss_8_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)
        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_ss_16_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        var0.get_reg().set_size(opC_size)
        var1.get_reg().set_size(opC_size)

        # return [
        # CmmBinAsInstruction->new([ 'add', $zeroA, $opA, $opA ]),
        # CmmBinAsInstruction->new([ 'addc', $var0, $oneC, $oneC ]),
        # CmmBinAsInstruction->new([ 'shli', $var0, $var0, CmmConst->new($opA_size) ]),
        # CmmBinAsInstruction->new([ 'or', $var0, $var0, $opA ]),
        # CmmBinAsInstruction->new([ 'add', $zeroB, $opB, $opB ]),
        # CmmBinAsInstruction->new([ 'addc', $opC, $oneC, $oneC ]),
        # CmmBinAsInstruction->new([ 'shli', $opC, $opC, CmmConst->new($opB_size) ]),
        # CmmBinAsInstruction->new([ 'or', $opC, $opC, $opB ]),
        # CmmBinAsInstruction->new([ $op, $opC, $var0, $opC ]),
        # ]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var1, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var1, var1, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opB ]),
            CmmBinAsInstruction([ op, opC, var0, var1 ]),
            ]

    def fix_64__alu_ss_16_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB ]),
            ]

    def fix_64__alu_ss_16_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        opB.set_size(opC_size)
        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB ]),
            ]

    def fix_64__alu_ss_16_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        opB_0.set_size(opC_size)
        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB_0 ]),
            ]

    def fix_64__alu_ss_16_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opB_size = opB.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        var0.get_reg().set_size(opC_size)

        print ("16_16_8 opC_size is " + str(opC_size) +\
               " and var0 size is " + str(var0.get_size()) + " " + str(var0.new().get_size()) + "\n")

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opB ]),
            CmmBinAsInstruction([ op, opC, opA, var0 ]),
            ]

    def fix_64__alu_ss_16_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opB_size = opB.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        opA.set_size(opC_size)
        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opB ]),
            CmmBinAsInstruction([ op, opC, opA, var0 ]),
            ]

    def fix_64__alu_ss_16_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opB_size = opB.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        opA_0.set_size(opC_size)
        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opB ]),
            CmmBinAsInstruction([ op, opC, opA_0, var0 ]),
            ]

    def fix_64__alu_ss_32_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_ss_64_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()
        var3 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        # return [
        # CmmBinAsInstruction->new([ 'add', $zeroA, $opA, $opA ]),
        # CmmBinAsInstruction->new([ 'addc', $var0, $oneC_0, $oneC_0 ]),
        # CmmBinAsInstruction->new([ 'shli', $opC_0, $var0, CmmConst->new($opA_size) ]),
        # CmmBinAsInstruction->new([ 'or', $opC_0, $opC_0, $opA ]),
        # CmmBinAsInstruction->new([ 'add', $zeroB, $opB, $opB ]),
        # CmmBinAsInstruction->new([ 'addc', $var1, $oneC_1, $oneC_1 ]),
        # CmmBinAsInstruction->new([ 'shli', $opC_1, $var1, CmmConst->new($opB_size) ]),
        # CmmBinAsInstruction->new([ 'or', $opC_1, $opC_1, $opB ]),
        # CmmBinAsInstruction->new([ $op, $opC_0, $opC_0, $opC_1 ]),
        # CmmBinAsInstruction->new([ $sec_op, $opC_1, $var0, $var1 ]),
        # ]

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]
        elif var2.get_reg().intersects_with(opC_0) :
            [var2, var1] = [var1, var2]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opA ]),
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var2, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ 'shli', var3, var2, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var3, var3, opB ]),
            CmmBinAsInstruction([ op, opC_0, var1, var3 ]),
            CmmBinAsInstruction([ sec_op, opC_1, var0, var2 ]),
            ]

    def fix_64__alu_ss_64_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        # return [
        # CmmBinAsInstruction->new([ 'add', $zeroA, $opA, $opA ]),
        # CmmBinAsInstruction->new([ 'addc', $var0, $oneC_0, $oneC_0 ]),
        # CmmBinAsInstruction->new([ 'shli', $opC_0, $var0, CmmConst->new($opA_size) ]),
        # CmmBinAsInstruction->new([ 'or', $opC_0, $opC_0, $opA ]),
        # CmmBinAsInstruction->new([ 'add', $zeroB, $opB, $opB ]),
        # CmmBinAsInstruction->new([ 'addc', $var1, $oneC_1, $oneC_1 ]),
        # CmmBinAsInstruction->new([ $op, $opC_0, $opC_0, $opB ]),
        # CmmBinAsInstruction->new([ $sec_op, $opC_1, $var0, $var1 ]),
        # ]

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]
        elif var2.get_reg().intersects_with(opC_0) :
            [var2, var1] = [var1, var2]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opA ]),
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var2, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ op, opC_0, var1, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, var0, var2 ]),
            ]

    def fix_64__alu_ss_64_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()

        opA_size = opA.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opA ]),
            CmmBinAsInstruction([ op, opC_0, var1, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, var0, opB_1 ]),
            ]

    def fix_64__alu_ss_64_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        # return [
        # CmmBinAsInstruction->new([ 'add', $zeroA, $opA, $opA ]),
        # CmmBinAsInstruction->new([ 'addc', $var0, $oneC_0, $oneC_0 ]),
        # CmmBinAsInstruction->new([ 'add', $zeroB, $opB, $opB ]),
        # CmmBinAsInstruction->new([ 'addc', $var1, $oneC_1, $oneC_1 ]),
        # CmmBinAsInstruction->new([ 'shli', $opC_1, $var1, CmmConst->new($opB_size) ]),
        # CmmBinAsInstruction->new([ 'or', $opC_1, $opC_1, $opB ]),
        # CmmBinAsInstruction->new([ $op, $opC_0, $opA, $opC_1 ]),
        # CmmBinAsInstruction->new([ $sec_op, $opC_1, $var0, $var1 ]),
        # ]

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]
        elif var2.get_reg().intersects_with(opC_0) :
            [var2, var1] = [var1, var2]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var2, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opB ]),
            CmmBinAsInstruction([ op, opC_0, opA, var1 ]),
            CmmBinAsInstruction([ sec_op, opC_1, var2, var0 ]),
            ]

    def fix_64__alu_ss_64_32_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', opC_1, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ op, opC_0, opA, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, var0, opC_1 ]),
            ]

    def fix_64__alu_ss_64_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()

        opA_size = opA.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ op, opC_0, opA, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, var0, opB_1 ]),
            ]

    def fix_64__alu_ss_64_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        opB_size = opB.get_size()

        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opB ]),
            CmmBinAsInstruction([ op, opC_0, opA_0, var1 ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, var0 ]),
            ]

    def fix_64__alu_ss_64_64_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        opB_size = opB.get_size()

        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ op, opC_0, opA_0, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, var0 ]),
            ]

    def fix_64__alu_ss_64_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        return [
            CmmBinAsInstruction([ op, opC_0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, opB_1 ]),
            ]

################################################################################

    def fix_64__alu_us_8_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        return [
            ]

    def fix_64__alu_us_8_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_us_8_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_us_8_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()

        opC_size = opC.get_size()

        opA.set_size(opC_size)

        return [
            ]

    def fix_64__alu_us_8_16_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opA.set_size(opC_size)
        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_us_8_16_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opA.set_size(opC_size)
        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_us_8_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)

        return [
            ]

    def fix_64__alu_us_8_64_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)
        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_us_8_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)
        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_us_16_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        var0.get_reg().set_size(opC_size)
        var1.get_reg().set_size(opC_size)

        if op == 'sub' :
            return [
                CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', var1, oneC, oneC ]),
                CmmBinAsInstruction([ 'shli', var1, var1, CmmConst(opB_size) ]),
                CmmBinAsInstruction([ 'or', var1, var1, opB ]),
                CmmBinAsInstruction([ op, opC, var0, var1 ]),
                ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', var1, oneC, oneC ]),
                CmmBinAsInstruction([ 'shli', var1, var1, CmmConst(opB_size) ]),
                CmmBinAsInstruction([ 'or', var1, var1, opB ]),
                CmmBinAsInstruction([ op, opC, var1, opA ]),
                ]

    def fix_64__alu_us_16_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        if op == 'sub' :
            var0.get_reg().set_size(opC_size)

            return [
                CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
                CmmBinAsInstruction([ op, opC, var0, opB ]),
                ]
        else :
            self.set_opA(opB)
            self.set_opB(opA)

            return [
                ]

    def fix_64__alu_us_16_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opB.set_size(opC_size)

        if op == 'sub' :
            var0.get_reg().set_size(opC_size)

            return [
                CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
                CmmBinAsInstruction([ op, opC, var0, opB ]),
                ]
        else :
            self.set_opA(opB)
            self.set_opB(opA)

            return [
                ]

    def fix_64__alu_us_16_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opB_0.set_size(opC_size)

        if op == 'sub' :
            var0.get_reg().set_size(opC_size)

            return [
                CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
                CmmBinAsInstruction([ op, opC, var0, opB_0 ]),
                ]
        else :
            self.set_opA(opB_0)
            self.set_opB(opA)

            return [
                ]

    def fix_64__alu_us_16_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opB_size = opB.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opB ]),
            CmmBinAsInstruction([ op, opC, opA, var0 ]),
            ]

    def fix_64__alu_us_16_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opB_size = opB.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        opA.set_size(opC_size)
        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opB ]),
            CmmBinAsInstruction([ op, opC, opA, var0 ]),
            ]

    def fix_64__alu_us_16_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opB_size = opB.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        opA_0.set_size(opC_size)
        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opB ]),
            CmmBinAsInstruction([ op, opC, opA_0, var0 ]),
            ]

    def fix_64__alu_us_32_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_us_64_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()
        var3 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if var2.get_reg().intersects_with(opC_0) :
            [var2, var1] = [var1, var2]

        if op == 'sub' :
            return [
                CmmBinAsInstruction([ 'or', var1, zeroC_0, opA ]),
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', var2, oneC_1, oneC_1 ]),
                CmmBinAsInstruction([ 'shli', var3, var2, CmmConst(opB_size) ]),
                CmmBinAsInstruction([ 'or', var3, var3, opB ]),
                CmmBinAsInstruction([ op, opC_0, var1, var3 ]),
                CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, var2 ]),
                ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', var2, oneC_1, oneC_1 ]),
                CmmBinAsInstruction([ 'shli', var3, var2, CmmConst(opB_size) ]),
                CmmBinAsInstruction([ 'or', var3, var3, opB ]),
                CmmBinAsInstruction([ op, opC_0, var3, opA ]),
                CmmBinAsInstruction([ sec_op, opC_1, var2, zeroC_1 ]),
                ]

    def fix_64__alu_us_64_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if var2.get_reg().intersects_with(opC_0) :
            [var2, var1] = [var1, var2]

        if op == 'sub' :
            return [
                CmmBinAsInstruction([ 'or', var1, zeroC_0, opA ]),
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', var2, oneC_1, oneC_1 ]),
                CmmBinAsInstruction([ op, opC_0, var1, opB ]),
                CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, var2 ]),
                ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
                CmmBinAsInstruction([ 'addc', var2, oneC_1, oneC_1 ]),
                CmmBinAsInstruction([ op, opC_0, opB, opA ]),
                CmmBinAsInstruction([ sec_op, opC_1, var2, zeroC_1 ]),
                ]

    def fix_64__alu_us_64_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        if op == 'sub' :
            opA_size = opA.get_size()

            oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
            zeroA = CmmSR('zero', CmmType('u', opA_size))

            if var0.get_reg().intersects_with(opC_0) :
                [var0, var1] = [var1, var0]

            return [
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
                CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
                CmmBinAsInstruction([ 'or', var1, var1, opA ]),
                CmmBinAsInstruction([ op, opC_0, var1, opB_0 ]),
                CmmBinAsInstruction([ sec_op, opC_1, var0, opB_1 ]),
                ]
        else :
            zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

            return [
                CmmBinAsInstruction([ op, opC_0, opB_0, opA ]),
                CmmBinAsInstruction([ sec_op, opC_1, opB_1, zeroC_1 ]),
                ]

    def fix_64__alu_us_64_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opB ]),
            CmmBinAsInstruction([ op, opC_0, opA, var1 ]),
            CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, var0 ]),
            ]

    def fix_64__alu_us_64_32_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var1, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ op, opC_0, opA, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, var1 ]),
            ]

    def fix_64__alu_us_64_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_1_size = opC_0.get_size()   # bug ? shouldn't it be opC_1.get_size() ?

        opA_size = opA.get_size()

        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        return [
            CmmBinAsInstruction([ op, opC_0, opA, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, opB_1 ]),
            ]

    def fix_64__alu_us_64_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        opB_size = opB.get_size()

        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opB_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opB ]),
            CmmBinAsInstruction([ op, opC_0, opA_0, var1 ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, var0 ]),
            ]

    def fix_64__alu_us_64_64_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        opB_size = opB.get_size()

        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroB, opB, opB ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ op, opC_0, opA_0, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, var0 ]),
            ]

    def fix_64__alu_us_64_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        return [
            CmmBinAsInstruction([ op, opC_0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, opB_1 ]),
            ]

################################################################################

    def fix_64__alu_su_8_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        return [
            ]

    def fix_64__alu_su_8_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_su_8_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_su_8_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()

        opC_size = opC.get_size()

        opA.set_size(opC_size)

        return [
            ]

    def fix_64__alu_su_8_16_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opA.set_size(opC_size)
        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_su_8_16_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opA.set_size(opC_size)
        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_su_8_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)

        return [
            ]

    def fix_64__alu_su_8_64_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)
        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_su_8_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)
        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_su_16_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zeroB = CmmSR('zero', CmmType('u', opB_size))

        var0.get_reg().set_size(opC_size)
        var1.get_reg().set_size(opC_size)

        # return [
        # CmmBinAsInstruction->new([ 'add', $zeroA, $opA, $opA ]),
        # CmmBinAsInstruction->new([ 'addc', $var0, $oneC, $oneC ]),
        # CmmBinAsInstruction->new([ 'shli', $var0, $var0, CmmConst->new($opA_size) ]),
        # CmmBinAsInstruction->new([ 'or', $var0, $var0, $opA ]),
        # CmmBinAsInstruction->new([ 'add', $zeroB, $opB, $opB ]),
        # CmmBinAsInstruction->new([ 'addc', $opC, $oneC, $oneC ]),
        # CmmBinAsInstruction->new([ 'shli', $opC, $opC, CmmConst->new($opB_size) ]),
        # CmmBinAsInstruction->new([ 'or', $opC, $opC, $opB ]),
        # CmmBinAsInstruction->new([ $op, $opC, $var0, $opC ]),
        # ]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB ]),
            ]

    def fix_64__alu_su_16_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        opB.set_size(opC_size)
        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB ]),
            ]

    def fix_64__alu_su_16_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        opB_0.set_size(opC_size)
        var0.get_reg().set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB_0 ]),
            ]

    def fix_64__alu_su_32_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_su_64_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opA ]),
            CmmBinAsInstruction([ op, opC_0, var1, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, var0, zeroC_1 ]),
            ]

    def fix_64__alu_su_64_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()

        opA_size = opA.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opA ]),
            CmmBinAsInstruction([ op, opC_0, var1, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, var0, opB_1 ]),
            ]

    def fix_64__alu_su_64_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        opA_size = opA.get_size()
        opB_size = opB.get_size()

        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ op, opC_0, opA, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, var0, zeroC_1 ]),
            ]

    def fix_64__alu_su_64_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()

        opA_size = opA.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var1] = [var1, var0]

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_0, oneC_0 ]),
            CmmBinAsInstruction([ op, opC_0, opA, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, var0, opB_1 ]),
            ]

    def fix_64__alu_su_64_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        return [
            CmmBinAsInstruction([ op, opC_0, opA_0, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, zeroC_1 ]),
            ]

    def fix_64__alu_su_64_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        return [
            CmmBinAsInstruction([ op, opC_0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, opB_1 ]),
            ]

################################################################################

    def fix_64__alu_uu_8_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        return [
            ]

    def fix_64__alu_uu_8_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opB.set_size(opC_size)

        return [
            ]

    def fix_64__alu_uu_8_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_uu_8_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()

        opC_size = opC.get_size()

        opA.set_size(opC_size)
        # $opA->get_reg()->set_size($opC_size) ;

        return [
            ]

    def fix_64__alu_uu_8_16_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opA.set_size(opC_size)
        # $opA->get_reg()->set_size($opC_size) ;
        opB.set_size(opC_size)
        # $opB->get_reg()->set_size($opC_size) ;

        return [
            ]

    def fix_64__alu_uu_8_16_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()
        # my [$opB_1, $opB_0] = $self->get_opB()->get_reg()->splt_into_2() ;

        opC_size = opC.get_size()

        opA.set_size(opC_size)
        # $opA->get_reg()->set_size($opC_size) ;
        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_uu_8_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        # my [$opA_1, $opA_0] = $self->get_opA()->get_reg()->splt_into_2() ;

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)

        return [
            ]

    def fix_64__alu_uu_8_64_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        # my [$opA_1, $opA_0] = $self->get_opA()->get_reg()->splt_into_2() ;
        opB = self.get_opB()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)
        opB.set_size(opC_size)
        # $opB->get_reg()->set_size($opC_size) ;

        return [
            ]

    def fix_64__alu_uu_8_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        # my [$opA_1, $opA_0] = $self->get_opA()->get_reg()->splt_into_2() ;
        [opB_1, opB_0] = self.get_opB().splt_into_2()
        # my [$opB_1, $opB_0] = $self->get_opB()->get_reg()->splt_into_2() ;

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)
        opB_0.set_size(opC_size)
        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_uu_16_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        var0.set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB ]),
            ]

    def fix_64__alu_uu_16_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        if op == 'sub' :
            zeroC = CmmSR('zero', CmmType('u', opC_size))

            var0.set_size(opC_size)

            return [
                CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
                CmmBinAsInstruction([ op, opC, var0, opB ]),
                ]
        else :
            self.set_opA(opB)
            self.set_opB(opA)

            return [
                ]

    def fix_64__alu_uu_16_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        opB.set_size(opC_size)
        # $opB->get_reg()->set_size($opC_size) ;

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        var0.set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB ]),
            ]

    def fix_64__alu_uu_16_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()
        # my [$opB_1, $opB_0] = $self->get_opB()->get_reg()->splt_into_2() ;

        opC_size = opC.get_size()

        opB_0.set_size(opC_size)
        # $opB_0->get_reg()->set_size($opC_size) ;

        if op == 'sub' :
            zeroC = CmmSR('zero', CmmType('u', opC_size))

            self.set_opB(opB_0)
            var0.set_size(opC_size)

            return [
                CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
                CmmBinAsInstruction([ op, opC, var0, opB_0 ]),
                ]
        else :
            self.set_opA(opB_0)
            self.set_opB(opA)

            return [
                ]

    def fix_64__alu_uu_32_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()
        # my [$opB_1, $opB_0] = $self->get_opB()->get_reg()->splt_into_2() ;

        self.set_opB(opB_0)

        return [
            ]

    def fix_64__alu_uu_64_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        # my [$opC_1, $opC_0] = $self->get_opC()->get_reg()->splt_into_2() ;
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        var0.set_size(opC_0_size)

        return [
            CmmBinAsInstruction([ 'or', var0, zeroC_0, opA ]),
            CmmBinAsInstruction([ op, opC_0, var0, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, zeroC_1 ]),
            ]

    def fix_64__alu_uu_64_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        # my [$opC_1, $opC_0] = $self->get_opC()->get_reg()->splt_into_2() ;
        opA = self.get_opA()
        opB = self.get_opB()

        if op == 'sub' :
            opC_0_size = opC_0.get_size()
            opC_1_size = opC_1.get_size()

            zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
            zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

            var0.set_size(opC_0_size)

            return [
                CmmBinAsInstruction([ 'or', var0, zeroC_0, opA ]),
                CmmBinAsInstruction([ op, opC_0, var0, opB ]),
                CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, zeroC_1 ]),
                ]
        else :
            opC_1_size = opC_1.get_size()

            zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

            return [
                CmmBinAsInstruction([ op, opC_0, opB, opA ]),
                CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, zeroC_1 ]),
                ]

    def fix_64__alu_uu_64_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        # my [$opC_1, $opC_0] = $self->get_opC()->get_reg()->splt_into_2() ;
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()
        # my [$opB_1, $opB_0] = $self->get_opB()->get_reg()->splt_into_2() ;

        if op == 'sub' :
            opC_0_size = opC_0.get_size()
            opC_1_size = opC_1.get_size()

            zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
            zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

            var0.set_size(opC_0_size)

            return [
                CmmBinAsInstruction([ 'or', var0, zeroC_0, opA ]),
                CmmBinAsInstruction([ op, opC_0, var0, opB_0 ]),
                CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, opB_1 ]),
                ]
        else :
            opC_1_size = opC_1.get_size()

            zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

            return [
                CmmBinAsInstruction([ op, opC_0, opB_0, opA ]),
                CmmBinAsInstruction([ sec_op, opC_1, opB_1, zeroC_1 ]),
                ]

    def fix_64__alu_uu_64_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        # my [$opC_1, $opC_0] = $self->get_opC()->get_reg()->splt_into_2() ;
        opA = self.get_opA()
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        return [
            CmmBinAsInstruction([ op, opC_0, opA, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, zeroC_1 ]),
            ]

    def fix_64__alu_uu_64_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        # my [$opC_1, $opC_0] = $self->get_opC()->get_reg()->splt_into_2() ;
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()
        # my [$opB_1, $opB_0] = $self->get_opB()->get_reg()->splt_into_2() ;

        opC_1_size = opC_1.get_size()

        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        return [
            CmmBinAsInstruction([ op, opC_0, opA, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, zeroC_1, opB_1 ]),
            ]

    def fix_64__alu_uu_64_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        # my [$opC_1, $opC_0] = $self->get_opC()->get_reg()->splt_into_2() ;
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        # my [$opA_1, $opA_0] = $self->get_opA()->get_reg()->splt_into_2() ;
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        return [
            CmmBinAsInstruction([ op, opC_0, opA_0, opB ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, zeroC_1 ]),
            ]

    def fix_64__alu_uu_64_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()
        sec_op = CmmBinAsInstruction.second_op[op]

        [opC_1, opC_0] = self.get_opC().splt_into_2()
        # my [$opC_1, $opC_0] = $self->get_opC()->get_reg()->splt_into_2() ;
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        # my [$opA_1, $opA_0] = $self->get_opA()->get_reg()->splt_into_2() ;
        [opB_1, opB_0] = self.get_opB().splt_into_2()
        # my [$opB_1, $opB_0] = $self->get_opB()->get_reg()->splt_into_2() ;

        return [
            CmmBinAsInstruction([ op, opC_0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ sec_op, opC_1, opA_1, opB_1 ]),
            ]

################################################################################

    def fix_64__shl_su_8_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        return [
            ]

    def fix_64__shl_su_8_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA.set_size(opC_size)

        new_opB = opB.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA, new_opB ]),
            ]

    def fix_64__shl_su_8_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA.set_size(opC_size)

        new_opB = opB_0.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA, new_opB ]),
            ]

    def fix_64__shl_su_8_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()

        opC_size = opC.get_size()

        opA.set_size(opC_size)

        return [
            ]

    def fix_64__shl_su_8_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)

        return [
            ]

    def fix_64__shl_su_8_64_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA_0.set_size(opC_size)

        new_opB = opB.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA_0, new_opB ]),
            ]

    def fix_64__shl_su_8_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA_0.set_size(opC_size)

        new_opB = opB_0.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA_0, new_opB ]),
            ]

    def fix_64__shl_su_16_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        var0.set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB ]),
            ]

    def fix_64__shl_su_16_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        new_opB = opB.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(6), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, opC, var0, new_opB ]),
            ]

    def fix_64__shl_su_16_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        new_opB = opB_0.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(6), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, opC, var0, new_opB ]),
            ]

    def fix_64__shl_su_64_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()
        var3 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()
        opA_size = opA.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zero64 = CmmSR('zero', CmmType('u', 64))
        zero32 = CmmSR('zero', CmmType('u', 32))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zero64 ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
                    CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
                    CmmBinAsInstruction([ 'or', opC_1, var1, opA ]),
                    ] +\
                    ([CmmBinAsInstruction([ op, opC_1, opC_1, opB - 32 ]))] if opB - 32 else []) +\
                    [
                    CmmAssgnInstruction([ 'move', opC_0, zero32 ]),
                    ]
            elif opB > 32 - opA_size :
                if var0.get_reg().intersects_with(opC_0) :
                    [var0, var1] = [var1, var0]

                if var0.get_reg().intersects_with(opC_1) :
                    [var0, var2] = [var2, var0]

                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
                    CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
                    CmmBinAsInstruction([ 'or', opC_0, var1, opA ]),
                    CmmBinAsInstruction([ 'shri', opC_1, opC_0, 32 - opB ]),
                    CmmBinAsInstruction([ op, opC_0, opC_0, opB ]),
                    CmmBinAsInstruction([ 'shri', var0, var0, opB ]),
                    CmmBinAsInstruction([ 'or', opC_1, opC_1, var0 ]),
                    ]
            else :
                if var0.get_reg().intersects_with(opC_0) :
                    [var0, var1] = [var1, var0]

                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
                    CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
                    CmmBinAsInstruction([ 'or', opC_0, var1, opA ]),
                    ] +\
                    ([CmmBinAsInstruction([ op, opC_0, opC_0, opB ])] if opB else []) +\
                    [
                    CmmAssgnInstruction([ 'move', opC_1, var0 ]),
                    ]
        else : # opB is a reg
            instr = []
            new_opB = opB.new()
            new_opB.set_size(8)

            if var0.get_reg().intersects_with(opC_1) :
                [var0, var2] = [var2, var0]

            if var1.get_reg().intersects_with(opC_1) :
                [var1, var2] = [var2, var1]

            if opB.intersects_with(opC_1) :
                if var3.get_reg().intersects_with(opC_1) :
                    [var3, var2] = [var2, var3]

                instr = [CmmAssgnInstruction([ 'move', var3, opB ])]
                new_opB = var3

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zero64 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(16 + len(instr)), CmmConst(0), CmmConst(0) ]),
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
                CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
                CmmBinAsInstruction([ 'or', var1, var1, opA ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ op, opC_1, var1, new_opB ]),
                CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(7 + len(instr)), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var2, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var2, var2, opB ]),
                ] +\
                instr +\
                [
                CmmBinAsInstruction([ 'shr', opC_1, var1, var2 ]),
                CmmBinAsInstruction([ 'shr', var0, var0, new_opB ]),
                CmmBinAsInstruction([ 'or', opC_1, opC_1, var0 ]),
                CmmBinAsInstruction([ op, opC_0, var1, new_opB ]),
                ]

    def fix_64__shl_su_64_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()
        var3 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()
        opA_size = opA.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zero64 = CmmSR('zero', CmmType('u', 64))

        instr = []
        new_opB = opB_0.new()
        new_opB.set_size(8)

        if var0.get_reg().intersects_with(opC_1) :
            [var0, var2] = [var2, var0]

        if var1.get_reg().intersects_with(opC_1) :
            [var1, var2] = [var2, var1]

        if opB_0.intersects_with(opC_1) :
            if var3.get_reg().intersects_with(opC_1) :
                [var3, var2] = [var2, var3]

            instr = [CmmAssgnInstruction([ 'move', var3, opB_0 ])]
            new_opB = var3

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zero64 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(16 + len(instr)), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
            CmmBinAsInstruction([ 'shli', var1, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var1, var1, opA ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC_1, var1, new_opB ]),
            CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(7 + len(instr)), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var2, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var2, var2, opB_0 ]),
            ] +\
            instr +\
            [
            CmmBinAsInstruction([ 'shr', opC_1, var1, var2 ]),
            CmmBinAsInstruction([ 'shr', var0, var0, new_opB ]),
            CmmBinAsInstruction([ 'or', opC_1, opC_1, var0 ]),
            CmmBinAsInstruction([ op, opC_0, var1, new_opB ]),
            ]

    def fix_64__shl_su_64_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()
        var3 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()
        opA_size = opA.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zero64 = CmmSR('zero', CmmType('u', 64))
        zero32 = CmmSR('zero', CmmType('u', 32))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zero64 ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ op, opC_1, opA, opB - 32 ]),
                    CmmAssgnInstruction([ 'move', opC_0, zero32 ]),
                    ]
            elif opB > 32 - opA_size :
                if var0.get_reg().intersects_with(opC_0) :
                    [var0, var1] = [var1, var0]

                if var0.get_reg().intersects_with(opC_1) :
                    [var0, var2] = [var2, var0]

                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
                    CmmAssgnInstruction([ 'move', opC_0, opA ]),
                    CmmBinAsInstruction([ 'shri', opC_1, opC_0, 32 - opB ]),
                    CmmBinAsInstruction([ op, opC_0, opC_0, opB ]),
                    CmmBinAsInstruction([ 'shri', var0, var0, opB ]),
                    CmmBinAsInstruction([ 'or', opC_1, opC_1, var0 ]),
                    ]
            else :
                if var0.get_reg().intersects_with(opC_0) :
                    [var0, var1] = [var1, var0]

                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
                    CmmBinAsInstruction([ op, opC_0, opA, opB ]),
                    CmmAssgnInstruction([ 'move', opC_1, var0 ]),
                    ]
        else : # opB is a reg
            instr = []
            new_opB = opB.new()
            new_opB.set_size(8)

            if var0.get_reg().intersects_with(opC_1) :
                [var0, var2] = [var2, var0]

            if var1.get_reg().intersects_with(opC_1) :
                [var1, var2] = [var2, var1]

            if opB.intersects_with(opC_1) :
                if var3.get_reg().intersects_with(opC_1) :
                    [var3, var2] = [var2, var3]

                instr = [CmmAssgnInstruction([ 'move', var3, opB ])]
                new_opB = var3

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zero64 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(15 + len(instr)), CmmConst(0), CmmConst(0) ]),
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
                CmmAssgnInstruction([ 'move', var1, opA ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ op, opC_1, var1, new_opB ]),
                CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(7 + len(instr)), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var2, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var2, var2, opB ]),
                ] +\
                instr +\
                [
                CmmBinAsInstruction([ 'shr', opC_1, var1, var2 ]),
                CmmBinAsInstruction([ 'shr', var0, var0, new_opB ]),
                CmmBinAsInstruction([ 'or', opC_1, opC_1, var0 ]),
                CmmBinAsInstruction([ op, opC_0, var1, new_opB ]),
                ]

    def fix_64__shl_su_64_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()
        var3 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()
        opA_size = opA.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        oneC_1 = CmmSR('one', CmmType('u', opC_1_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        zero64 = CmmSR('zero', CmmType('u', 64))

        instr = []
        new_opB = opB_0.new()
        new_opB.set_size(8)

        if var0.get_reg().intersects_with(opC_1) :
            [var0, var2] = [var2, var0]

        if var1.get_reg().intersects_with(opC_1) :
            [var1, var2] = [var2, var1]

        if opB_0.intersects_with(opC_1) :
            if var3.get_reg().intersects_with(opC_1) :
                [var3, var2] = [var2, var3]

            instr = [CmmAssgnInstruction([ 'move', var3, opB_0 ])]
            new_opB = var3

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zero64 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(15 + len(instr)), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC_1, oneC_1 ]),
            CmmAssgnInstruction([ 'move', var1, opA ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC_1, var1, new_opB ]),
            CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(7 + len(instr)), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var2, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var2, var2, opB_0 ]),
            ] +\
            instr +\
            [
            CmmBinAsInstruction([ 'shr', opC_1, var1, var2 ]),
            CmmBinAsInstruction([ 'shr', var0, var0, new_opB ]),
            CmmBinAsInstruction([ 'or', opC_1, opC_1, var0 ]),
            CmmBinAsInstruction([ op, opC_0, var1, new_opB ]),
            ]

    def fix_64__shl_su_64_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zero64 = CmmSR('zero', CmmType('u', 64))
        zero32 = CmmSR('zero', CmmType('u', 32))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zero64 ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ op, opC_1, opA_0, opB - 32 ]) if opB - 32\
               else CmmAssgnInstruction([ 'move', opC_1, opA_0 ]),
                    CmmAssgnInstruction([ 'move', opC_0, zero32 ]),
                    ]
            else :
                if opB :
                    if var0.get_reg().intersects_with(opC_0) :
                        [var0, var1] = [var1, var0]

                    return [
                        CmmBinAsInstruction(['shri', var0, opA_0, 32 - opB ]),
                        CmmBinAsInstruction([ op, opC_1, opA_1, opB ]),
                        CmmBinAsInstruction([ op, opC_0, opC_0, opB ]),
                        CmmBinAsInstruction([ 'or', opC_1, opC_1, var0 ]),
                        ]
                else :
                    return [
                        CmmAssgnInstruction([ 'move', opC, opA ]),
                        ]
        else : # opB is a reg
            new_opB = opB.new()
            new_opB.set_size(8)

            instr = []

            if opA_0.intersects_with(opC_0) and opB.intersects_with(opC_1) :
                if var0.get_reg().intersects_with(opC_0) :
                    [var0, var2] = [var2, var0]

                if var1.get_reg().intersects_with(opC_0) :
                    [var1, var2] = [var2, var1]

                instr =\
                    [CmmAssgnInstruction([ 'move', var0, opA_0 ]),
                    CmmBinAsInstruction([ op, opC_0, var0, new_opB ]),
                    CmmBinAsInstruction([ 'shr', opC_1, var0, var1 ])]
            elif opA_0.intersects_with(opC_0) :
                instr =\
                    [CmmBinAsInstruction([ 'shr', opC_1, opA_0, var1 ]),
                    CmmBinAsInstruction([ op, opC_0, opA_0, new_opB ])]
            else :
                instr =\
                    [CmmBinAsInstruction([ op, opC_0, opA_0, new_opB ]),
                    CmmBinAsInstruction([ 'shr', opC_1, opA_0, var1 ])]

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zero64 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(8 + len(instr)), CmmConst(0), CmmConst(0) ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ op, opC_1, opA_0, new_opB ]),
                CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(3 + len(instr)), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                ] +\
                instr

    def fix_64__shl_su_64_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zero64 = CmmSR('zero', CmmType('u', 64))

        new_opB = opB_0.new()
        new_opB.set_size(8)

        instr = []

        if opA_0.intersects_with(opC_0) and opB_0.intersects_with(opC_1) :
            if var0.get_reg().intersects_with(opC_0) :
                [var0, var2] = [var2, var0]

            if var1.get_reg().intersects_with(opC_0) :
                [var1, var2] = [var2, var1]

            instr =\
                [CmmAssgnInstruction([ 'move', var0, opA_0 ]),
                CmmBinAsInstruction([ op, opC_0, var0, new_opB ]),
                CmmBinAsInstruction([ 'shr', opC_1, var0, var1 ])]
        elif opA_0.intersects_with(opC_0) :
            instr =\
                [CmmBinAsInstruction([ 'shr', opC_1, opA_0, var1 ]),
                CmmBinAsInstruction([ op, opC_0, opA_0, new_opB ])]
        else :
            instr =\
                [CmmBinAsInstruction([ op, opC_0, opA_0, new_opB ]),
                CmmBinAsInstruction([ 'shr', opC_1, opA_0, var1 ])]

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zero64 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(8 + len(instr)), CmmConst(0), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC_1, opA_0, new_opB ]),
            CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(3 + len(instr)), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            ] +\
            instr

################################################################################

    def fix_64__shl_uu_8_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        return [
            ]

    def fix_64__shl_uu_8_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA.set_size(opC_size)

        new_opB = opB.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA, new_opB ]),
            ]

    def fix_64__shl_uu_8_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA.set_size(opC_size)

        new_opB = opB_0.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA, new_opB ]),
            ]

    def fix_64__shl_uu_8_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        opA = self.get_opA()

        opC_size = opC.get_size()

        opA.set_size(opC_size)

        return [
            ]

    def fix_64__shl_uu_8_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()

        opC_size = opC.get_size()

        opA_0.set_size(opC_size)
        self.set_opA(opA_0)

        return [
            ]

    def fix_64__shl_uu_8_64_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA_0.set_size(opC_size)

        new_opB = opB.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA_0, new_opB ]),
            ]

    def fix_64__shl_uu_8_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA_0.set_size(opC_size)

        new_opB = opB_0.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA_0, new_opB ]),
            ]

    def fix_64__shl_uu_16_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        var0.set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB ]),
            ]

    def fix_64__shl_uu_16_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        new_opB = opB.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(3), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
            CmmBinAsInstruction([ op, opC, var0, new_opB ]),
            ]

    def fix_64__shl_uu_16_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        new_opB = opB_0.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(3), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
            CmmBinAsInstruction([ op, opC, var0, new_opB ]),
            ]

    def fix_64__shl_uu_64_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()
        opA_size = opA.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zero64 = CmmSR('zero', CmmType('u', 64))
        zero32 = CmmSR('zero', CmmType('u', 32))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zero64 ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ 'or', opC_1, zeroC_1, opA ]),
                    ] +\
                    ([CmmBinAsInstruction([ op, opC_1, opC_1, opB - 32 ])] if opB - 32 else []) +\
                    [
                    CmmAssgnInstruction([ 'move', opC_0, zero32 ]),
                    ]
            elif opB > 32 - opA_size :
                return [
                    CmmBinAsInstruction([ 'or', opC_0, zeroC_0, opA ]),
                    CmmBinAsInstruction([ 'shri', opC_1, opC_0, 32 - opB ]),
                    CmmBinAsInstruction([ op, opC_0, opC_0, opB ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ 'or', opC_0, zeroC_0, opA ]),
                    ] +\
                    ([CmmBinAsInstruction([ op, opC_0, opC_0, opB ])] if opB else []) +\
                    [
                    CmmAssgnInstruction([ 'move', opC_1, zero32 ]),
                    ]
        else : # opB is a reg
            new_opB = opB.new()
            new_opB.set_size(8)

            if var0.get_reg().intersects_with(opC_0) :
                [var0, var2] = [var2, var0]

            if var1.get_reg().intersects_with(opC_0) :
                [var1, var2] = [var2, var1]

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zero64 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(11), CmmConst(0), CmmConst(0) ]),
                CmmBinAsInstruction([ 'or', var0, zeroC_1, opA ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ op, opC_1, var0, new_opB ]),
                CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(5), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ op, opC_0, var0, new_opB ]),
                CmmBinAsInstruction([ 'shr', opC_1, var0, var1 ]),
                ]

    def fix_64__shl_uu_64_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zero64 = CmmSR('zero', CmmType('u', 64))

        new_opB = opB_0.new()
        new_opB.set_size(8)

        if var0.get_reg().intersects_with(opC_0) :
            [var0, var2] = [var2, var0]

        if var1.get_reg().intersects_with(opC_0) :
            [var1, var2] = [var2, var1]

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zero64 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(11), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'or', var0, zeroC_1, opA ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC_1, var0, new_opB ]),
            CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(5), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ op, opC_0, var0, new_opB ]),
            CmmBinAsInstruction([ 'shr', opC_1, var0, var1 ]),
            ]

    def fix_64__shl_uu_64_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()
        opA_size = opA.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zero64 = CmmSR('zero', CmmType('u', 64))
        zero32 = CmmSR('zero', CmmType('u', 32))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zero64 ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ op, opC_1, opA, opB - 32 ]),
                    CmmAssgnInstruction([ 'move', opC_0, zero32 ]),
                    ]
            elif opB > 32 - opA_size :
                if opA.intersects_with(opC_1) :
                    return [
                        CmmBinAsInstruction([ op, opC_0, opA, opB ]),
                        CmmBinAsInstruction([ 'shri', opC_1, opA, 32 - opB ]),
                        ]
                else :
                    return [
                        CmmBinAsInstruction([ 'shri', opC_1, opA, 32 - opB ]),
                        CmmBinAsInstruction([ op, opC_0, opA, opB ]),
                        ]
            else :
                return [
                    CmmBinAsInstruction([ op, opC_0, opA, opB ]),
                    CmmAssgnInstruction([ 'move', opC_1, zero32 ]),
                    ]
        else : # opB is a reg
            instr = []
            new_opB = opB.new()
            new_opB.set_size(8)

            if opA.intersects_with(opC_0) :
                instr = [CmmAssgnInstruction([ 'move', var0, opA ])]
            else :
                var0 = opA

            if var1.get_reg().intersects_with(opC_0) :
                [var1, var2] = [var2, var1]


            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zero64 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(10 + len(instr)), CmmConst(0), CmmConst(0) ]),
                ] +\
                instr +\
                [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ op, opC_1, var0, new_opB ]),
                CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(5), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ op, opC_0, var0, new_opB ]),
                CmmBinAsInstruction([ 'shr', opC_1, var0, var1 ]),
                ]

    def fix_64__shl_uu_64_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zero64 = CmmSR('zero', CmmType('u', 64))

        instr = []
        new_opB = opB_0.new()
        new_opB.set_size(8)

        if opA.intersects_with(opC_0) :
            instr = [CmmAssgnInstruction([ 'move', var0, opA ])]
        else :
            var0 = opA

        if var1.get_reg().intersects_with(opC_0) :
            [var1, var2] = [var2, var1]

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zero64 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(10 + len(instr)), CmmConst(0), CmmConst(0) ]),
            ] +\
            instr +\
            [
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC_1, var0, new_opB ]),
            CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(5), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ op, opC_0, var0, new_opB ]),
            CmmBinAsInstruction([ 'shr', opC_1, var0, var1 ]),
            ]

    def fix_64__shl_uu_64_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zero64 = CmmSR('zero', CmmType('u', 64))
        zero32 = CmmSR('zero', CmmType('u', 32))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zero64 ]),
                    ]
            elif opB > 31 :
                return [
                    (CmmBinAsInstruction([ op, opC_1, opA_0, opB - 32 ]) if opB - 32\
               else CmmAssgnInstruction([ 'move', opC_1, opA_0 ])),
                    CmmAssgnInstruction([ 'move', opC_0, zero32 ]),
                    ]
            else :
                if opB :
                    if var0.get_reg().intersects_with(opC_0) :
                        [var0, var1] = [var1, var0]

                    return [
                        CmmBinAsInstruction(['shri', var0, opA_0, 32 - opB ]),
                        CmmBinAsInstruction([ op, opC_1, opA_1, opB ]),
                        CmmBinAsInstruction([ op, opC_0, opC_0, opB ]),
                        CmmBinAsInstruction([ 'or', opC_1, opC_1, var0 ]),
                        ]
                else :
                    return [
                        CmmAssgnInstruction([ 'move', opC, opA ]),
                        ]
        else : # opB is a reg
            new_opB = opB.new()
            new_opB.set_size(8)

            instr = []

            if opA_0.intersects_with(opC_0) and opB.intersects_with(opC_1) :
                if var0.get_reg().intersects_with(opC_0) :
                    [var0, var2] = [var2, var0]

                if var1.get_reg().intersects_with(opC_0) :
                    [var1, var2] = [var2, var1]

                instr =\
                    [CmmAssgnInstruction([ 'move', var0, opA_0 ]),
                    CmmBinAsInstruction([ op, opC_0, var0, new_opB ]),
                    CmmBinAsInstruction([ 'shr', opC_1, var0, var1 ])]
            elif opA_0.intersects_with(opC_0) :
                instr =\
                    [CmmBinAsInstruction([ 'shr', opC_1, opA_0, var1 ]),
                    CmmBinAsInstruction([ op, opC_0, opA_0, new_opB ])]
            else :
                instr =\
                    [CmmBinAsInstruction([ op, opC_0, opA_0, new_opB ]),
                    CmmBinAsInstruction([ 'shr', opC_1, opA_0, var1 ])]

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zero64 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(8 + len(instr)), CmmConst(0), CmmConst(0) ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ op, opC_1, opA_0, new_opB ]),
                CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(3 + len(instr)), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                ] +\
                instr

    def fix_64__shl_uu_64_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        [opA_1, opA_0] = self.get_opA().splt_into_2()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zero64 = CmmSR('zero', CmmType('u', 64))

        new_opB = opB_0.new()
        new_opB.set_size(8)

        instr = []

        if opA_0.intersects_with(opC_0) and opB_0.intersects_with(opC_1) :
            if var0.get_reg().intersects_with(opC_0) :
                [var0, var2] = [var2, var0]

            if var1.get_reg().intersects_with(opC_0) :
                [var1, var2] = [var2, var1]

            instr =\
                [CmmAssgnInstruction([ 'move', var0, opA_0 ]),
                CmmBinAsInstruction([ op, opC_0, var0, new_opB ]),
                CmmBinAsInstruction([ 'shr', opC_1, var0, var1 ])]
        elif opA_0.intersects_with(opC_0) :
            instr =\
                [CmmBinAsInstruction([ 'shr', opC_1, opA_0, var1 ]),
                CmmBinAsInstruction([ op, opC_0, opA_0, new_opB ])]
        else :
            instr =\
                [CmmBinAsInstruction([ op, opC_0, opA_0, new_opB ]),
                CmmBinAsInstruction([ 'shr', opC_1, opA_0, var1 ])]

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zero64 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(8 + len(instr)), CmmConst(0), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC_1, opA_0, new_opB ]),
            CmmAssgnInstruction([ 'move', opC_0, zeroC_0 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(3 + len(instr)), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            ] +\
            instr

################################################################################

    def fix_64__shr_su_8_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        zeroA = CmmSR('zero', CmmType('u', opA_size))

        CA_size = opC_size + opA_size

        if CA_size > 16 :
            CA_size = 32

        oneCA = CmmSR('one', CmmType('u', CA_size))

        var0_CA = var0.new()
        var0_CA.set_size(CA_size)

        var0_C = var0.new()
        var0_C.set_size(opC_size)

        if opB_size > CA_size :
            opB.set_size(CA_size)

        return [
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0_CA, oneCA, oneCA ]),
            CmmBinAsInstruction([ 'shli', var0_CA, var0_CA, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0_CA, var0_CA, opA ]),
            CmmBinAsInstruction([ op, var0_CA, var0_CA, opB ]),
            CmmAssgnInstruction([ 'move', opC, var0_C ]),
            ]

    def fix_64__shr_su_8_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        zeroA = CmmSR('zero', CmmType('u', opA_size))

        CA_size = opC_size + opA_size

        if CA_size > 16 :
            CA_size = 32

        oneCA = CmmSR('one', CmmType('u', CA_size))

        var0_CA = var0.new()
        var0_CA.set_size(CA_size)

        var0_C = var0.new()
        var0_C.set_size(opC_size)

        new_opB = opB_0.new()
        new_opB.set_size(CA_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(7), CmmConst(29), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0_CA, oneCA, oneCA ]),
            CmmBinAsInstruction([ 'shli', var0_CA, var0_CA, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0_CA, var0_CA, opA ]),
            CmmBinAsInstruction([ op, var0_CA, var0_CA, new_opB ]),
            CmmAssgnInstruction([ 'move', opC, var0_C ]),
            ]

    def fix_64__shr_su_8_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneA = CmmSR('one', CmmType('u', opA_size))

        var0_C = var0.new()
        var0_C.set_size(opC_size)

        if opB.is_a_constant() :
            if opB >= opA_size :
                return [
                    CmmBinAsInstruction([ op, opC, zeroC, opB ]),
                    ]
            elif opB > 32 - opC_size :
                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var0, oneA, oneA ]),
                    CmmBinAsInstruction([ 'shl', var0, var0, 32 - opB ]),
                    CmmBinAsInstruction([ op, var1, opA, opB ]),
                    CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
                    CmmAssgnInstruction([ 'move', opC, var0_C ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ op, var0, opA, opB ]),
                    CmmAssgnInstruction([ 'move', opC, var0_C ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var0, oneA, oneA ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var0, var0, var1 ]),
                CmmBinAsInstruction([ op, var1, opA, opB ]),
                CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
                CmmAssgnInstruction([ 'move', opC, var0_C ]),
                ]

    def fix_64__shr_su_8_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneA = CmmSR('one', CmmType('u', opA_size))

        var0_C = var0.new()
        var0_C.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(9), CmmConst(29), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneA, oneA ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var0, var0, var1 ]),
            CmmBinAsInstruction([ op, var1, opA, opB_0 ]),
            CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
            CmmAssgnInstruction([ 'move', opC, var0_C ]),
            ]

    def fix_64__shr_su_8_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_1_size = opA_1.get_size()
        opB_size = opB.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))
        oneA_1 = CmmSR('one', CmmType('u', opA_1_size))

        new_var0 = var0.new()
        new_var0.set_size(opC_size)

        if opB.is_a_constant() :
            if opB > 63 :
                print ("shri " + str(self.body) + "\n")
                return [
                    CmmAssgnInstruction([ 'move', opC, zeroC ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
                    CmmBinAsInstruction([ 'addc', var1, oneA_1, oneA_1 ]),
                    CmmBinAsInstruction([ 'shli', var1, var1, 32 - (opB - 32) ]),
                    ] +\
                    [CmmBinAsInstruction([ op, var0, opA_1, opB - 32 ]) if opB - 32\
                else CmmAssgnInstruction([ 'move', var0, opA_1 ])] +\
                    [
                    CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
                    CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                    ]
            elif opB > 32 - opC_size :
                return [
                    CmmBinAsInstruction([ 'shli', var1, opA_1, 32 - opB ]),
                    CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                    CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
                    CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                    CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                    ]
        else : # opB is a reg
            var0_B = var0.new()
            var0_B.set_size(opB_size)

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zeroC ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(18), CmmConst(0), CmmConst(0) ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(10), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ 'sbi', var0_B, opB, CmmConst(32) ]),
                CmmAssgnInstruction([ 'mviw', var2, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var2, var2, var0_B ]),
                CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
                CmmBinAsInstruction([ 'addc', var1, oneA_1, oneA_1 ]),
                CmmBinAsInstruction([ 'shli', var1, var1, var2 ]),
                CmmBinAsInstruction([ op, var0, opA_1, var0_B ]),
                CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(7), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
                CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
                CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                ]

    def fix_64__shr_su_8_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.splt_into_2()

        opC_size = opC.get_size()
        opA_1_size = opA_1.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))
        oneA_1 = CmmSR('one', CmmType('u', opA_1_size))

        new_var0 = var0.new()
        new_var0.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(18), CmmConst(0), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(10), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'sbi', var0, opB_0, CmmConst(32) ]),
            CmmAssgnInstruction([ 'mviw', var2, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var2, var2, var0 ]),
            CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
            CmmBinAsInstruction([ 'addc', var1, oneA_1, oneA_1 ]),
            CmmBinAsInstruction([ 'shli', var1, var1, var2 ]),
            CmmBinAsInstruction([ op, var0, opA_1, var0 ]),
            CmmAssgnInstruction([ 'move', opC, new_var0 ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(7), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
            CmmBinAsInstruction([ op, var0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
            CmmAssgnInstruction([ 'move', opC, new_var0 ]),
            ]

    def fix_64__shr_su_32_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        var0_C = var0.new()
        var0_C.set_size(opC_size)

        if opB.is_a_constant() :
            if opB >= opA_size :
                return [
                    CmmAssgnInstruction([ 'move', opC, zeroC ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
                    CmmBinAsInstruction([ 'shl', var1, var0, 32 - opB ]),
                    CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
                    CmmBinAsInstruction([ 'or', var0, var0, opA ]),
                    CmmBinAsInstruction([ op, var0, var0, opB ]),
                    CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
                    CmmAssgnInstruction([ 'move', opC, var0_C ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var1, var0, var1 ]),
                CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
                CmmBinAsInstruction([ 'or', var0, var0, opA ]),
                CmmBinAsInstruction([ op, var0, var0, opB ]),
                CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
                CmmAssgnInstruction([ 'move', opC, var0_C ]),
                ]

    def fix_64__shr_su_32_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.splt_into_2()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        oneC = CmmSR('one', CmmType('u', opC_size))
        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        var0_C = var0.new()
        var0_C.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(11), CmmConst(29), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneC, oneC ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var1, var0, var1 ]),
            CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, var0, var0, opB_0 ]),
            CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
            CmmAssgnInstruction([ 'move', opC, var0_C ]),
            ]

    def fix_64__shr_su_32_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneA = CmmSR('one', CmmType('u', opA_size))

        if opB.is_a_constant() :
            if opB >= opA_size :
                return [
                    CmmBinAsInstruction([ op, opC, zeroC, opB ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var0, oneA, oneA ]),
                    CmmBinAsInstruction([ 'shl', var0, var0, 32 - opB ]),
                    CmmBinAsInstruction([ op, var1, opA, opB ]),
                    CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var0, oneA, oneA ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var0, var0, var1 ]),
                CmmBinAsInstruction([ op, var1, opA, opB ]),
                CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
                ]

    def fix_64__shr_su_32_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneA = CmmSR('one', CmmType('u', opA_size))

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(9), CmmConst(29), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var0, oneA, oneA ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var0, var0, var1 ]),
            CmmBinAsInstruction([ op, var1, opA, opB_0 ]),
            CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
            ]

    def fix_64__shr_su_32_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_1_size = opA_1.get_size()
        opB_size = opB.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))
        oneA_1 = CmmSR('one', CmmType('u', opA_1_size))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zeroC ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
                    CmmBinAsInstruction([ 'addc', var1, oneA_1, oneA_1 ]),
                    CmmBinAsInstruction([ 'shli', var1, var1, 32 - (opB - 32) ]),
                    ] +\
                    [CmmBinAsInstruction([ op, var0, opA_1, opB - 32 ]) if opB - 32\
               else CmmAssgnInstruction([ 'move', var0, opA_1 ])] +\
                    [
                    CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ 'shli', var1, opA_1, 32 - opB ]),
                    CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                    CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
                    ]
        else : # opB is a reg
            var0_B = var0.new()
            var0_B.set_size(opB_size)

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zeroC ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(17), CmmConst(0), CmmConst(0) ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(10), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ 'sbi', var0_B, opB, CmmConst(32) ]),
                CmmAssgnInstruction([ 'mviw', var2, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var2, var2, var0_B ]),
                CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
                CmmBinAsInstruction([ 'addc', var1, oneA_1, oneA_1 ]),
                CmmBinAsInstruction([ 'shli', var1, var1, var2 ]),
                CmmBinAsInstruction([ op, var0, opA_1, var0_B ]),
                CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(6), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
                CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
                ]

    def fix_64__shr_su_32_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.splt_into_2()

        opC_size = opC.get_size()
        opA_1_size = opA_1.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))
        oneA_1 = CmmSR('one', CmmType('u', opA_1_size))

        new_var0 = var0.new()
        new_var0.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(17), CmmConst(0), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(10), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'sbi', var0, opB_0, CmmConst(32) ]),
            CmmAssgnInstruction([ 'mviw', var2, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var2, var2, var0 ]),
            CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
            CmmBinAsInstruction([ 'addc', var1, oneA_1, oneA_1 ]),
            CmmBinAsInstruction([ 'shli', var1, var1, var2 ]),
            CmmBinAsInstruction([ op, var0, opA_1, var0 ]),
            CmmAssgnInstruction([ 'move', opC, new_var0 ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(7), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
            CmmBinAsInstruction([ op, var0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
            ]

    def fix_64__shr_su_64_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()
        opA_size = opA.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))
        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        var0_C = var0.new()
        var0_C.set_size(opC_0_size)

        if opB.is_a_constant() :
            if opB >= opA_size :
                return [
                    CmmAssgnInstruction([ 'move', opC, zeroC ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var2, oneC_0, oneC_0 ]),
                    CmmBinAsInstruction([ 'shl', var1, var2, 32 - opB ]),
                    CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
                    CmmBinAsInstruction([ 'or', var0, var0, opA ]),
                    CmmBinAsInstruction([ op, var0, var0, opB ]),
                    CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
                    CmmAssgnInstruction([ 'move', opC_1, var2 ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var2, oneC_0, oneC_0 ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var1, var2, var1 ]),
                CmmBinAsInstruction([ 'shli', var0, var0, CmmConst(opA_size) ]),
                CmmBinAsInstruction([ 'or', var0, var0, opA ]),
                CmmBinAsInstruction([ op, var0, var0, opB ]),
                CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
                CmmAssgnInstruction([ 'move', opC_1, var2 ]),
                ]

    def fix_64__shr_su_64_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.splt_into_2()

        opC_0_size = opC_0.get_size()
        opA_size = opA.get_size()

        oneC_0 = CmmSR('one', CmmType('u', opC_0_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))

        var0_C = var0.new()
        var0_C.set_size(opC_0_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(11), CmmConst(29), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var2, oneC_0, oneC_0 ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var1, var2, var1 ]),
            CmmBinAsInstruction([ 'shli', var0, var2, CmmConst(opA_size) ]),
            CmmBinAsInstruction([ 'or', var0, var0, opA ]),
            CmmBinAsInstruction([ op, var0, var0, opB_0 ]),
            CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
            CmmAssgnInstruction([ 'move', opC_1, var2 ]),
            ]

    def fix_64__shr_su_64_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneA = CmmSR('one', CmmType('u', opA_size))

        if opB.is_a_constant() :
            if opB >= opA_size :
                return [
                    CmmBinAsInstruction([ op, opC, zeroC, opB ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                    CmmBinAsInstruction([ 'addc', var2, oneA, oneA ]),
                    CmmBinAsInstruction([ 'shl', var0, var2, 32 - opB ]),
                    CmmBinAsInstruction([ op, var1, opA, opB ]),
                    CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
                    CmmAssgnInstruction([ 'move', opC_1, var2 ]),
                    ]
        else :
            return [
                CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
                CmmBinAsInstruction([ 'addc', var2, oneA, oneA ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var0, var2, var1 ]),
                CmmBinAsInstruction([ op, var1, opA, opB ]),
                CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
                CmmAssgnInstruction([ 'move', opC_1, var2 ]),
                ]

    def fix_64__shr_su_64_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opA_size = opA.get_size()

        zeroA = CmmSR('zero', CmmType('u', opA_size))
        oneA = CmmSR('one', CmmType('u', opA_size))

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(9), CmmConst(29), CmmConst(0) ]),
            CmmBinAsInstruction([ 'add', zeroA, opA, opA ]),
            CmmBinAsInstruction([ 'addc', var2, oneA, oneA ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var0, var2, var1 ]),
            CmmBinAsInstruction([ op, var1, opA, opB_0 ]),
            CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
            CmmAssgnInstruction([ 'move', opC_1, var2 ]),
            ]

    def fix_64__shr_su_64_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()
        var3 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_1_size = opA_1.get_size()
        opB_size = opB.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))
        oneA_1 = CmmSR('one', CmmType('u', opA_1_size))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zeroC ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
                    CmmBinAsInstruction([ 'addc', var2, oneA_1, oneA_1 ]),
                    CmmBinAsInstruction([ 'shli', var1, var2, 32 - (opB - 32) ]),
                    ] +\
                    [CmmBinAsInstruction([ op, var0, opA_1, opB - 32 ]) if opB - 32\
                else CmmAssgnInstruction([ 'move', var0, opA_1 ])] +\
                    [
                    CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
                    CmmAssgnInstruction([ 'move', opC_1, var2 ]),
                    ]
            else :
                if var1.get_reg().intersects_with(opC_0) :
                    [var1, var0] = [var0, var1]

                if var2.get_reg().intersects_with(opC_0) :
                    [var2, var0] = [var0, var2]

                return [
                    CmmBinAsInstruction([ 'shli', var1, opA_1, 32 - opB ]),
                    CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                    CmmBinAsInstruction([ op, var2, opA_1, opB ]),
                    CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
                    CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
                    CmmBinAsInstruction([ 'addc', var1, oneA_1, oneA_1 ]),
                    CmmBinAsInstruction([ 'shli', var1, var1, 32 - opB ]),
                    CmmBinAsInstruction([ 'or', opC_1, var1, var2 ]),
                    ]
        else : # opB is a reg
            var0_B = var0.new()
            var0_B.set_size(opB_size)

            if var1.get_reg().intersects_with(opC_0) :
                [var1, var0] = [var0, var1]

            if var2.get_reg().intersects_with(opC_0) :
                [var2, var0] = [var0, var2]

            if var3.get_reg().intersects_with(opC_0) :
                [var3, var0] = [var0, var3]

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zeroC ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(23), CmmConst(0), CmmConst(0) ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(11), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ 'sbi', var0_B, opB, CmmConst(32) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, var0_B ]),
                CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
                CmmBinAsInstruction([ 'addc', var2, oneA_1, oneA_1 ]),
                CmmBinAsInstruction([ 'shli', var1, var2, var1 ]),
                CmmBinAsInstruction([ op, var0, opA_1, var0_B ]),
                CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
                CmmAssgnInstruction([ 'move', opC_1, var2 ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(6), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var3, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var3, var3, opB ]),
                CmmBinAsInstruction([ 'shl', var1, opA_1, var3 ]),
                CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                CmmBinAsInstruction([ op, var2, opA_1, opB ]),
                CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
                CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
                CmmBinAsInstruction([ 'addc', var1, oneA_1, oneA_1 ]),
                CmmBinAsInstruction([ 'shl', var1, var1, var3 ]),
                CmmBinAsInstruction([ 'or', opC_1, var1, var2 ]),
                ]

    def fix_64__shr_su_64_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()
        var2 = self._get_auxiliary_var()
        var3 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.splt_into_2()

        opC_size = opC.get_size()
        opA_1_size = opA_1.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroA_1 = CmmSR('zero', CmmType('u', opA_1_size))
        oneA_1 = CmmSR('one', CmmType('u', opA_1_size))

        if var1.get_reg().intersects_with(opC_0) :
            [var1, var0] = [var0, var1]

        if var2.get_reg().intersects_with(opC_0) :
            [var2, var0] = [var0, var2]

        if var3.get_reg().intersects_with(opC_0) :
            [var3, var0] = [var0, var3]

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(23), CmmConst(0), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(11), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'sbi', var0, opB_0, CmmConst(32) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, var0 ]),
            CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
            CmmBinAsInstruction([ 'addc', var2, oneA_1, oneA_1 ]),
            CmmBinAsInstruction([ 'shli', var1, var2, var1 ]),
            CmmBinAsInstruction([ op, var0, opA_1, var0 ]),
            CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
            CmmAssgnInstruction([ 'move', opC_1, var2 ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(6), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var3, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var3, var3, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var1, opA_1, var3 ]),
            CmmBinAsInstruction([ op, var0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ op, var2, opA_1, opB_0 ]),
            CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
            CmmBinAsInstruction([ 'add', zeroA_1, opA_1, opA_1 ]),
            CmmBinAsInstruction([ 'addc', var1, oneA_1, oneA_1 ]),
            CmmBinAsInstruction([ 'shl', var1, var1, var3 ]),
            CmmBinAsInstruction([ 'or', opC_1, var1, var2 ]),
            ]

################################################################################

    def fix_64__shr_uu_8_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        return [
            ]

    def fix_64__shr_uu_8_8_16(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA.set_size(opC_size)

        new_opB = opB.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shr', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA, new_opB ]),
            ]

    def fix_64__shr_uu_8_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        opA.set_size(opC_size)

        new_opB = opB_0.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shr', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(2), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC, opA, new_opB ]),
            ]

    def fix_64__shr_uu_8_16_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()

        new_var0 = var0.new()
        new_var0.set_size(opC_size)

        var0.set_size(opA_size)

        return [
            CmmBinAsInstruction([ op, var0, opA, opB ]),
            CmmAssgnInstruction([ 'move', opC, new_var0 ]),
            ]

    def fix_64__shr_uu_8_16_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        zeroB = CmmSR('zero', CmmType('u', opB_size))

        new_var0 = var0.new()
        new_var0.set_size(opC_size)

        var0.set_size(opB_size)

        return [
            CmmBinAsInstruction([ 'or', var0, zeroB, opA ]),
            CmmBinAsInstruction([ op, var0, var0, opB ]),
            CmmAssgnInstruction([ 'move', opC, new_var0 ]),
            ]

    def fix_64__shr_uu_8_16_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_0_size = opB_0.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroB_0 = CmmSR('zero', CmmType('u', opB_0_size))

        new_var0 = var0.new()
        new_var0.set_size(opC_size)

        var0.set_size(opB_0_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(1) ]),
            CmmBinAsInstruction([ 'shl', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(4), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'or', var0, zeroB_0, opA ]),
            CmmBinAsInstruction([ op, var0, var0, opB_0 ]),
            CmmAssgnInstruction([ 'move', opC, new_var0 ]),
            ]

    def fix_64__shr_uu_8_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opB_size = opB.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        new_var0 = var0.new()
        new_var0.set_size(opC_size)

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zeroC ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ op, var0, opA_1, opB - 32 ]) if opB - 32\
               else CmmAssgnInstruction([ 'move', var0, opA_1 ])] +\
                    [
                    CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                    ]
            elif opB > 32 - opC_size :
                return [
                    CmmBinAsInstruction([ 'shli', var1, opA_1, 32 - opB ]),
                    CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                    CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
                    CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                    CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                    ]
        else : # opB is a reg
            var0_B = var0.new()
            var0_B.set_size(opB_size)

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zeroC ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(13), CmmConst(0), CmmConst(0) ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(5), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ 'sbi', var0_B, opB, CmmConst(32) ]),
                CmmBinAsInstruction([ op, var0, opA_1, var0_B ]),
                CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(7), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
                CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
                CmmAssgnInstruction([ 'move', opC, new_var0 ]),
                ]

    def fix_64__shr_uu_8_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        new_var0 = var0.new()
        new_var0.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(13), CmmConst(0), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(5), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'sbi', var0, opB_0, CmmConst(32) ]),
            CmmBinAsInstruction([ op, var0, opA_1, var0 ]),
            CmmAssgnInstruction([ 'move', opC, new_var0 ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(7), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
            CmmBinAsInstruction([ op, var0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ 'or', var0, var0, var1 ]),
            CmmAssgnInstruction([ 'move', opC, new_var0 ]),
            ]

    def fix_64__shr_uu_16_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        var0.set_size(opC_size)

        return [
            CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
            CmmBinAsInstruction([ op, opC, var0, opB ]),
            ]

    def fix_64__shr_uu_16_8_32(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        new_opB = opB.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shr', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(3), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
            CmmBinAsInstruction([ op, opC, var0, new_opB ]),
            ]

    def fix_64__shr_uu_16_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        new_opB = opB_0.new()
        new_opB.set_size(opC_size)

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'shr', opC, zeroC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(3), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'or', var0, zeroC, opA ]),
            CmmBinAsInstruction([ op, opC, var0, new_opB ]),
            ]

    def fix_64__shr_uu_32_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opB_size = opB.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zeroC ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ op, opC, opA_1, opB - 32 ]) if opB - 32\
               else CmmAssgnInstruction([ 'move', opC, opA_1 ])
                    ]
            elif opB > 32 - opC_size :
                return [
                    CmmBinAsInstruction([ 'shli', var1, opA_1, 32 - opB ]),
                    CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                    CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
                    ]
            else :
                return [
                    CmmBinAsInstruction([ op, opC, opA_0, opB ]),
                    ]
        else : # opB is a reg
            var0_B = var0.new()
            var0_B.set_size(opB_size)

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zeroC ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(11), CmmConst(0), CmmConst(0) ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ 'sbi', var0_B, opB, CmmConst(32) ]),
                CmmBinAsInstruction([ op, opC, opA_1, var0_B ]),
                CmmGotoInstruction([ 'btjr', zeroC, CmmConst(6), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
                CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
                ]

    def fix_64__shr_uu_32_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.splt_into_2()

        opC_size = opC.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(11), CmmConst(0), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(4), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'sbi', var0, opB_0, CmmConst(32) ]),
            CmmBinAsInstruction([ op, opC, opA_1, var0 ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(6), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
            CmmBinAsInstruction([ op, var0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
            ]

    def fix_64__shr_uu_64_8_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opC_0_size = opC_0.get_size()
        opC_1_size = opC_1.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroC_0 = CmmSR('zero', CmmType('u', opC_0_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        var0.set_size(opC_0_size)

        if opB.is_a_constant() and opB >= 8 :
            return [
                CmmAssgnInstruction([ 'move', opC, zeroC ]),
                ]
        else : # opB is a reg
            return [
                CmmBinAsInstruction([ 'or', var0, zeroC_0, opA ]),
                CmmBinAsInstruction([ op, opC_0, var0, opB ]),
                CmmAssgnInstruction([ 'move', opC_1, zeroC_1 ]),
                ]

    def fix_64__shr_uu_64_8_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()
        opC_1_size = opC_1.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(4), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ 'or', var0, zeroC_1, opA ]),
            CmmBinAsInstruction([ op, opC_0, var0, opB_0 ]),
            CmmAssgnInstruction([ 'move', opC_1, zeroC_1 ]),
            ]

    def fix_64__shr_uu_64_32_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        opB = self.get_opB()

        opC_1_size = opC_1.get_size()

        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        return [
            CmmBinAsInstruction([ op, opC_0, opA, opB ]),
            CmmAssgnInstruction([ 'move', opC_1, zeroC_1 ]),
            ]

    def fix_64__shr_uu_64_32_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opB_1, opB_0] = self.get_opB().splt_into_2()

        opC_size = opC.get_size()
        opC_1_size = opC_1.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))
        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(0xff) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC, CmmConst(4), CmmConst(0), CmmConst(0) ]),
            CmmBinAsInstruction([ op, opC_0, opA, opB_0 ]),
            CmmAssgnInstruction([ 'move', opC_1, zeroC_1 ]),
            ]

    def fix_64__shr_uu_64_64_8(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()

        opC_size = opC.get_size()
        opC_1_size = opC_1.get_size()
        opB_size = opB.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        if opB.is_a_constant() :
            if opB > 63 :
                return [
                    CmmAssgnInstruction([ 'move', opC, zeroC ]),
                    ]
            elif opB > 31 :
                return [
                    CmmBinAsInstruction([ op, opC_0, opA_1, opB - 32 ]) if opB - 32\
               else CmmAssgnInstruction([ 'move', opC_0, opA_1 ])] +\
                    [
                    CmmAssgnInstruction([ 'move', opC_1, zeroC_1 ]),
                    ]
            elif opB :
                return [
                    CmmBinAsInstruction([ 'shli', var1, opA_1, 32 - opB ]),
                    CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                    CmmBinAsInstruction([ 'or', opC_0, var0, var1 ]),
                    CmmBinAsInstruction([ op, opC_1, opA_1, opB ]),
                    ]
            else :
                return [
                    CmmAssgnInstruction([ 'move', opC, opA ]),
                    ]
        else : # opB is a reg
            var0_B = var0.new()
            var0_B.set_size(opB_size)

            return [
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(63) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
                CmmAssgnInstruction([ 'move', opC, zeroC ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(13), CmmConst(0), CmmConst(0) ]),
                CmmCmpInstruction([ 'cpiw', opB, CmmConst(31) ]),
                CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(5), CmmConst(31), CmmConst(0) ]),
                CmmBinAsInstruction([ 'sbi', var0_B, opB, CmmConst(32) ]),
                CmmBinAsInstruction([ op, opC_0, opA_1, var0_B ]),
                CmmAssgnInstruction([ 'move', opC_1, zeroC_1 ]),
                CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(7), CmmConst(0), CmmConst(0) ]),
                CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
                CmmBinAsInstruction([ 'sub', var1, var1, opB ]),
                CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
                CmmBinAsInstruction([ op, var0, opA_0, opB ]),
                CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
                CmmBinAsInstruction([ op, opC_1, opA_1, opB ]),
                ]

    def fix_64__shr_uu_64_64_64(self) :
        var0 = self._get_auxiliary_var()
        var1 = self._get_auxiliary_var()

        op = self.get_op()

        opC = self.get_opC()
        [opC_1, opC_0] = opC.splt_into_2()
        opA = self.get_opA()
        [opA_1, opA_0] = opA.splt_into_2()
        opB = self.get_opB()
        [opB_1, opB_0] = opB.splt_into_2()

        opC_size = opC.get_size()
        opC_1_size = opC_1.get_size()

        zeroC = CmmSR('zero', CmmType('u', opC_size))

        zeroC_1 = CmmSR('zero', CmmType('u', opC_1_size))

        return [
            CmmCmpInstruction([ 'cpiw', opB_1, CmmConst(0) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(29), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(63) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(3), CmmConst(31), CmmConst(0) ]),
            CmmAssgnInstruction([ 'move', opC, zeroC ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(13), CmmConst(0), CmmConst(0) ]),
            CmmCmpInstruction([ 'cpiw', opB_0, CmmConst(31) ]),
            CmmGotoInstruction([ 'btjr', CmmSR('pc'), CmmConst(5), CmmConst(31), CmmConst(0) ]),
            CmmBinAsInstruction([ 'sbi', var0, opB_0, CmmConst(32) ]),
            CmmBinAsInstruction([ op, opC_0, opA_1, var0 ]),
            CmmAssgnInstruction([ 'move', opC_1, zeroC_1 ]),
            CmmGotoInstruction([ 'btjr', zeroC_1, CmmConst(7), CmmConst(0), CmmConst(0) ]),
            CmmAssgnInstruction([ 'mviw', var1, CmmConst(32) ]),
            CmmBinAsInstruction([ 'sub', var1, var1, opB_0 ]),
            CmmBinAsInstruction([ 'shl', var1, opA_1, var1 ]),
            CmmBinAsInstruction([ op, var0, opA_0, opB_0 ]),
            CmmBinAsInstruction([ 'or', opC, var0, var1 ]),
            CmmBinAsInstruction([ op, opC_1, opA_1, opB_0 ]),
            ]

    fix_64 = {
        'alu' : {
            'ss' : {
                8  : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_8_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_8_16_16,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_8_16_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_8_16_16,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_8_64_16,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_8_64_64,
                    },
                },
                16 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_16_8_16,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_16_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_16_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_16_32_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_8_16_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_16_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_8_64_64,
                    },
                },
                32 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_16_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_16_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_16_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_16_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_16_16_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_32_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_16_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_16_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_8_64_64,
                    },
                },
                64 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_64_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_64_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_64_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_64_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_64_32_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_64_32_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_64_32_32,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_64_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_ss_64_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_ss_64_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_ss_64_64_32,
                        64 : CmmBinAsInstruction.fix_64__alu_ss_64_64_64,
                    },
                },
            },
            'us' : {
                8  : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__alu_us_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_us_8_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_8_16_16,
                        32 : CmmBinAsInstruction.fix_64__alu_us_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_us_8_16_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_8_16_16,
                        32 : CmmBinAsInstruction.fix_64__alu_us_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_us_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_8_64_16,
                        32 : CmmBinAsInstruction.fix_64__alu_us_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__alu_us_8_64_64,
                    },
                },
                16 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_16_8_16,
                        32 : CmmBinAsInstruction.fix_64__alu_us_16_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_us_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_16_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_us_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_16_32_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_8_16_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_us_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_16_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__alu_us_8_64_64,
                    },
                },
                32 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_16_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_us_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_16_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_us_16_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_16_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_16_16_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__alu_us_32_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_16_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_16_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__alu_us_8_64_64,
                    },
                },
                64 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_64_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_us_64_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_64_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_us_64_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_64_32_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_64_32_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_64_32_32,
                        64 : CmmBinAsInstruction.fix_64__alu_us_64_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_us_64_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_us_64_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_us_64_64_32,
                        64 : CmmBinAsInstruction.fix_64__alu_us_64_64_64,
                    },
                },
            },
            'su' : {
                8  : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__alu_su_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_su_8_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_8_16_16,
                        32 : CmmBinAsInstruction.fix_64__alu_su_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_su_8_16_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_8_16_16,
                        32 : CmmBinAsInstruction.fix_64__alu_su_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_su_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_8_64_16,
                        32 : CmmBinAsInstruction.fix_64__alu_su_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__alu_su_8_64_64,
                    },
                },
                16 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_16_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_su_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_su_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_8_16_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_su_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__alu_su_8_64_64,
                    },
                },
                32 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_16_8_8,
                        64 : CmmBinAsInstruction.fix_64__alu_su_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_16_8_8,
                        64 : CmmBinAsInstruction.fix_64__alu_su_16_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__alu_su_32_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__alu_su_8_64_64,
                    },
                },
                64 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__alu_su_64_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__alu_su_64_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_64_32_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_64_32_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_64_32_8,
                        64 : CmmBinAsInstruction.fix_64__alu_su_64_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_su_64_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_su_64_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_su_64_64_8,
                        64 : CmmBinAsInstruction.fix_64__alu_su_64_64_64,
                    },
                },
            },
            'uu' : {
                8  : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_8_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_8_16_16,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_8_16_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_8_16_16,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_8_64_16,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_8_64_64,
                    },
                },
                16 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_16_8_16,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_16_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_8_16_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_8_16_16,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_8_64_64,
                    },
                },
                32 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_16_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_16_8_16,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_16_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_32_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_8_64_64,
                    },
                },
                64 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_64_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_64_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_64_8_32,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_64_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_64_32_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_64_32_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_64_32_8,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_64_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__alu_uu_64_64_8,
                        16 : CmmBinAsInstruction.fix_64__alu_uu_64_64_8,
                        32 : CmmBinAsInstruction.fix_64__alu_uu_64_64_8,
                        64 : CmmBinAsInstruction.fix_64__alu_uu_64_64_64,
                    },
                },
            },
        },
        'shl' : {
            'su' : {
                8  : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__shl_su_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_su_8_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__shl_su_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_su_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__shl_su_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_su_8_8_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_8_64_16,
                        32 : CmmBinAsInstruction.fix_64__shl_su_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__shl_su_8_64_64,
                    },
                },
                16 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_16_8_32,
                        64 : CmmBinAsInstruction.fix_64__shl_su_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_su_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_8_16_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_su_8_8_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__shl_su_8_64_64,
                    },
                },
                32 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_16_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_su_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_16_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_su_16_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_su_8_8_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__shl_su_8_64_64,
                    },
                },
                64 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_su_64_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_su_64_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_64_32_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_64_32_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_64_32_8,
                        64 : CmmBinAsInstruction.fix_64__shl_su_64_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shl_su_64_64_8,
                        16 : CmmBinAsInstruction.fix_64__shl_su_64_64_8,
                        32 : CmmBinAsInstruction.fix_64__shl_su_64_64_8,
                        64 : CmmBinAsInstruction.fix_64__shl_su_64_64_64,
                    },
                },
            },
            'uu' : {
                8  : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_8_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_8_8_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_8_64_16,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_8_64_64,
                    },
                },
                16 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_16_8_32,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_8_16_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_8_8_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_8_64_16,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_8_64_64,
                    },
                },
                32 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_16_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_16_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_16_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_8_8_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_8_64_64,
                    },
                },
                64 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_64_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_64_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_64_32_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_64_32_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_64_32_8,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_64_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shl_uu_64_64_8,
                        16 : CmmBinAsInstruction.fix_64__shl_uu_64_64_8,
                        32 : CmmBinAsInstruction.fix_64__shl_uu_64_64_8,
                        64 : CmmBinAsInstruction.fix_64__shl_uu_64_64_64,
                    },
                },
            },
        },
        'shr' : {
            'su' : {
                8  : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_8_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_8_32_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_8_32_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_8_32_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_8_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_8_64_64,
                    },
                },
                16 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_8_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_8_32_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_8_32_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_8_32_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_8_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_8_64_64,
                    },
                },
                32 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_32_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_32_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_32_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_32_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_32_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_32_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_32_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_32_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_32_32_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_32_32_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_32_32_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_32_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_32_64_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_32_64_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_32_64_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_32_64_64,
                    },
                },
                64 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_64_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_64_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_64_32_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_64_32_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_64_32_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_64_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shr_su_64_64_8,
                        16 : CmmBinAsInstruction.fix_64__shr_su_64_64_8,
                        32 : CmmBinAsInstruction.fix_64__shr_su_64_64_8,
                        64 : CmmBinAsInstruction.fix_64__shr_su_64_64_64,
                    },
                },
            },
            'uu' : {
                8  : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_8_8_16,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_8_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_8_16_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_8_16_32,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_8_16_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_8_16_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_8_16_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_8_64_64,
                    },
                },
                16 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_16_8_32,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_8_8_16,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_8_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_8_16_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_8_16_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_8_16_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_8_16_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_8_64_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_8_64_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_8_64_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_8_64_64,
                    },
                },
                32 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_16_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_16_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_16_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_16_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_16_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_16_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_8_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_8_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_8_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_8_8_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_32_64_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_32_64_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_32_64_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_32_64_64,
                    },
                },
                64 : {
                    8  : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_64_8_64,
                    },
                    16 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_64_8_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_64_8_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_64_8_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_64_8_64,
                    },
                    32 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_64_32_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_64_32_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_64_32_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_64_32_64,
                    },
                    64 : {
                        8  : CmmBinAsInstruction.fix_64__shr_uu_64_64_8,
                        16 : CmmBinAsInstruction.fix_64__shr_uu_64_64_8,
                        32 : CmmBinAsInstruction.fix_64__shr_uu_64_64_8,
                        64 : CmmBinAsInstruction.fix_64__shr_uu_64_64_64,
                    },
                },
            },
        },
    }

    def fix_64_bit_arithmetic(self) :
        op = self.get_op()

        if (op == 'shr' or op == 'shri') :
            op = 'shr'
        elif (op == 'shl' or op == 'shli') :
            op = 'shl'
        else :
            op = 'alu'

        opC = self.get_opC()
        opA = self.get_opA()
        opB = self.get_opB()
        opC_size = opC.get_size()
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        if op == 'shl' or op == 'shr' :
            if opA.is_signed() :
                sign = 'su'
            else :
                sign = 'uu'
        else :
            if opA.is_signed() and opB.is_signed() :
                sign = 'ss'
            elif opA.is_signed() and opB.is_unsigned() :
                sign = 'su'
            elif opA.is_unsigned() and opB.is_signed() :
                sign = 'us'
            else :
                sign = 'uu'

        opC_size = self.get_normalized_size(opC_size)
        opA_size = self.get_normalized_size(opA_size)
        opB_size = self.get_normalized_size(opB_size)

        [opC, opA, opB] = [ x.get_reg() if x.is_a_var() else x for x in [opC, opA, opB] ]

        self.set_opC(opC)
        self.set_opA(opA)
        self.set_opB(opB)

        instr = CmmBinAsInstruction.fix_64[op][sign][opC_size][opA_size][opB_size](self)

        self.set_final_asm_form( instr )

    def _alu_reg_reg_reg_fix_size(self) :
        [opA, opB] = self.body[2 : 4]
        opA_size = opA.get_size()
        opB_size = opB.get_size()

        print ("fix_size_operand !!! size of " + str(opA_size) + ", " + str(opA_size) + ", " + str(opB_size) + "\n")
        self.body += [self.get_asm_operand_size(opA_size, opA_size, opB_size)]

    def _alu_mem_reg_reg_fix_size(self) :
        [opA, opB, mp] = [ x.get_reg() if x.is_a_var() else x for x in self.body[2 : 5] ]
        opA_size = opA.get_size()
        opB_size = opB.get_size()
        mp_index = mp.get_index()

        self.body.pop()

        self.body += [self.get_asm_operand_size(opA_size, opA_size, opB_size), mp_index]

    def _alu_reg_reg_imm_fix_size(self) :
        opA = self.body[2]
        opA_size = opA.get_size()

        self.body += [self.get_asm_operand_size(opA_size, opA_size, '_')]

    def _alu_mem_reg_imm_fix_size(self) :
        [opA, mp] = [ x.get_reg() if x.is_a_var() else x for x in [self.body[2], self.body[4]] ]
        opA_size = opA.get_size()
        mp_index = mp.get_index()

        self.body.pop()

        self.body += [self.get_asm_operand_size(opA_size, opA_size, '_'), mp_index]

    def fix_size_operand(self) :
        instr = self.get_final_asm_form()

        if len(instr) :
            for lv in instr :
                lv.fix_size_operand()
        else :
            print ("fix_size_operand 0 " + str(self.body) + "\n")
            print ("fix_size_operand 1 " + str(self.body[1]) + " " + str(self.body[3]) + "\n")

            if self.body[3].is_a_var() or self.body[3].is_a_reg() :
                if self.body[1].is_a_var() or self.body[1].is_a_reg() :
                    self._alu_reg_reg_reg_fix_size()
                else :
                    self._alu_mem_reg_reg_fix_size()
            else :
                if self.body[1].is_a_var() or self.body[1].is_a_reg() :
                    self._alu_reg_reg_imm_fix_size()
                else :
                    self._alu_mem_reg_imm_fix_size()

    def compute_left(self) :
        return [self.body[1]]

    def _is_addition(self) :
        op = self.body[0]

        return op == 'add' or op == 'addm' or op == 'adi' or op == 'adim'

    def _is_subtraction(self) :
        op = self.body[0]

        return op == 'sub' or op == 'subm' or op == 'sbi' or op == 'sbim'

    def _get_flags_set(self) :
        return ['E', 'C'] if self._is_addition() or self._is_subtraction() else ['E']

    evaluate = {
        'add'     :    lambda fst, sec: fst + sec ,
        'addm'    :    lambda fst, sec: fst + sec ,
        'adi'     :    lambda fst, sec: fst + sec ,
        'adim'    :    lambda fst, sec: fst + sec ,
        'sub'     :    lambda fst, sec: fst - sec ,
        'subm'    :    lambda fst, sec: fst - sec ,
        'sbi'     :    lambda fst, sec: fst - sec ,
        'sbim'    :    lambda fst, sec: fst - sec ,
        'and'     :    lambda fst, sec: fst & sec ,
        'andm'    :    lambda fst, sec: fst & sec ,
        'andi'    :    lambda fst, sec: fst & sec ,
        'andim'   :    lambda fst, sec: fst & sec ,
        'or'      :    lambda fst, sec: fst | sec ,
        'orm'     :    lambda fst, sec: fst | sec ,
        'ori'     :    lambda fst, sec: fst | sec ,
        'orim'    :    lambda fst, sec: fst | sec ,
        'xor'     :    lambda fst, sec: fst ^ sec ,
        'xorm'    :    lambda fst, sec: fst ^ sec ,
        'xori'    :    lambda fst, sec: fst ^ sec ,
        'xorim'   :    lambda fst, sec: fst ^ sec ,
        'shl'     :    lambda fst, sec: fst << sec ,
        'shlm'    :    lambda fst, sec: fst << sec ,
        'shli'    :    lambda fst, sec: fst << sec ,
        'shlim'   :    lambda fst, sec: fst << sec ,
        'shr'     :    lambda fst, sec: fst >> sec ,
        'shrm'    :    lambda fst, sec: fst >> sec ,
        'shri'    :    lambda fst, sec: fst >> sec ,
        'shrim'   :    lambda fst, sec: fst >> sec ,
    }

    def evaluate(self) :
        op = self.body[0]
        [left] = self.get_left()
        [opA, opB] = [ x if x.is_a_constant() else x.get_value() for x in self.get_operands() ]
        lattice_elm = None

        if opA.is_ordinary_constant() and opB.is_ordinary_constant() :
            lattice_elm = CmmBinAsInstruction.evaluate[op](opA, opB)
        elif opA.is_max() or opB.is_max() :
            lattice_elm = CmmValueLattice('multiple_data')
        # else {
        # bug : can be done better, e.g. a & 0 = 0
        # $lattice_elm = CmmValueLattice->new('no_data') ;
        # }

        if lattice_elm :
            left.set_value(lattice_elm)

            opC = left.get_value() ; # mask with size of left

            if opC.is_ordinary_constant() :
                if opC :
                    self._reset_E()
                else :
                    self._set_E()

                if self._is_addition() :
                    if lattice_elm < 1 << left.get_size() :
                        self._reset_C()
                    else :
                        self._set_C()
                elif self._is_subtraction() :
                    if opA >= opB :
                        self._reset_C()
                    else :
                        self._set_C()
            elif opC.is_max() :
                for lv in self.get_left_side_effect() :
                    lv.set_value(opC.new())

            # $self->set_value($opC->new()) ;

    spec_opcode = {
        'add' : 0,
        'adi' : 0,
        'sub' : 0,
        'sbi' : 0,
        'shl' : 0,
        'shli': 0,
        'shr' : 0,
        'shri': 0,
    }

    def get_asm_code(self) :
        instr = self.get_final_asm_form()

        if len(instr) :
            return [ y for x in instr for y in y.get_asm_code() ]
        else :
            code = ''

            body = [ x.get_reg() if hasattr(x, 'is_a_var') and x.is_a_var() else x for x in self.body ]

            if body[-2].is_a_constant() and not CmmBinAsInstruction.spec_opcode.has_key(body[0]) :
                body[-2] = "0x%02x" % body[-2]

            code = "\t" + str(body.pop(0)) + "\t"

            if len(body) < 5 :
                if (filter(lambda x: not x, body)) :
                    print "blya 0 " + str([ (">" + str(x) + "<")  for x in body]) + "%%% " + str(self.body) + "\n")
                code .= ",\t".join([ str(x) for x in body])
            else :
                if (filter(lambda x: not x, body)) :
                    print "blya 1 " + str([ (">" + str(x) + "<")  for x in body]) + str(self.body + "\n")

                code += " ,\t".join([ str(x) for x in body[0:5] ])
                code += ",  " + ",  ".join([ str(x) for x in body[5:] ])

            return [code]
